/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/*
 * This SIP module is based on and is an extension of the modules mod_echo
 * and mod_cgi included as standard modules with the Apache 2.0 distribution
 */

#include "ap_config.h"
#include "ap_mmn.h"
#include "httpd.h"
#include "http_config.h"
#include "http_connection.h"
#include "apr_fnmatch.h"
#include "http_log.h"
#include "http_core.h"

API_VAR_EXPORT module sip_module;

/* from httpd.h (same value as well) so not sure why I'm not getting it
   when I include that file. */
#define DEFAULT_TIMEOUT 120000 
#define BUF_SIZE 8192
#define MAX_ENV 100

/* linked list of the things we read since we read, parse, and then need
   to pass to the client script */
typedef struct StringNode
{
    char *payload;
    struct StringNode *next;
} StringNode;


typedef struct
{
    int bEnabled;
    char *path;
} SipConfig;


char* sip_header_transform (ap_pool_t *p, char *input) {

    /* according to the spec, we need to take all headers and
       translate by prepending SIP_, capitalizing all chars, and
       replacing - with _, so.... */
    
    char *returnval, *current, *temp, *temp2;
    temp = current = ap_psprintf(p, "%s", input);
    while (*current != '\0') {
        if (*current == '-')
            *current = '_';
        else
            *current = (char)toupper(*current);
        current++;
    }
    
    temp2 = ap_getword_nc(p, &temp, ':');
    returnval = ap_psprintf(p, "SIP_%s", temp2);

    return returnval;
}


int sip_check_scriptfile(char *filename, conn_rec *c) {

    /* this is UNIX specific, unlike all the other code. I didn't have an
       apache file structure or a request record to pass (far too many of the
       routines need request_rec which makes doing other protocols a pain */

    struct stat filestat;
    extern int errno;

    if (stat(filename, &filestat) != 0) {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                      "Unable to check status of script %s (may not exist).",
                      filename);
        return -1;
    }


    if (!S_ISREG(filestat.st_mode)) {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                      "SIP script name %s cooresponds to directory, not file.",
                      filename);
        return -1;
    }
    
    if (access(filename, X_OK) != 0) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                          "SIP script name %s not executable.", filename);
        return -1;
    }

    return 0;

}


void sip_error_400 (BUFF *buffer) {

    /* write error to the buff and flush */
    ap_bprintf(buffer, "SIP/2.0 400 Server Received Invalid Message\r\n"); 
    ap_bflush(buffer);
    
}


void sip_error_500 (BUFF *buffer) {

    /* write error to the buff and flush */
    ap_bprintf(buffer, "SIP/2.0 500 Internal Server Error\r\n"); 
    ap_bflush(buffer);
    
}


void sip_error_505 (BUFF *buffer) {

    /* write error to the buff and flush */
    ap_bprintf(buffer, "SIP/2.0 505 This server supports only SIP/2.0\r\n"); 
    ap_bflush(buffer);
    
}


int parse_request_line (char *message, char **method, char **uri,
                      char **user, char **protocol, ap_pool_t *p) {

    /* on failure all four strings will be null, return -1,
       else set to correct values and return 0 */
    
    char *temp, *temp2;

    /* use these as temps so no results passed if parse fails */
    char *t_method, *t_protocol, *t_uri, *t_user;
    *method = *uri = *user = *protocol = NULL;
    
    t_method = ap_getword_white_nc(p, ((char **)&message));
    /* if the line is empty or doesn't start w/a method, return error
       invite is most likely, so list it first so we save if we
       have lazy evaluation */
    if((ap_fnmatch("", t_method, 0) != FNM_NOMATCH) ||
       (!((ap_fnmatch("INVITE", t_method, 0) != FNM_NOMATCH) ||
          (ap_fnmatch("ACK", t_method, 0) != FNM_NOMATCH) ||
          (ap_fnmatch("OPTIONS", t_method, 0) != FNM_NOMATCH) ||
          (ap_fnmatch("BYE", t_method, 0) != FNM_NOMATCH) ||
          (ap_fnmatch("CANCEL", t_method, 0) != FNM_NOMATCH) ||
          (ap_fnmatch("REGISTER", t_method, 0) != FNM_NOMATCH)))) {
        return -1;
    }
    
    /* next parse the URI, which should start w/sip
     keep the copy t_uri unmodified, and use temp to check
     that it starts with sip: and get user name*/
    t_uri = ap_getword_white_nc(p, ((char **)&message));
    if(ap_fnmatch("", t_uri, 0) != FNM_NOMATCH) {
        return -1;
    }
    temp = ap_psprintf(p, "%s", t_uri);
    /* this should start with the word sip followed by colon */
    temp2 = ap_getword_nc(p, &temp, ':');
    if(ap_fnmatch("sip", temp2, 0) == FNM_NOMATCH) {
        return -1;
    }

    /* want to strip the port, (if there is one, which will have a :)
     and then any additional info (which will have ;) */
    temp2 = ap_getword_nc(p, &temp, ':');
    t_user = ap_getword_nc(p, &temp2, ';');
    
    /* should see the SIP version next */
    t_protocol = ap_getword_white_nc(p, ((char **)&message));
    if((ap_fnmatch("", t_protocol, 0) != FNM_NOMATCH) ||
       (ap_fnmatch("SIP*", t_protocol, 0) == FNM_NOMATCH))
        return -1;

    /* copy into the strings provided by the caller */
    *method = ap_psprintf(p, "%s", t_method);
    *uri = ap_psprintf(p, "%s", t_uri);
    *user = ap_psprintf(p, "%s", t_user);
    *protocol = ap_psprintf(p, "%s", t_protocol);

    return 0;

}


/* copied directly over from http_protocol.c w/no modification */

/* Get a line of protocol input, including any continuation lines
 * caused by MIME folding (or broken clients) if fold != 0, and place it
 * in the buffer s, of size n bytes, without the ending newline.
 *
 * Returns -1 on error, or the length of s.
 *
 * Note: Because bgets uses 1 char for newline and 1 char for NUL,
 *       the most we can get is (n - 2) actual characters if it
 *       was ended by a newline, or (n - 1) characters if the line
 *       length exceeded (n - 1).  So, if the result == (n - 1),
 *       then the actual input line exceeded the buffer length,
 *       and it would be a good idea for the caller to puke 400 or 414.
 */
static int getline(char *s, int n, BUFF *in, int fold)
{
    char *pos, next;
    int retval;
    int total = 0;

    pos = s;

    do {
        retval = ap_bgets(pos, n, in);
       /* retval == -1 if error, 0 if EOF */

        if (retval <= 0)
            return ((retval < 0) && (total == 0)) ? -1 : total;

        /* retval is the number of characters read, not including NUL      */

        n -= retval;            /* Keep track of how much of s is full     */
        pos += (retval - 1);    /* and where s ends                        */
        total += retval;        /* and how long s has become               */

        if (*pos == '\n') {     /* Did we get a full line of input?        */
            /*
             * Trim any extra trailing spaces or tabs except for the first
             * space or tab at the beginning of a blank string.  This makes
             * it much easier to check field values for exact matches, and
             * saves memory as well.  Terminate string at end of line.
             */
            while (pos > (s + 1) && (*(pos - 1) == ' ' || *(pos - 1) == '\t'))
            {
                --pos;          /* trim extra trailing spaces or tabs      */
                --total;        /* but not one at the beginning of line    */
                ++n;
            }
            *pos = '\0';
            --total;
            ++n;
        }
        else
            return total;       /* if not, input line exceeded buffer size */

        /* Continue appending if line folding is desired and
         * the last line was not empty and we have room in the buffer and
         * the next line begins with a continuation character.
         */
    } while (fold && (retval != 1) && (n > 1)
                  && (next = ap_blookc(in))
                  && ((next == ' ') || (next == '\t')));

    return total;
}


/* copied and modified from run_cgi_child in mod_cgi.c */
static ap_status_t run_sip_cgi_child(BUFF **script_out, BUFF **script_in,
                                     BUFF **script_err, const char *command,
                                     const char *path, char *const argv[],
                                     ap_pool_t *p, char *env[])
{

    ap_procattr_t *procattr;
    ap_proc_t procnew;
    ap_status_t rc = APR_SUCCESS;
    ap_file_t *file = NULL;
    ap_iol *iol;
    long int timeout = DEFAULT_TIMEOUT;

    RAISE_SIGSTOP(CGI_CHILD);

    /* Transumute ourselves into the script.
     * NB only ISINDEX scripts get decoded arguments.
     */
    if (((rc = ap_createprocattr_init(&procattr, p)) != APR_SUCCESS) ||
        ((rc = ap_setprocattr_io(procattr, 
                                 APR_CHILD_BLOCK, 
                                 APR_CHILD_BLOCK,
                                 APR_CHILD_BLOCK)) != APR_SUCCESS) ||
        ((rc = ap_setprocattr_dir(procattr, path))
         != APR_SUCCESS) ||
        ((rc = ap_setprocattr_cmdtype(procattr, APR_PROGRAM)) != APR_SUCCESS))
    {
        /* Something bad happened, tell the world. */
        ap_log_perror(APLOG_MARK, APLOG_ERR,
                      0, p, "couldn't create child process: %s", command);
    }
    else {
        rc = ap_create_process(&procnew, command, argv, env, procattr, p);
        
        if (rc != APR_SUCCESS) {
            /* Bad things happened. Everyone should have cleaned up. */
            ap_log_perror(APLOG_MARK, APLOG_ERR,
                          0, p, "couldn't create child process: %s", command);
        } else {
            
            ap_note_subprocess(p, &procnew, kill_after_timeout);
            
            /* Fill in BUFF structure for parents pipe to child's stdout */
            //ap_get_childout(&file, procnew);
            file = procnew.out;
            iol = ap_create_file_iol(file);
            if (!iol)
                return APR_EBADF;
            *script_in = ap_bcreate(p, B_RD);
            ap_bpush_iol(*script_in, iol);
            ap_bsetopt(*script_in, BO_TIMEOUT, &timeout);

            /* Fill in BUFF structure for parents pipe to child's stdin */
            //ap_get_childin(&file, procnew);
            file = procnew.in;
            iol = ap_create_file_iol(file);
            if (!iol)
                return APR_EBADF;
            *script_out = ap_bcreate(p, B_WR);
            ap_bpush_iol(*script_out, iol);
            ap_bsetopt(*script_out, BO_TIMEOUT, &timeout);

            /* Fill in BUFF structure for parents pipe to child's stderr */
            //            ap_get_childerr(&file, procnew);
            file = procnew.err;
            iol = ap_create_file_iol(file);
            if (!iol)
                return APR_EBADF;
            *script_err = ap_bcreate(p, B_RD);
            ap_bpush_iol(*script_err, iol);
            ap_bsetopt(*script_err, BO_TIMEOUT, &timeout);
        }
    }

    return (rc);
}


static int process_sip_connection(conn_rec *c)
{

    
#define HACK_TILL_FIXED 1
#ifdef HACK_TILL_FIXED
    ap_ssize_t r;
#endif

    
    int len, numread, env_index = 0;
    char *buf = NULL, *word = NULL, *translated = NULL, *enviro_value = NULL;
    char *method = NULL, *protocol = NULL, *user = NULL, *uri = NULL;
    char *command = NULL;
    char *env[MAX_ENV], *argv[2];
    BUFF *script_out = NULL, *script_in = NULL, *script_err = NULL;
    StringNode *listBase = NULL, *listItem = NULL;

    /* get the module configuration */
    SipConfig *pConfig=ap_get_module_config(c->base_server->module_config,
					     &sip_module);

    /* if the module isn't active, return DECLINED */
    if(!pConfig->bEnabled) {
	return DECLINED;
    }

    buf = (char*)ap_palloc(c->pool, (BUF_SIZE + 2));


    
#ifdef HACK_TILL_FIXED
    ap_blookstring(c->client, buf, BUF_SIZE, &r);
    if(ap_fnmatch("*SIP/2.0*", buf, 0) == FNM_NOMATCH) {
        return DECLINED;
    }
#endif    

    
    /* if it is SIP, we should have the string "SIP/2.0" somewhere in the
       first line. We know that this port is only for SIP, so look for
       SIP/2.0. If we don't find it, try SIP. If not, return Bad Request,
       else SIP version not supported */
    
    len = getline(buf, BUF_SIZE, c->client, 1);
    
    if (len == 0) {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                      "Malformed SIP message received.");
        
        sip_error_500(c->client);
        return OK;
    }

    if(ap_fnmatch("*SIP/2.0*", buf, 0) == FNM_NOMATCH) {
        
        if(ap_fnmatch("*SIP*", buf, 0) == FNM_NOMATCH) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                          "Malformed SIP message received.");
            while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
            /* check for error len != numread here? */
            sip_error_400(c->client);
            return OK;
        } else {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                          "SIP message not SIP/2.0.");
            while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
            sip_error_505(c->client);
            return OK;
        }
    }
    
    /* If we get here it is SIP and the correct version (2.0)
       Extract some information out of the headers and put it into
       environment variables and pass them to the CGI script. */
    
    listBase = ap_palloc(c->pool, sizeof(StringNode));
    listBase->payload = ap_psprintf(c->pool, "%s\n", buf);
    listItem = listBase;
    listItem->next = NULL; 
    
    /* this should be the first line, so parse out the method,
       destination user and protocol version */
    
    if (pConfig->path == NULL) {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                      "Path not configured for SIP module.");
        while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
        sip_error_500(c->client);
        return OK;
    }
    
    if ((parse_request_line (buf, &method, &uri, &user, &protocol,
                             c->pool) != 0)) {
        ap_log_perror(APLOG_MARK, APLOG_NOTICE,
                      0, c->pool,
                      "Received Malformed SIP request from client.");
        while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
        sip_error_400(c->client);
        return OK;
    }
    
    /* strip out all .. and . to prevent hacking */
    ap_getparents(user);
    command = ap_psprintf(c->pool, "%s%s", pConfig->path, user);
    
    if (sip_check_scriptfile(command, c) != 0) {
        /* try default */
        command = ap_psprintf(c->pool, "%sdefault.cgi", pConfig->path);
            if (sip_check_scriptfile(command, c) != 0) {
                while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
                sip_error_500(c->client);
                return OK;
            }
    }

    /* from the message, then message independent */
    env[env_index++] = ap_psprintf(c->pool, "REQUEST_METHOD=%s", method);
    env[env_index++] = ap_psprintf(c->pool, "REQUEST_URI=%s", uri);
    env[env_index++] = ap_psprintf(c->pool, "SERVER_PROTOCOL=%s",protocol);
    env[env_index++] = ap_psprintf(c->pool, "REMOTE_ADDR=%s", c->remote_ip);
    env[env_index++] = ap_psprintf(c->pool, "GATEWAY_INTERFACE=%s",
                                   "SIP-CGI/1.1");
    env[env_index++] = ap_psprintf(c->pool, "SERVER_SOFTWARE=%s",
                                   ap_get_server_version());
    env[env_index++] = ap_psprintf(c->pool, "SERVER_NAME=%s",
                                   c->local_ip);
    env[env_index++] = ap_psprintf(c->pool, "SERVER_PORT=%d",
                                   (int)((c->local_addr).sin_port));

    /* now read the rest of the lines */
    while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0) {

        listItem->next = ap_palloc(c->pool, sizeof(StringNode));
        listItem = listItem->next;
        listItem->payload = ap_psprintf(c->pool, "%s\n", buf);
        listItem->next = NULL;
        
        word = ap_getword_white_nc(c->pool, ((char **)&buf));


        if(ap_fnmatch("Content-Length*", word, 0) != FNM_NOMATCH) { 
            enviro_value = ap_getword_nc(c->pool, ((char **)&buf), '\r');
            env[env_index++] = ap_psprintf(c->pool, "CONTENT_LENGTH=%s",
                                           enviro_value); 
            env[env_index++] = ap_psprintf(c->pool, "SIP_CONTENT_LENGTH=%s",
                                           enviro_value); 
            break; /* stop parsing -- body follows, so read no more lines */
        }

        if(ap_fnmatch("Content-Type*", word, 0) != FNM_NOMATCH) { 
            enviro_value = ap_getword_nc(c->pool, ((char **)&buf), '\r');
            env[env_index++] = ap_psprintf(c->pool, "SIP_CONTENT_TYPE=%s",
                                           enviro_value); 
            env[env_index++] = ap_psprintf(c->pool, "CONTENT_TYPE=%s",
                                           enviro_value); 
        } else {
            /* translate into proper format for CGI protocol */
            translated = sip_header_transform(c->pool, word);
            enviro_value = ap_getword_nc(c->pool, ((char **)&buf), '\r');
            env[env_index++] = ap_psprintf(c->pool, "%s=%s", translated, 
                                           enviro_value); 
        }
            
        if (env_index >= (MAX_ENV - 2)) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                          "Too many SIP header lines in message (%s).",
                          "compile time limit");
            while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
            sip_error_400(c->client);
            return OK;
        }

    }

    /* finished building environment */
    env[env_index] = NULL;
    
    /* No command line arguments */
    argv[0] = ap_psprintf(c->pool, "%s", command);
    argv[1] = NULL;

    /* spawn and run the child */
    if ((run_sip_cgi_child(&script_out, &script_in, &script_err, command,
                           pConfig->path, argv, c->pool, env))
        != APR_SUCCESS) {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, c->pool,
                      "Failure spawning child for script %s.", command);
        while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0);
        sip_error_500(c->client);
        return OK;
    }

    /* send the stuff we read into the linked list */
    listItem = listBase;
    while (listItem != NULL) {
        
        ap_bprintf(script_out, "%s", listItem->payload); 
        listItem = listItem->next;
            
    }
    

    /* send the remainder of the input to the client's socket */
    while ((len = getline(buf, BUF_SIZE, c->client, 1)) > 0) {

        ap_bwrite(script_out, buf, len, &numread);
        /* check for error len != numread here? */
    }

    ap_bflush(script_out);
    ap_bclose(script_out);
            
    /* get the standard out from the script and write it to the
       socket of the client, then flush the output to the client
       this is essentially anything after Content-Length right
       now. Should use the length to know how much to read/send */
    while (ap_bgets(buf, 8100, script_in) > 0) {
        ap_bputs(buf, c->client);
    }
    ap_bflush(c->client); 
    
    return OK;

}


/* ************************************************************************
                            Standard Module Stuff
   ************************************************************************ */

static void *create_sip_server_config(ap_pool_t *p,server_rec *s)
{
    SipConfig *pConfig=ap_pcalloc(p,sizeof *pConfig);
    pConfig->bEnabled=0;
    pConfig->path = NULL;
    return pConfig;
}

static const char *sip_on(cmd_parms *cmd, void *dummy, char *arg)
{
    SipConfig *pConfig=ap_get_module_config(cmd->server->module_config,
                                             &sip_module);
    pConfig->bEnabled=1;
    return NULL;
}

static const char *sip_directory(cmd_parms *cmd, void *dummy, char *arg)
{
    SipConfig *pConfig=ap_get_module_config(cmd->server->module_config,
                                             &sip_module);
    pConfig->path=arg;
    return NULL;
}

static const command_rec sip_cmds[] = {
{ "ProtocolSip", sip_on, NULL, RSRC_CONF, RAW_ARGS,
  "Run an SIP server on this host" },
{ "SipDirectory", sip_directory, NULL, RSRC_CONF, RAW_ARGS,
  "Directory for SIP CGI scripts" },
{ NULL }
};

static void register_hooks(void)
{
    ap_hook_process_connection(process_sip_connection,NULL,NULL,
                               AP_HOOK_MIDDLE);
}

API_VAR_EXPORT module sip_module = {
    STANDARD20_MODULE_STUFF,
    NULL,			/* create per-directory config structure */
    NULL,			/* merge per-directory config structures */
    create_sip_server_config,	/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    sip_cmds,			/* command ap_table_t */
    NULL,			/* handlers */
    register_hooks		/* register hooks */
};
