
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Endpoint_cxx_Version =
    "$Id: Endpoint.cxx,v 1.46 2001/01/23 22:22:34 cktam Exp $";
#include <VTime.hxx>
#include <signal.h>
#include <sys/types.h>
#include <typeinfo>
#include <unistd.h>
#include "MgAgent.hxx"
//#include "sdp.hxx"
#include "Connection.h"
#include "DigitMap.h"
#include "Hardware.h"
#include "Endpoint.h"


Endpoint::Endpoint ( string endpointId,
                     int numConnections,
                     MgAgent* agent,
                     Hardware& hardware,
                     Fifo < EndpointMsg > * inputQ,
                     Fifo < HardwareMsg > * outputQ )
        : myId(endpointId)

{
    myAgent = agent;
    myQ = inputQ;
    hardwareQ = outputQ;

    for (int i = 0; i < TimeoutMaxTimer; i++)
    {
        myTimerEventId[i] = (FifoEventId) NULL;
    }

    if (hardware.isOffhook())
    {
        myHookstate = HookstateOnhook;
    }
    else
    {
        myHookstate = HookstateOffhook;
    }

    activeReqEvs = new list < ReqEv* > ;
    activeReqSigs = new list < ReqSig* > ;

    // preallocate all connections
    myMaxConnections = numConnections;
    for (int i = 0; i < myMaxConnections; i++)
    {
        myConnections[i] = new Connection(myAgent, hardwareQ, endpointId);
        assert(myConnections[i] != NULL);
    }
    myDigits = new DigitMap();

    // initialize myReqEv arrays
    initializeGenericEvents();
    initializeLineEvents();
    initializeDtmfEvents();

    // initialize myReqSig arrays
    initializeGenericSignals(hardware);
    initializeLineSignals(hardware);
    initializeDtmfSignals(hardware);

} // end Endpoint::Endpoint()


Endpoint::~Endpoint (void)
{
    int i;

    for (i = 0; i < GenericReqEvMaxEvent; i++)
    {
        delete myGenericReqEv[i];
        myGenericReqEv[i] = NULL;
    }
    for (i = 0; i < GenericReqSigMaxSignal; i++)
    {
        delete myGenericReqSig[i];
        myGenericReqSig[i] = NULL;
    }

    for (i = 0; i < LineReqEvMaxEvent; i++)
    {
        delete myLineReqEv[i];
        myLineReqEv[i] = NULL;
    }
    for (i = 0; i < LineReqSigMaxSignal; i++)
    {
        delete myLineReqSig[i];
        myLineReqSig[i] = NULL;
    }

    for (i = 0; i < DtmfReqEvMaxEvent; i++)
    {
        delete myDtmfReqEv[i];
        myDtmfReqEv[i] = NULL;
    }
    for (i = 0; i < DtmfReqSigMaxSignal; i++)
    {
        delete myDtmfReqSig[i];
        myDtmfReqSig[i] = NULL;
    }

    // individual Connections should already be deleted but may need
    // to double check here by iterating through list
    delete myConnections;

}


void*
Endpoint::endpointMain (void* parms)
{
    EndpointMsg msg;

    while (1)
    {

        msg = myQ->getNext();
        cerr << "Endpt:  getting message from myQ\n";

        switch (msg.type)
        {
            case EndpointAgentCommandType:
            {
                cpLog(LOG_DEBUG, "got AgentCmd");
                // cpLog (LOG_DEBUG, "%s <- Agent Cmd", myId.data());
                processAgentCommand(msg.command);
            }
            break;

            case EndpointHardwareEventType:
            {
                cpLog(LOG_DEBUG, "got HW Event");
                // cpLog (LOG_DEBUG, "%s <- HW Event", myId.data());
                processHardwareEvent(msg.event);
            }
            break;

            case EndpointTimeoutType:
            {
                cpLog(LOG_DEBUG, "got Timeout");
                // cpLog (LOG_DEBUG, "%s <- Timeout", myId.data());
                processTimeout(msg.timeout);
            }
            break;

            default:
            {
                cpLog (LOG_DEBUG, "%s <- Unknown Msg", myId.data());
            }
            break;
        } // end switch

    }
}


void
Endpoint::processAgentCommand (EndpointAgentCommand& command)
{
    switch (command.type)
    {
        case MgEndpointConfigurationType:
        {
            EPCF epcf;
            epcf.dynamicCast(command.cmd);
            processEPCF( epcf );
        }
        break;

        case MgCreateConnectionType:
        {
            CRCX crcx;
            crcx.dynamicCast(command.cmd);
            processCRCX( crcx );
        }
        break;

        case MgModifyConnectionType:
        {
            MDCX mdcx;
            mdcx.dynamicCast(command.cmd);
            processMDCX( mdcx );
        }
        break;

        case MgDeleteConnectionAgentType:
        {
            DLCX dlcx;
            dlcx.dynamicCast(command.cmd);
            processDLCX( dlcx );
        }
        break;

        case MgNotificationRequestType:
        {
            RQNT rqnt;
            rqnt.dynamicCast(command.cmd);
            processRQNT( rqnt );
        }
        break;

        case MgAuditEndpointType:
        {
            AUEP auep;
            auep.dynamicCast(command.cmd);
            processAUEP( auep );
        }
        break;

        case MgAuditConnectionType:
        {
            AUCX aucx;
            aucx.dynamicCast(command.cmd);
            processAUCX( aucx );
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Rcvd unknown cmd from gateway, type = %d",
                  command.type);
        }
        break;

    }

}



void
Endpoint::processHardwareEvent (HardwareEvent event)
{
    ReqEv* reqEv = 0;

    switch (event)
    {
        case EventOffhook:
        {

            cpLog(LOG_DEBUG, "Received offhook");

            reqEv = myLineReqEv[LineReqEvOffhook];

            if (myHookstate == HookstateFlashhook)
            {
                gettimeofday(&finishDebounceTime, 0);

                diffTime = finishDebounceTime.tv_usec -
                           startDebounceTime.tv_usec +
                           1000000 * ( finishDebounceTime.tv_sec -
                                       startDebounceTime.tv_sec);

                if (diffTime > DEBOUNCE_DELAY)
                {
                    cpLog(LOG_DEBUG, "Report flashhook and stop timer");
                    stopTimer(TimeoutFlashhook);
                    reqEv = myLineReqEv[LineReqEvFlashhook];
                }
                else
                {
                    cpLog(LOG_DEBUG, "Debounce check < 20ms    not reporting onhook");
                    stopTimer(TimeoutFlashhook);
                    myHookstate = HookstateOffhook;
                }

            }
            else
            {
                reqEv = myLineReqEv[LineReqEvOffhook];
            }
        }
        break;

        case EventOnhook:
        {
            gettimeofday(&startDebounceTime, 0);

            cpLog(LOG_DEBUG, "Received onhook");

            myHookstate = HookstateFlashhook;
            cpLog(LOG_DEBUG, "Hookstate -> flashhook");
            cpLog(LOG_DEBUG, "Start flashhook timer");
            startTimer(TimeoutFlashhook, FLASHHOOK_TIMER_DURATION);
            myDigits->clearBuffer();
        }
        break;

        case EventFlashhook:
        {
            cpLog(LOG_DEBUG, "Received flashhook");
            reqEv = myLineReqEv[LineReqEvFlashhook];
            myHookstate = HookstateOffhook;
            cpLog(LOG_DEBUG, "Hookstate -> offhook");
            myDigits->clearBuffer();
        }
        break;

        case EventDigit0:
        {
            cpLog(LOG_DEBUG, "Received dtmf 0");
            reqEv = myDtmfReqEv[DtmfReqEv0];
        }
        break;

        case EventDigit1:
        {
            cpLog(LOG_DEBUG, "Received dtmf 1");
            reqEv = myDtmfReqEv[DtmfReqEv1];
        }
        break;

        case EventDigit2:
        {
            cpLog(LOG_DEBUG, "Received dtmf 2");
            reqEv = myDtmfReqEv[DtmfReqEv2];
        }
        break;

        case EventDigit3:
        {
            cpLog(LOG_DEBUG, "Received dtmf 3");
            reqEv = myDtmfReqEv[DtmfReqEv3];
        }
        break;

        case EventDigit4:
        {
            cpLog(LOG_DEBUG, "Received dtmf 4");
            reqEv = myDtmfReqEv[DtmfReqEv4];
        }
        break;

        case EventDigit5:
        {
            cpLog(LOG_DEBUG, "Received dtmf 5");
            reqEv = myDtmfReqEv[DtmfReqEv5];
        }
        break;

        case EventDigit6:
        {
            cpLog(LOG_DEBUG, "Received dtmf 6");
            reqEv = myDtmfReqEv[DtmfReqEv6];
        }
        break;

        case EventDigit7:
        {
            cpLog(LOG_DEBUG, "Received dtmf 7");
            reqEv = myDtmfReqEv[DtmfReqEv7];
        }
        break;

        case EventDigit8:
        {
            cpLog(LOG_DEBUG, "Received dtmf 8");
            reqEv = myDtmfReqEv[DtmfReqEv8];
        }
        break;

        case EventDigit9:
        {
            cpLog(LOG_DEBUG, "Received dtmf 9");
            reqEv = myDtmfReqEv[DtmfReqEv9];
        }
        break;

        case EventDigitStar:
        {
            cpLog(LOG_DEBUG, "Received dtmf *");
            reqEv = myDtmfReqEv[DtmfReqEvStar];
        }
        break;

        case EventDigitHash:
        {
            cpLog(LOG_DEBUG, "Received dtmf #");
            reqEv = myDtmfReqEv[DtmfReqEvHash];
        }
        break;

        case EventAudioFailed:
        {
        }

        break;

        case EventAudioClosed:
        {
        }

        break;

        default:
        {
            cpLog(LOG_ERR, "Rcvd unknown HW event = %d", event);
        }
        break;

    }

    if (reqEv != 0)
    {
        switch (reqEv->getAction())
        {
            case ActionNotify:
            {
                // create and send a Notify msg regarding the event
                cpLog(LOG_DEBUG, "Send Notify to agent");

                MgNotify ntfy(myId, reqEv->getRequestId(),
                              reqEv->getObservedEvent());
                myAgent->send(ntfy);

                // TBD, process response
                //cerr << "Agent rspCode: " << ntfy.getResponseCode() << '\n';

                //if (!(rqnt->keepSignalsActive()))
                clearAllSignals();

                // TBD, for now, continue to process future events based
                // on previously received RQNT, but eventually need to
                // enhance to store events until next RQNT arrives, then
                // process them according to that list of ReqEv

            }

            break;

            case ActionAccumulate:
            {
                cpLog(LOG_DEBUG, "Accumulate");
            }
            break;

            case ActionDigitMap:
            {
                cpLog(LOG_DEBUG, "Accumulate according to digit map");

                myDigits->handledigits(myAgent,
                                       myDtmfReqEv,
                                       myId,
                                       event);

                //                stopTimer (TimeoutInterdigit);
                //                startTimer(TimeoutInterdigit, 4000);

            }

            break;

            case ActionIgnore:
            {
                cpLog(LOG_DEBUG, "ActionIgnore");
            }
            break;

            default:
            {
                cpLog(LOG_DEBUG, "Invalid Action %d", reqEv->getAction());
            }
            break;
        }

    } 
}



void
Endpoint::processTimeout (EndpointTimeout timeout)
{

    ReqEv* reqEv = 0;

    myTimerEventId[timeout] = (FifoEventId) NULL;
    switch (timeout)
    {
        case TimeoutInterdigit:
        {
            cpLog(LOG_DEBUG, "TimeoutInterdigit");

            myDigits->handleDigitTimer(myAgent,
                                       myDtmfReqEv,
                                       myId);
        }
        break;

        case TimeoutFlashhook:
        {
            cpLog(LOG_DEBUG, "TimeoutFlashhook");
            if (myHookstate == HookstateFlashhook)
            {
                // treat previously received onhook as a real onhook
                cpLog(LOG_DEBUG, "Hookstate -> onhook");
                reqEv = myLineReqEv[LineReqEvOnhook];
                myHookstate = HookstateOnhook;
            }
            else
            {
                // ignore timeout because timer should have been stopped
                cpLog(LOG_DEBUG, "Ignore TimeoutFlashhook");
            }
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Rcvd unknown timeout = %d", timeout);
        }
        break;

    }

    if (reqEv != 0)
    {
        if (reqEv->isRequested())
        {
            if (reqEv->getAction() == ActionNotify)
            {
                // create and send a Notify msg regarding the event
                cpLog(LOG_DEBUG, "Send Notify to agent");
                MgNotify ntfy(myId, reqEv->getRequestId(),
                              reqEv->getObservedEvent());
                myAgent->send(ntfy);

                // TBD, comment out until blocking API is working
                // check response
                // cerr << "Agent Response Code: " << ntfy.getResponseCode() << '\n';
                //if (!(rqnt->keepSignalsActive()))
                clearAllSignals();
            } // end if ActionNotify
            else
            {
                // TBD need to add handling for this case
                cpLog(LOG_ERR, "Event requested but not ActionNotify");
            }
        }
    }
}



void
Endpoint::processEPCF (EPCF epcf)
{
    MgcpResponseCode responseCode;

    cpLog(LOG_DEBUG, "Endpoint::processEPCF");

    // TBD
    //bearerInfo = epcf->getBearerInformation();
    responseCode = TransactionExecuted;

    // respond to agent with result
    //epcf->setResponseCode(responseCode);
    myAgent->reply(epcf);

} 


void
Endpoint::processCRCX (CRCX crcx)
{
    MgcpResponseCode responseCode;
    int i;

    // search for available connection
    for (i = 0; i < myMaxConnections; i++)
    {
        if (myConnections[i]->getState() == ConnectionClosed)
            break;
    }

    if (i == myMaxConnections )
    {
        cpLog(LOG_ERR, "No connections available");
        responseCode = ReasonTransientInsufficientResources;
    }
    else
    {
        // give to appropriate connection
        responseCode = myConnections[i]->open(crcx);
    }

    // respond to agent with result
    crcx->setResponseCode(responseCode);
    myAgent->reply(crcx);

} 


void
Endpoint::processMDCX (MDCX mdcx)
{
    MgcpResponseCode responseCode;
    int i;

    // find appropriate connection
    for (i = 0; i < myMaxConnections; i++)
    {
        if (myConnections[i]->getId() == mdcx->getConnectionId())
            break;
    }

    if (i == myMaxConnections )
    {
        cpLog(LOG_ERR, "Connection %s does not exist",
              (mdcx->getConnectionId().c_str()));
        responseCode = ReasonIncorrectConnectionId;
    }
    else
    {
        // give to appropriate connection
        responseCode = myConnections[i]->modify(mdcx);
    }

    // respond to agent with result
    mdcx->setResponseCode(responseCode);
    myAgent->reply(mdcx);

} 


void
Endpoint::processDLCX (DLCX dlcx)
{
    MgcpResponseCode responseCode;
    int i;

    cpLog(LOG_DEBUG, "Endpoint::processDLCX");

    // find appropriate connection
    for (i = 0; i < myMaxConnections; i++)
    {
        if (myConnections[i]->getId() == dlcx->getConnectionId())
            break;
    }

    if (i == myMaxConnections )
    {
        cpLog(LOG_ERR, "Connection %s does not exist",
              (dlcx->getConnectionId().c_str()));
        responseCode = ReasonIncorrectConnectionId;
    }
    else
    {
        // give to appropriate connection
        responseCode = myConnections[i]->close(dlcx);
    }

    // respond to agent with result
    dlcx->setResponseCode(responseCode);
    myAgent->reply(dlcx);

}


void
Endpoint::clearAllReqEvs (void)
{
    // clear any outstanding requested events
    while (!(activeReqEvs->empty()))
    {
        ( activeReqEvs->front() )->stop();
        activeReqEvs->pop_front();
    }
}


void
Endpoint::clearAllSignals (void)
{
    // clear any outstanding requested events
    while (!(activeReqSigs->empty()))
    {
        ( activeReqSigs->front() )->stop();
        activeReqSigs->pop_front();
    }
}



void
Endpoint::restartInProgress(const string & restartMethod, 
                            const string& restartDelay = "none")
{
    MgRestartInProgress myRestart(myId, restartMethod, restartDelay);

    myAgent->send (myRestart);
}



void
Endpoint::processRQNT (RQNT rqnt)
{
    string requestId = rqnt->getRequestIdentifier();
    int numRequests;
    int i;  // index to access requested events

    cpLog(LOG_DEBUG, "Endpoint::processRQNT");

    // clear any outstanding requested events
    clearAllReqEvs();

    // If the call agent sent down a digit map, use it.
    if (rqnt->getDigitMap() != "")
    {
        myDigits->setDigitMap(rqnt->getDigitMap() );
    }

    // process new requested events
    numRequests = rqnt->getNumRequestedEvents();
    for (i = 0; i < numRequests; i++)
    {
        //addReqEv((rqnt->getItemRequestedEvents(i))->getEvent(), rqnt);
        addReqEv(requestId, rqnt->getItemRequestedEvents(i));
    }

    // clear any active signals
    clearAllSignals();

    // process new requested signals
    numRequests = rqnt->getNumSignalRequests();
    for (i = 0; i < numRequests; i++)
    {
        activateSignal((rqnt->getItemSignalRequests(i))->getEvent(), rqnt);
    }

    // reply with successful response code and string,
    rqnt->setResponseCode(ReasonTransactionExecuted);  // TBD MGCP API constant
    rqnt->setResponseString(OKResponse);
    myAgent->reply(rqnt);

} 


void
Endpoint::processAUEP (AUEP auep)
{
    cpLog(LOG_DEBUG, "Endpoint::processAUEP");
}


void
Endpoint::processAUCX (AUCX aucx)
{
    cpLog(LOG_DEBUG, "Endpoint::processAUCX");
}


void
Endpoint::addReqEv (string requestId, Sptr < MgRequestedEvent > event)
{
    ReqEv* reqEv = 0;

    cpLog(LOG_DEBUG, "Endpoint::addReqEv");
    switch (event->getPackageType())
    {
        case MgcpGenericMediaPackageType:
        {
            GenericReqEv index = getGenericReqEvIndex(
                                     (event->getEvent()).getGenericPackageType());

            if (index != GenericReqEvInvalidEvent)
            {
                reqEv = myGenericReqEv[index];
            }
        }
        break;

        case MgcpLinePackageType:
        {
            LineReqEv index = getLineReqEvIndex(
                                  (event->getEvent()).getLinePackageType());

            if (index != LineReqEvInvalidEvent)
            {
                reqEv = myLineReqEv[index];
            }
        }
        break;

        case MgcpDTMFPackageType:
        {
            DtmfReqEv index = getDtmfReqEvIndex(
                                  (event->getEvent()).getDTMFPackageType());

            if (index != DtmfReqEvInvalidEvent)
            {
                reqEv = myDtmfReqEv[index];
            }

            // check for special case of wildcard (aka any digit 0-9)
            if (index == DtmfReqEvAnyDigit)
            {
                // request all digits 0-9
                myDtmfReqEv[DtmfReqEv0]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv0]);
                myDtmfReqEv[DtmfReqEv1]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv1]);
                myDtmfReqEv[DtmfReqEv2]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv2]);
                myDtmfReqEv[DtmfReqEv3]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv3]);
                myDtmfReqEv[DtmfReqEv4]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv4]);
                myDtmfReqEv[DtmfReqEv5]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv5]);
                myDtmfReqEv[DtmfReqEv6]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv6]);
                myDtmfReqEv[DtmfReqEv7]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv7]);
                myDtmfReqEv[DtmfReqEv8]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv8]);
                myDtmfReqEv[DtmfReqEv9]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEv9]);
                myDtmfReqEv[DtmfReqEvHash]->start(requestId, event);
                activeReqEvs->push_back(myDtmfReqEv[DtmfReqEvHash]);
            }
        }
        break;

        case MgcpMFPackageType:
        case MgcpTrunkPackageType:
        case MgcpHandsetPackageType:
        case MgcpRTPPackageType:
        case MgcpNetworkAccessServerPackageType:
        case MgcpAnnouncementPackageType:
        case MgcpScriptPackageType:
        case MgcpXVoiceMailPackageType:
        default:
        {
            cpLog(LOG_ERR, "Unsupported req event package = %d",
                  event->getPackageType());
        }
        break;
    }   

    if (reqEv != 0)
    {
        // mark event as requested and store associated paramters
        reqEv->start(requestId, event);

        // add to list of currently requested events
        activeReqEvs->push_back(reqEv);
    }

}


void
Endpoint::activateSignal (const MgcpGenericEvent& event,
                          const RQNT rqnt)
{
    ReqSig* signal = 0;

    cpLog(LOG_DEBUG, "Endpoint::activateSignal");
    switch (event.getPackageType())
    {
        case MgcpGenericMediaPackageType:
        {
            GenericReqSig index =
                getGenericReqSigIndex(event.getGenericPackageType());
            if (index != GenericReqSigInvalidSignal)
            {
                signal = myGenericReqSig[index];
            }
        }
        break;

        case MgcpLinePackageType:
        {
            LineReqSig index = getLineReqSigIndex(event.getLinePackageType());
            if (index != LineReqSigInvalidSignal)
            {
                signal = myLineReqSig[index];
            }
        }
        break;

        case MgcpDTMFPackageType:
        {
            DtmfReqSig index = getDtmfReqSigIndex(event.getDTMFPackageType());
            if (index != DtmfReqSigInvalidSignal)
            {
                signal = myDtmfReqSig[index];
            }
        }
        break;

        case MgcpMFPackageType:
        case MgcpTrunkPackageType:
        case MgcpHandsetPackageType:
        case MgcpRTPPackageType:
        case MgcpNetworkAccessServerPackageType:
        case MgcpAnnouncementPackageType:
        case MgcpScriptPackageType:
        case MgcpXVoiceMailPackageType:
        default:
        {
            cpLog(LOG_ERR, "Unsupported req signal package = %d",
                  event.getPackageType());
        }
        break;
    } 

    if (signal != 0)
    {
        // mark signal as active and configure hardware
        signal->start(rqnt);

        // add to list of active signals
        activeReqSigs->push_back(signal);
    }

} 


GenericReqEv
Endpoint::getGenericReqEvIndex (MgcpGenericPackageItemType type)
{
    GenericReqEv index = GenericReqEvInvalidEvent;

    cpLog(LOG_DEBUG, "Endpoint::getGenericReqEvIndex");
    switch (type)
    {
        case MgcGenericPackageModemDetect:
        {
            index = GenericReqEvModemDetected;
        }
        break;

        case MgcGenericPackageFaxToneDetect:
        {
            index = GenericReqEvFaxToneDetected;
        }
        break;

        case MgcGenericPackageLongDurationConnection:
        {
            index = GenericReqEvLongDurationConnection;
        }
        break;

        case MgcGenericPackagePatternDetected:
        {
            index = GenericReqEvPatternDetected;
        }
        break;

        case MgcGenericPackageReportFailure:
        {
            index = GenericReqEvReportFailure;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown generic req event = %d", type);
        }
    } 

    return index;

} 


GenericReqSig
Endpoint::getGenericReqSigIndex (MgcpGenericPackageItemType type)
{
    GenericReqSig index = GenericReqSigInvalidSignal;

    cpLog(LOG_DEBUG, "Endpoint::getGenericSignalIndex");
    switch (type)
    {
        case MgcGenericPackagePatternDetected:
        {
            index = GenericReqSigPatternDetected;
        }
        break;

        case MgcGenericPackageRingBackTone:
        {
            index = GenericReqSigRingbackTone;
        }
        break;

        case MgcGenericPackageRingBackOnConnection:
        {
            index = GenericReqSigRingbackOnConnection;
        }
        break;

        case MgcGenericPackageConfirmTone:
        {
            index = GenericReqSigConfirmTone;
        }
        break;

        case MgcGenericPackageNetworkCongestionTone:
        {
            index = GenericReqSigNetworkCongestionTone;
        }
        break;

        case MgcGenericPackageInterceptTone:
        {
            index = GenericReqSigInterceptTone;
        }
        break;

        case MgcGenericPackagePreemtionTone:
        {
            index = GenericReqSigPreemptionTone;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown generic req signal = %d", type);
        }
    } 

    return index;

} 


LineReqEv
Endpoint::getLineReqEvIndex (MgcpLinePackageItemType type)
{
    LineReqEv index = LineReqEvInvalidEvent;

    cpLog(LOG_DEBUG, "Endpoint::getLineReqEvIndex");
    switch (type)
    {
        case MgcLinePackageOffHookTransition:
        {
            index = LineReqEvOffhook;
        }
        break;

        case MgcLinePackageOnHookTransition:
        {
            index = LineReqEvOnhook;
        }
        break;

        case MgcLinePackageFlashHook:
        {
            index = LineReqEvFlashhook;
        }
        break;

        case MgcLinePackageAnswerTone:
        {
            index = LineReqEvAnswerTone;
        }
        break;

        case MgcLinePackageNetworkBusy:
        {
            index = LineReqEvNetworkBusy;
        }
        break;

        case MgcLinePackagePromptTone:
        {
            index = LineReqEvPromptTone;
        }
        break;

        case MgcLinePackageErrorTone:
        {
            index = LineReqEvErrorTone;
        }
        break;

        case MgcLinePackageReportOnCompletion:
        {
            index = LineReqEvRptOnComplete;
        }
        break;

        case MgcLinePackageDistinctiveTonePattern:
        {
            index = LineReqEvDsnctTonePattern;
        }
        break;

        case MgcLinePackageReportFailure:
        {
            index = LineReqEvFailureRpt;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown line req event = %d", type);
        }
    } 

    return index;

} 


LineReqSig
Endpoint::getLineReqSigIndex (MgcpLinePackageItemType type)
{
    LineReqSig index = LineReqSigInvalidSignal;

    cpLog(LOG_DEBUG, "Endpoint::getLineReqSigIndex");
    switch (type)
    {
        case MgcLinePackageAdsiDisplay:
        {
            index = LineReqSigAdsiDisplay;
        }
        break;

        case MgcLinePackageVisualMessageWaitingIndicator:
        {
            index = LineReqSigVisualMsg;
        }
        break;

        case MgcLinePackageAnswerTone:
        {
            index = LineReqSigAnswerTone;
        }
        break;

        case MgcLinePackageBusyTone:
        {
            index = LineReqSigBusyTone;
        }
        break;

        case MgcLinePackageCallerId:
        {
            index = LineReqSigCallerId;
        }
        break;

        case MgcLinePackageCallWaitingTone:
        {
            index = LineReqSigCallWaitTone;
        }
        break;

        case MgcLinePackageDialTone:
        {
            index = LineReqSigDialTone;
        }
        break;

        case MgcLinePackageMessageWaitingInd:
        {
            index = LineReqSigMsgWaitInd;
        }
        break;

        case MgcLinePackageNetworkBusy:
        {
            index = LineReqSigNetworkBusy;
        }
        break;

        case MgcLinePackageRinging:
        {
            index = LineReqSigRinging;
        }
        break;

        case MgcLinePackageDistinctiveRinging:
        {
            index = LineReqSigDsnctRing0;
        }
        break;

        case MgcLinePackageRingsplash:
        {
            index = LineReqSigRingSplash;
        }
        break;

        case MgcLinePackagePromptTone:
        {
            index = LineReqSigPromptTone;
        }
        break;

        case MgcLinePackageErrorTone:
        {
            index = LineReqSigErrorTone;
        }
        break;

        case MgcLinePackageStutterDialtone:
        {
            index = LineReqSigStutterTone;
        }
        break;

        case MgcLinePackageAlertingTone:
        {
            index = LineReqSigAlertTone;
        }
        break;

        case MgcLinePackageRecorderWarningTone:
        {
            index = LineReqSigRecordWarnTone;
        }
        break;

        case MgcLinePackageCallingCardServiceTone:
        {
            index = LineReqSigCallCardSrvTone;
        }
        break;

        case MgcLinePackageOffHookWarningTone:
        {
            index = LineReqSigOffhookWarnTone;
        }
        break;

        case MgcLinePackageDistinctiveTonePattern:
        {
            index = LineReqSigDsnctTonePattern;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown line req signal = %d", type);
        }
    } 

    return index;

} 


DtmfReqEv
Endpoint::getDtmfReqEvIndex (MgcpDTMFPackageItemType type)
{
    DtmfReqEv index = DtmfReqEvInvalidEvent;

    cpLog(LOG_DEBUG, "Endpoint::getDtmfReqEvIndex");
    switch (type)
    {
        case MgcDTMFPackageDTMF_0:
        {
            index = DtmfReqEv0;
        }
        break;

        case MgcDTMFPackageDTMF_1:
        {
            index = DtmfReqEv1;
        }
        break;

        case MgcDTMFPackageDTMF_2:
        {
            index = DtmfReqEv2;
        }
        break;

        case MgcDTMFPackageDTMF_3:
        {
            index = DtmfReqEv3;
        }
        break;

        case MgcDTMFPackageDTMF_4:
        {
            index = DtmfReqEv4;
        }
        break;

        case MgcDTMFPackageDTMF_5:
        {
            index = DtmfReqEv5;
        }
        break;

        case MgcDTMFPackageDTMF_6:
        {
            index = DtmfReqEv6;
        }
        break;

        case MgcDTMFPackageDTMF_7:
        {
            index = DtmfReqEv7;
        }
        break;

        case MgcDTMFPackageDTMF_8:
        {
            index = DtmfReqEv8;
        }
        break;

        case MgcDTMFPackageDTMF_9:
        {
            index = DtmfReqEv9;
        }
        break;

        case MgcDTMFPackageDTMF_Hash:
        {
            index = DtmfReqEvHash;
        }
        break;

        case MgcDTMFPackageDTMF_Star:
        {
            index = DtmfReqEvStar;
        }
        break;

        case MgcDTMFPackageDTMF_A:
        {
            index = DtmfReqEvA;
        }
        break;

        case MgcDTMFPackageDTMF_B:
        {
            index = DtmfReqEvB;
        }
        break;

        case MgcDTMFPackageDTMF_C:
        {
            index = DtmfReqEvC;
        }
        break;

        case MgcDTMFPackageDTMF_D:
        {
            index = DtmfReqEvD;
        }
        break;

        case MgcDTMFPackageLongDurationIndicator:
        {
            index = DtmfReqEvLongDurInd;
        }
        break;

        case MgcDTMFPackageWildcardMatchAnyDigit_0_9:
        {
            index = DtmfReqEvAnyDigit;
        }
        break;

        case MgcDTMFPackageInterdigitTimer:
        {
            index = DtmfReqEvInterdigitTmr;
        }
        break;

        case MgcDTMFPackageReportFailure:
        {
            index = DtmfReqEvRptFailure;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown dtmf req event = %d", type);
        }
    } 

    return index;

}


DtmfReqSig
Endpoint::getDtmfReqSigIndex (MgcpDTMFPackageItemType type)
{
    DtmfReqSig index = DtmfReqSigInvalidSignal;

    cpLog(LOG_DEBUG, "Endpoint::getDtmfReqSigIndex");
    switch (type)
    {
        case MgcDTMFPackageDTMF_0:
        {
            index = DtmfReqSig0;
        }
        break;

        case MgcDTMFPackageDTMF_1:
        {
            index = DtmfReqSig1;
        }
        break;

        case MgcDTMFPackageDTMF_2:
        {
            index = DtmfReqSig2;
        }
        break;

        case MgcDTMFPackageDTMF_3:
        {
            index = DtmfReqSig3;
        }
        break;

        case MgcDTMFPackageDTMF_4:
        {
            index = DtmfReqSig4;
        }
        break;

        case MgcDTMFPackageDTMF_5:
        {
            index = DtmfReqSig5;
        }
        break;

        case MgcDTMFPackageDTMF_6:
        {
            index = DtmfReqSig6;
        }
        break;

        case MgcDTMFPackageDTMF_7:
        {
            index = DtmfReqSig7;
        }
        break;

        case MgcDTMFPackageDTMF_8:
        {
            index = DtmfReqSig8;
        }
        break;

        case MgcDTMFPackageDTMF_9:
        {
            index = DtmfReqSig9;
        }
        break;

        case MgcDTMFPackageDTMF_Hash:
        {
            index = DtmfReqSigHash;
        }
        break;

        case MgcDTMFPackageDTMF_Star:
        {
            index = DtmfReqSigStar;
        }
        break;

        case MgcDTMFPackageDTMF_A:
        {
            index = DtmfReqSigA;
        }
        break;

        case MgcDTMFPackageDTMF_B:
        {
            index = DtmfReqSigB;
        }
        break;

        case MgcDTMFPackageDTMF_C:
        {
            index = DtmfReqSigC;
        }
        break;

        case MgcDTMFPackageDTMF_D:
        {
            index = DtmfReqSigD;
        }
        break;

        default:
        {
            cpLog(LOG_ERR, "Unknown dtmf req event = %d", type);
        }
    }

    return index;

} 


void
Endpoint::initializeGenericEvents (void)
{
    // create all requested events
    myGenericReqEv[GenericReqEvModemDetected] =
        new ReqEv(MgGenericPackage::notifyModemDetect());
    myGenericReqEv[GenericReqEvFaxToneDetected] =
        new ReqEv(MgGenericPackage::notifyFaxToneDetect());
    myGenericReqEv[GenericReqEvLongDurationConnection] =
        new ReqEv(MgGenericPackage::notifyLongDurationConnection());
    // TBD, how do we handle initialization with unknown parameter?
    myGenericReqEv[GenericReqEvPatternDetected] =
        // TBD need to fake until API works for parameterized observed events
        //new ReqEv(MgGenericPackage::notifyPatternDetected(*pattern));
        new ReqEv(MgGenericPackage::notifyReportFailure());
    myGenericReqEv[GenericReqEvReportFailure] =
        new ReqEv(MgGenericPackage::notifyReportFailure());

    // check that all events were created successfully
    for (int i = 0 ; i < GenericReqEvMaxEvent; i++)
    {
        assert(myGenericReqEv[i] != NULL);
    }

}


void
Endpoint::initializeLineEvents (void)
{
    // initialize line signals
    myLineReqEv[LineReqEvOffhook] =
        new ReqEv(MgLinePackage::notifyOffHookTransition());
    myLineReqEv[LineReqEvOnhook] =
        new ReqEv(MgLinePackage::notifyOnHookTransition());
    myLineReqEv[LineReqEvFlashhook] =
        new ReqEv(MgLinePackage::notifyFlashHook());
    myLineReqEv[LineReqEvAnswerTone] =
        new ReqEv(MgLinePackage::notifyAnswerTone());
    myLineReqEv[LineReqEvNetworkBusy] =
        new ReqEv(MgLinePackage::notifyNetworkBusy());
    myLineReqEv[LineReqEvPromptTone] =
        new ReqEv(MgLinePackage::notifyPromptTone());
    myLineReqEv[LineReqEvErrorTone] =
        new ReqEv(MgLinePackage::notifyErrorTone());
    myLineReqEv[LineReqEvRptOnComplete] =
        new ReqEv(MgLinePackage::notifyReportOnCompletion());
    //myLineReqEv[LineReqEvDsnctTonePattern] =
    //new ReqEv(MgLinePackage::notifyDistinctiveTonePattern());
    myLineReqEv[LineReqEvFailureRpt] =
        new ReqEv(MgLinePackage::notifyReportFailure());

} 


void
Endpoint::initializeDtmfEvents (void)
{
    // create all requested events
    myDtmfReqEv[DtmfReqEv0] = new ReqEv(MgDTMFPackage::notifyDTMF_0());
    myDtmfReqEv[DtmfReqEv1] = new ReqEv(MgDTMFPackage::notifyDTMF_1());
    myDtmfReqEv[DtmfReqEv2] = new ReqEv(MgDTMFPackage::notifyDTMF_2());
    myDtmfReqEv[DtmfReqEv3] = new ReqEv(MgDTMFPackage::notifyDTMF_3());
    myDtmfReqEv[DtmfReqEv4] = new ReqEv(MgDTMFPackage::notifyDTMF_4());
    myDtmfReqEv[DtmfReqEv5] = new ReqEv(MgDTMFPackage::notifyDTMF_5());
    myDtmfReqEv[DtmfReqEv6] = new ReqEv(MgDTMFPackage::notifyDTMF_6());
    myDtmfReqEv[DtmfReqEv7] = new ReqEv(MgDTMFPackage::notifyDTMF_7());
    myDtmfReqEv[DtmfReqEv8] = new ReqEv(MgDTMFPackage::notifyDTMF_8());
    myDtmfReqEv[DtmfReqEv9] = new ReqEv(MgDTMFPackage::notifyDTMF_9());
    myDtmfReqEv[DtmfReqEvHash] = new ReqEv(MgDTMFPackage::notifyDTMF_Hash());
    myDtmfReqEv[DtmfReqEvStar] = new ReqEv(MgDTMFPackage::notifyDTMF_Star());
    myDtmfReqEv[DtmfReqEvA] = new ReqEv(MgDTMFPackage::notifyDTMF_A());
    myDtmfReqEv[DtmfReqEvB] = new ReqEv(MgDTMFPackage::notifyDTMF_B());
    myDtmfReqEv[DtmfReqEvC] = new ReqEv(MgDTMFPackage::notifyDTMF_C());
    myDtmfReqEv[DtmfReqEvD] = new ReqEv(MgDTMFPackage::notifyDTMF_D());
    myDtmfReqEv[DtmfReqEvLongDurInd] =
        new ReqEv(MgDTMFPackage::notifyLongDurationIndicator());
    myDtmfReqEv[DtmfReqEvAnyDigit] =
        new ReqEv(MgDTMFPackage::notifyWildcardMatch());
    myDtmfReqEv[DtmfReqEvInterdigitTmr] =
        new ReqEv(MgDTMFPackage::notifyInterdigitTimer());
    myDtmfReqEv[DtmfReqEvRptFailure] =
        new ReqEv(MgDTMFPackage::notifyReportFailure());

    // check that all events were created successfully
    for (int i = 0 ; i < DtmfReqEvMaxEvent; i++)
    {
        assert(myDtmfReqEv[i] != NULL);
    }

}


void
Endpoint::initializeGenericSignals (Hardware& hw)
{
    // initialize signals
    myGenericReqSig[GenericReqSigPatternDetected] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigPatternDetected, SignalOnOff);
    myGenericReqSig[GenericReqSigRingbackTone] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigRingbackTone, SignalTimeout);
    myGenericReqSig[GenericReqSigRingbackOnConnection] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigRingbackOnConnection, SignalTimeout, 180);
    myGenericReqSig[GenericReqSigConfirmTone] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigConfirmTone, SignalBrief);
    myGenericReqSig[GenericReqSigNetworkCongestionTone] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigNetworkCongestionTone, SignalTimeout);
    myGenericReqSig[GenericReqSigInterceptTone] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigInterceptTone, SignalOnOff);
    myGenericReqSig[GenericReqSigPreemptionTone] =
        new ReqSig(hw, myId, MgcpGenericMediaPackageType,
                   GenericReqSigPreemptionTone, SignalOnOff);

    // check that all events were created successfully
    for (int i = 0 ; i < GenericReqSigMaxSignal; i++)
    {
        assert(myGenericReqSig[i] != NULL);
    }

} 


void
Endpoint::initializeLineSignals (Hardware& hw)
{
    // initialize signals
    myLineReqSig[LineReqSigAdsiDisplay] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigAdsiDisplay, SignalBrief);
    myLineReqSig[LineReqSigVisualMsg] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigVisualMsg, SignalTimeout);
    myLineReqSig[LineReqSigAnswerTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigAnswerTone, SignalOnOff);
    myLineReqSig[LineReqSigBusyTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigBusyTone, SignalTimeout, 30);
    myLineReqSig[LineReqSigCallerId] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigCallerId, SignalBrief);
    myLineReqSig[LineReqSigCallWaitTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigCallWaitTone, SignalTimeout, 30);
    myLineReqSig[LineReqSigDialTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDialTone, SignalTimeout, 16);
    myLineReqSig[LineReqSigMsgWaitInd] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigMsgWaitInd, SignalTimeout, 16);
    myLineReqSig[LineReqSigNetworkBusy] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigNetworkBusy, SignalOnOff);
    myLineReqSig[LineReqSigRinging] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigRinging, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing0] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing0, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing1] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing1, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing2] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing2, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing3] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing3, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing4] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing4, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing5] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing5, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing6] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing6, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctRing7] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctRing7, SignalTimeout, 180);
    myLineReqSig[LineReqSigRingSplash] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigRingSplash, SignalBrief);
    myLineReqSig[LineReqSigPromptTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigPromptTone, SignalBrief);
    myLineReqSig[LineReqSigErrorTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigErrorTone, SignalBrief);
    myLineReqSig[LineReqSigStutterTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigStutterTone, SignalTimeout, 16);
    myLineReqSig[LineReqSigAlertTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigAlertTone, SignalOnOff);
    myLineReqSig[LineReqSigRecordWarnTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigRecordWarnTone, SignalOnOff);
    myLineReqSig[LineReqSigCallCardSrvTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigCallCardSrvTone, SignalOnOff);
    myLineReqSig[LineReqSigOffhookWarnTone] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigOffhookWarnTone, SignalTimeout, 180);
    myLineReqSig[LineReqSigDsnctTonePattern] =
        new ReqSig(hw, myId, MgcpLinePackageType,
                   LineReqSigDsnctTonePattern, SignalBrief);
    // check that all events were created successfully
    for (int i = 0 ; i < LineReqSigMaxSignal; i++)
    {
        assert(myLineReqSig[i] != NULL);
    }

}


void
Endpoint::initializeDtmfSignals (Hardware& hw)
{
    // initialize signals
    myDtmfReqSig[DtmfReqSig0] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig0, SignalBrief);
    myDtmfReqSig[DtmfReqSig1] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig1, SignalBrief);
    myDtmfReqSig[DtmfReqSig2] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig2, SignalBrief);
    myDtmfReqSig[DtmfReqSig3] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig3, SignalBrief);
    myDtmfReqSig[DtmfReqSig4] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig4, SignalBrief);
    myDtmfReqSig[DtmfReqSig5] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig5, SignalBrief);
    myDtmfReqSig[DtmfReqSig6] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig6, SignalBrief);
    myDtmfReqSig[DtmfReqSig7] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig7, SignalBrief);
    myDtmfReqSig[DtmfReqSig8] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig8, SignalBrief);
    myDtmfReqSig[DtmfReqSig9] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSig9, SignalBrief);
    myDtmfReqSig[DtmfReqSigHash] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                   DtmfReqSigHash, SignalBrief);
    myDtmfReqSig[DtmfReqSigStar] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                   DtmfReqSigStar, SignalBrief);
    myDtmfReqSig[DtmfReqSigA] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSigA, SignalBrief);
    myDtmfReqSig[DtmfReqSigB] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSigB, SignalBrief);
    myDtmfReqSig[DtmfReqSigC] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSigC, SignalBrief);
    myDtmfReqSig[DtmfReqSigD] = new ReqSig(hw, myId, MgcpDTMFPackageType,
                                           DtmfReqSigD, SignalBrief);

    // check that all events were created successfully
    for (int i = 0 ; i < DtmfReqSigMaxSignal; i++)
    {
        assert(myDtmfReqSig[i] != NULL);
    }

}


void
Endpoint::startTimer (EndpointTimeout timer, int msec)
{
    EndpointMsg msg;

    // stop timer if already running
    if (myTimerEventId[timer])
    {
        stopTimer(timer);
    }

    // send message to start timer
    msg.type = EndpointTimeoutType;
    msg.timeout = timer;
    myTimerEventId[timer] = myQ->addDelayMs(msg, msec);
    cpLog(LOG_DEBUG, "Start timer %d as event %d",
          timer, myTimerEventId[timer] );
} 


void
Endpoint::stopTimer (EndpointTimeout timer)
{
    // stop timer if running
    if (myTimerEventId[timer] != (FifoEventId) NULL)
    {
        cpLog(LOG_DEBUG, "Stop timer %d as event %d",
              timer, myTimerEventId[timer] );
        myQ->cancel( myTimerEventId[timer] );
        myTimerEventId[timer] = (FifoEventId) NULL;
    }
} 
