#ifndef ENDPOINT_H_
#define ENDPOINT_H_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const Endpoint_h_Version =
    "$Id: Endpoint.h,v 1.28 2001/01/23 22:22:34 cktam Exp $";

#include <iostream>
#include <list>
#include "common.h"
#include "ReqEv.h"
#include "ReqSig.h"

class MgAgent;
class Connection;
class Hardware;
class DigitMap;

// timer durations in terms of milliseconds
// #define FLASHHOOK_TIMER_DURATION 1000
#define FLASHHOOK_TIMER_DURATION 2000

// Debounce time (in microseconds)
#define DEBOUNCE_DELAY 120000

// absolute maximum number of connections any endpoint can support
#define ENDPOINT_CONNECTION_MAX 3

/** Enumeration to track hookstate. HookstateFlashhook is a transient
 ** state while running a timer to dinstinquish between between
 ** flashhook and onhook.
 */
enum Hookstate
{
    HookstateOffhook,
    HookstateOnhook,
    HookstateFlashhook
};

/** Base class for all endpoints.  Specifies the MGCP endpoint API.
 ** Defines a basic level of service provided by all MGCP endpoints.
 */
class Endpoint
{
    public:

        /** Creates a new endpoint with the specified id which uses the
         ** provided message queues for asynchronous communication.
         */
        Endpoint ( string endpointId,
                   int numConnections,
                   MgAgent* agent,
                   Hardware& hardware,
                   Fifo < EndpointMsg > * inputQ,
                   Fifo < HardwareMsg > * outputQ );

        /** Deletes a endpoint, including deallocating all heap memory
         ** allocated by the endpoint.
         */
        ~Endpoint (void);

        /** Starts the main processing loop of the endpoint
         */
        void*
        endpointMain (void* parms);

        void
        restartInProgress (const string& restartMethod,
                           const string& restartDelay = "none");

    private:
        /**
         */
        void
        initializeGenericEvents (void);

        void
        initializeLineEvents (void);

        void
        initializeDtmfEvents (void);

        /**
         */
        void
        initializeGenericSignals (Hardware& hardware);

        void
        initializeLineSignals (Hardware& hardware);

        void
        initializeDtmfSignals (Hardware& hardware);

        /** Processes an MGCP command sent by the call agent. This command
         ** has been previously determined to be valid and for this
         ** endpoint.
         */
        void
        processAgentCommand (EndpointAgentCommand& cmd);

        /** Processes an event sent by the hardware.
         */
        void
        processHardwareEvent (HardwareEvent event);

        /** Processes a timeout
         */
        void
        processTimeout (EndpointTimeout timeout);

        void
        processEPCF (EPCF epcf);

        void
        processCRCX (CRCX crcx);

        void
        processMDCX (MDCX mdcx);

        void
        processDLCX (DLCX dlcx);

        void
        processRQNT (RQNT rqnt);

        void
        processAUEP (AUEP auep);

        void
        processAUCX (AUCX aucx);

        void
        addReqEv (string requestId, Sptr < MgRequestedEvent > event);

        void
        activateSignal (const MgcpGenericEvent& event,
                        const RQNT rqnt);

        void
        clearAllReqEvs (void);

        void
        clearAllSignals (void);

        /** If the parameter is a known MgcpXXXPackageItemType
         ** requestable event, returns its location in the myXXXReqEv array,
         ** otherwise returns XXXReqEvInvalidEvent.  */
        GenericReqEv
        getGenericReqEvIndex (MgcpGenericPackageItemType type);

        LineReqEv
        getLineReqEvIndex (MgcpLinePackageItemType type);

        DtmfReqEv
        getDtmfReqEvIndex (MgcpDTMFPackageItemType type);

        /** If the parameter is a known MgcpXXXPackageItemType
         ** requestable signal, returns its location in the myXXXReqSig
         ** array, otherwise returns XXXReqSigInvalidSignal.  */
        GenericReqSig
        getGenericReqSigIndex (MgcpGenericPackageItemType type);

        LineReqSig
        getLineReqSigIndex (MgcpLinePackageItemType type);

        DtmfReqSig
        getDtmfReqSigIndex (MgcpDTMFPackageItemType type);

        /** Start a timer that expires in specified number of
         ** milliseconds. The expiration of the timer results in reception
         ** of EndpointMsg in Q. If timer is already running, stop it, then
         ** restart it for the specified number of milliseconds.
         */
        void
        startTimer (EndpointTimeout timer, int msec);

        /** Stop specified timer (if running).
         */
        void
        stopTimer (EndpointTimeout timer);

        // package events
        ReqEv* myGenericReqEv[GenericReqEvMaxEvent];
        ReqEv* myLineReqEv[LineReqEvMaxEvent];
        ReqEv* myDtmfReqEv[DtmfReqEvMaxEvent];

        // package signals
        ReqSig* myGenericReqSig[GenericReqSigMaxSignal];
        ReqSig* myLineReqSig[LineReqSigMaxSignal];
        ReqSig* myDtmfReqSig[DtmfReqSigMaxSignal];

        /// endpoint id supplied by gateway
        string myId;

        /// maximum number of number of connections
        int myMaxConnections;

        /// call agent endpoint communicates with
        MgAgent* myAgent;

        /// A-law or mu-law
        MgcpBearerInformation bearerInfo;

        /// input queue of messages requiring processing
        Fifo < EndpointMsg > * myQ;

        /// output queue to send requests to control hardware signals
        Fifo < HardwareMsg > * hardwareQ;

        /// list of the ReqEvs corresponding to last call agent request
        list < ReqEv* > * activeReqEvs;

        /// list of the ReqSigs corresponding to the last call agent request
        list < ReqSig* > * activeReqSigs;

        /// connection(s) associated with endpoint
        Connection* myConnections[ENDPOINT_CONNECTION_MAX];

        /// store if endpoint is onhook, offhook, timing for flashhook, etc.
        Hookstate myHookstate;

        /// store running timers so they can be cancelled (if necessary)
        FifoEventId myTimerEventId[TimeoutMaxTimer];

        /// For digit dialing
        DigitMap* myDigits;

        string myEndpointId;
        struct timeval startDebounceTime;
        struct timeval finishDebounceTime;
        long diffTime;
};

#endif // ENDPOINT_H_
