#ifndef HARDWARE_H_
#define HARDWARE_H_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const Hardware_h_Version =
    "$Id: Hardware.h,v 1.16 2001/01/23 22:22:34 cktam Exp $";

#include "common.h"
#include "RtpSession.hxx"
#include "Gateway.h"

#define HW_ENDPOINT_MAX  1

enum HardwareCodecType
{
    CodecTypeULaw,
    CodecTypeALaw,
    CodecTypeLinear16,
    CodecTypeG729A
};



struct ChannelData
{
    RtpSession* audioStack;
    RtpPacket* inRtpPkt;
    RtpPacket* outRtpPkt;

    bool hookStateOffhook;
    bool audioActive;       // true if RTP audio packets are being processed
    bool audioHalfActive;   // true if only sending RTP audio packets
    bool hasPlayed;
    HardwareCodecType myCodecType;
};



/** Abstract base class for all hardware.  Specifies the MGCP gateway
 ** hardware API.  Defines a basic level of service provided by all
 ** supported hardware interfaces.
 **/
class Hardware
{
    public:

        ///
        Hardware (void)
        {};

        /** Creates a hardware object to control a single port gateway.
         */
        Hardware ( char* deviceName,
                   Fifo < HardwareMsg > * inputQ,
                   Fifo < EndpointMsg > * outputQ )
        {};

        /** Creates a hardware object to control a multi port gateway.
         */
        Hardware (char* deviceName, int numEndpoints,
                  Fifo < HardwareMsg > * inputQ,
                  Fifo < EndpointMsg > * outputQ[] )
        {};



        /** Destructor
         */
        virtual
        ~Hardware (void)
        {};

        /** Starts the main processing loop of the hardware
         */
        virtual void*
        hardwareMain (void* parms) = 0;

        /** Add device's socket to the fd; returns 0 if successful,
         ** errorcode otherwise 
         */
        virtual int
        addToFdSet (fd_set* fd) = 0;

        /** Process all pending hardware events; returns 0 if successful,
         ** otherwise returns an errorcode.  
         */
        virtual int
        process (fd_set* fd) = 0;

        /** Start audio channel on device; returns 0 if successful,
         ** errorcode otherwise 
         */
        virtual int
        audioStart (const HardwareAudioRequest& request, const int channelNum) = 0;

        /** Stops audio channel on device; returns 0 if successful,
         ** errorcode otherwise.
         */
        virtual int
        audioStop (const int channelNum) = 0;

        /** Returns the HardwareSignal supported by the hardware which
         ** corresponds to turning on the mgcp encoded signal.
         */
        virtual HardwareSignal
        getSignalOn (MgcpPackageType package, int index) const = 0;

        /** Returns the HardwareSignal supported by the hardware which
         ** corresponds to turning off the mgcp encoded signal.
         */
        virtual HardwareSignal
        getSignalOff (MgcpPackageType package, int index) const = 0;

        /** access functions to private data
         */
        inline Fifo < HardwareMsg > *
        getQ (void) const
        {
            return myQ;
        };

        /** returns the hookstate
         */
        inline bool
        isOffhook (void) const
        {
            return hookStateOffhook;
        };


        void
        setEndpointIndex(string id[], int numEndpoints);


        int
        getEndpointIndex(string id);



    protected:
        /** Trigger the device to send the specified signal; returns 0 if
         ** successful, errorcode otherwise.
         */
        virtual int
        sendSignal (HardwareSignal signal, int endpointIndex) = 0;

        /** Process a msg received from an endpoint serves by the hardware
         */
        void
        processEndpointMsg (const HardwareMsg& msg);

        /** Report hardware event by adding it to specified queue
         */
        void
        reportEvent (Fifo < EndpointMsg > * outputQ, const HardwareEvent& event);


        /// file descriptor for hardware device
        int myFD;

        // input queue to allow endpoints to control hardware signals
        Fifo < HardwareMsg > * myQ;

        // true if last reported hook state was offhook
        bool hookStateOffhook;


        int myNumEndpoints;

        string myEndpointId[GW_ENDPOINT_MAX];

        ChannelData myChannel[GW_ENDPOINT_MAX];

};

#endif // HARDWARE_H_
