
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSDecisionMsg_cxx_Version =
    "$Id: COPSDecisionMsg.cxx,v 1.10 2000/12/18 23:42:53 bko Exp $";
#include <assert.h>

#include "COPSHandle.hxx"
#include "COPSContext.hxx"
#include "COPSDecision.hxx"
#include "COPSClientSI.hxx"
#include "COPSIntegrity.hxx"
#include "COPSDecisionMsg.hxx"
#include "cpLog.h"


COPSDecisionMsg::COPSDecisionMsg()
        : _clientHandle(0),
        _error(0),
        _integrity(0),
        _decContext(0)
{}


COPSDecisionMsg::COPSDecisionMsg(const char* data)
        : _clientHandle(0),
        _error(0),
        _integrity(0),
        _decContext(0)
{
    parse(data);
}

void
COPSDecisionMsg::parse(const COPSHeader& hdr, const char* data) throw (COPSBadDataException&)
{
    _hdr = new COPSHeader(hdr);
    parse(data);
    setMsgLength();
}

COPSDecisionMsg::~COPSDecisionMsg()
{}


void
COPSDecisionMsg::parse(const char* data) throw (COPSBadDataException&)
{
    COPSMsg::parseHeader(data);
    while (_dataStart < _dataLength)
    {
        COPSObjHeader objHdr;
        objHdr.parse(data + _dataStart);
        switch (objHdr.getCNum())
        {
            case COPS_HANDLE:
            {
                _clientHandle = new COPSHandle(data + _dataStart);
                _dataStart += _clientHandle->getDataLength();
            }
            break;
            case COPS_CONTEXT:
            {
                //dec context
                _decContext = new COPSContext(data + _dataStart);
                _dataStart += _decContext->getDataLength();
            }
            break;
            case COPS_ERROR:
            {
                _error = new COPSError(data + _dataStart);
                _dataStart += _error->getDataLength();
            }
            break;
            case COPS_DEC:
            {
                COPSDecision decs(data + _dataStart);
                _dataStart += decs.getDataLength();
                addDecision(decs, *_decContext);
            }
            break;
            case COPS_MSG_INTEGRITY:
            {
                _integrity = new COPSIntegrity(data + _dataStart);
                _dataStart += _integrity->getDataLength();
            }
            break;
            default:
            {
                cpLog(LOG_ALERT, "Bad Message format, unknown object type");
                throw COPSBadDataException("Bad Message format", __FILE__, __LINE__, 0);
            }
            break;
        }
    }
    checkSanity();
}

void
COPSDecisionMsg::checkSanity() const throw (COPSBadDataException&)
{
    if (!_hdr || !_clientHandle || ( !_error && !_decisions.size()))
    {
        cpLog(LOG_ALERT, "Bad Message format");
        throw COPSBadDataException("Bad message format",
                                   __FILE__, __LINE__, 0);
    }
}

void
COPSDecisionMsg::dump(ostream& os)const
{
    if (_hdr.getPtr()) os << *_hdr << endl;
    if (_clientHandle.getPtr()) os << *_clientHandle << endl;

    if (_error.getPtr()) os << *_error << endl;

    //Display any local decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        const COPSContext& cntxt = (*itr).first;
        const vector < COPSDecision > & decVec = (*itr).second;
        os << cntxt << endl;
        for (vector < COPSDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            if (cntxt.getRequestType() == COPSContext::CONFIG)
            {
                (*vItr).dumpPrObject(os);
            }
            else
            {
                os << (*vItr) << endl;
            }
        }
    }

    if (_integrity.getPtr()) os << *_integrity << endl;
}

COPSDecisionMsg&
COPSDecisionMsg::operator<<(const COPSError& obj)
{
    assert(!_decisions.size());
    assert(!_error);
    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);
    _error = new COPSError(obj);
    setMsgLength();
    return (*this);
}

void
COPSDecisionMsg::addDecision(const COPSDecision& obj, const COPSContext& cntxt)
{
    //Either error or decision can be added
    //If error is aleady there assert
    assert(!_error);

    assert(!obj.isLocalDecision());
    if (obj.isFlagSet())
    {
        if (_decisions[cntxt].size())
        {
            //Only one set of decision flags is allowed
            //for each context
            cpLog(LOG_ALERT, "Bad Message format, only one set of decision flags is allowed.");
            throw COPSBadDataException("Bad message format",
                                       __FILE__, __LINE__, 0);
        }
    }
    else
    {
        if (_decisions[cntxt].size() == 0)
        {
            //The flags decision must precede any other
            //decision message, since the decision is not
            //flags throw exception
            cpLog(LOG_ALERT, "Bad Message format, flags decision must precede any other decision object.");
            throw COPSBadDataException("Bad message format",
                                       __FILE__, __LINE__, 0);
        }
    }
    _decisions[cntxt].push_back(obj);
    setMsgLength();
}

COPSDecisionMsg&
COPSDecisionMsg::operator<<(const COPSIntegrity& obj)
{
    assert(obj.isMessageIntegrity());
    assert(!_integrity);
    _integrity = new COPSIntegrity(obj);
    setMsgLength();
    return (*this);
}

COPSDecisionMsg&
COPSDecisionMsg::operator<<(const COPSHeader& obj)
{
    assert(!_hdr);
    assert(obj.getOpCode() == COPS_OP_DEC);
    _hdr = new COPSHeader(obj);
    setMsgLength();
    return (*this);
}

COPSDecisionMsg&
COPSDecisionMsg::operator<<(const COPSHandle& obj)
{
    assert(!_clientHandle);
    _clientHandle = new COPSHandle(obj);
    setMsgLength();
    return (*this);
}

void
COPSDecisionMsg::setMsgLength()
{
    u_int16_t len = 0;
    if (_clientHandle.getPtr()) len += _clientHandle->getDataLength();
    if (_error.getPtr()) len += _error->getDataLength();

    //Display any local decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        const COPSContext& cntxt = (*itr).first;
        const vector < COPSDecision > & decVec = (*itr).second;
        len += cntxt.getDataLength();
        for (vector < COPSDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            len += (*vItr).getDataLength();
        }
    }

    if (_integrity.getPtr())
    {
        len += _integrity->getDataLength();
    }
    _hdr->setMsgLength(len);
}

void
COPSDecisionMsg::writeData(int id) const throw(VNetworkException&)
{
    checkSanity();
    if (_hdr.getPtr()) _hdr->writeData(id);
    if (_clientHandle.getPtr()) _clientHandle->writeData(id);
    if (_error.getPtr()) _error->writeData(id);

    //Display decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        const COPSContext& cntxt = (*itr).first;
        const vector < COPSDecision > & decVec = (*itr).second;
        cntxt.writeData(id);
        for (vector < COPSDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            (*vItr).writeData(id);
        }
    }

    if (_integrity.getPtr()) _integrity->writeData(id);
}
