
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSReportMsg_cxx_Version =
    "$Id: COPSReportMsg.cxx,v 1.7 2000/12/18 23:42:55 bko Exp $";
#include <assert.h>

#include "COPSReportMsg.hxx"
#include "cpLog.h"

COPSReportMsg::COPSReportMsg()
        : _clientHandle(0),
        _report(0),
        _integrity(0)
{}


COPSReportMsg::COPSReportMsg(const char* data)
        : _clientHandle(0),
        _report(0),
        _integrity(0)
{
    parse(data);
}

void
COPSReportMsg::parse(const COPSHeader& hdr, const char* data) throw (COPSBadDataException&)
{
    assert(hdr.getOpCode() == COPS_OP_RPT);
    _hdr = new COPSHeader(hdr);
    parse(data);
    setMsgLength();
}

COPSReportMsg::~COPSReportMsg()
{}


void
COPSReportMsg::parse(const char* data) throw (COPSBadDataException&)
{
    COPSMsg::parseHeader(data);

    while (_dataStart < _dataLength)
    {
        COPSObjHeader objHdr;
        objHdr.parse(data + _dataStart);
        switch (objHdr.getCNum())
        {
            case COPS_HANDLE:
            {
                _clientHandle = new COPSHandle(data + _dataStart);
                _dataStart += _clientHandle->getDataLength();
            }
            break;
            case COPS_RPT:
            {
                _report = new COPSReportType(data + _dataStart);
                _dataStart += _report->getDataLength();
            }
            break;
            case COPS_CSI:
            {
                COPSClientSI csi(data + _dataStart);
                _dataStart += csi.getDataLength();
                _clientSI.push_back(csi);
            }
            break;

            case COPS_MSG_INTEGRITY:
            {
                _integrity = new COPSIntegrity(data + _dataStart);
                _dataStart += _integrity->getDataLength();
            }
            break;

            default:
            {
                cpLog(LOG_ALERT, "Bad Message format, unknown object type");
                throw COPSBadDataException("Bad Message format", __FILE__, __LINE__, 0);
            }
            break;
        }
    }
    checkSanity();
}

void
COPSReportMsg::checkSanity() const throw (COPSBadDataException&)
{
    if (!_hdr || !_clientHandle || !_report )
    {
        cpLog(LOG_ALERT, "Bad Message format");
        throw COPSBadDataException("Bad message format",
                                   __FILE__, __LINE__, 0);
    }
}

void
COPSReportMsg::dump(ostream& os)const
{
    if (_hdr.getPtr()) os << *_hdr << endl;
    if (_clientHandle.getPtr()) os << *_clientHandle << endl;
    if (_report.getPtr()) os << *_report << endl;

    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        os << (*sItr) << endl;
    }

    if (_integrity.getPtr()) os << *_integrity << endl;
}

COPSReportMsg&
COPSReportMsg::operator<<(const COPSHandle& obj)
{
    assert(!_clientHandle);

    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);
    _clientHandle = new COPSHandle(obj);
    setMsgLength();
    return (*this);
}

COPSReportMsg&
COPSReportMsg::operator<<(const COPSReportType& obj)
{
    assert(!_report);

    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);
    _report = new COPSReportType(obj);
    setMsgLength();
    return (*this);
}

COPSReportMsg&
COPSReportMsg::operator<<(const COPSIntegrity& obj)
{
    assert(!_integrity);
    assert(obj.isMessageIntegrity());
    _integrity = new COPSIntegrity(obj);
    setMsgLength();
    return (*this);
}

COPSReportMsg&
COPSReportMsg::operator<<(const COPSHeader& obj)
{
    assert(!_hdr);
    assert(obj.getOpCode() == COPS_OP_RPT);
    _hdr = new COPSHeader(obj);
    setMsgLength();
    return (*this);
}

COPSReportMsg&
COPSReportMsg::operator<<(const COPSClientSI& obj)
{
    assert(obj.isClientSI());
    _clientSI.push_back(obj);
    setMsgLength();
    return (*this);
}

void
COPSReportMsg::setMsgLength()
{
    u_int16_t len = 0;
    if (_clientHandle.getPtr()) len += _clientHandle->getDataLength();
    if (_report.getPtr()) len += _report->getDataLength();

    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        len += (*sItr).getDataLength();
    }

    if (_integrity.getPtr()) len += _integrity->getDataLength();
    _hdr->setMsgLength(len);
}

void
COPSReportMsg::writeData(int id) const throw(VNetworkException&)
{
    checkSanity();
    if (_hdr.getPtr()) _hdr->writeData(id);
    if (_clientHandle.getPtr()) _clientHandle->writeData(id);
    if (_report.getPtr()) _report->writeData(id);
    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        (*sItr).writeData(id);
    }
    if (_integrity.getPtr()) _integrity->writeData(id);
}
