#ifndef OSPProvider_HXX
#define OSPProvider_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const OSPProvider_hxx_Version =
    "$Id: OSPProvider.hxx,v 1.7 2001/01/20 01:33:06 wbreu Exp $";




#include <string>
#include <list>

#include <ospprovider.h>
#include <ospdatatypes.h>
#include <OSPDefines.hxx>

/** Private Key structure
 **/
typedef struct OSPT_PRIVATEKEY
{
    unsigned char *PrivateKeyData;
    unsigned PrivateKeyLength;
}
OSP_PRIVATEKEY;


/// Configurable information for OSP Server
struct OSPProviderInfo
{
    /// number of OSP Server url's
    unsigned int numOfServicePoints;
    /// list of OSP Server url's
    list < string > servicePoints;
    /// audit url - not used but required to be configured by SDK
    string auditURL;
    /// RSA private key used for signing msg
    OSPTPRIVATEKEY localPrivateKey;   
    /** X.509 formatted cert. containing RSA public key 
     * corresponding to local private key
     */
    unsigned char* localCertificate;
    /// # of CA certificates
    unsigned int numOfAuthorityCertificates;
    /// array of X.509 formatted cert containing CA authority public keys
    list < unsigned char* > authorityCertificates; 
    /// should auth tokens be validated locally?
    unsigned int localValidation;   
    /// lifetime in seconds of a single SSL session key
    unsigned int sslLifetime;       
    /// max # of simultaneous connections
    unsigned int httpMaxConnections;   
    /// time in sec an HTTP connection is maintained after a comm exchange
    unsigned int httpPersistence;
    /// delay time in sec for HTTP connection retry
    unsigned int httpRetryDelay;
    /// # of HTTP connection retries are allowed
    unsigned int httpRetryLimit;
    /// max time in millisec to wait for a response from the OSP Server
    unsigned int httpTimeout;
    // How to handle:
    /// Customer ID - TransNexus Extension
    string customerId;      // TransNexus Extension
    /// Device ID - TransNexus Extension
    string deviceId;        // TransNexus Extension
}
;

/// C++ OSP Provider wrapper object
class OSPProvider
{
    public: 
       /// call to OSPPProviderDelete() osplib api
       virtual ~OSPProvider();

        /// invoke OSPPProviderNew() in osplib:
        int connect( OSPProviderInfo& info );

	/** attempt to connect to OSP Server.
	 *  if connection is successful, update _providerHandle
	 */
        int connect( unsigned int numOfServicePoints,
                     const char * servicePoints[],
                     const char * auditURL,
                     const OSPTPRIVATEKEY * localPrivateKey,
                     unsigned char * localCertificate,
                     unsigned int numOfAuthorityCertificates,
                     const void * authorityCertificates[],
                     unsigned int localValidation,
                     unsigned int sslLifetime,
                     unsigned int httpMaxConnections,
                     unsigned int httpPersistence,
                     unsigned int httpRetryDelay,
                     unsigned int httpRetryLimit,
                     unsigned int httpTimeout,
                     const char* customerId,
                     const char* deviceId ) ;

	/// Is OSP Provider ready to deploy?
        // True if providerHandle is not 0
        bool isProviderReady( void ) const
        {
            return _providerReady;
        }

        /// get ProviderHandle
        OSPProviderHandle getHandle( void ) const
        {
            return _providerHandle;
        }

        /// Debug only: display obj content
        void print( void );

        ///
        // access functions to get & set configuration of a Provider
        // object.  The set methods do not interrupt any active
        // communication exchanges, but subsequent exchanges will
        // use the new values.
	//
	// OSPPProviderGet/SetAuthorityCertificates()
          
	/// get number of CA Certificates
        int getNumberOfAuthorityCertificates(
            unsigned int *numOfAuthorityCertificates );
	/// retrieve CA Certificates
        int getAuthorityCeritficates(
            unsigned int sizeOfCertificate,    // max size of any ind. cert.
            unsigned int *numOfAuthorityCertificates,
            void * authorityCertificatesPtr[] );
	/// set CA Certificates
        int setAuthorityCeritficates( unsigned int numOfAuthorityCertificates,
                                      const void * authorityCertificatesPtr[] );

        // OSPPProviderGet/SetHTTPMaxConnections()
	/// get max # of HTTP Connections
        int getHttpMaxConnections( unsigned int *httpMaxConnections );
	/// set max # of HTTP Connections
        int setHttpMaxConnections( unsigned int httpMaxConnectons);

        // OSPPProviderGet/SetHTTPPersistence() - in seconds
	/// get HTTP Persistence time - in seconds
        int getHttpPersistence( unsigned int *httpPersistence );
	/// set HTTP Persistence time - in seconds
        int setHttpPersistence( unsigned int httpPersistence );

        // OSPPProviderGet/SetHTTPRetryDelay() - in seconds
	/// get retry delay for HTTP connection - in seconds
        int getHttpRetryDelay( unsigned int *httpRetryDelay );
	/// set retry delay for HTTP connection - in seconds
        int setHttpRetryDelay( unsigned int httpRetryDelay );

        // OSPPProviderGet/SetHTTPRetryLimit()
	/// get max number allowed to do HTTP reconnection retry
        int getHttpRetryLimit( unsigned int *httyRetryLimit );
	/// set max number allowed to do HTTP reconnection retry
        int setHttpRetryLimit( unsigned int httyRetryLimit );

        // OSPPProviderGet/SetHTTPTimeout() - in millisecs
	/// get HTTP connection timeout value - in millisecs
        int getHttpTimeout( unsigned int *httpTimeout );
	/// set HTTP connection timeout value - in millisecs
        int setHttpTimeout( unsigned int httpTimeout );

        // OSPPProviderGet/SetLocalKeys() - public & private key pair
        // used by the provider to sign its requests and indications.
        //
	/** get public & private key pair.
	 *  used by the provider to sign its requests and indications
	 */
        int getLocalKeys( OSPTPRIVATEKEY *localPrivateKey,
                          unsigned int sizeOfCertificate,
                          unsigned char * localCertificate );
	/// set public & private key pair
        int setLocalKeys( const OSPTPRIVATEKEY *localPrivateKey,
                          const unsigned char* localCertificate );

        // OSPPProviderGet/SetLocalValidation
        // if localValidation == 0, remote validation
        //                    != 0, local validation
	/// get validation mode
        int getLocalValidation( unsigned int *localValidation );
	/// set validation mode
        int setLocalValidation( unsigned int localValidation );

        // OSPPProviderGet/SetServicePoints()
	/// 
        int getNumOfServicePoints( unsigned int * numOfServicePoints );
	///
        int getServicePoints( unsigned int numOfServicePoints,
                              unsigned int sizeOfServicePoint,
                              char * servicePoints[] );
	///
        int setServicePoints( unsigned int numOfServicePoints,
                              const char * servicePoints[] );

        // OSPPProviderGet/SetSSLLifetime - in seconds
	///
        int getSslLifetime( unsigned int *sslLifetime );
	///
        int setSslLifetime( unsigned int sslLifetime );

	friend class OSPIntf;

	// for testing: wbr - 1/19/01
	OSPProvider( void );

    protected:


    private:

        /** private constructor instantiated by OSPIntf.
         * need to call setData() to make the obj well-formed
         * e.g. OSPProivder *ospProvider = new OSPProvider();
         *      OSPProviderInfo  data;
         *      data.numOfServicePoints = ....;
         *      data.....
         *      ospProvider->connect( data );
	 */
        // OSPProvider ( void );

        /**
         * True if this object is initialized and connected to
         * the OSP Server.
	 */
        bool _providerReady;

        // OSPPProviderNew() returns a Provider Handle
        // OSPTPROVHANDLE  _providerHandle;
        // u_int32_t _providerHandle;
	/** Provider Handle returned by OSPPProviderNew()
	 */
        OSPProviderHandle _providerHandle;

        /// equal operator - not in use
        OSPProvider& operator=( const OSPProvider& rhs );

        /// copy constructor - not in use
        OSPProvider( const OSPProvider& rhs );
};

#endif

// End of File
