
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const PolicyTranslator_cxx_Version =
    "$Id: PolicyTranslator.cxx,v 1.19 2001/01/20 01:33:06 wbreu Exp $";
#include "cpLog.h"
#include "VMissingDataException.hxx"
#include "NamedDataParser.hxx"
#include "Cops_def.hxx"
#include "COPSReqMsg.hxx"
#include "COPSReportMsg.hxx"
#include "OSPAuthInfo.hxx"
#include "OSPValInfo.hxx"
#include "PolicyTranslator.hxx"
#include "OSPIntf.hxx"
#include "OSPToken.hxx"

PolicyTranslator::PolicyTranslator(void)
{
    cpLog( LOG_DEBUG, "PolicyTranslator ctor invoked" );
}

///
PolicyTranslator::~PolicyTranslator()
{
    cpLog(LOG_DEBUG, "PolicyTranslator dtor invoked" );
}

///
int PolicyTranslator::sendReq( const COPSMsg & cMsg, PolicyTransaction & pol ) throw (PolicyBadDataException &)
{
    cpLog( LOG_DEBUG, "sendReq() called ");

    // extract info from the Cops msg and translate it to an
    // Authorization Request or Authorization Validation Request
    const COPSHeader& hdr = cMsg.getHeader();
    if (hdr.isARequest() == false)
    {
        cpLog( LOG_ALERT, "sendReq() does not receive a COPS request");

        throw PolicyBadDataException("sendReq: not a COPS request message",
                                     __FILE__, __LINE__, 0);
    }

    int errCode = -1;

    const COPSReqMsg* reqMsg = dynamic_cast < const COPSReqMsg* > (&cMsg);
    assert(reqMsg);

    const COPSContext & ctxt = reqMsg->getContext();
    if (ctxt.isOutgoingMessage())   // Outgoing Gateway
    {
        try
        {
            errCode = translateRequest( cMsg, pol );
        }
        catch (OSPException& e)
        {
            cpLog(LOG_ALERT,
                  "translateRequest() exception: %s",
                  e.getDescription().c_str());
        }

    }
    else     // Incoming Gateway: validate the authorization
    {
        try
        {
            errCode = translateValRequest( cMsg, pol );
        }
        catch (OSPException& e)
        {
            cpLog(LOG_ALERT,
                  "translateValRequest() exception: %s",
                  e.getDescription().c_str());
        }
    }

    return errCode;
}   // PolicyTranslator::sendReq()

///
int PolicyTranslator::sendUsage( const COPSMsg & cMsg, PolicyTransaction & pol ) throw
( PolicyBadDataException & , OSPException& )
{
    cpLog( LOG_DEBUG, "sendUsage() called");
    // extract info from the Cops msg and translate it to an
    // Usage Info object
    const COPSHeader& hdr = cMsg.getHeader();
    if (hdr.isAReport() == false)
    {
        cpLog( LOG_ALERT, "sendUsage() does not receive a COPS Report");
        throw PolicyBadDataException("sendUsage: not a COPS report message",
                                     __FILE__, __LINE__, 0);
    }
    int errCode = -1;

    try
    {
        errCode = translateUsage( cMsg, pol );
    }
    catch (OSPException& e)
    {
        cpLog(LOG_ALERT,
              "PolicyTranslator::sendUsage exception: %s",
              e.getDescription().c_str());
    }

    return errCode;
}   // PolicyTranslator::sendUsage()

///
int PolicyTranslator::abandonTrans( const string & reason, PolicyTransaction & pol )
{
    cpLog(LOG_DEBUG, "abandonTrans() - id = %d", pol.getID());
    // send a 0 usage message to OSP to get the transaction in the
    // right state:

    int errCode = -1;
    try
    {
        errCode = OSPIntf::Instance().abandonTrans( reason, pol );
    }
    catch (OSPException& e)
    {
        cpLog(LOG_ALERT,
              "PolicyTranslator::abandonTrans exception: %s",
              e.getDescription().c_str());
    }

    return errCode;

}  // PolicyTranslator::abandonTrans()

///
int PolicyTranslator::translateRequest( const COPSMsg & cMsg,
                                        PolicyTransaction & pol ) throw
( PolicyBadDataException & , OSPException&)
{

    Sptr<OSPAuthInfo> ai = new OSPAuthInfo;
    if (ai == NULL)
    {
        cpLog(LOG_ALERT, "translateRequest: failure to new OSPAuthInfo");
        throw PolicyBadDataException("failure to new OSPAuthInfo",
                                     __FILE__, __LINE__, 0);

        return -1;
    }
    // set up the Authorization Request:
    //
    // look for data given the keys/ tags:
    const COPSReqMsg* reqMsg = dynamic_cast < const COPSReqMsg* > (&cMsg);
    assert(reqMsg);

    NamedDataParser parser(reqMsg->getClientSI());

    ai->setNumOfCallIds(1);
    try
    {
        ai->setSource( parser.getValue( COPS_key::source.c_str() ).c_str());
        ai->setSourceDevice( parser.getValue( COPS_key::sourceDev.c_str() ).c_str());
        ai->setCallingNum( parser.getValue( COPS_key::callingNumber.c_str() ).c_str());
        ai->setCalledNum( parser.getValue( COPS_key::calledNumber.c_str() ).c_str());
        ai->setUser( parser.getValue( COPS_key::user.c_str() ).c_str());
        string callIdVal = parser.getValue( COPS_key::callId.c_str() );

        // wbr: workaround - use md5CallId instead
        ai->setCallId(1, callIdVal.length() + 1, callIdVal.c_str() );
        // set the CallId in the policy transaction
        pol.setCallId( callIdVal.c_str() );
        // use md5 Call Id in OSP Settlement request message:
        // ai->setCallId(1, pol.getMd5CallIdLen()+1,
        //               pol.getMd5CallId() );
        cpLog(LOG_ALERT, "translateRequest: policy transaction updates --");
        pol.print();
    }
    catch (VMissingDataException & e)
    {
        cpLog(LOG_ERR, "%s", e.getDescription().c_str());
        return -1;
    }
    catch (OSPException& e)
    {
        cpLog(LOG_ERR, "%s", e.getDescription().c_str());
        return -1;
    }
    ai->setPreferDests( reinterpret_cast < const char** > (NULL) );    // no preferred destination
    ai->setNumDest( 1 );

    // let OSP interface handle the request message
    OSPIntf::Instance().sendReq( *ai, pol );

    return 0;

}

///

int PolicyTranslator::translateUsage( const COPSMsg & cMsg,
                                      PolicyTransaction & pol ) throw
( PolicyBadDataException & , OSPException& )
{
    cpLog( LOG_DEBUG, "translateUsage() called");

    // parse the COPS report message:
    const COPSReportMsg* repMsg = dynamic_cast < const COPSReportMsg* > (&cMsg);
    assert(repMsg);

    int id = pol.getID();          // get transaction id
    int d = pol.getDuration();    // call duration in seconds

    cpLog(LOG_DEBUG, "id = %d,  duration = %d", id, d);
    Sptr<OSPUsageInfo> u = new OSPUsageInfo( id, d, 0, -1, 0, -1 );
    assert( u != 0 );

    OSPIntf::Instance().sendUsage( *u, pol );

    return 0;
}

///
int PolicyTranslator::translateValRequest( const COPSMsg & cMsg,
        PolicyTransaction & pol ) throw
( PolicyBadDataException & )
{

    Sptr<OSPValInfo> vi = new OSPValInfo;
    int errorcode = -1;

    cpLog( LOG_DEBUG, "translateValRequest() called");
    if (vi == NULL)
    {
        cpLog(LOG_ALERT, "translateValRequest: failure to new OSPValInfo");
        throw PolicyBadDataException("failure to new OSPValInfo",
                                     __FILE__, __LINE__, 0);

    }
    // set up the ValidateAuthorization Request:
    //
    // look for data given the keys/ tags:
    const COPSReqMsg* reqMsg = dynamic_cast < const COPSReqMsg* > (&cMsg);
    assert(reqMsg);

    NamedDataParser parser(reqMsg->getClientSI());

    try
    {
        vi->setSource( parser.getValue( COPS_key::source.c_str() ).c_str());
        vi->setDestination( parser.getValue( COPS_key::destination.c_str() ).c_str());

        vi->setSourceDevice( parser.getValue( COPS_key::sourceDev.c_str() ).c_str());
	// save for future:
        // vi->setDestDevice( parser.getValue( COPS_key::destDev.c_str() ).c_str());
        vi->setCallingNum( parser.getValue( COPS_key::callingNumber.c_str() ).c_str());
        vi->setCalledNum( parser.getValue( COPS_key::calledNumber.c_str() ).c_str());
        string callIdVal = parser.getValue( COPS_key::callId.c_str() );

        vi->setCallId( callIdVal.c_str() );
        // set the CallId in the policy transaction
        pol.setCallId( callIdVal.c_str() );

        // the token is always at the end of COPS data:
        const vector < COPSClientSI > & vec = reqMsg->getClientSI();
        const COPSData& cData = vec[vec.size() - 1].getData();
        const char* token = cData.getData();
        unsigned int tlen = cData.length();

        // convert token (base64) from the COPS message into binary:
        OSPToken tok;
        string btok;     // binary token
        if (tok.decodeToken( token, tlen, btok ))
        {
            vi->setToken( btok );
            cpLog(LOG_ALERT,
                  "translateValRequest: retrieved token, size = %u (b64 %u)",
                  vi->getTokenSize(), cData.length() );
        }
        else
        {
            return errorcode;
        }

        cpLog(LOG_ALERT, "translateValRequest: policy transaction updates --");
        pol.print();
    }
    catch (VMissingDataException & e)
    {
        cpLog(LOG_DEBUG, "%s", e.getDescription().c_str());
        return errorcode;
    }
    catch (OSPException& e)
    {
        cpLog(LOG_DEBUG, "%s", e.getDescription().c_str());
        return errorcode;
    }

    // let OSP interface handle the auth validation request message
    OSPIntf::Instance().sendValReq( *vi, pol );


    return 0;

}

//  end of file
