/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const PepAgent_cxx_Version =
    "$Id: PepAgent.cxx,v 1.54 2002/05/15 01:26:26 sprajpat Exp $";

#include <stdio.h>
#include <strstream.h>



#include "VThreadPool.hxx"
#include "VFunctor.hxx"
#include "VCondition.h"
#include "Tcp_ClientSocket.hxx"
#include "NetworkAddress.h"
#include "PepAgent.hxx"
#include "COPSMsgParser.hxx"
#include "PepAgent.hxx"
#include "OSPTime.hxx"
#include "Cops_def.hxx"
#include "PolRequest.hxx"
#include "LockHelper.hxx"
#include "PolicyCfg.hxx"
#include "RoundRobinServers.hxx"
#include "COPSTransceiver.hxx"
#ifdef HAS_OSP
#include "PolicyTransaction.hxx"
#endif

using namespace Vocal;

static const int WELL_KNOWN_PDP_PORT = 3288;
static const int BUFSIZE             = 256;
static const int BIG_BUFSIZE         = 4256;
static const int SYNC_SLEEP          = 20000;

PepAgent* PepAgent::_instance = 0;
int PepAgent::PDPPort = WELL_KNOWN_PDP_PORT;


PepAgent&
PepAgent::instance(const char* psHost, int psPort,
                   const char* psAltHost, int psAltPort,
                   int nThreads )
{
    if (!_instance)
    {
        _instance = new PepAgent(psHost, psPort, psAltHost, psAltPort, nThreads);
    }
    return (*_instance);
}

PepAgent::PepAgent(const char* psHost, int psPort,
                   const char* psAltHost, int psAltPort,
                   int numThreads)
{
    char buf[BUFSIZE]; 
    _callBackFunc = 0;
    _clientType = C_SIP;
    _clientPort = 0;
    _clientHost = "Unknown";
    //Default response time is 10 sec, however applications can set it
    //to the value they want by calling PepAgent::instance().setResonseTime(newTime);
    _responseTime = 10;



#if defined (__linux__)
    if (atexit(PepAgent::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };
    if (getdomainname(buf, BUFSIZE-1) != 0)
    {
        sprintf(buf, "Failed to get the domain name, reason:%s", strerror(errno));
        cpLog(LOG_ALERT, buf);
        exit( -1);
    }
    _domainName = buf;
#else
     _domainName = "vovida.com";
#endif
    if (gethostname(buf, BUFSIZE-1) != 0)
    {
        sprintf(buf, "Failed to get the host name, reason:%s", strerror(errno));
        cpLog(LOG_ALERT, buf);
        exit( -1);
    }
    _hostName = buf;

    PolicyCfg::Instance(psHost, psPort, psAltHost, psAltPort);
    _threadPool = new VThreadPool(numThreads);

    VFunctor functor(PepAgent::garbageCollectFunc, this);
    _threadPool->addFunctor(functor);
    VFunctor functor2(PepAgent::checkSendQueue, this);
    _threadPool->addFunctor(functor2);
    VFunctor functor3(PepAgent::checkUsageQueue, this);
    _threadPool->addFunctor(functor3);
    VFunctor functor4(PepAgent::checkResponseQueue, this);
    _threadPool->addFunctor(functor4);
}

// establish connection to PDP's IP address
void
PepAgent::connect(const string& hashKey) throw (VNetworkException&)
{
    char buf[BIG_BUFSIZE];

    cpLog(LOG_DEBUG, "Thread (%d) - Connecting to PDP", pthread_self());
    COPSHeader hdr(COPS_OP_OPN, C_SIP);

    COPSPepId pepId;
    sprintf(buf, "Vovida-Pep-%lu", pthread_self());
    COPSData d(buf);
    pepId.setData(d);

    COPSClientOpenMsg msg;
    msg << hdr ;
    msg << pepId ;

    ostrstream ostr(buf, BIG_BUFSIZE);
    ostr << msg << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());

    cpLog(LOG_DEBUG, "PolicyServer:%s, port:%d",
          getActivePdp(hashKey).c_str(), PepAgent::PDPPort);
    NetworkAddress pdpAddress (getActivePdp(hashKey), PepAgent::PDPPort);

    TcpClientSocket cSock(pdpAddress, false);
    cSock.connect();
    Connection& conn = cSock.getConn();
    msg.writeData(conn.getConnId());
    processData(conn, hashKey);
}

void
PepAgent::garbageCollectFunc(void* args)
{
    PepAgent& self = *(static_cast < PepAgent* > (args));
    for (; ; )
    {
        ///check to see if PolicyCfg has changed
        if (PolicyCfg::Instance().isPdpDataChanged())
        {
            cpLog(LOG_DEBUG, "***** Policy configuration changed ******");
            LockHelper guard(self._connlock);
            for ( ConnectionMap::iterator itr = self._connectionMap.begin();
                    itr != self._connectionMap.end(); itr++)
            {
                if ((*itr).second.isLive())
                {
                    (*itr).second.close();
                }
            }
            PolicyCfg::Instance().pdpDataChanged(false);
        }

        //Garbage collect any completed/failed request
        {
            LockHelper guard(self._lock);
            for (PolReqMap::iterator itr = self._reqMap.begin();
                    itr != self._reqMap.end(); itr++)
            {
                if (((*itr).second)->toBeDeleted())
                {
                    cpLog(LOG_DEBUG, "**Deleting request with handle %s", ((*itr).second)->getClientHandle().str().c_str());
                    self._reqMap.erase(itr);
                    //Do only one, so in each iteration try to free as much
                    break;
                }

            }
        }

        {
            LockHelper guard(self._lock);
            //Check if any of the responses have timed-out
            for (PolReqMap::iterator itr = self._reqMap.begin();
                    itr != self._reqMap.end(); itr++)
            {
                Sptr < PolRequest > & pReq = (*itr).second;
                pReq->checkIfTimedOut();
            }
        }
        sleep(2);
    }
}

void
PepAgent::checkSendQueue(void* args)
{
    PepAgent& self = *(static_cast < PepAgent* > (args));
    for (; ; )
    {
        cpLog(LOG_DEBUG, "PepAgent::checkSendQueue");
        PolRequest* pReq = self._sendQueue.getNext();
        cpLog(LOG_DEBUG, "Got request:%s", pReq->getClientHandle().str().c_str());
        if (pReq->hasFailed()) continue;

	// schedule to the thread pool
        VFunctor functor(PepAgent::sendPolicyRequest, pReq);
        self._threadPool->addFunctor(functor);
    }
}

void
PepAgent::checkUsageQueue(void* args)
{
    PepAgent& self = *(static_cast < PepAgent* > (args));
    for (; ; )
    {
        cpLog(LOG_DEBUG, "PepAgent::checkUsageQueue");
        PolRequest* pReq = self._usageQueue.getNext();
        cpLog(LOG_DEBUG, "Got request:%s", pReq->getClientHandle().str().c_str());
        VFunctor functor(PepAgent::sendUsage, pReq);
        self._threadPool->addFunctor(functor);
    }
}

void
PepAgent::checkResponseQueue(void* args)
{
    PepAgent& self = *(static_cast < PepAgent* > (args));
    for (; ; )
    {
        cpLog(LOG_DEBUG, "PepAgent::checkResponseQueue");
        PolRequest* pReq = self._responseQueue.getNext();
        cpLog(LOG_DEBUG, "Got response:%s", pReq->getClientHandle().str().c_str());
        self.notifyClient(*pReq);
    }
}

void
PepAgent::sendPolicyRequest(void* args)
{
    PolRequest* pReq = static_cast < PolRequest* > (args);
    string clientHandle = pReq->getClientHandle().str();
    cpLog(LOG_DEBUG, "PepAgent::sendPolicyRequest:%s", clientHandle.c_str());

    try
    {
        Connection& conn = PepAgent::instance().getConnection(pReq->getCallId());
        int respTime = PepAgent::instance().getResponseTime() / 1000;
        pReq->sendCOPSMsg(conn);
        cpLog(LOG_DEBUG, "Sent request:%s", clientHandle.c_str());
        cpLog(LOG_DEBUG, "No response time:%d", respTime );
        if (conn.isReadReady(respTime))
        {
            PepAgent::instance().processData(conn);
            cpLog(LOG_DEBUG, "Received response:%s", clientHandle.c_str());
        }
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Failed to send Policy request, reason:%s" , e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_DEBUG, "Failed to send Policy request, reason:%s" , e.getDescription().c_str());
    }
}

Connection&
PepAgent::getConnection(const string& hashKey)
{
    LockHelper guard(_connlock);
    string lKey = getConnectionLookupKey(hashKey);
    pthread_t tid = pthread_self();
    if (!_connectionMap.count(lKey))
    {
        connect(hashKey);
    }

    if (!_connectionMap[lKey].isLive())
    {
        cpLog(LOG_ALERT, "Thread (%d) lost connection, connecting again using key (%s)", tid, lKey.c_str());
        connect(hashKey);
    }
    cpLog(LOG_DEBUG, "Thread (%d) got connection (%s) with key (%s).", tid, _connectionMap[lKey].getDescription().c_str(), lKey.c_str());
    return (_connectionMap[lKey]);
}

void
PepAgent::sendUsage(void* args)
{
    PolRequest* pReq = static_cast < PolRequest* > (args);
    try
    {
        pReq->sendUsage(PepAgent::instance().getConnection(pReq->getCallId()));
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to send usage for handle %s, reason:%s",
              pReq->getClientHandle().str().c_str(), e.getDescription().c_str());
        pReq->resendUsage();
    }
}

void
PepAgent::checkForData()
{
    for (ConnectionMap::iterator itr = _connectionMap.begin();
            itr != _connectionMap.end(); itr++)
    {
        Connection& conn = (*itr).second;
        checkForData(conn);
    }
}

void
PepAgent::checkForData(Connection& conn)
{
    if (conn.isLive() && conn.isReadReady(1))
    {
        cpLog(LOG_DEBUG , "Data ready on %s", conn.getDescription().c_str());
        processData(conn);
    }
}

void
PepAgent::processData(Connection& conn, const string& hashKey )
{
    try
    {
        Sptr < COPSMsg > msg = COPSTransceiver::receiveMsg(conn.getConnId());
        processCOPSMsg(*msg, conn, hashKey);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "CLOSING Connection:%s, reason %s",
              conn.getDescription().c_str(),
              e.getDescription().c_str());
        conn.close();
        return ;
    }
}

void
PepAgent::processCOPSMsg(const COPSMsg& msg, Connection& conn, const string& hashKey)
{
    if (msg.getHeader().isADecision())
    {
        PolicyTransaction* polT = new PolicyTransaction;
        const COPSDecisionMsg* dMsg = dynamic_cast < const COPSDecisionMsg* > (&msg);
        const COPSDecisionMsg::DecisionMap& dMap = dMsg->getDecisions();

        for (COPSDecisionMsg::DecisionMap::const_iterator itr = dMap.begin();
                itr != dMap.end(); itr++)
        {
            const COPSContext& ctxt = (*itr).first;
            if (ctxt.getMessageType() == M_OSP)
            {
                const vector < COPSDecision > & vDec = (*itr).second;
                assert(vDec.size() == 2);
                string dStr(vDec[1].getData().str());
                polT->decode(dStr);
                break;
            }
        }
        //get the Policy request context
        const COPSHandle& handle = dMsg->getClientHandle();

        _lock.lock();
        assert(_reqMap.count(handle.getId()));
        Sptr < PolRequest > & pReq = _reqMap[handle.getId()];
        _lock.unlock();

        pReq->setPolicyTransaction(polT);
        if (polT->getStatus() != 0)
        {
            pReq->fail();
        }
        else
        {
            pReq->received();
        }
    }
    else if (msg.getHeader().isAReport())
    {
        const COPSReportMsg* dMsg = dynamic_cast < const COPSReportMsg* > (&msg);
        handleReportMsg(*dMsg);
    }
    else if(msg.getHeader().isAClientAccept())
    {
        const COPSClientAcceptMsg* cMsg =
            dynamic_cast < const COPSClientAcceptMsg* > (&msg);
        const Sptr < COPSKATimer > & kt = cMsg->getKATimer();

        assert(kt.getPtr() != 0);

        _kaTimeVal = kt->getTimerVal();

        string lKey = getConnectionLookupKey(hashKey);
        if (_connectionMap.count(lKey) != 0)
        {
            cpLog(LOG_DEBUG, "Already connected.");
        }
        cpLog(LOG_DEBUG, "Thread (%d) - Inserting connection (%s) for PDP using key (%s)", pthread_self(), conn.getDescription().c_str(), lKey.c_str());
        _connectionMap[lKey] = conn;
    }
    else if(msg.getHeader().isAClientClose())
    {
        cpLog(LOG_INFO, "Closing connection for (%s)", conn.getDescription().c_str());
        conn.close(); 
    }
}

void
PepAgent::auth(const string& calledNumber,
               const string& callingNumber,
               const string& callId)
throw (VDuplicateIdException&)

{
    PolRequest* pReq = new PolRequest(calledNumber,
                                      callingNumber,
                                      callId, M_OSP);

    _lock.lock();
    if (_reqMap.count(pReq->getClientHandle()))
    {
        char buf[BUFSIZE];
        sprintf(buf, "Duplicate call Id %s, rejecting", callId.c_str());
        cpLog(LOG_DEBUG, buf);
        delete pReq;
        throw VDuplicateIdException(buf, __FILE__, __LINE__, 0);
    }
    _reqMap[pReq->getClientHandle()] = pReq;
    _lock.unlock();

    pReq->auth();
}

void
PepAgent::enableQoS(const string& callId,
                    const string& srcIp, int srcPort,
                    const string& destIp, int destPort)
{
    Sptr<PolRequest> pReq = new PolRequest(callId, M_QOS);

    cpLog(LOG_DEBUG, "Enabling QoS for callId %s [%s_%d -> %s_%d]",
          callId.c_str(), srcIp.c_str(), srcPort, destIp.c_str(), destPort);

    pReq->enableQoS(srcIp, srcPort, destIp, destPort);
}

void
PepAgent::disableQoS( const string& callId )
{
    Sptr<PolRequest> pReq = new PolRequest(callId, M_QOS);

    cpLog(LOG_DEBUG, "Disabling QoS for callId %s", callId.c_str());

    pReq->disableQoS();
}

void
PepAgent::queueForSend(PolRequest& req)
{
    cpLog(LOG_DEBUG, "Queing for send:%s", req.getClientHandle().str().c_str());
    _sendQueue.add(&req);
}

void
PepAgent::queueForResponse(PolRequest& req)
{
    cpLog(LOG_DEBUG, "Queing after response:%s", req.getClientHandle().str().c_str());
    _responseQueue.add(&req);
}

void
PepAgent::queueForUsage(PolRequest& req)
{
    cpLog(LOG_DEBUG, "Queing for usage:%s", req.getClientHandle().str().c_str());
    _usageQueue.add(&req);
}

void
PepAgent::notifyClient(PolRequest& pReq)
{
    if (!_callBackFunc)
    {
        cpLog(LOG_DEBUG, "No callback function is registered.");
        return ;
    }
    _callBackFunc(pReq);
}

void
PepAgent::handleReportMsg(const COPSReportMsg& msg)
{
    const COPSHandle& handle = msg.getClientHandle();

    Sptr < PolRequest > pReq;
    _lock.lock();
    assert(_reqMap.count(handle.getId()));
    pReq = _reqMap[handle.getId()];
    _lock.unlock();

    const COPSContext& cntxt = pReq->getContext();
    const COPSReportType& repType = msg.getReport();
    if (cntxt.getMessageType() == M_QOS)
    {
        const vector < COPSClientSI > & csi = msg.getClientSI();
        assert(csi.size());
        if (repType.isSuccess())
        {
            cpLog(LOG_ALERT, "%s.", csi[0].getData().str().c_str());
        }
        else
        {
            cpLog(LOG_ALERT, "%s , failed", csi[0].getData().str().c_str());
        }
    }
}

void
PepAgent::setClientInfo(const string& cHost, int cPort, COPS_ClientType cType)
{
    _clientType = static_cast < u_int16_t > (cType);
    _clientHost = cHost;
    _clientPort = cPort;
}


PepAgent::~PepAgent()
{
    cpLog(LOG_DEBUG, "PepAgent::~PepAgent");
}

void
PepAgent::destroy()
{
    delete PepAgent::_instance;
    PepAgent::_instance = 0;
}

PolRequest&
PepAgent::getPolReq(const string& callId) throw (VMissingDataException&)
{
    LockHelper guard(_lock);
    COPSData key(callId);
    if (_reqMap.count(key))
    {
        return (*(_reqMap[key]));
    }
    else
    {
        char buf[BUFSIZE];
        sprintf(buf, "Failed to get PolicyRequest object for callId:%s", callId.c_str());
        cpLog(LOG_ALERT, buf);
        throw VMissingDataException(buf, __FILE__, __LINE__, 0);
    }
}

PolRequest&
PepAgent::getPolReq(const string& calledNumber,
                    const string& callingNumber,
                    const string& callId)
throw (VMissingDataException&)
{
    return (getPolReq(callId));
}

void
PepAgent::cleanUp(const PolRequest& req)
{
    LockHelper guard(_lock);
    for (PolReqMap::iterator itr = _reqMap.begin();
            itr != _reqMap.end(); itr++)
    {
        if (((*itr).second).getPtr() == &req)
        {
            cpLog(LOG_DEBUG, "Deleting request with handle %s", req.getClientHandle().str().c_str());
            _reqMap.erase(itr);
            break;
        }
    }
}

void
PepAgent::validateToken(const string& calledNumber,
                        const string& callingNumber,
                        const string& callId, const COPSData& token) throw (VDuplicateIdException&)
{
    PolRequest* pReq = new PolRequest(calledNumber,
                                      callingNumber,
                                      callId,
                                      token,
                                      M_OSP);

    _lock.lock();
    if (_reqMap.count(pReq->getClientHandle()))
    {
        char buf[BUFSIZE];
        sprintf(buf, "Duplicate call Id %s, rejecting", callId.c_str());
        cpLog(LOG_DEBUG, buf);
        delete pReq;
        throw VDuplicateIdException(buf, __FILE__, __LINE__, 0);
    }
    _reqMap[pReq->getClientHandle()] = pReq;
    _lock.unlock();

    pReq->validate();
}

string
PepAgent::getConnectionLookupKey(const string& hashKey) const
{
    char buf[BUFSIZE];
    sprintf(buf, "%lu.%s", pthread_self(), getActivePdp(hashKey).c_str());
    return buf;
}

string
PepAgent::getActivePdp(const string& hashKey) const
{
    Data hashKeyData = hashKey;
    Data activePdp = RoundRobinServers::instance( SERVER_POS ).getHost( hashKeyData );
    cpLog(LOG_DEBUG, "PolicyServer: %s", activePdp.logData());
    return activePdp.convertString();
}

int
PepAgent::sendSyncWithRetries(const PolRequest& req, int nRetries)
{
    int numTries = 1;
    while (1)
    {
        try
        {
            Connection& conn = PepAgent::instance().getConnection(req.getCallId());
            req._copsMsg->writeData(conn.getConnId());

            //Wait for _responseTime seconds for the response
            if (conn.isReadReady(_responseTime))
            {
                PepAgent::instance().processData(conn, req.getCallId());
                return 0;
            }
            else
            {
                return -1;
            }
        }
        catch (VException& e)
        {
            cpLog(LOG_ALERT, "Failed to send COPS req, reason:%s",
                  e.getDescription().c_str());
            cpLog(LOG_INFO, "Retrying...");
            if (numTries == nRetries) return -1;
            numTries++;
            vusleep(SYNC_SLEEP);
        }
    }
    return -1;
}
