// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef AssignmentList_hxx
#define AssignmentList_hxx

#include <vector>
#include <map>
#include <string>

#include "Assignment.hxx"
#include "Sptr.hxx"

namespace Epoisses {

/**
   An ordered collection of Assignments. 
   This is the implementation class of a pointer hiding scheme. */
class IAssignmentList : public PSDBInterface
{
      friend class AssignmentList;
   public:
      void stringify(ostream &str) const;
      
   private:
      IAssignmentList(const BaseAssignment& assignment)
         : contents()
      {
         contents.push_back(assignment);
      }

      IAssignmentList()
         : contents()
      {}
         
      void
      add(const BaseAssignment &assignment)
      {
         contents.push_back(assignment);
      }

      vector<BaseAssignment> contents;
};

class AssignmentList : public PSDBInterface
{
      friend class BaseAssignment;
   public:
      typedef vector<BaseAssignment>::const_iterator const_iterator;
      
      /** Implicitly construct an AssignmentList from a static array of
          Assignments.  AssignmentList<void> means allow any type of
          Assignment. Need to know the size of the statically allocated array,
          so templatized. */
      template <int SZ>
      AssignmentList(const Assignment<void> (&arr)[SZ])
         : assignmentList(new IAssignmentList())
      {
         fromArray(arr, SZ);
      }
      
      // Copy constructor
      AssignmentList(const AssignmentList &rhs)
         : assignmentList(rhs.assignmentList)
      {}

      // implicit conversion from Assignment to AssignmentList
      AssignmentList(const BaseAssignment &assignment)
         : assignmentList(new IAssignmentList(assignment))
      {}

      // Empty contructor for var initialization
      AssignmentList()
         : assignmentList(new IAssignmentList())
      {}         

      int size() const
      {
         return assignmentList->contents.size();
      }

      bool
      empty() const
      {
         return assignmentList->contents.empty();
      }

      const_iterator 
      begin() const
      {
         return assignmentList->contents.begin();
      }

      const_iterator 
      end() const
      {
         return assignmentList->contents.end();
      }
      
      // add the assignment to contents
      AssignmentList
      add(const BaseAssignment &assignment)
      {
         assignmentList->add(assignment);
         return *this;
      }

      // add the assignment to contents, as operator
      AssignmentList
      operator,(const BaseAssignment& assignment)
      {
         add(assignment);
         return *this;
      }

      virtual void stringify(ostream &str) const
      {
         assignmentList->stringify(str);
      }

      // allow exercise of interactive test
      static AssignmentList
      fromMap(const map < string, string > &table);

   private:
      // helper for converting a static array 
      void
      fromArray(const Assignment<void> arr[], int sz);

      // share the list
      Sptr<IAssignmentList> assignmentList;
};

// Global operator
AssignmentList
operator,(const BaseAssignment& assignment1,
          const BaseAssignment& assignment2);

} // namespace Epoisses

#endif // AssignmentList_hxx
