// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef BaseField_hxx
#define BaseField_hxx

#include "IField.hxx"
#include "Sptr.hxx"

namespace Epoisses {

// forward reference all Fields for later friend declaration
template <class T> class Field;
class Constant;

/**
   Common class for all Fields. Provides reference counting
   for the internal fields. */
class BaseField : public FieldInterface
{
   public:
      BaseField(const BaseField &field)
         : internalField(field.internalField)
      {}

      const string &
      getName() const
      {
         return internalField->getName();
      }

      friend class BaseIAssignment;
      friend class FieldList;
      template <class T> friend class Field;

      static const Constant None;

      bool 
      operator==(const BaseField& rhs) const
      {
         return getRawInternal() == rhs.getRawInternal();
      }

      bool 
      operator!=(const BaseField& rhs) const
      {
         return getRawInternal() != rhs.getRawInternal();
      }

      Sptr<BaseIField> getInternal() const
      {
         return internalField;
      }

      
   protected:
      BaseField(BaseIField *field)
         : internalField(field)
      {}
      virtual ~BaseField() {}

      BaseIField* getRawInternal() const
      {
         return internalField.getPtr();
      }

      // wrapper to heap allocated implementation field
      Sptr<BaseIField> internalField;
};

/**
   Base class for all Fields that may be bound. Corresponds to columns and does
   not include Constants. */
class BaseColumnField : public BaseField
{
   public:
      BaseColumnField(const BaseColumnField &field)
         : BaseField(field)
      {}

      /// convert self to string
      virtual void stringify(ostream &str) const
      {
         internalField->stringify(str);
      }

      virtual ~BaseColumnField() {}

   protected:
      BaseColumnField(BaseIField *field)
         : BaseField(field)
      {}

      friend class PSDBresult_BindIterator;
   private:
      /// give the field a new value (used by PSDBresult_BindIterator
      virtual void bind(const string &s) throw (BadValueException&)
      {
         internalField->bind(s);
      }
};

/**
   Explicit instantiation of Field<void> to mean  Field of any type. */
class Field<void> : public BaseColumnField
{
   public:
      Field<void>(const BaseColumnField &field)
         : BaseColumnField(field)
      {}
   private:
      Field<void>();
};

} // namespace Epoisses

#endif // BaseField_hxx

