// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef BaseIField_hxx
#define BaseIField_hxx

#include <string>

#include "PSDBInterface.hxx"
#include "VException.hxx"
#include "global.h"

namespace Epoisses {

class BaseAssignment;

/**
   Thrown when the backend determines that the value type does not match
   the Field type.  */
class BadValueException : public VException
{
   public:
      BadValueException(const string &s,
                        const string& file, 
                        const int line)
         : VException(s, file, line)
      {}

      virtual string getName() const
      {
         return "BadValueException";
      }
};

/**
   Thrown when the front end determines that the field to bind does not
   exist. */
class NoSuchFieldException : public VException
{
   public:
      NoSuchFieldException(const string &s,
                           const string& file, 
                           const int line)
         : VException(s, file, line)
      {}

      virtual string getName() const
      {
         return "NoSuchFieldException";
      }
};

/**
   Thrown when the frontend tries to get the value of a Field that has
   not been bound to a result. */
class NotBoundException : public VException
{
   public:
      NotBoundException(const string &s,
                        const string& file, 
                        const int line)
         : VException(s, file, line)
      {}

      virtual string getName() const
      {
         return "NotBoundException";
      }
};

/// Provides type unity for Field and Filter.
class FilterableInterface : public PSDBInterface
{
   public:
      // don't allow a standard comma operator
      void operator,(const BaseAssignment&)
      {
      }
};

class FieldInterface : public FilterableInterface
{
   public:
      virtual const string & getName() const = 0;
      /// give this instance a set of values
      virtual void bind(const string &s) throw (BadValueException&) = 0;
};

/** 
    Common base class for all IFields; the data structures that hold
    names and values. This class is used as the interface for binding
    values to fields. */
class BaseIField : public FieldInterface
{
   public:
      virtual const string & 
      getName() const
      {
         return _name;
      }
      
      /// give this IField a value
      virtual void bind(const string &s)  throw (BadValueException&) = 0;
      mutable bool isBound;

   protected:
      BaseIField(const string &name);

      string _name;
};

/**
   Template for internal Field representations.
   Template for name sytax only. All classes of this template are
   explicitly specialized.  
*/
template <typename T>
class IField
{
   private:
      IField()
      {
         // don't allow implicit instantiation
         BOOST_STATIC_ASSERT("you may use only the pre-defined classes"[0]);
      }
};

} // namespace Epoisses

#endif // BaseIField_hxx
