// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef Field_hxx
#define Field_hxx

#include "BaseField.hxx"
#include "IField.hxx"
#include "Filter.hxx"
#include "Assignment.hxx"

#include "global.h"
#include <string>
#include <loki/TypeTraits.h>

namespace Epoisses {

/**
   Not visible to user. Provides shared pointer wrapper and uniformity
   for Filters and Assignments. */
class Constant : public BaseField
{
   public:
      template <class T> friend class Field;

      explicit Constant(int v)
         : BaseField(new IConstant(v))
      {}

      // .dlb. hack to support -=
      Constant(const string& fieldName,
               int v)
         : BaseField(new IConstant(fieldName, v))
      {}

      explicit Constant(bool v)
         : BaseField(new IConstant(v))
      {}

      explicit Constant(long v)
         : BaseField(new IConstant(v))
      {}

      explicit Constant(double v)
         : BaseField(new IConstant(v))
      {}

      explicit Constant(const string &v)
         : BaseField(new IConstant(v))
      {}

      explicit Constant(const char *v)
         : BaseField(new IConstant(v))
      {}

      explicit Constant(const vector<string> &v)
         : BaseField(new IConstant(v))
      {}

      // need a serialization format...
      //Constant(const DateTime &v);

      virtual void stringify(ostream &str) const
      {
         internalField->stringify(str);
      }

      virtual void bind(const string &s) throw (BadValueException &)
      {
         internalField->bind(s);
      }
};


/*
// other Field types, each with operators that produce Filters and Assignments
class Field<DateTime>;
class Field<Date>;
class Field<Time>;
class Field<Interval>;
 -- others?
*/

template <class T>
class Field : public BaseColumnField
{
      friend class Filter;
   public:
      explicit Field(const string &name)
         : BaseColumnField(new IField<T>(name))
      {}

      Field(const Field &field)
         : BaseColumnField(field)
      {}

      virtual ~Field() {}

      // will compile only for T==BOOL
      operator Filter()
      {
         static T *test;
         static bool *test1;
         test1 = test;

         //return Filter(internalField,
         //              Filter::Self);
         return *this == true;
      }

      // will compile only for T==BOOL
      Filter operator!()
      {
         static T *test;
         static bool *test1;
         test1 = test;
         //return Filter(internalField,
         //              Filter::Not);
         return *this == false;
      }  
      
      Filter operator==(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Equal, field.internalField);
      }

      Filter operator==( const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Equal, constant.internalField);
      }

      Filter operator!=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::NotEqual, field.internalField);
      }

      Filter operator!=(const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::NotEqual, constant.internalField);
      }

      Filter operator<(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Less, field.internalField);
      }

      Filter operator<(const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Less, constant.internalField);
      }

      Filter operator>(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Greater, field.internalField);
      }

      Filter operator>(const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Greater, constant.internalField);
      }

      Filter operator<=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::LessEqual, field.internalField);
      }
      Filter operator<=(const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::LessEqual, constant.internalField);
      }

      Filter operator>=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::GreaterEqual, field.internalField);
      }
      Filter operator>=(const typename Loki::TypeTraits<T>::ParameterType val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::GreaterEqual, constant.internalField);
      }

      // !dlb! add Regexp type?
      // LIKE operator -- works on all types, strangely enough
      Filter operator*=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Like, constant.internalField);
      }

      // Assignment
      Assignment<T> operator=(const Field &field) const
      {
         return Assignment<T>(*this, field);
      }

      Assignment<T> operator=(const typename Loki::TypeTraits<T>::ParameterType v) const
      {
         return Assignment<T>(*this, v);
      }

      Assignment<T> operator-=(const typename Loki::TypeTraits<T>::ParameterType v) const
      {
         //                             dummy arg
         return Assignment<T>(*this, v, true);
      }

      /// convert bound self to Type
      T // Loki::TypeTraits<T>::ParameterType   // !dlb! this didn't work???
      operator*() const throw (NotBoundException&)
      {
         // ugh
         return static_cast<const IField<T>*>(getRawInternal())->value();
      }

   protected:
      Field(BaseIField *field)
         : BaseColumnField(field)
      {}

      friend class PSDBresult_BindIterator;
   private:
      /// give the field a new value (used by PSDBresult_BindIterator)
      virtual void bind(const string &s) throw (BadValueException&)
      {
         internalField->bind(s);
      }
};

// I got tired of fighting with the compiler over TypeTraits -- try again someday
class Field< vector<string> > : public BaseColumnField
{
      friend class Filter;
   public:
      explicit Field(const string &name)
         : BaseColumnField(new IField< vector<string> >(name))
      {}

      Field(const Field &field)
         : BaseColumnField(field)
      {}

      /// convert self to string
      virtual void stringify(ostream &str) const
      {
         internalField->stringify(str);
      }

      virtual ~Field() {}

      Filter operator==(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Equal, field.internalField);
      }

      Filter operator==(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Equal, constant.internalField);
      }

      Filter operator!=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::NotEqual, field.internalField);
      }

      Filter operator!=(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::NotEqual, constant.internalField);
      }

      Filter operator<(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Less, field.internalField);
      }

      Filter operator<(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Less, constant.internalField);
      }

      Filter operator>(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Greater, field.internalField);
      }

      Filter operator>(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Greater, constant.internalField);
      }

      Filter operator<=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::LessEqual, field.internalField);
      }
      Filter operator<=(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::LessEqual, constant.internalField);
      }

      Filter operator>=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::GreaterEqual, field.internalField);
      }
      Filter operator>=(const vector<string>& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::GreaterEqual, constant.internalField);
      }

      // !dlb! add Regexp type?
      // LIKE operator -- works on all types, strangely enough
      Filter operator*=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Like, constant.internalField);
      }

      // Assignment
      Assignment< vector<string> > operator=(const Field &field) const
      {
         return Assignment< vector<string> >(*this, field);
      }

      Assignment< vector<string> > operator=(const vector<string>& v) const
      {
         return Assignment< vector<string> >(*this, v);
      }

      /// convert bound self to Type
      const vector<string>& // Loki::TypeTraits<T>::ParameterType   // !dlb! this didn't work???
      operator*() const throw (NotBoundException&)
      {
         // ugh
         return static_cast<const IField< vector<string> >*>(getRawInternal())->value();
      }

   protected:
      Field(BaseIField *field)
         : BaseColumnField(field)
      {}

      friend class PSDBresult_BindIterator;
   private:
      /// give the field a new value (used by PSDBresult_BindIterator)
      virtual void bind(const string &s) throw (BadValueException&)
      {
         internalField->bind(s);
      }
};

// string is weird because it needs to accept static char*
// in the form of "string"
// this is exactly what the compiler would produce, except
// that templates won't match after type conversions.
// this is NOT A TEMPLATE, so type conversions work. Sigh.
// More weirdness; the TypeTrait stuff didn't work,
// probably because string is a typedef.
// Finally, the bool trickery can't be used here because
// again, this is not a template.
class Field<string> : public BaseColumnField
{
      friend class Filter;
   public:
      explicit Field(const string &name)
         : BaseColumnField(new IField<string>(name))
      {}

      Field(const Field &field)
         : BaseColumnField(field)
      {}

      /// convert self to string
      virtual void stringify(ostream &str) const
      {
         internalField->stringify(str);
      }

      virtual ~Field() {}

      Filter operator==(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Equal, field.internalField);
      }

      Filter
      operator==(const string &val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Equal, constant.internalField);
      }

      Filter operator!=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::NotEqual, field.internalField);
      }

      Filter operator!=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::NotEqual, constant.internalField);
      }

      Filter operator<(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Less, field.internalField);
      }

      Filter operator<(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Less, constant.internalField);
      }

      Filter operator>(const Field& field) const
      {
          return Filter(internalField,
                        Filter::Greater, field.internalField);
      }

      Filter operator>(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Greater, constant.internalField);
      }

      Filter operator<=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::LessEqual, field.internalField);
      }
      Filter operator<=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::LessEqual, constant.internalField);
      }

      Filter operator>=(const Field& field) const
      {
          return Filter(internalField,
                        Filter::GreaterEqual, field.internalField);
      }
      Filter operator>=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::GreaterEqual, constant.internalField);
      }

      // !dlb! add Regexp type?
      // LIKE operator -- works on all types, strangely enough
      Filter operator*=(const string& val) const
      {
         Constant constant(val);
         return Filter(internalField,
                       Filter::Like, constant.internalField);
      }

      // Assignment
      Assignment<string> operator=(const Field &field) const
      {
         return Assignment<string>(*this, field);
      }

      Assignment<string> operator=(const string& v) const
      {
         return Assignment<string>(*this, v);
      }

      /// convert bound self to Type
      const string&
      operator*() const throw (NotBoundException&)
      {
         // ugh
         return static_cast<IField<string>*>(getRawInternal())->value();
      }

   protected:
      Field(BaseIField *field)
         : BaseColumnField(field)
      {}

      friend class PSDBresult_BindIterator;
   private:
      /// give the field a new value (used by PSDBresult_BindIterator)
      virtual void bind(const string &s) throw (BadValueException&)
      {
         internalField->bind(s);
      }
};

} // namespace Epoisses

#endif // Field_hxx
