// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef FieldList_hxx
#define FieldList_hxx

#include <vector>
#include "Sptr.hxx"
#include "BaseField.hxx"
#include "PSDBCommonType.h"

namespace Epoisses {

class BaseColumnField;

/**
   An ordered collection of Fields. This class is not seen by the
   user. */
class IFieldList : public PSDBInterface
{
      friend class FieldList;
      // synonym for PSDBResult::BindIterator (see PSDBResult.hxx)
      friend class PSDBResult_BindIterator;
   public:
      void stringify(ostream &str) const;

   private:
      IFieldList(const Sptr<BaseIField> &field)
         : contents()
      {
         contents.push_back(field);
      }

      IFieldList()
         : contents()
      {}
         
      void
      add(const Sptr <BaseIField> &field)
      {
         contents.push_back(field);
      }

      typedef vector< Sptr <BaseIField> > Contents;
      
      void
      addAll(const IFieldList &fieldList)
      {
         contents.reserve(contents.size() + fieldList.contents.size());
         for (Contents::const_iterator it = fieldList.contents.begin();
              it != fieldList.contents.end();
              it++)
         {
            add(*it);
         }
      }

      vector< Sptr <BaseIField> > contents;
};

/**
   An unordered collection of non-constant Fields suitable for passing to
   ProvisioningInterface and PSDBResult::BindIterator. FieldLists can be
   contructed on the fly from Fields with operator,. The Field collection is
   destructively altered. Assignment of FieldLists results in two references to
   the same list, which could be confusing. */
class FieldList : public PSDBInterface
{
      friend class BaseField;
      friend class PSDBResult_BindIterator;
   public:
      /** Implicitly construct a FieldList from a static array of Fields.
          Field<void> means allow any type of Field. Need to know the size of
          the statically allocated array, so templatized. */
      template <int SZ>
      FieldList(const Field<void> (&arr)[SZ])
         : fieldList(new IFieldList())
      {
         fromArray(arr, SZ);
      }

      // Copy constructor
      FieldList(const FieldList &rhs)
         : fieldList(rhs.fieldList)
      {}

      // implicit conversion from Field to FieldList
      FieldList(const BaseColumnField &_field);

      /// Empty Field list for programmatic construction
      FieldList();

      typedef vector< Sptr <BaseIField> > Contents;
      
      int size() const
      {
         return fieldList->contents.size();
      }

      bool empty() const
      {
         return fieldList->contents.empty();
      }

      // add the field to contents
      FieldList add(const BaseColumnField &field);

      // add the field to contents
      FieldList addAll(const FieldList &fieldList);

      // append the field to contents, as operator
      FieldList operator,(const BaseColumnField& field);

      // concatenate field lists, as operator
      FieldList operator,(const FieldList& fieldList);
      
      virtual void stringify(ostream &str) const;

      /// Allow exercise of interactive test
      static FieldList fromList(const FieldListtype& l);

      Contents::iterator begin()
      {
         return fieldList->contents.begin();
      }
      
      Contents::iterator end()
      {
         return fieldList->contents.end();
      }

      Contents::const_iterator begin() const
      {
         return fieldList->contents.begin();
      }
      
      Contents::const_iterator end() const
      {
         return fieldList->contents.end();
      }

      static const FieldList Empty;
      
   private:
      // helper for converting a static array 
      void
      fromArray(const Field<void> arr[], int sz);

      Sptr<IFieldList> fieldList;
};

// Global operators

/// Create a FieldList from two Fields
FieldList
operator,(const BaseColumnField &field1,
          const BaseColumnField &field2);

/// prepend comma operator
FieldList
operator,(const BaseColumnField &field,
          const FieldList &fieldList);

} // namespace Epoisses

#endif // FieldList_hxx
