/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const PSInterface_cxx_Version ="$Id: PSInterface.cxx,v 1.1 2002/09/30 04:55:14 jason Exp $";

#include <unistd.h>
#include <boost/lexical_cast.hpp>

#include "PSCommon.hxx"
#include "PSInterface.hxx"
#include "Logger.hxx"
#include "Mutex.hxx"
#include "Lock.hxx"
#include "VNetworkException.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Epoisses;

const unsigned int
PSInterface::NumSenders = 5;

PSInterface::Sender::Sender(const ACE_INET_Addr& addr) 
   : _addr(addr),
     _disconnected(true)
{
}

void
PSInterface::Sender::disconnect()
{
   _disconnected = true;
   _server.close();
}

string
PSInterface::Sender::sendRequest(const string& httpMessage)
   throw (VException&)
{
   if (_disconnected)
   {
      if (_connector.connect (_server, _addr) == -1)
      {
         DebugLog(<< strerror(errno));
         DebugLog (<< "Trying to connect to " 
                   << _addr.get_host_name() << ":" << _addr.get_port_number() 
                   << " and failed");
         throw VNetworkException("Can't connect to pserver", __FILE__,__LINE__);
      }
      _disconnected = false;
   }

   // !jf! may want to timeout here - also on the receive
   if (_server.send_n (httpMessage.c_str(), httpMessage.size()) == -1)
   {
      DebugLog (<< "Failed sending to pserver: <" << httpMessage << ">");
      throw VNetworkException("Failed send to pserver", __FILE__,__LINE__);
   }

   return PSCommon::readHttpMessage(_server);
}


PSInterface::Sender*
PSInterface::checkout()
{
   _sema.acquire();
   ACE_Guard<ACE_Thread_Mutex> g(_mutex);
   Sender* sender = _senders.top();
   _senders.pop();
   //DebugLog (<< "*********  " << _senders.size() << "  *************");
   return sender;
}

void
PSInterface::checkin(PSInterface::Sender* sender)
{
   ACE_Guard<ACE_Thread_Mutex> g(_mutex);
   _senders.push(sender);
   _sema.release();
}



PSInterface::PSInterface(const char* host, const int port) 
   : _address(port, host),
     _na(_address.get_host_name(), _address.get_port_number()),
     _sema(NumSenders)
{
   InfoLog(<< "Provisioning server is " << _na );
   for (unsigned int i=0; i<NumSenders; i++)
   {
      _senders.push(new Sender(_address));
   }
}

PSInterface::~PSInterface()
{
   while (!_senders.empty())
   {
      delete _senders.top();
      _senders.pop();
   }
}

// synchronous connection
string
PSInterface::sendRequest(const string& httpMessage)
   throw (VException&)
{
   Sender* sender=0;
   string result;
   try
   {
      sender = checkout();
      result = sender->sendRequest(httpMessage);
      checkin(sender); 
   }
   catch (...)
   {
      sender->disconnect();
      checkin(sender);
      throw;
   }
   return result;
}


