// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

static const char* const ProvisioningInterface_cxx_Version = "$Id: ProvisioningInterface.cxx,v 1.1 2002/09/30 04:55:14 jason Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <strstream.h>
#include <unistd.h>
#include <errno.h>
#include <algorithm>

#include "LockHelper.hxx"
#include "Logger.hxx"
#include "NetworkAddress.h"
#include "PSIFCreator.hxx"
#include "PSListener.hxx"
#include "ProvisioningInterface.hxx"
#include "VBadDataException.hxx"
#include "VException.hxx"
#include "VMissingDataException.hxx"
#include "VNetworkException.hxx"



#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Epoisses;

ProvisioningInterface* ProvisioningInterface::myInstance = 0;

ProvisioningInterface&
ProvisioningInterface::make(const char* host,const int port)
{
   try
   {
      PSIFCreator creator;
      ProvisioningInterface& pif = creator.getPSIFInstance(host, port,
                                                           host, port,
                                                           PSCommon::HTTPXML);
      return pif;
   }
   catch (VException& e)
   {
      DebugLog (<< "Caught: " << e);
      throw;
   }
}


ProvisioningInterface::ProvisioningInterface(const char* primHostname,
                                             const int primPort) :
   myInterface(primHostname, primPort)
{

   DebugLog(<< "ProvisioningInterface constructor invoked");
    myListener = new PSListener();

#if defined(__linux__)
    //Destroy when program exits
    if (atexit(ProvisioningInterface::destroy) < 0)
    {
       // was ERR
       DebugLog(<< "Failed to register ProvisioningInterface::destroy");
    }
#endif

    char buf[256];
    gethostname(buf, 256);
    myHostName = buf;

}

void
ProvisioningInterface::destroy()
{
    delete ProvisioningInterface::myInstance;
    ProvisioningInterface::myInstance = 0;
}

ProvisioningInterface::~ProvisioningInterface()
{
    DebugLog(<< "ProvisioningInterface destructor invoked");
    delete myListener;

    // clean up the lease for each functor !dlb!
}

///
int
ProvisioningInterface::getListenPort() const
{
   return (myListener->getListenPort());
}

string
ProvisioningInterface::getHostPort()
{
   if (hostPort.empty()) 
   {
      // cache the host
      char buf[1024];
      gethostname(buf, 1024);
      
      strstream hp;
      hp << string(buf) << ":" << myListener->getListenPort() << ends;
      hp.freeze(0);
      
      hostPort = hp.str();
   }
   
   return hostPort;
}

void
ProvisioningInterface::registerDomainForUpdate(const Domain& domain,
                                               Sptr<UpdateFunctor> func) throw (VException&)
{
   static const KeySet emptyKeylist;
   // client bookkeeping
   registerUpdate(domain,
                  emptyKeylist,
                  func);
         
   // send the register to the server
   sendRegisterDomainForUpdate(domain,
                               func->getUId());
}

void
ProvisioningInterface::unregisterForUpdate(const Domain &domain,
                                           Sptr<UpdateFunctor> fun) throw (VException&)
{
   // client bookkeeping
   unregisterUpdate(domain, fun);
         
   // send the register to the server
   sendUnregisterForUpdate(domain, fun->getUId());
}

void
ProvisioningInterface::registerDomainForForeignUpdate(const Domain& foreignDomain,
                                                      const Field<void>& foreignKeyField,
                                                      const Domain& primaryDomain,
                                                      Sptr<ForeignUpdateFunctor> func)
{
   // client bookkeeping
   registerForeignUpdate(func);
         
   // send the register to the server
   sendRegisterDomainForForeignUpdate(foreignDomain,
                                      foreignKeyField.toString(),
                                      primaryDomain,
                                      func->getUId());
}

void
ProvisioningInterface::unregisterDomainForForeignUpdate(const Domain& foreignDomain,
                                                        Sptr<ForeignUpdateFunctor> fun)
{
   // client bookkeeping
   unregisterForeignUpdate(fun);
         
   // send the register to the server
   sendUnregisterForForeignUpdate(foreignDomain, fun->getUId());
}

void
ProvisioningInterface::triggerDomainUpdates(const string& domainName,
                                            const KeySet &keySet)
{
   Vocal::Threads::ReadLock lock(updatesMutex);
   
   dumpRegistration();

   // get the set of KeySetUpdates for this domain
   const list<KeySetUpdate> &kss(updatesByDomain[domainName]);

   Domain domain(domainName);
   for (list<KeySetUpdate>::const_iterator ksIt = kss.begin();
        ksIt != kss.end(); ksIt++)
   {
      const KeySetUpdate ksu(*ksIt);
      
      if (ksu.keySet.empty())
      {
         attemptUpdate(ksu.fun, keySet);
      } 
   }
}

void
ProvisioningInterface::triggerForeignDomainUpdates(const string& uid,
                                                   const KeySet& keySet)
{
   Vocal::Threads::ReadLock lock(updatesMutex);
   
   // look up the functor with the given uid
   ForeignUpdatesByUId::iterator it = foreignUpdatesByUId.find(uid);
   // on off chance that domain has no callbacks
   if (it == foreignUpdatesByUId.end())
   {
      // was ERR
      DebugLog(<< "Provisioning foreign update (#" << 
               uid << ") with no registered function (OK if rare).");
   }
   else
   {
      attemptForeignUpdate(it->second, keySet);
   }
}

void
ProvisioningInterface::attemptUpdate(Sptr<UpdateFunctor> fun,
                                     const KeySet &keySet)
{
   // keep the user code from ruining everything
   try
   {
      (*fun)(keySet);
   }
   catch (VException &ve)
   {
      // was ERR
      DebugLog(<< "Error in UpdateFunctor -- " 
               << ve.getDescription().data() 
               << " (UID=" 
               << fun->getUId() << ")");
   }
   catch(...)
   {
      DebugLog(<< "Unknown error in callback (UID=" << fun->getUId() << ")");
   }
}

void
ProvisioningInterface::attemptForeignUpdate(Sptr<ForeignUpdateFunctor> fun,
                                            const KeySet &keySet)
{
   // keep the user code from ruining everything
   try
   {
      (*fun)(keySet);
   }
   catch (VException &ve)
   {
      // was ERR
      DebugLog(<< "Error in ForeignUpdateFunctor -- "
               << ve.getDescription().data()
               << " (UID=" << fun->getUId() << ")");
   }
   catch(...)
   {
      // was ERR
      DebugLog(<< "Unknown error in ForeignUpdateFunctor UID=" << fun->getUId());
   }
}

void
ProvisioningInterface::registerUpdate(const Domain &domain,
                                      const KeySet& keySet, 
                                      Sptr<UpdateFunctor> func)
{
   list<KeySetUpdate> &kss(updatesByDomain[domain.getName()]);
   kss.push_front(KeySetUpdate(keySet, func));
}

// remove the functor from the mappings
void
ProvisioningInterface::unregisterUpdate(const Domain &domain,
                                        Sptr<UpdateFunctor> fun)
{
   Vocal::Threads::WriteLock lock(updatesMutex);

   UpdatesByDomain::iterator it = updatesByDomain.find(domain.getName());
   // on off chance that domain has no callbacks
   if (it != updatesByDomain.end())
   {
      list<KeySetUpdate> &kss(it->second);

      // remove the uid's KeySetUpdate structure from the domain's list
      for (list<KeySetUpdate>::iterator ksuIt = kss.begin();
           ksuIt != kss.end(); ksuIt++)
      {
         if ((*ksuIt->fun) == (*fun))
         {
            kss.erase(ksuIt);
            break;
         }
      }
   }
}

void
ProvisioningInterface::registerForeignUpdate(Sptr<ForeignUpdateFunctor> fun)
{
   Vocal::Threads::WriteLock lock(updatesMutex);

   DebugLog(<< "ProvisioningInterface::registerForeignUpdate(" << fun->getUId() << ")");

   // .dlb. this should compile out eventually
   ForeignUpdatesByUId::iterator it = foreignUpdatesByUId.find(fun->getUId());
   // on off chance that domain has no callbacks
   if (it != foreignUpdatesByUId.end())
   {
      // don't use a foreign update functor twice
      assert(false);
   }

   foreignUpdatesByUId[fun->getUId()] = fun;
}

// remove the functor from the mappings
void
ProvisioningInterface::unregisterForeignUpdate(Sptr<ForeignUpdateFunctor> fun)
{
   DebugLog(<< "ProvisioningInterface::unregisterForeignUpdate(" << fun->getUId() << ")");

   Vocal::Threads::WriteLock lock(updatesMutex);

   foreignUpdatesByUId.erase(fun->getUId());
}

void
ProvisioningInterface::dumpRegistration()
{
   strstream str;
   
   for (UpdatesByDomain::const_iterator regIt = updatesByDomain.begin();
        regIt != updatesByDomain.end(); regIt++)
   {
      str << "(" << regIt->first << ")\n[";
      for (list<KeySetUpdate>::const_iterator ksusIt = regIt->second.begin();
           ksusIt != regIt->second.end(); ksusIt++)
      {
         str << ksusIt->fun->getUId() << "{";
         for (vector<string>::const_iterator ksIt = ksusIt->keySet._keys.begin();
              ksIt != ksusIt->keySet._keys.end(); ksIt++)
         {
            str << *ksIt << ", ";
         }
         str << "}" << endl;
      }
      str << "]" << endl;
   }
   str << ends;
   str.freeze(0);
}
