// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#include <boost/lexical_cast.hpp>
#include <strstream>

#include "Logger.hxx"
#include "Users.hxx"
#include "VectorKeyCache.hxx"
#include "cppunit/TestCase.h"
#include "cppunit/TestCaller.h"
#include "cppunit/TestSuite.h"
#include "cppunit/TextTestRunner.h"
#include "cppunit/TextTestResult.h"
#include "ProvisioningInterface.hxx"

#define VOCAL_SUBSYSTEM Subsystem::TEST

using namespace CppUnit;
using namespace Epoisses;

class DerivedVectorKeyCacheTest : public TestCase
{
   public:

      DerivedVectorKeyCacheTest()
         : pif(ProvisioningInterface::make())
      {}
        
      virtual ~DerivedVectorKeyCacheTest()
      {}

      // sadly, cppunit makes an instance of the test on each call
      VectorKeyCache<SpecUser::Db, Feature::Db>&
      getCache()
      {
         if (cache == NULL)
         {
            cache = new VectorKeyCache<SpecUser::Db, Feature::Db>(20);
         }
         return *cache;
      }

      virtual void setUp()
      {
      }

      virtual void tearDown()
      {
      }

      void empty()
      {
         cerr << "DerivedVectorKeyCacheTest::empty " << this << endl;

         pif.deleteRecordsInDomain(specusers,
                                   Filter::None);

         pif.deleteRecordsInDomain(features,
                                   Filter::None);
      }

      void populatePrimary()
      {
         // create an assignment list from an array

         cerr << "DerivedVectorKeyCacheTest::populate " << this << endl;
         
         Assignment<string> setUserid(specusers.userid = "");
         Assignment<string> setFirstname(specusers.firstname = "");
         Assignment<string> setLastname(specusers.lastname = "");
         Assignment<bool> setExpired(specusers.expired = false);
         Assignment<int> setCount(specusers.count = 0);
         Assignment<int> setCredits(specusers.credits = 0);
         Assignment<double> setSpeed(specusers.speed = 0.0);

         AssignmentList specusersList((setUserid, setFirstname, setLastname,
                                       setExpired, setCount, setCredits, setSpeed));

         int i = 1;
         for (i = 0; i < 10; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "Berry");
            setExpired.set((bool)(i%2));
            setCount.set(i);
            setCredits.set(10*i);
            setSpeed.set(10.0*i);

            pif.addRecordInDomain(specusers,
                                  specusersList);
         }

         Assignment<string> setPermission(specusers.permission = "");
         AssignmentList otherspecusersList(specusersList.add(setPermission));
         
         for (; i < 20; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "SpecialBerry");
            setExpired.set((bool)(i%2));
            setCount.set(i);
            setCredits.set(10*i);
            setSpeed.set(10.0*i);
            setPermission.set("perm_" + boost::lexical_cast<string>(i));
            
            pif.addRecordInDomain(otherspecusers,
                                  otherspecusersList);
         }         
      }
      
      void populateForeign()
      {
         Assignment<string> setFeatureid(features.featureid = "");
         Assignment<string> setUserid(features.userid = "");
         Assignment<string> setName(features.name = "");

         AssignmentList assignments((setFeatureid, setUserid, setName));

         // every user whose userid is divisible by 3 gets a feature
         for (int i = 3; i < 20; i+=3) {
            // slam values into setters
            setFeatureid.set("f1_" + boost::lexical_cast<string>(i));
            setUserid.set(boost::lexical_cast<string>(i));
            setName.set("feature1");

            // insert the feature
            pif.addRecordInDomain(features,
                                  assignments);
         }

         // every user whose userid is divisible by 5 gets another feature
         for (int i = 5; i < 20; i+=5) {
            // slam values into setters
            setFeatureid.set("f2_" + boost::lexical_cast<string>(i));
            setUserid.set(boost::lexical_cast<string>(i));
            setName.set("feature2");

            // insert the feature
            pif.addRecordInDomain(features,
                                  assignments);
         }
      }

      // convert vector of Feature::Db::KeyType to string
      static string
      toString(const vector<Feature::Db::KeyType>& featRefs)
      {
         string s("{");
         for (vector<Feature::Db::KeyType>::const_iterator frIt = featRefs.begin();
              frIt != featRefs.end(); frIt++)
         {
            s = s + boost::lexical_cast<string>(*frIt) + ", ";
         }
         s = s + "}";
         return s;
      }

      void testFetch()
      {
         sleep(1);
         cerr << "DerivedVectorKeyCacheTest::testFetch " << this << endl;

         vector<Feature::Db::KeyType> d1 = getCache()["1"];
         vector<Feature::Db::KeyType> d3 = getCache()["3"];
         vector<Feature::Db::KeyType> d15 = getCache()["15"];

         // get everything into cache
         for (int i = 1; i < 20; i++)
         {
            getCache()[boost::lexical_cast<string>(i)];
         }

         // should have no features
         string res = toString(d1);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{}");
            
         // should have 1 feature
         res = toString(d3);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{f1_3, }");

         // should have 2 features
         res = toString(d15);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{f1_15, f2_15, }");
            
         cerr << getCache().size() << endl;
         CPPUNIT_ASSERT(getCache().size() == 19);
      }
      
      void testModify()
      {
         cerr << "DerivedVectorKeyCacheTest::testModify " << this << endl;

         pif.setRecordsInDomain(features,
                                features.featureid == "f1",
                                features.name = "modified 1");
         pif.setRecordsInDomain(features,
                                features.featureid == "f15",
                                features.name = "modified 15");
         sleep(1);
         cerr << "!! " << getCache().size() << endl;
         // should have no effect on cache
         CPPUNIT_ASSERT(getCache().size() == 19);
      }

      void testRemove()
      {
         cerr << "DerivedVectorKeyCacheTest::testRemove " << this << endl;
         cerr << getCache() << endl;
         
         pif.deleteRecordsInDomain(features,
                                   features.featureid < "f2_10");
         sleep(1);
         cerr << "!! " << getCache().size() << endl;
         CPPUNIT_ASSERT(getCache().size() == 13);
      }

      void testAdd()
      {
         cerr << "DerivedVectorKeyCacheTest::testAdd " << this << endl;
         Assignment<string> setFeatureid(features.featureid = "");
         Assignment<string> setUserid(features.userid = "");
         Assignment<string> setName(features.name = "");

         AssignmentList assignments((setFeatureid, setUserid, setName));

         // every user whose userid is divisible by 4 gets a feature
         for (int i = 4; i < 20; i+=4) {
            // slam values into setters
            setFeatureid.set("f3_" + boost::lexical_cast<string>(i));
            setUserid.set(boost::lexical_cast<string>(i));
            setName.set("feature3");

            // insert the feature
            pif.addRecordInDomain(features,
                                  assignments);
         }

         sleep(1);
         cerr << getCache();
         cerr << "!! " << getCache().size() << endl;
         CPPUNIT_ASSERT(getCache().size() == 13);
      }

      void testPostAdd()
      {
         cerr << "DerivedVectorKeyCacheTest::testPostAdd " << this << endl;
         vector<Feature::Db::KeyType> d = getCache()["12"];
         string res = toString(d);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{f3_12, }");
      }

   private:
      ProvisioningInterface& pif;
      static VectorKeyCache<SpecUser::Db, Feature::Db>* cache;
      SpecUser::Db specusers;
      OtherSpecUser::Db otherspecusers;
      Feature::Db features;
};

VectorKeyCache<SpecUser::Db, Feature::Db>* DerivedVectorKeyCacheTest::cache = NULL;

// local macros
#define cacheTEST(_method) \
   suite->addTest(new TestCaller<DerivedVectorKeyCacheTest>(#_method, \
                                                            &DerivedVectorKeyCacheTest::_method))

int main(int argc, char** argv)
{
   //cpLogSetPriority(LOG_DEBUG);

   ProvisioningInterface::make("localhost", 6005);

   TestSuite* suite = new TestSuite;
   assert(suite->countTestCases() == 0);
   
   TextTestResult result;
   cacheTEST(empty);
   cacheTEST(populatePrimary);
   cacheTEST(populateForeign);
   cacheTEST(testFetch);
   cacheTEST(testModify);
   cacheTEST(testRemove);
   cacheTEST(testAdd);
   cacheTEST(testPostAdd);

   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();
   return 0;
}

