// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#include <boost/lexical_cast.hpp>
#include <strstream>
#include <iostream>

#include "Logger.hxx"
#include "Users.hxx"
#include "VectorKeyCache.hxx"
#include "cppunit/TestCase.h"
#include "cppunit/TestCaller.h"
#include "cppunit/TestSuite.h"
#include "cppunit/TextTestRunner.h"
#include "cppunit/TextTestResult.h"
#include "ProvisioningInterface.hxx"
#include "VDbException.hxx"
#include "VectorCache.hxx"
#include "Inserter.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::TEST

using namespace CppUnit;
using namespace Epoisses;

/* Managing a mapping table; 
                       GroupMaps 
                      /         \  
                     v           v   
                  Groups      Members
*/

namespace Epoisses
{
class CacheImpl<Group> : public Cache<Group>
{
   public:
      CacheImpl<Group>()
         : Cache<Group>()
      {}

      Group operator[](const string& groupId)
      {
         IdentityFunctor fun;
         applyRead(groupId, fun);
         return fun.get();
      }
};

typedef SingletonHolder< Epoisses::CacheImpl<Group> > GroupCache;
}

// group to group relation storage
class GroupMapData
{
   public:
      typedef GroupToMemberMap Db;

      GroupMapData(const GroupToMemberMap& map)
         : memberId(*map.memberId),
           groupId(*map.groupId)
      {}

      GroupMapData()
      {}

      static string
      className()
      {
         return "GroupMapData";
      }

      string memberId;
      string groupId;

      // easier to use than global << because of nested/forward
      string
      toString() const
      {
         return ("[" + memberId + " " + groupId + "]");
      }
};

// wrapped in Singleton in real use
class GroupMapRelationCache : public Cache<GroupMapData>
{
   public:
      GroupMapRelationCache()
         : Cache<GroupMapData>()
      {}

      GroupMapData operator[](const GroupToMemberMap::KeyType& groupRelId)
      {
         IdentityFunctor fun;
         applyRead(groupRelId, fun);
         return fun.get();
      }
};

// GroupMapCache allows mapping between group member <=> group
class GroupMapCache
{
   public:
      GroupMapCache()
         : groupToMemberCache(100),
           memberToGroupCache(100)
      {
      }

      bool empty() const
      {
         return groupToMemberCache.empty();
      }

      unsigned int size() const
      {
         return groupToMemberCache.size();
      }

      vector<Group>
      getMembers(const string& groupId)
      {
         vector<Group> members;
         vector<MemberToGroupMap::KeyType> groupRelIds = memberToGroupCache[groupId];
         for (vector<MemberToGroupMap::KeyType>::const_iterator i = groupRelIds.begin();
              i != groupRelIds.end(); i++)
         {
            DebugLog(<< "relation id: " << *i);
            try 
            {
               // an indirection
               members.push_back(groupCache[groupMapRelationCache[*i].memberId]);
            }
            catch (NotFoundInProvisioningException& e)
            {
               DebugLog(<< "Caught: " << e);
               DebugLog(<< "Failed to find group " << groupId);
            }
         }
         return members;
      }

      vector<Group>
      getGroups(const string& groupId)
      {
         vector<Group> groups;
         vector<GroupToMemberMap::KeyType> groupRelIds = groupToMemberCache[groupId];
         for (vector<GroupToMemberMap::KeyType>::const_iterator i = groupRelIds.begin();
              i != groupRelIds.end(); i++)
         {
            try 
            {
               groups.push_back(groupCache[groupMapRelationCache[*i].groupId]);
            }
            catch (NotFoundInProvisioningException& e)
            {
               DebugLog(<< "Caught: " << e);
               DebugLog(<< "Failed to find group " << groupId);
            }
         }
         return groups;
      }

   private:
      friend ostream& operator<<(ostream& s, const GroupMapCache& groupMapCache);
      // shared group cache (singleton in real life)
      CacheImpl<Group> groupCache;
      // relation cache from group map => members
      VectorKeyCache<Group::Db, GroupToMemberMap> groupToMemberCache;
      // relation cache from group map => groups
      VectorKeyCache<Group::Db, MemberToGroupMap> memberToGroupCache;
      // shared relation data cache (singleton in real life)
      GroupMapRelationCache groupMapRelationCache;
};

ostream&
operator<<(ostream& s, const GroupMapCache& groupMapCache)
{
   s << "GroupMapCache[" << groupMapCache.groupCache << ", "
     <<  groupMapCache.groupToMemberCache << ", " 
     <<  groupMapCache.memberToGroupCache << "]";
   return s;
}

class GroupMapCacheTest : public TestCase
{
   public:

      GroupMapCacheTest()
         : pif(ProvisioningInterface::make())
      {}
        
      virtual ~GroupMapCacheTest()
      {}

      // sadly, cppunit makes an instance of the test on each call
      GroupMapCache&
      getCache()
      {
         if (cache == NULL)
         {
            cache = new GroupMapCache();
         }
         return *cache;
      }

      virtual void setUp()
      {
      }

      virtual void tearDown()
      {
      }

      void empty()
      {
         cerr << "GroupMapCacheTest::empty " << this << endl;

         pif.deleteRecordsInDomain(groupMaps,
                                   Filter::None);

         pif.deleteRecordsInDomain(groups,
                                   Filter::None);

      }

      void populateGroups()
      {
         // create an assignment list from an array

         cerr << "GroupMapCacheTest::populateGroups " << this << endl;
         
         Assignment<string> setGroupId(groups.groupId = "");
         Assignment<string> setBusinessGroupId(groups.companyId = "");

         AssignmentList groupsList((setGroupId, setBusinessGroupId));

         int i = 1;
         for (; i < 4; i++)
         {
            // slam values into assignments
            setGroupId.set(boost::lexical_cast<string>(i));
            setBusinessGroupId.set("BizGroup" + boost::lexical_cast<string>(i % 2));

            pif.addRecordInDomain(groups,
                                  groupsList);
         }

         // add some groups that will be members
         for (; i < 20; i++)
         {
            // slam values into assignments
            setGroupId.set(boost::lexical_cast<string>(i));
            setBusinessGroupId.set("Member");

            pif.addRecordInDomain(groups,
                                  groupsList);
            

            // add map between member and group
            pif.addRecordInDomain(groupMaps,
                                  (groupMaps.memberId = boost::lexical_cast<string>(i),
                                   groupMaps.groupId = boost::lexical_cast<string>((i % 5) + 1)));
         }

         // give some groups an additional member
         for (; i < 30; i++)
         {
            // slam values into assignments
            setGroupId.set(boost::lexical_cast<string>(i));
            setBusinessGroupId.set("Member");

            pif.addRecordInDomain(groups,
                                  groupsList);
            

            // add map between member and group
            pif.addRecordInDomain(groupMaps,
                                  (groupMaps.memberId = boost::lexical_cast<string>(i),
                                   groupMaps.groupId = boost::lexical_cast<string>((i % 5) + 1)));
         }
      }
      
      void testFetch()
      {
         cerr << "GroupMapCacheTest::testFetch " << this << endl;

         vector<Group> g1 = getCache().getGroups("1");
         vector<Group> g3 = getCache().getGroups("3");
         vector<Group> g21 = getCache().getGroups("21");

         vector<Group> m1 = getCache().getMembers("1");
         vector<Group> m3 = getCache().getMembers("3");
         vector<Group> m21 = getCache().getMembers("21");

         cerr << "!!" << getCache() << endl;
         
         cerr << "!! 1's groups = " << Inserter(g1) << endl;
         cerr << "!! 3's groups = " << Inserter(g3) << endl;
         cerr << "!! 21's groups = " << Inserter(g21) << endl;

         cerr << "!! 1's members = " << Inserter(m1) << endl;
         cerr << "!! 3's members = " << Inserter(m3) << endl;
         cerr << "!! 21's members = " << Inserter(m21) << endl;
         
         strstream resg1;
         resg1 << Inserter(g1) << ends;
         resg1.freeze(0);

         strstream resg3;
         resg3 << Inserter(g3) << ends;
         resg3.freeze(0);

         strstream resg21;
         resg21 << Inserter(g21) << ends;
         resg21.freeze(0);

         strstream resm1;
         resm1 << Inserter(m1) << ends;
         resm1.freeze(0);

         strstream resm3;
         resm3 << Inserter(m3) << ends;
         resm3.freeze(0);

         strstream resm21;
         resm21 << Inserter(m21) << ends;
         resm21.freeze(0);

         CPPUNIT_ASSERT(resg1.str() == string("[]"));
         CPPUNIT_ASSERT(resg3.str() == string("[]"));
         CPPUNIT_ASSERT(resg21.str() == string("[Group[groupId=2 companyId=BizGroup0]]"));

         CPPUNIT_ASSERT(resm1.str() == string("[Group[groupId=5 companyId=Member], Group[groupId=10 companyId=Member], Group[groupId=15 companyId=Member], Group[groupId=20 companyId=Member], Group[groupId=25 companyId=Member]]"));
         CPPUNIT_ASSERT(resm3.str() == string("[Group[groupId=7 companyId=Member], Group[groupId=12 companyId=Member], Group[groupId=17 companyId=Member], Group[groupId=22 companyId=Member], Group[groupId=27 companyId=Member]]"));
         CPPUNIT_ASSERT(resm21.str() == string("[]"));
      }
      
      void testRemove()
      {
         cerr << "GroupMapCacheTest::testRemove " << this << endl;
         pif.deleteRecordsInDomain(groups,
                                   groups.groupId > "22");
         // remove from mappings as well
         pif.deleteRecordsInDomain(groupMaps,
                                   (groupMaps.groupId > "22" ||
                                    groupMaps.memberId > "22"));
         sleep(1);

         vector<Group> m1 = getCache().getMembers("1");

         strstream resm1;
         resm1 << Inserter(m1) << ends;
         resm1.freeze(0);

         cerr << "!!" << Inserter(m1) << endl;
         CPPUNIT_ASSERT(resm1.str() == string("[Group[groupId=10 companyId=Member], Group[groupId=15 companyId=Member], Group[groupId=20 companyId=Member]]"));
      }

      void testAdd()
      {
         cerr << "GroupMapCacheTest::testAdd " << this << endl;

         // add new group member
         pif.addRecordInDomain(groups,
                               (groups.groupId = "25",
                                groups.companyId = "new member"));

         // add to two groups
         pif.addRecordInDomain(groupMaps,
                               (groupMaps.memberId = "25",
                                groupMaps.groupId = "1"));
         pif.addRecordInDomain(groupMaps,
                               (groupMaps.memberId = "25",
                                groupMaps.groupId = "2"));

         sleep(1);
         
         vector<Group> g25 = getCache().getGroups("25");
         vector<Group> m1 = getCache().getMembers("1");
         vector<Group> m2 = getCache().getMembers("2");

         strstream resg25;
         resg25 << Inserter(g25) << ends;
         resg25.freeze(0);

         strstream resm1;
         resm1 << Inserter(m1) << ends;
         resm1.freeze(0);

         strstream resm2;
         resm2 << Inserter(m2) << ends;
         resm2.freeze(0);

         cerr << "!! 25's groups: " << Inserter(g25) << endl;
         cerr << "!! 1's members: " << Inserter(m1) << endl;
         cerr << "!! 2's members: " << Inserter(m2) << endl;

         CPPUNIT_ASSERT(resg25.str() == string("[Group[groupId=1 companyId=BizGroup1], Group[groupId=2 companyId=BizGroup0]]"));
         CPPUNIT_ASSERT(resm1.str() == string("[Group[groupId=10 companyId=Member], Group[groupId=15 companyId=Member], Group[groupId=20 companyId=Member], Group[groupId=25 companyId=new member]]"));
         CPPUNIT_ASSERT(resm2.str() == string("[Group[groupId=11 companyId=Member], Group[groupId=16 companyId=Member], Group[groupId=21 companyId=Member], Group[groupId=25 companyId=new member]]"));
      }

      void testUnregister()
      {
         cerr << "GroupMapCacheTest::testunRegister " << this << endl;
         delete cache;
         cache = NULL;
      }

   private:
      ProvisioningInterface& pif;
      static GroupMapCache* cache;
      Group::Db groups;
      GroupMap groupMaps;
};

GroupMapCache* GroupMapCacheTest::cache = NULL;

// local macros
#define cacheTEST(_method) \
   suite->addTest(new TestCaller<GroupMapCacheTest>(#_method, \
                                                    &GroupMapCacheTest::_method))

int main(int argc, char** argv)
{
   //cpLogSetPriority(LOG_DEBUG);
   Log::initialize(Log::COUT, Log::DEBUG, "testGroupMapCache");
   ProvisioningInterface::make("localhost", 6005);

   TestSuite* suite = new TestSuite;
   assert(suite->countTestCases() == 0);
   
   TextTestResult result;
   cacheTEST(empty);
   cacheTEST(populateGroups);
   cacheTEST(testFetch);
   cacheTEST(testRemove);
   cacheTEST(testAdd);

   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();
   return 0;
}

