// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#include <strstream>
#include <boost/lexical_cast.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/TestCaller.h>
#include <cppunit/TestSuite.h>
#include <cppunit/TextTestRunner.h>
#include <cppunit/TextTestResult.h>

#include "CommandLineParser.hxx"
#include "Logger.hxx"
#include "PSIFCreator.hxx"
#include "ProvisioningInterface.hxx"
#include "TimeAccumulate.hxx"
#include "TimeGuard.hxx"
#include "Users.hxx"
#include "VDbException.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::TEST

using namespace CppUnit;
using namespace Vocal;
using namespace Epoisses;

#ifdef NOTES
Plan:
Add bulk update.
Add partitioned register?
#endif

/// Base for customized test case classes.
class ResultStringTestCast : public TestCase
{
   public:
      virtual ~ResultStringTestCast()
      {}
      
      virtual void setUp()
      {
         resultString.clear();
      }

      virtual void tearDown()
      {
         resultString.freeze(0);
      }

   protected:
      strstream resultString;
      
};

/// Test the generation of interface elements to the provisioning interface.
class ProvisioningInterfaceTest : public ResultStringTestCast
{
   public:
      // need not be tested -- just compiled
      void compile()
      {
      }

      void testFields()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFields" );
         
         // output the fields
         resultString << users.userid << ", "
                      << users.firstname << ", "
                      << users.lastname << ", "
                      << users.expired << ", "
                      << users.count << ", "
                      << users.credits << ", "
                      << users.speed << ", "
                      << users.aliases << ends;
         // .dlb. don't know why I have to create a string...
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "userid, firstname, lastname, expired, count, credits, speed, aliases");
      }

#if 0
      // used to compile...
      // added a hack to prevent the built-in operator, 
      void testBadAssignmentLis()
      {
         AssignmentList al = 
            // wrong type, ignored by built-in operator,
            (users.userid == "1", 
             // right type, returned by operator, then implicitly constructs AssignmentList
             users.firstname = "newJohn");  
      }
#endif

      void testAssignmentConstantConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testAssignmentConstantConstruct" );

         vector<string> aliases;
         aliases.push_back("dlb");
         aliases.push_back("lazar");
         // output the assignments
         resultString << (users.firstname = "david") << ", "
                      << (users.lastname = "Butcher") << ", "
                      << (users.expired = true) << ", "
                      << (users.count = 5) << ", "
                      << (users.credits = 15) << ", "
                      << (users.speed = 0.02) << ", "
                      << (users.aliases = aliases)  << ", "
                      << (users.count -= 17)
                      << ends;
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "firstname = 'david', lastname = 'Butcher', expired = true, count = 5, credits = 15, speed = 0.02, aliases = 'dlb,lazar', count = count - (17)");
      }
      
      void testAssignmentFieldConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testAssignmentFieldConstruct" );

         // output the assignments
         resultString << (users.firstname = Field<string>("sfield")) << ", "
                      << (users.lastname = Field<string>("sfield")) << ", "
                      << (users.expired = Field<bool>("bfield")) << ", "
                      << (users.count = Field<int>("ifield")) << ", "
                      << (users.credits = Field<int>("ifield")) << ", "
                      << (users.speed = Field<double>("dfield")) << ", "
                      << (users.aliases = Field< vector<string> >("csvfield"))
                      << ends;
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "firstname = sfield, lastname = sfield, expired = bfield, count = ifield, credits = ifield, speed = dfield, aliases = csvfield");
      }
         
      void testAssignmentArrayConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testAssignmentArrayConstruct" );

         // create an assignment list from an array
         const Assignment<void> al[] = 
            {Field<int>("ff") = 5,
             Field<string>("bar") = "baz",
             Field<bool>("expired") = false};
         
         resultString << AssignmentList(al) << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "ff = 5, bar = 'baz', expired = false, ");
      }

      void testFieldListConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFieldListConstruct" );

         resultString << (Field<int>("ff"),
                          Field<string>("bar"),
                          Field<bool>("expired")) << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "ff, bar, expired, ");
      }
      
      void testFieldListArrayConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFieldListArrayConstruct" );

         // use an array of Fields to initialize the field list
         const Field<void> fl[] = 
            {Field<int>("ff"),
             Field<string>("bar"),
             Field<bool>("expired")};

         // try an inline FieldList
         resultString << FieldList(fl) << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "ff, bar, expired, ");
      }
      
      void testFieldListAdd()
      {
         FieldList l = ((Field<int>("foo"), Field<int>("bar")));
         l.add(Field<int>("baz"));
         
         resultString << l << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "foo, bar, baz, ");
      }
      
      void testFieldListAddChain()
      {
         FieldList l = ((Field<int>("foo"), Field<int>("bar")));
         l.add(Field<int>("quux")).add(Field<int>("oof"));
         resultString << l << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "foo, bar, quux, oof, ");
      }
      
      void
      testAssignmentListAdd()
      {
         AssignmentList l = ((Field<int>("foo") = 1, Field<int>("bar") = 2));
         l.add(Field<int>("baz") = 3);
         
         resultString << l << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "foo = 1, bar = 2, baz = 3, ");

      }

      void
      testAssignmentListAddChain()
      {
         AssignmentList l = ((Field<int>("foo") = 1, Field<int>("bar") = 2));
         l.add(Field<int>("quux") = 4).add(Field<int>("oof") = 5);
         resultString << l << ends;
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "foo = 1, bar = 2, quux = 4, oof = 5, ");
      }
      
      void testFilterSimpleConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFilterSimpleConstruct" );

         resultString << (Field<int>("foo") < Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") <= Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") == Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") > Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") >= Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") != Field<int>("bar")) << ", ";
         resultString << (Field<int>("foo") *= "11*") << ", ";

         resultString << (Field<bool>("foo") < Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") <= Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") == Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") > Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") >= Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") != Field<bool>("bar")) << ", ";
         resultString << (Field<bool>("foo") *= "f*") << ", ";

         resultString << (Field<double>("foo") < Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") <= Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") == Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") > Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") >= Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") != Field<double>("bar")) << ", ";
         resultString << (Field<double>("foo") *= "17.*") << ", ";

         resultString << (Field<string>("foo") < Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") <= Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") == Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") > Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") >= Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") != Field<string>("bar")) << ", ";
         resultString << (Field<string>("foo") *= "b*r") << ", ";

         // try constants
         resultString << (Field<int>("foo") < 17) << ", ";
         resultString << (Field<int>("foo") <= 17) << ", ";
         resultString << (Field<int>("foo") == 17) << ", ";
         resultString << (Field<int>("foo") > 17) << ", ";
         resultString << (Field<int>("foo") >= 17) << ", ";
         resultString << (Field<int>("foo") != 17) << ", ";
         resultString << (Field<int>("foo") *= "1*") << ", ";

         resultString << (Field<bool>("foo") < false) << ", ";
         resultString << (Field<bool>("foo") <= false) << ", ";
         resultString << (Field<bool>("foo") == false) << ", ";
         resultString << (Field<bool>("foo") > false) << ", ";
         resultString << (Field<bool>("foo") >= false) << ", ";
         resultString << (Field<bool>("foo") != false) << ", ";
         resultString << (Field<bool>("foo") *= "fal*") << ", ";

         resultString << (Field<double>("foo") < 17.71) << ", ";
         resultString << (Field<double>("foo") <= 17.71) << ", ";
         resultString << (Field<double>("foo") == 17.71) << ", ";
         resultString << (Field<double>("foo") > 17.71) << ", ";
         resultString << (Field<double>("foo") >= 17.71) << ", ";
         resultString << (Field<double>("foo") != 17.71) << ", ";
         resultString << (Field<double>("foo") *= "1*") << ", ";

         resultString << (Field<string>("foo") < "bar") << ", ";
         resultString << (Field<string>("foo") <= "bar") << ", ";
         resultString << (Field<string>("foo") == "bar") << ", ";
         resultString << (Field<string>("foo") > "bar") << ", ";
         resultString << (Field<string>("foo") >= "bar") << ", ";
         resultString << (Field<string>("foo") != "bar") << ", ";
         resultString << (Field<string>("foo") *= "ba*") << ", ";

         vector<string> bar;
         bar.push_back("bar1");
         bar.push_back("bar2");
         bar.push_back("bar3");
            
         resultString << (Field<vector< string > >("foo") < bar) << ", ";
         resultString << (Field<vector< string > >("foo") <= bar) << ", ";
         resultString << (Field<vector< string > >("foo") == bar) << ", ";
         resultString << (Field<vector< string > >("foo") > bar) << ", ";
         resultString << (Field<vector< string > >("foo") >= bar) << ", ";
         resultString << (Field<vector< string > >("foo") != bar) << ends;

         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "(foo &lt; bar), (foo &lt;= bar), (foo = bar), (foo &gt; bar), (foo &gt;= bar), (foo != bar), (foo like '11*'), (foo &lt; bar), (foo &lt;= bar), (foo = bar), (foo &gt; bar), (foo &gt;= bar), (foo != bar), (foo like 'f*'), (foo &lt; bar), (foo &lt;= bar), (foo = bar), (foo &gt; bar), (foo &gt;= bar), (foo != bar), (foo like '17.*'), (foo &lt; bar), (foo &lt;= bar), (foo = bar), (foo &gt; bar), (foo &gt;= bar), (foo != bar), (foo like 'b*r'), (foo &lt; 17), (foo &lt;= 17), (foo = 17), (foo &gt; 17), (foo &gt;= 17), (foo != 17), (foo like '1*'), (foo &lt; false), (foo &lt;= false), (foo = false), (foo &gt; false), (foo &gt;= false), (foo != false), (foo like 'fal*'), (foo &lt; 17.71), (foo &lt;= 17.71), (foo = 17.71), (foo &gt; 17.71), (foo &gt;= 17.71), (foo != 17.71), (foo like '1*'), (foo &lt; 'bar'), (foo &lt;= 'bar'), (foo = 'bar'), (foo &gt; 'bar'), (foo &gt;= 'bar'), (foo != 'bar'), (foo like 'ba*'), (foo &lt; 'bar1,bar2,bar3'), (foo &lt;= 'bar1,bar2,bar3'), (foo = 'bar1,bar2,bar3'), (foo &gt; 'bar1,bar2,bar3'), (foo &gt;= 'bar1,bar2,bar3'), (foo != 'bar1,bar2,bar3')");
      }

      void testFilterCompositeConstruct()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFilterCompositeConstruct" );

         resultString << ((!users.expired &&
                           users.count > 0 &&
                           (users.lastname *= "Shan*")) ||
                          (users.expired &&
                           users.speed >= 0.5) ||
                          (users.expired &&
                           users.firstname == "david")) << ends;

         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "(((((expired = false) and (count &gt; 0)) and (lastname like 'Shan*')) or ((expired = true) and (speed &gt;= 0.5))) or ((expired = true) and (firstname = 'david')))");
      }

      void testFilterCompositeConstruct1()
      {
         InfoLog (<< "ProvisioningInterfaceTest::testFilterCompositeConstruct1" );

         resultString << !((!users.expired &&
                            users.count > 0 &&
                            (users.lastname *= "Shan*")) ||
                           (users.expired &&
                            users.speed >= 0.5) ||
                           (users.expired &&
                            users.firstname == "david")) << ends;

         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "(not(((((expired = false) and (count &gt; 0)) and (lastname like 'Shan*')) or ((expired = true) and (speed &gt;= 0.5))) or ((expired = true) and (firstname = 'david'))))");
      }
      
   protected:
      // hierarchy of user types
      User::Db users;
      SpecUser::Db specUsers;
      ExtraSpecUser::Db extraSpecUsers;
      OtherSpecUser::Db otherSpecUsers;

      // foreign table
      Feature::Db features;
};

/// base for ProvisioningInterface back-end
class ProvisioningInterfaceMixin
{
   public:
      ProvisioningInterfaceMixin()
         : pif(ProvisioningInterface::make("localhost", 6005))
      {
      }

   protected:
      ProvisioningInterface& pif;
};

/// Test the registration against the back-end
class ProvisioningInterfaceRegistrationTest : public ProvisioningInterfaceTest, public ProvisioningInterfaceMixin
{
   public:
      class DomainUpdator : public UpdateFunctor
      {
         public:
            DomainUpdator(const string& _message)
               : message(_message)
            {}

            virtual string
            getClassName() const
            {
               static const string tmp("ProvisioningInterfaceRegistrationTest::DomainUpdator");
               return tmp;
            }
            
            virtual void
            operator()(const KeySet& keySet)
            {
               InfoLog (<< keySet );

               // put some stuff in lastResults to allow some automated checking
               lastResults = (lastResults + message + " D(" + 
                              boost::lexical_cast<string>(keySet._keys.size()) +
                              "), ");
            }
         private:
            const string message;
      };

      class ForeignDomainUpdator : public ForeignUpdateFunctor
      {
         public:
            ForeignDomainUpdator(const string& _message)
               : message(_message)
            {}
            virtual string
            getClassName() const
            {
               static const string tmp("ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator");
               return tmp;
            }
            
            virtual void
            operator()(const KeySet& keySet)
            {
               InfoLog (<< message << " " << keySet );

               // put some stuff in lastResults to allow some automated checking
               lastResults = (lastResults + message + " F(" + 
                              boost::lexical_cast<string>(keySet._keys.size()) +
                              "), ");
            }
         private:
            const string message;
      };
      
      ProvisioningInterfaceRegistrationTest()
         : ProvisioningInterfaceMixin()
      {}

      void
      testDomainRegistration()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testDomainRegistration(#"
                  << userDomUpdator->getUId() << ")"
            );
         pif.registerDomainForUpdate(users, userDomUpdator);
      }

      void
      testDomainUnregistration()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testDomainUnregistration(#"
                  << userDomUpdator->getUId() << ")"
            );
         pif.unregisterForUpdate(users, userDomUpdator);
      }

      void
      testForeignBadDomainRegister1()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testForeignBadDomainRegister1"
            );

         string res;

         Sptr<ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator> fu = 
            new ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator("(bad1)Features->Users");

         try 
         {
            Domain bad("!bad!");
            pif.registerDomainForForeignUpdate(features, 
                                               features.getForeignKeyField(),
                                               bad,
                                               fu);
         }
         catch (VException& e)
         {
            res = e.getDescription();
         }

         CPPUNIT_ASSERT(res == "");
      }

      void
      testForeignBadDomainRegister2()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testForeignBadDomainRegister2"
            );

         string res;

         Sptr<ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator> fu = 
            new ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator("(bad2)Features->Users");

         try 
         {
            Domain bad("!bad!");
            pif.registerDomainForForeignUpdate(bad, 
                                               features.getForeignKeyField(),
                                               users,
                                               fu);
         }
         catch (VException& e)
         {
            res = e.getDescription();
         }
         
         CPPUNIT_ASSERT(res == "");
      }
         
      void
      testForeignDomainRegister()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testForeignDomainRegister(#"
                  << foreignUpdator->getUId() << ")"
            );

         pif.registerDomainForForeignUpdate(features, 
                                            features.getForeignKeyField(),
                                            users,
                                            foreignUpdator);
      }

      void
      testForeignDomainUnregister()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testForeignDomainUnregister(#"
                  << foreignUpdator->getUId() << ")"
            );
         pif.unregisterDomainForForeignUpdate(features, foreignUpdator);
      }

      void
      testInsertReturn()
      {
         pif.deleteRecordsInDomain(users, Filter::None);
         
         Sptr<PSDBResult> ires = 
            pif.addRecordInDomain(users,
                                  (users.firstname = "david"));
         
         // extract added id
         ires->begin(users.userid);
         string id = *users.userid;

         // get id by other means
         Sptr<PSDBResult> res = pif.getRecordsInDomain(users,
                                                       users.firstname == "david",
                                                       (users.userid));
         res->begin(users.userid);
         
         CPPUNIT_ASSERT(id == *users.userid);
         
         pif.deleteRecordsInDomain(users, Filter::None);
      }

      void
      testForeignDomainPopulate()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testForeignDomainPopulate" );

         // get users ids to give features
         Sptr<PSDBResult> userids = pif.getRecordsInDomain(users,
                                                           Filter::None,
                                                           users.userid);
         // all users get mapping to feature1
         for (PSDBResult::BindIterator bind = userids->begin(users.userid);
              bind != userids->end(); bind++)
         {
            pif.addRecordInDomain(features,
                                  (features.userid = *users.userid,
                                   features.name = "feature1"));
         }

         // grab three users
         string userid1;
         string userid2;
         string userid3;
         PSDBResult::BindIterator bind = userids->begin(users.userid);
         userid1 = *users.userid;
         bind++;
         userid2 = *users.userid;
         bind++;
         userid3 = *users.userid;

         // some users get additional features
         pif.addRecordInDomain(features,
                               (features.userid = userid1,
                                features.name = "feature2"));

         pif.addRecordInDomain(features,
                               (features.userid = userid2,
                                features.name = "feature2"));

         pif.addRecordInDomain(features,
                               (features.userid = userid3,
                                features.name = "feature2"));
         pif.addRecordInDomain(features,
                               (features.userid = userid3,
                                features.name = "feature3"));
         pif.addRecordInDomain(features,
                               (features.userid = userid3,
                                features.name = "feature4"));
         
         // look at updates' side effects
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), Features->Users F(1), ");
      }

      void
      testInsertForeign()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testInsertForeign" );

         // get users ids to give feature
         Sptr<PSDBResult> userids = pif.getRecordsInDomain(users,
                                                           Filter::None,
                                                           users.userid);

         // grab one user
         string userid1;
         PSDBResult::BindIterator bind = userids->begin(users.userid);
         userid1 = *users.userid;

         // this user gets an additional feature
         pif.addRecordInDomain(features,
                               (features.userid = userid1,
                                features.name = "feature2"));

         // check updates' side effects
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Features->Users F(1), ");
      }

      void
      testRemoveForeign1()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testRemoveForeign1" );

         // remove all instances of feature 2
         pif.deleteRecordsInDomain(features, 
                                   features.name == "feature2");

         // check updates' side effects
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Features->Users F(3), ");
      }

      void
      testRemoveForeign2()
      {
         InfoLog (<< "ProvisioningInterfaceRegistrationTest::testRemoveForeign2" );

         // remove all instances of features 3 and 4
         pif.deleteRecordsInDomain(features, 
                                   (features.name == "feature3" ||
                                    features.name == "feature4"));

         // check updates' side effects
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Features->Users F(1), ");
      }

      static string getLastResults()
      {
         sleep(1); // make sure callbacks get processed
         const string resultString(lastResults);

         lastResults = "";
         return resultString;
      }
      
   private:
      static Sptr<DomainUpdator> userDomUpdator;

      static Sptr<ForeignDomainUpdator> foreignUpdator;
      // mechanism for testabilty
      static string lastResults;
};

string ProvisioningInterfaceRegistrationTest::lastResults;
Sptr<ProvisioningInterfaceRegistrationTest::DomainUpdator>
ProvisioningInterfaceRegistrationTest::userDomUpdator =
new ProvisioningInterfaceRegistrationTest::DomainUpdator("Users");

Sptr<ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator>
ProvisioningInterfaceRegistrationTest::foreignUpdator = 
new ProvisioningInterfaceRegistrationTest::ForeignDomainUpdator("Features->Users");

/// Test use of the database actively
class ProvisioningInterfaceUseTest : public ProvisioningInterfaceTest, public ProvisioningInterfaceMixin
{
   public:
      ProvisioningInterfaceUseTest()
         : ProvisioningInterfaceMixin()
      {}
      
      void
      testEmpty()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testEmpty" );

         pif.deleteRecordsInDomain(users,
                                   Filter::None);

         // clean up foreign domains too
         pif.deleteRecordsInDomain(features,
                                   Filter::None);

         resultString << retrieveAll();
         InfoLog (<< "empty: " << resultString.str() );
         
         CPPUNIT_ASSERT(string(resultString.str()) ==  "");
      }

      void
      testRetrieveNoFilterNoContent()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveNoFilterNoContent" );
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          FieldList::Empty);
         for (PSDBResult::BindIterator binder = result->begin(FieldList::Empty);
              binder != result->end(); binder++)
         {
            // should not iterate
            CPPUNIT_ASSERT(false);
         }
         
         CPPUNIT_ASSERT(result->size() == 0);
      }

      void
      testNoSuchDomain()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testNoSuchDomain" );
         try {
            pif.getRecordsInDomain(Domain("nosuchdomain"),
                                   Filter::None,
                                   Field<string>("key"));
         } 
         catch (VDbException& vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "VDbException: VDbException: DBSelect Error: PostgreSQL error: ERROR:  Relation \"nosuchdomain\" does not exist\n"
                        "(in SELECT key FROM nosuchdomain )");
      }

      void
      testBadDomain()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testBadDomain" );
         try {
            pif.getRecordsInDomain(Domain("no such @$^#@$ domain"),
                                   Filter::None,
                                   Field<string>("key"));
         }
         catch (VDbException& vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }

         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "VDbException: VDbException: DBSelect Error: PostgreSQL error: ERROR:  parser: parse error at or near \"@$^#@$\"\n"
                        "(in SELECT key FROM no such @$^#@$ domain )");
      }

      void
      testPopulate()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testPopulate" );

         try 
         {
            TimeGuard<StandardErrorOutput> timeGuard("ProvisioningInterfaceUseTest::testPopulate");
            vector<string> noAliases;
            vector<string> aliases1;
            vector<string> aliases2;

            aliases1.push_back("alias1");

            aliases2.push_back("a.1"); // make sure odd characters parse
            aliases2.push_back("a.2");
            aliases2.push_back("a.3");

            pif.addRecordInDomain(users,
                                  (users.firstname = "david",
                                   users.lastname = "Butcher",
                                   users.expired = true,
                                   users.count = 7,
                                   users.credits = 1000,
                                   users.speed = 0.02,
                                   users.aliases = aliases2));

            sleep(2);

            pif.addRecordInDomain(users,
                                  (users.firstname = "Enlai",
                                   users.lastname = "Chu",
                                   users.expired = false,
                                   users.count = 19,
                                   users.credits = 10000,
                                   users.speed = 1.0,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Wei",
                                   users.lastname = "Lee",
                                   users.expired = false,
                                   users.count = 27,
                                   users.credits = 12000,
                                   users.speed = 1.02,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Jason",
                                   users.lastname = "Fischl",
                                   users.expired = true,
                                   users.count = 71,
                                   users.credits = 10000,
                                   users.speed = 1.05,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Ludo",
                                   users.lastname = "Pierre",
                                   users.expired = false,
                                   users.count = 37,
                                   users.credits = 10001,
                                   users.speed = 0.2,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.userid = "SpeedyId",
                                   users.firstname = "Speedy",
                                   users.lastname = "Shannon",
                                   users.expired = false,
                                   users.count = 0,
                                   users.credits = 0,
                                   users.speed = 0.00,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Belle",
                                   users.lastname = "Shannon",
                                   users.expired = false,
                                   users.count = 7,
                                   users.credits = 10000,
                                   users.speed = 2.00,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Oort",
                                   users.lastname = "Shannon",
                                   users.expired = true,
                                   users.count = 17,
                                   users.credits = 0,
                                   users.speed = 2.02,
                                   users.aliases = aliases2));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Isaac",
                                   users.lastname = "Shannon",
                                   users.expired = true,
                                   users.count = 77,
                                   users.credits = 1000,
                                   users.speed = 1.6,
                                   users.aliases = aliases1));
            pif.addRecordInDomain(users,
                                  (users.firstname = "Sam",
                                   users.lastname = "Shannon",
                                   users.expired = true,
                                   users.count = 777,
                                   users.credits = 1000,
                                   users.speed = 1.82,
                                   users.aliases = aliases2));

            resultString << retrieveAll();
            InfoLog (<< "!! " << resultString.str() );
            CPPUNIT_ASSERT(string(resultString.str()) == 
                           "david Butcher 1 7 1000 0.02\n"
                           "Enlai Chu 0 19 10000 1\n"
                           "Wei Lee 0 27 12000 1.02\n"
                           "Jason Fischl 1 71 10000 1.05\n"
                           "Ludo Pierre 0 37 10001 0.2\n"
                           "Speedy Shannon 0 0 0 0\n"
                           "Belle Shannon 0 7 10000 2\n"
                           "Oort Shannon 1 17 0 2.02\n"
                           "Isaac Shannon 1 77 1000 1.6\n"
                           "Sam Shannon 1 777 1000 1.82\n");
         }
         catch (VException& e)
         {
            DebugLog(<< "Caught: " << e.getDescription());
         }

         // also look at call back results
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), ");
      }

      void
      testNoSuchColumn()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testNoSuchColumn" );
         try
         {
            pif.getRecordsInDomain(users, 
                                   Filter::None,
                                   Field<string>("foobar"));
         }
         catch (VDbException &vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "VDbException: VDbException: DBSelect Error: PostgreSQL error: ERROR:  Attribute 'foobar' not found\n"
                        "(in SELECT foobar FROM testusers )");
      }

      void
      testTypeMismatch()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testTypeMismatch" );
         try
         {
            pif.setRecordsInDomain(users, 
                                   Filter::None,
                                   Field<string>("expired") = "not a boolean");
         }
         catch (VDbException &vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "VDbException: VDbException: DBUpdate Error: PostgreSQL error: "
                        "ERROR:  Bad boolean external representation 'not a boolean'\n"
                        "(in UPDATE testusers SET expired = 'not a boolean' )");
      }

      /// try to insert an existing unique key
      void 
      testBadInsert()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testBadInsert" );

         // get a key
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          users.userid);
         PSDBResult::BindIterator binder = result->begin(users.userid);

         try
         {
            pif.addRecordInDomain(users,
                                  users.userid = *users.userid);
         }
         catch (VDbException &vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }

         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "VDbException: VDbException: DBInsert Error: PostgreSQL error: "
                        "ERROR:  Cannot insert a duplicate key into unique index testusers_pkey\n"
                        "(in INSERT INTO testusers ( userid  ) VALUES ( '" + boost::lexical_cast<string>(*users.userid) + "'  ) )");
      }      

      void
      testCount()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testCount" );
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          FieldList::Empty);
         for (PSDBResult::BindIterator binder = result->begin(FieldList::Empty);
              binder != result->end(); binder++)
         {
         }

         CPPUNIT_ASSERT(result->getNumMatches() == 10);
      }

      void
      testRetrieveBool()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveBool" );

         FieldList retrieve((users.firstname, 
                             users.lastname,
                             users.expired,
                             users.speed));

         FieldList output((users.firstname, 
                           users.expired,
                           users.speed));

         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          !users.expired,
                                                          output);

         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(output);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.expired << " "
                         << *users.speed 
                         << endl;
         }
      
         resultString << ends;
         
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Enlai 0 1\n"
                        "Wei 0 1.02\n"
                        "Ludo 0 0.2\n"
                        "Speedy 0 0\n"
                        "Belle 0 2\n");
      }
      
      void
      testRetrieveString()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveString" );
         
         FieldList retrieve((users.firstname, 
                             users.lastname,
                             users.expired,
                             users.speed));
         
         FieldList output((users.firstname, 
                           users.expired,
                           users.speed));

         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          users.lastname == "Shannon",
                                                          retrieve);

         // retrieve some results and output
         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(output);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.expired << " "
                         << *users.speed
                         << endl;
         }
         resultString << ends;
   
         InfoLog (<< "!! " << resultString.str() );
         InfoLog (<< "newline\nenilwen");
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Speedy 0 0\n"
                        "Belle 0 2\n"
                        "Oort 1 2.02\n"
                        "Isaac 1 1.6\n"
                        "Sam 1 1.82\n");
      }

      void
      testRetrieveStringSorted()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveStringSorted" );
         
         FieldList retrieve((users.firstname, 
                             users.lastname,
                             users.expired,
                             users.speed));

         FieldList output((users.firstname, 
                           users.expired,
                           users.speed));

         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          users.lastname == "Shannon",
                                                          retrieve,
                                                          users.speed);

         // retrieve some results and output
         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(output);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.expired << " "
                         << *users.speed 
                         << endl;
         }
         resultString << ends;

         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Speedy 0 0\n"
                        "Isaac 1 1.6\n"
                        "Sam 1 1.82\n"
                        "Belle 0 2\n"
                        "Oort 1 2.02\n");
      }

      void
      testRetrieveVectorString()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveVectorString" );

         FieldList retrieve((users.firstname, 
                             users.lastname,
                             users.aliases));

         FieldList output((users.firstname, 
                           users.aliases));

         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          users.lastname == "Shannon",
                                                          retrieve);

         // retrieve some results and output
         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(output);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " [";
            for(vector<string>::const_iterator it = (*users.aliases).begin();
                it != (*users.aliases).end(); it++)
            {
               resultString << *it << " ";
            }
            resultString << "]" << endl;
         }
         resultString << ends;
         
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Speedy [alias1 ]\n"
                        "Belle [alias1 ]\n"
                        "Oort [a.1 a.2 a.3 ]\n"
                        "Isaac [alias1 ]\n"
                        "Sam [a.1 a.2 a.3 ]\n");
      }
      
      void
      testEmptyRetrieve()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testEmptyRetrieve" );
         
         FieldList all = ((users.userid,
                           users.firstname,
                           users.lastname,
                           users.expired,
                           users.count,
                           users.credits,
                           users.speed));
         
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          (users.count > 1000 &&
                                                           users.speed > 5.0),
                                                          all);
         
         // retrieve some results and output
         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(all);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.count << " "
                         << *users.speed << endl;
         }
         resultString << ends;
         
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "");
      }
      
      void
      testRetrieveComposite()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveComposite" );

         FieldList all = ((users.userid,
                           users.firstname,
                           users.lastname,
                           users.expired,
                           users.count,
                           users.credits,
                           users.speed));

         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          (users.count > 10 &&
                                                           users.speed > 1.0),
                                                          all);

         // retrieve some results and output
         // output the matches
         for (PSDBResult::BindIterator binder = result->begin(all);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.count << " "
                         << *users.speed << endl;
         }
         resultString << ends;
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Wei 27 1.02\n"
                        "Jason 71 1.05\n"
                        "Oort 17 2.02\n"
                        "Isaac 77 1.6\n"
                        "Sam 777 1.82\n");
      }
      
      void
      testRetrieveNoFilter()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRetrieveNoFilter" );
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          FieldList::Empty);
         
         unsigned int i = 0;
         for (PSDBResult::BindIterator binder = result->begin(FieldList::Empty);
              binder != result->end(); binder++)
         {
            i++;
         }
         
         CPPUNIT_ASSERT(result->size() == 10);
         CPPUNIT_ASSERT(result->size() == i);
      }
      
      void testBadBindIterator()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testBadBindIterator" );
         
         FieldList all = ((users.firstname,  // leave out userid -- not stable
                           users.lastname,
                           users.expired,
                           users.count,
                           users.credits,
                           users.speed));
         
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          all);
         
         FieldList unrelated = ((Field<int>("foo"),
                                 Field<string>("bar"),
                                 Field<string>("baz")));
         try
         {
            for (PSDBResult::BindIterator binder = result->begin(unrelated);
                 binder != result->end(); binder++)
            {
               CPPUNIT_ASSERT(false);
               InfoLog (<< *users.firstname );
            }
         } 
         catch (NoSuchFieldException &e)
         {
         }
      }
      
      void
      testModify()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testModify" );
         
         AssignmentList al = ((users.count = 123,
                               users.credits = 10000,
                               users.speed = 0.5));
         Filter filter = (users.lastname != "Shannon" &&
                          users.count < 15);
         
         pif.setRecordsInDomain(users,
                                filter,
                                al);

         resultString << retrieveAll();
         InfoLog (<< "!!" << resultString.str() );
         
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Enlai Chu 0 19 10000 1\n"
                        "Wei Lee 0 27 12000 1.02\n"
                        "Jason Fischl 1 71 10000 1.05\n"
                        "Ludo Pierre 0 37 10001 0.2\n"
                        "Speedy Shannon 0 0 0 0\n"
                        "Belle Shannon 0 7 10000 2\n"
                        "Oort Shannon 1 17 0 2.02\n"
                        "Isaac Shannon 1 77 1000 1.6\n"
                        "Sam Shannon 1 777 1000 1.82\n"
                        "david Butcher 1 123 10000 0.5\n");
         // also look at call back results
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Users D(1), ");
      }

      void
      testRemove()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRemove" );

         Filter filter = users.lastname == "Shannon";
         
         pif.deleteRecordsInDomain(users,
                                   filter);

         resultString << retrieveAll();
         CPPUNIT_ASSERT(string(resultString.str()) == 
                        "Enlai Chu 0 19 10000 1\n"
                        "Wei Lee 0 27 12000 1.02\n"
                        "Jason Fischl 1 71 10000 1.05\n"
                        "Ludo Pierre 0 37 10001 0.2\n"
                        "david Butcher 1 123 10000 0.5\n");

         // also look at call back results
         const string res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res);
         CPPUNIT_ASSERT(res == "Users D(5), ");
      }
      
      string
      retrieveAll()
      {
         FieldList all = ((users.firstname,  // leave out userid -- not stable
                           users.lastname,
                           users.expired,
                           users.count,
                           users.credits,
                           users.speed));
         
         Sptr<PSDBResult> result = pif.getRecordsInDomain(users,
                                                          Filter::None,
                                                          all);
         
         for (PSDBResult::BindIterator binder = result->begin(all);
              binder != result->end(); binder++)
         {
            resultString << *users.firstname << " "
                         << *users.lastname << " "
                         << *users.expired << " "
                         << *users.count << " "
                         << *users.credits << " "
                         << *users.speed << endl;
         }
         resultString << ends;
         
         return resultString.str();
      }
};

/// test registration with inheritance
class ProvisioningInterfaceInheritanceRegistrationTest : public ProvisioningInterfaceRegistrationTest
{
   public:
      class DomainUpdator : public UpdateFunctor
      {
         public:
            DomainUpdator(const string& _message)
               : message(_message)
            {}
            
            virtual string
            getClassName() const
            {
               static const string tmp("ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator");
               return tmp;
            }
            
            virtual void
            operator()(const KeySet& keySet)
            {
               InfoLog (<< message << " " << keySet );

               // put some stuff in lastResults to allow some automated checking
               lastResults = (lastResults + message + " D(" +
                              boost::lexical_cast<string>(keySet._keys.size()) +
                              "), ");
            }
         private:
            const string message;
      };

      ProvisioningInterfaceInheritanceRegistrationTest()
         : ProvisioningInterfaceRegistrationTest()
      {}

      void
      testPopulate()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testPopulate" );

         // add some records to each of the derived tables
         pif.addRecordInDomain(specusers,
                               (specusers.firstname = "First",
                                specusers.lastname = "SpecUser",
                                specusers.expired = true,
                                specusers.count = 7,
                                specusers.credits = 1000,
                                specusers.speed = 0.02));
         pif.addRecordInDomain(specusers,
                               (specusers.firstname = "Second",
                                specusers.lastname = "SpecUser",
                                specusers.expired = false,
                                specusers.count = 19,
                                specusers.credits = 10000,
                                specusers.speed = 1.0));

         pif.addRecordInDomain(extraspecusers,
                               (extraspecusers.firstname = "First",
                                extraspecusers.lastname = "ExtraSpecUser",
                                extraspecusers.expired = false,
                                extraspecusers.count = 27,
                                extraspecusers.credits = 12000,
                                extraspecusers.speed = 1.02));
         pif.addRecordInDomain(extraspecusers,
                               (extraspecusers.firstname = "Second",
                                extraspecusers.lastname = "ExtraSpecUser",
                                extraspecusers.expired = true,
                                extraspecusers.count = 71,
                                extraspecusers.credits = 10000,
                                extraspecusers.speed = 1.05));

         pif.addRecordInDomain(otherspecusers,
                               (otherspecusers.firstname = "First",
                                otherspecusers.lastname = "OtherSpecUser",
                                otherspecusers.expired = false,
                                otherspecusers.count = 37,
                                otherspecusers.credits = 10001,
                                otherspecusers.speed = 0.2));
         pif.addRecordInDomain(otherspecusers,
                               (otherspecusers.firstname = "Second",
                                otherspecusers.lastname = "OtherSpecUser",
                                otherspecusers.expired = false,
                                otherspecusers.count = 0,
                                otherspecusers.credits = 0,
                                otherspecusers.speed = 0.00));

         // also look at call back results
         string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "");
         res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), Users D(1), ");
      }

      // helper to register one record from each derive domain, 
      // and each derived domain
      void registerHelper(const User::Db& u,
                          Sptr<DomainUpdator> domUpdator)
      {
         pif.registerDomainForUpdate(u, domUpdator);
      }

      void testRegisterNoSuchDomain()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRegisterNoSuchDomain" );
         try
         {
            pif.registerDomainForUpdate(Domain("NoSuchDomain"), userDomUpdator);
         }
         catch (VDbException& vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == "VDbException: RegisterDomainForUpdate failed for domain 'nosuchdomain'");
      }

      void
      testRegister()
      {
         // get a key from each derived domain
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testRegister" );

         registerHelper(specusers, specUserDomUpdator);
         registerHelper(extraspecusers, extraSpecUserDomUpdator);
         registerHelper(otherspecusers, otherSpecUserDomUpdator);
      }

      void
      testDeregister()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testDeregister" );

         pif.unregisterForUpdate(users, userDomUpdator);
         pif.unregisterForUpdate(specusers, specUserDomUpdator);
         pif.unregisterForUpdate(extraspecusers, extraSpecUserDomUpdator);
         pif.unregisterForUpdate(otherspecusers, otherSpecUserDomUpdator);
      }

      void testDeregisterAgain()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testDeregisterAgain" );

         try
         {
            pif.unregisterForUpdate(users, userDomUpdator);
            pif.unregisterForUpdate(specusers, specUserDomUpdator);
            pif.unregisterForUpdate(extraspecusers, extraSpecUserDomUpdator);
            pif.unregisterForUpdate(otherspecusers, otherSpecUserDomUpdator);

            CPPUNIT_ASSERT(false);
         }
         catch (VDbException &e)
         {
         }
      }

      void
      testDeregisterNoSuchDomain()
      {
         InfoLog (<< "ProvisioningInterfaceUseTest::testRegisterNoSuchDomain" );
         try
         {
            pif.unregisterForUpdate(Domain("NoSuchDomain"), userDomUpdator);
         }
         catch (VDbException& vdbe)
         {
            resultString << vdbe.getDescription() << ends;
         }
         InfoLog (<< "!! " << resultString.str() );
         CPPUNIT_ASSERT(string(resultString.str()) == "VDbException: UnregisterDomainForUpdate failed for domain 'nosuchdomain'");
      }

      /// modify records that are not derived
      void testModifyUsersOnly()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testModifyUsersOnly" );

         AssignmentList al = ((users.count = 123,
                               users.credits = 10000,
                               users.speed = 0.5));
         Filter filter = (users.firstname != "First" &&
                          users.firstname != "Second");
         
         pif.setRecordsInDomain(users,
                                filter,
                                al);

       
         // also look at call back results
         string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "");
         res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Users D(5), ");
      }

      /// modify records that are derived
      void testModifyUsers()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testModifyUsers" );

         AssignmentList al = ((users.count = 123,
                               users.credits = 10000,
                               users.speed = 0.5));
         Filter filter = (users.firstname == "First" ||
                          users.firstname == "Second");
         
         pif.setRecordsInDomain(users,
                                filter,
                                al);
         // also look at call back results
         sleep(1);
         string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "SpecUsers D(6), ExtraSpecUsers D(2), OtherSpecUsers D(2), ");
         res = ProvisioningInterfaceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "Users D(6), ");
      }

      /// modify records that are specusers as users
      void testModifySpecUsers()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testModifySpecUsers" );

         AssignmentList al = ((specusers.count = 123,
                               specusers.credits = 10000,
                               specusers.speed = 0.5));
         Filter filter = (specusers.lastname == "SpecUser");
         
         pif.setRecordsInDomain(specusers,
                                filter,
                                al);
         // also look at call back results
         const string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "SpecUsers D(2), ");
      }

      /// modify records that are extraspecusers as users
      void testModifyExtraSpecUsers()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testModifyExtraSpecUsers" );

         AssignmentList al = ((extraspecusers.count = 123,
                               extraspecusers.credits = 10000,
                               extraspecusers.speed = 0.5));
         Filter filter = (extraspecusers.lastname == "ExtraSpecUser");
         
         pif.setRecordsInDomain(extraspecusers,
                                filter,
                                al);
         // also look at call back results
         const string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         
         CPPUNIT_ASSERT(res == "SpecUsers D(2), ExtraSpecUsers D(2), ");
      }
      
      /// modify records that are otherspecusers as users
      void testModifyOtherSpecUsers()
      {
         InfoLog (<< "ProvisioningInterfaceInheritanceRegistrationTest::testModifyOtherSpecUsers" );

         AssignmentList al = ((otherspecusers.count = 123,
                               otherspecusers.credits = 10000,
                               otherspecusers.speed = 0.5));
         Filter filter = (otherspecusers.lastname == "OtherSpecUser");
         
         pif.setRecordsInDomain(otherspecusers,
                                filter,
                                al);
         // also look at call back results
         const string res = ProvisioningInterfaceInheritanceRegistrationTest::getLastResults();
         InfoLog (<< "!! " << res );
         CPPUNIT_ASSERT(res == "SpecUsers D(2), OtherSpecUsers D(2), ");
      }

   private:      

      static string getLastResults()
      {
         sleep(1); // make sure callbacks get processed
         const string resultString(lastResults);

         lastResults = "";
         return resultString;
      }

      User::Db users;
      SpecUser::Db specusers;
      ExtraSpecUser::Db extraspecusers;
      OtherSpecUser::Db otherspecusers;
      
      static Sptr<DomainUpdator> userDomUpdator;

      static Sptr<DomainUpdator> specUserDomUpdator;

      static Sptr<DomainUpdator> extraSpecUserDomUpdator;

      static Sptr<DomainUpdator> otherSpecUserDomUpdator;

      static string lastResults;
};

string 
ProvisioningInterfaceInheritanceRegistrationTest::lastResults;

Sptr<ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator>
ProvisioningInterfaceInheritanceRegistrationTest::userDomUpdator =
new ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator("Users");

Sptr<ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator>
ProvisioningInterfaceInheritanceRegistrationTest::specUserDomUpdator =
new ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator("SpecUsers");

Sptr<ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator>
ProvisioningInterfaceInheritanceRegistrationTest::extraSpecUserDomUpdator =
new ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator("ExtraSpecUsers");

Sptr<ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator>
ProvisioningInterfaceInheritanceRegistrationTest::otherSpecUserDomUpdator =
new ProvisioningInterfaceInheritanceRegistrationTest::DomainUpdator("OtherSpecUsers");

// local macros
#define piTEST(_method) \
   suite->addTest(new TestCaller<ProvisioningInterfaceTest>(#_method, \
                                                            &ProvisioningInterfaceTest::_method))

#define pirTEST(_method) \
   suite->addTest(new TestCaller<ProvisioningInterfaceRegistrationTest>(#_method, \
                                                                        &ProvisioningInterfaceRegistrationTest::_method))

#define piuTEST(_method) \
   suite->addTest(new TestCaller<ProvisioningInterfaceUseTest>(#_method, \
                                                               &ProvisioningInterfaceUseTest::_method))

#define pi_inhTEST(_method) \
   suite->addTest(new TestCaller<ProvisioningInterfaceInheritanceRegistrationTest>(#_method, \
                                                                                   &ProvisioningInterfaceInheritanceRegistrationTest::_method))

int
main(int argc, char** argv)
{ 
   CommandLineParser args(argc, argv);
   Log::initialize(args.logType(), args.logLevel(), argv[0]);

   DebugLog(<< "example output more");

   TestSuite* suite = new TestSuite;
   assert(suite->countTestCases() == 0);
   
   TextTestResult result;
#if 1
   // interface elements
   piTEST(testFields);
   piTEST(testAssignmentConstantConstruct);
   piTEST(testAssignmentFieldConstruct);
   piTEST(testAssignmentArrayConstruct);
   piTEST(testFieldListConstruct);
   piTEST(testFieldListArrayConstruct);
   piTEST(testFieldListAdd);
   piTEST(testFieldListAddChain);
   piTEST(testAssignmentListAdd);
   piTEST(testAssignmentListAddChain);
   piTEST(testFilterSimpleConstruct);
   piTEST(testFilterCompositeConstruct);
   piTEST(testFilterCompositeConstruct1);

   // back-end invocation
   piuTEST(testEmpty);
   piuTEST(testRetrieveNoFilterNoContent);

   pirTEST(testInsertReturn);
   
   // test registration set-up and tear down
   pirTEST(testDomainRegistration);
   pirTEST(testDomainUnregistration);
   pirTEST(testDomainRegistration);

   // back-end invocation
   piuTEST(testPopulate);

   // test backend errors
   piuTEST(testNoSuchDomain);
   piuTEST(testBadDomain);
   piuTEST(testNoSuchColumn);
   piuTEST(testTypeMismatch);
   piuTEST(testBadInsert);

   piuTEST(testCount);
   piuTEST(testRetrieveBool);
   piuTEST(testRetrieveString);
   piuTEST(testRetrieveStringSorted);
   piuTEST(testRetrieveVectorString);
   piuTEST(testEmptyRetrieve);
   piuTEST(testRetrieveComposite);
   piuTEST(testRetrieveNoFilter);

   //pirTEST(testForeignBadDomainRegister1); // correctly aborts
   //pirTEST(testForeignBadDomainRegister2); // correctly aborts

   // set up foreign registration with read ids
   pirTEST(testForeignDomainRegister);
   pirTEST(testForeignDomainPopulate);

   pirTEST(testInsertForeign);
   pirTEST(testRemoveForeign1);
   pirTEST(testRemoveForeign2);

   piuTEST(testModify);
   piuTEST(testBadBindIterator);
   piuTEST(testRemove);

   // test inheritance callbacks
   pi_inhTEST(testPopulate);
   pi_inhTEST(testRegisterNoSuchDomain);   
   pi_inhTEST(testRegister);
   pi_inhTEST(testModifyUsersOnly);
   pi_inhTEST(testModifyUsers);
   pi_inhTEST(testModifySpecUsers);
   pi_inhTEST(testModifyExtraSpecUsers);
   pi_inhTEST(testModifyOtherSpecUsers);
   pi_inhTEST(testDeregister);
   pi_inhTEST(testDeregisterAgain);
   pi_inhTEST(testDeregisterNoSuchDomain);
   pirTEST(testDomainUnregistration);
   pirTEST(testForeignDomainUnregister);
   pirTEST(testInsertReturn);
#else
   piuTEST(testEmpty);
   pirTEST(testDomainRegistration);
   piuTEST(testPopulate);
   pirTEST(testDomainUnregistration);
#endif

   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();

   TimeAccumulate::dump();
   return 0;
}

#undef piTEST
#undef pirTEST
#undef piuTEST
#undef pi_inhTEST
