// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#include "LockedStreamMap.hxx"
#include "Lock.hxx"
#include "Logger.hxx"
#include "VNetworkException.hxx"

using namespace Epoisses;
using namespace Vocal;

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

LockedStreamMap::LockedStreamMap()
{}

LockedStreamMap::~LockedStreamMap()
{
   Threads::WriteLock lock(_mutex);
   
   for(hash_map<string, LockedStream*>::iterator i = _map.begin();
       i != _map.end(); i++)
   {
      delete i->second;
   }
}

LockedStreamMap::LockedStream::LockedStream(const string& hostport)
   : _disconnected(true)
{
   DebugLog(<< "LockedStreamMap::LockedStream::LockedStream(" << hostport << ")");
   ACE_INET_Addr addr(hostport.c_str());

   if (_connector.connect (_server, addr) == -1)
   {
      DebugLog(<< strerror(errno));
      DebugLog (<< "Trying to connect to " 
                << addr.get_host_name() << ":" << addr.get_port_number() 
                << " and failed");
      throw VNetworkException("Can't connect", __FILE__,__LINE__);
   }
   _disconnected = false;
   DebugLog(<< "LockedStreamMap::LockedStream::LockedStream completed");
}

LockedStreamMap::LockedStream::~LockedStream()
{
   if (!_disconnected)
   {
      _server.close();
   }
}

void
LockedStreamMap::LockedStream::write(const string& data)
{
   DebugLog(<< "LockedStream::LockedStream::write");
   ACE_Time_Value timeout(1, 0);
   
   Threads::Lock lock(_mutex);
   if (!_disconnected)
   {
      if (_server.send_n (data.c_str(), data.size(), &timeout) == -1)
      {
         _disconnected = true;
         DebugLog(<< "LockedStreamMap::LockedStream::write: Failed to write pserver client");
         throw VNetworkException("Failed to write pserver client", __FILE__, __LINE__);
      }
   }
   DebugLog(<< "LockedStream::LockedStream::write completed");
}
            
void 
LockedStreamMap::add(const string& hostPort)
{
   Vocal::Threads::WriteLock lock(_mutex);
   DebugLog(<< "LockedStreamMap::add(" << hostPort << ")");
   
   if (_map.find(hostPort) == _map.end())
   {
      LockedStream *ls = new LockedStream(hostPort); 
      _map[hostPort] = ls;
   }
}

void
LockedStreamMap::remove(const string& hostPort)
{
   Threads::WriteLock lock(_mutex);
   DebugLog(<< "LockedStreamMap::remove(" << hostPort << "!dlb!)");

   hash_map<string, LockedStream*>::iterator i = _map.find(hostPort);   
   if (i != _map.end())
   {
      delete i->second;
      _map.erase(i);
   }
}

void
LockedStreamMap::write(const string& hostPort,
                       const string& message)
{
   bool found = false;
   {
      Threads::ReadLock lock(_mutex);
      found = (_map.find(hostPort) != _map.end());
   }
   if (!found)
   {      
      add(hostPort);
   }

   DebugLog(<< "LockedStreamMap::write(" << hostPort << ", " << message << ")");
   Threads::ReadLock lock(_mutex);
   _map[hostPort]->write(message);
}
