// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

static const char* const ProvisionServer_cxx_Version = "$Id: ProvisionClientHandler.cxx,v 1.1 2002/09/30 04:55:16 jason Exp $";

#include <boost/lexical_cast.hpp>

#include "ClientAcceptor.hxx"
#include "ConnectionManager.hxx"
#include "DBConnection.hxx"
#include "DBResultset.hxx"
#include "Logger.hxx"
#include "PSDBRequest.hxx"
#include "PSMessage.hxx"
#include "PSRspXmlContentHandlers.hxx"
#include "ProvisionClientHandler.hxx"
#include "ProvisionServer.hxx"
#include "RegisterManager.hxx"
#include "TimeAccumulate.hxx"
#include "VBadDataException.hxx"
#include "VDbException.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace Epoisses;

ProvisionClientHandler::ProvisionClientHandler() 
{
}

ProvisionClientHandler::~ProvisionClientHandler()
{
}

int
ProvisionClientHandler::process() 
{
   TimeAccumulate t("ProvisionClientHandler::process");
   string input;
   try
   {
      input = PSCommon::readHttpMessage(peer());
   }
   catch (VNetworkException& e)
   {
      DebugLog (<< "Caught: " << e);
      return -1;
   }
   
   // partial messages not dealt with
   // create HttpMessage object to handle the http parsing.
   Http::Message myHttpMessage;
  
   try
   {
      DebugLog( << "process: Received Message : " << input );
      myHttpMessage.decode( input );
   }
   catch (VException& e)
   {
      DebugLog(<< "process: Error parsing Http Message");
      sendHttpFail("bad request");
      return 0;
   }

   switch (myHttpMessage.method().type())
   {
      case Http::Method::GET:
         handleGetRequest( myHttpMessage );
         break;
      case Http::Method::PUT:
         handlePutRequest( myHttpMessage );
         break;
      case Http::Method::POST:
         handlePostRequest( myHttpMessage );
         break;
      case Http::Method::DELETE:
         handleDeleteRequest( myHttpMessage );
         break;
      default:
         WarningLog(<< "Don't know how to handle http request: "
                    << input);
         break;
   }
   return 0; // done for now, but keep connection open
}

void
ProvisionClientHandler::handleGetRequest( const Http::Message& httpmsg)
{
   DebugLog(<< "handleGetRequest(" << httpmsg.url().url() << ")");

   // Get the url for the function request.

   if ( httpmsg.url().url() == PSCommon::GET_RECORDS_IN_DOMAIN_STR)
   {
      handleGetRecordsInDomain(httpmsg.body() );  
   }
   else if (httpmsg.url().url() == PSCommon::SESSION_OPEN_STR)
   {
      handleSessionOpen(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::SESSION_GET_DB_RESULT_STR)
   {
      handleSessionGetDBResult(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::GET_FIELDS_IN_DOMAIN_STR)
   {
      handleGetFieldsInDomain(httpmsg.body() );
   }
   else 
   {
      assert(0); // !jf!
   }
}


void
ProvisionClientHandler::handlePostRequest( const Http::Message& httpmsg)
{
   TimeAccumulate t("ProvisionClientHandler::handlePostRequest");
   DebugLog(<< "handlePostRequest(" << httpmsg.url().url() << ")");

   if (httpmsg.url().url() == PSCommon::SESSION_CLOSE_STR)
   {
      handleSessionClose(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::SESSION_SEARCH_DB_STR)
   {
      handleSessionSearchDB(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::SESSION_SORT_DB_RESULT_STR )
   {
      handleSessionSortDBResult(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::REGISTER_DOMAIN_FOR_UPDATE_STR )
   {
      handleRegisterDomainForUpdate(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::UN_REGISTER_DOMAIN_FOR_UPDATE_STR )
   {
      handleUnRegisterDomainForUpdate(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_STR )
   {
      handleRegisterDomainForForeignUpdate(httpmsg.body() );
   }
   else if (httpmsg.url().url() == PSCommon::UN_REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_STR )
   {
      handleUnRegisterDomainForForeignUpdate(httpmsg.body() );
   }
   else
   {
      ErrLog(<< "handlePostRequest: Unrecognized: " << httpmsg.url().url());
      return;
   }
}


void
ProvisionClientHandler::handlePutRequest( const Http::Message& httpmsg)
{
   DebugLog(<< "handlePutRequest(" << httpmsg.url().url() << ")");

   if ( httpmsg.url().url() ==  
               PSCommon::SET_RECORDS_IN_DOMAIN_STR)
   {
      handleSetRecordsInDomain(httpmsg.body() );
   }
   else if ( httpmsg.url().url() ==  PSCommon::ADD_RECORD_IN_DOMAIN_STR)
   {
      handleAddRecordInDomain(httpmsg.body() );
   }
   else if ( httpmsg.url().url() ==  PSCommon::SESSION_SET_DB_RESULT_STR )
   {
      handleSessionSetDBResult(httpmsg.body() );
   }
   else
   {
      ErrLog(<< "handlePutRequest: Unrecognized: " << httpmsg.url().url());
      return;
   }
};


void
ProvisionClientHandler::handleDeleteRequest( const Http::Message& httpmsg)
{
   TimeAccumulate t("ProvisionClientHandler::handleDeleteRequest");
   //DebugLog(<< "handleDeleteRequest(" << httpmsg.url().url() << ")");

   if ( httpmsg.url().url() ==  
               PSCommon::DELETE_RECORDS_IN_DOMAIN_STR )
   {
      handleDeleteRecordsInDomain(httpmsg.body() );
   }
   else if ( httpmsg.url().url() ==  PSCommon::SESSION_DELETE_ALL_STR )
   {
      handleSessionDeleteAll(httpmsg.body() );
   }
   else
   {
      DebugLog(<< "handleDeleteRequest: Unrecognized: " << httpmsg.url().url());
   }
};


/**
 **  Get Functions
 **/

void
ProvisionClientHandler::handleGetRecordsInDomain( const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleGetRecordsInDomain");
   DebugLog(<< "handleGetRecordsInDomain");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   // 1.  Parse Request
   try
   {
      request = proto.parseRequest(
         PSMessage::GET_RECORDS_IN_DOMAIN_REQ, httpBody );
   }
   catch (VException& e)
   {
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail( e.getDescription());
      return;
   }

   // 2.  Call DBSelect to get result set.
   std::auto_ptr<DBResultset> resultSet;
   PSRspXmlContentHandlers myXmlContentHandler;
   string myXmlResponse;

   FieldListtype fieldlist = request->getFieldList();

   if ( fieldlist.empty() )
   {
      // fieldlist is empty.  return only the result count.
      try
      {
         resultSet = auto_ptr<DBResultset>(ConnectionManager::DBSelect(request->getDomain(),
                                                                       request->getFilter()));
      }
      catch( VException& e )
      {
         DebugLog("Caught: " << e);
         WarningLog(<< "Failed call to provisioning service");
         sendHttpFail("Failed call to provisioning service");
         return;
      }

      // get the count out of the result
      myXmlResponse =
         myXmlContentHandler.buildResultCountRsp(request->getDomain(),
                                                 resultSet->GetRecord(0)[0]);
   }
   else
   {
      // fieldlist is not empty.  return the result set.
      try
      {
         if (request->getOrderBy() == "")
         {
            resultSet = auto_ptr<DBResultset>(ConnectionManager::DBSelect( request->getDomain(), 
                                                               request->getFilter(), 
                                                               request->getFieldList() ));
         }
         else
         {
            resultSet = auto_ptr<DBResultset>(ConnectionManager::DBSelect( request->getDomain(), 
                                                               request->getFilter(), 
                                                               request->getFieldList(),
                                                               ASCENDING,
                                                               request->getOrderBy()));
         }
      }
      catch( VException& e )
      {
         DebugLog(<< "Caught: " << e);
         WarningLog(<< "Failed getRecords on " << request->getDomain() << " with filter: " << request->getFilter());
         sendHttpFail( e.getDescription() );
         return;
      }

      // 3.  Build the ps response using the DBResultset

      myXmlResponse =
         myXmlContentHandler.buildResultSetRsp( request->getDomain(),
                                                resultSet->GetRecords(),
                                                request->getFieldList() );
   };
 
   sendHttpOk(myXmlResponse );
}


void
ProvisionClientHandler::handleGetFieldsInDomain( const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleGetFieldsInDomain");
   DebugLog(<< "handleGetFieldsInDomain");

   // FieldListtype fieldList;
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   // 1.  Parse Request
   try
   {
      request = proto.parseRequest(
         PSMessage::GET_FIELDS_IN_DOMAIN_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(e.getDescription() );
      return;
   }

   // 2.  Call DBSelect to get result set.

   ResultTabletype restable;

   try
   {
      restable = ConnectionManager::DBDescribe( request->getDomain() );
   }
   catch( VException& e )
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Failed call to provisioning service");
      sendHttpFail(e.getDescription() );
      return;
   }

   // 3.  Build Xml response.
   PSRspXmlContentHandlers myXmlContentHandler;

   string myXmlResponse =
      myXmlContentHandler.buildGetFieldsRsp( request->getDomain(),
					     restable );
   sendHttpOk(myXmlResponse );

}

void 
ProvisionClientHandler::handleSessionOpen( const string& httpBody)
{
   DebugLog(<< "handleSessionOpen");
   Sptr<PSDBRequest> request;
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();

   // 1.  Parse Request

   try
   {
      request = proto.parseRequest(
         PSMessage::SESSION_OPEN_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail( e.getDescription() );
      return;
   }

   // 2.  Get field value 
   //     o  expiretime
   string myExpireTime;
   request->getFieldValue( "expiretime", myExpireTime );

   // Don't use the expire time for now...

   // Q:  Should the client app be allowed to specify the expire time?
   // Q:  Should the expire time be determined by the pserver only?


   // 3. add into cache.
   string sessionid = ProvisionServer::Instance().mySessionManager.genSessionId();

   ProvisionServer::Instance().mySessionManager.addCacheRequest( sessionid,  // string sessionid
                                                                 "",         // string domain
                                                                 "",         // string filter
                                                                 "",         // string sortType
                                                                 "",         // string sortField
                                                                 0 );        // DBResultset* resultSet

   // 4.  Build Xml response.
   PSRspXmlContentHandlers myXmlContentHandler;

   string myXmlResponse =
      myXmlContentHandler.buildSessionOpenRsp ( sessionid );

   sendHttpOk( myXmlResponse );
}


void 
ProvisionClientHandler::handleSessionGetDBResult( const string& httpBody)
{
}


/**
 **  Post Functions
 **/

void
ProvisionClientHandler::handleSessionClose( const string& httpBody)
{
   DebugLog(<< "handleSessionClose");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();

   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::SESSION_CLOSE_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   if (ProvisionServer::Instance().mySessionManager.deleteCacheRequest(request->getSessionID()) == 0)
   {
      sendHttpFail(  "No Token Found");
   }
   else
   {
      string responseStr = "\0";
      sendHttpOk( responseStr );
   }

   return;
}


void
ProvisionClientHandler::handleSessionSearchDB( const string& httpBody)
{
}

void
ProvisionClientHandler::handleSessionSortDBResult( const string& httpBody)
{
}

void
ProvisionClientHandler::handleRegisterDomainForUpdate( const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleRegisterDomainForUpdate");
   DebugLog(<< "handleRegisterDomainForUpdate");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::REGISTER_DOMAIN_FOR_UPDATE_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   string responseStr = "\0";
   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
   if (registerMgr.registerDomainForUpdate(request->getDomain(),
                                           request->getRegisterHostport(),
                                           request->getFunctorUId()))
   {
      sendHttpOk(  responseStr );
   }
   else
   {
      sendHttpFail("RegisterDomainForUpdate failed for domain '" + 
                   request->getDomain() + "'");
   }
}

void
ProvisionClientHandler::handleUnRegisterDomainForUpdate( const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleUnRegisterDomainForUpdate");
   DebugLog(<< "handleUnRegisterDomainForUpdate");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::UN_REGISTER_DOMAIN_FOR_UPDATE_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   string responseStr = "\0"; 
   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
   if (registerMgr.unregisterDomainForUpdate(request->getDomain(),
                                             request->getRegisterHostport(),
                                             request->getFunctorUId()))
   {
      sendHttpOk(  responseStr );
   }
   else
   {
      sendHttpFail("UnregisterDomainForUpdate failed for domain '" + 
                   request->getDomain() + "'");
   }
}

void
ProvisionClientHandler::handleRegisterDomainForForeignUpdate
(const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleRegisterDomainForForeignUpdate");
   DebugLog(<< "handleRegisterDomainForForeignUpdate");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );

      return;
   }

   string responseStr = "\0";

   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
   if (registerMgr.registerDomainForForeignUpdate(request->getForeignDomain(),
                                                  request->getForeignField(),
                                                  request->getDomain(),
                                                  request->getRegisterHostport(),
                                                  request->getFunctorUId()))
   {
      sendHttpOk( responseStr);
   }
   else
   {
      sendHttpFail("RegisterDomainForForeignUpdate failed for foreign domain '" + 
                   request->getForeignDomain() +
                   "' to domain '" + request->getDomain() + "' via field '" + 
                   request->getForeignField());
   }
}


void
ProvisionClientHandler::handleUnRegisterDomainForForeignUpdate
(const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleUnRegisterDomainForForeignUpdate");
   DebugLog(<< "handleUnRegisterDomainForForeignUpdate");

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::UN_REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   string responseStr = "\0"; 

   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
   if (registerMgr.unregisterDomainForForeignUpdate(request->getForeignDomain(),
                                                    request->getRegisterHostport(),
                                                    request->getFunctorUId()))
   {
      sendHttpOk(responseStr);
   }
   else
   {
      sendHttpFail("UnregisterDomainForForeignUpdate failed for foreign domain '" + 
                   request->getForeignDomain() +
                   "' with UID " + request->getFunctorUId());
   }
}

/**
 **  Put Functions
 **/

void
ProvisionClientHandler::handleSetRecordsInDomain(const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleSetRecordsInDomain");
   DebugLog(<< "handleSetRecordsInDomain");

   FieldValuetype fieldvaluelist;

   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(PSMessage::SET_RECORDS_IN_DOMAIN_REQ, 
                                   httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   PSRspXmlContentHandlers myXmlContentHandler;
   auto_ptr<DBResultset> results;

   try
   {
      // are there listeners for updates?
      RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
      if (registerMgr.hasRegistered(request->getDomain()))
      {
         results = auto_ptr<DBResultset>(affectedRows(request->getDomain(),
                                                      request->getFilter()));
      }
      
      // Actual set call.
      ConnectionManager::DBUpdate(request->getDomain(), 
                                  request->getFilter(), 
                                  request->getTable());

      const string recordCount = ( results.get()
                                   ? boost::lexical_cast<string>(results->GetRecordCount())
                                   : "0");

      string myXmlResponse = 
         myXmlContentHandler.buildSetRecordsRsp(request->getDomain(),
                                                recordCount);
    
      if (results.get()) 
      {
         registerMgr.checkRegisterList(results.get(), PSCommon::SET_CHANGE);
      }

      // send the OK after the update check -- errors detected by client
      sendHttpOk(myXmlResponse);
   }
   catch( VException& e)
   {
      DebugLog(<< "Caught: " << e);
      InfoLog(<< "Failed call to provisioning service: " << httpBody);
      sendHttpFail(  e.getDescription() );
   }
}

void
ProvisionClientHandler::handleAddRecordInDomain(const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleAddRecordInDomain");
   DebugLog(<< "handleAddRecordInDomain");

   FieldListtype fieldlist;
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::ADD_RECORD_IN_DOMAIN_REQ, httpBody );
   }
   catch (VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   int oid;
   try
   {
      // !dlb! remember the returned oid
      oid = ConnectionManager::DBInsert(request->getDomain(), request->getTable());
   }
   catch ( VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Failed Add to domain '" << request->getDomain() << "'");
      sendHttpFail(  e.getDescription() );
      return;
   }

   string responseStr = "\0";

   const string domain = request->getDomain();
   
   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
   // does the inserted domain have foreign updates?
   if (registerMgr.hasForeignRegistered(domain))
   {
      try 
      {
         list<FieldResult> foreignResultSet;

         // get set of foreign fields for this domain
         const set<string> foreignFields = registerMgr.getForeignFields(domain);
         for (set<string>::const_iterator i = foreignFields.begin();
              i != foreignFields.end(); i++)
         {
            // foreign key field from inserted domain to primary domain
            const string& foreignField = *i;

            const string primaryDomain =
               registerMgr.getPrimaryDomain(domain, foreignField);

            DebugLog(<< "handleAddRecordInDomain primaryDomain = " << primaryDomain);

            // key field of primary
            const string primaryField = ConnectionManager::DBPrimaryKey(primaryDomain);

            // !dlb! complain if no value
            string primaryKey;
            request->getFieldValue(foreignField, primaryKey);

            // determine the leafDomain type of the referenced primary row
            // resulting output row looks like this:
            // domain primaryDomain primaryKey

            // !dlb! postgres specific 
           FieldListtype fieldList;
            // awful hack -- fake out the back end to produce sql
            fieldList.push_back("'" + domain + "'");
            fieldList.push_back("'" + primaryDomain + "'");
            fieldList.push_back(primaryKey); 

            auto_ptr<DBResultset> result;

            result = auto_ptr<DBResultset>(ConnectionManager::DBRetrieveByOid(domain,
                                                                              oid,
                                                                              fieldList));
            foreignResultSet.push_back(FieldResult(foreignField,
                                                   result));
         }
         foreignUpdateCheck(foreignResultSet, PSCommon::ADD_CHANGE);
      }
      catch (VBadDataException& e)
      {
         DebugLog("Caught: " << e);
         // shouldn't happen
         WarningLog(<< "Badly formed insertion; no foreign field referenced: " 
                    <<  httpBody);
      }
   }

   // key field of domain
   const string primaryField = ConnectionManager::DBPrimaryKey(domain);

   // primaryKey, domain
   FieldListtype fieldList;
   fieldList.push_back(primaryField); 
   fieldList.push_back("'" + domain + "'"); // add, so domain is leaf

   auto_ptr<DBResultset> result;
   result = auto_ptr<DBResultset>(ConnectionManager::DBRetrieveByOid(domain, 
                                                                     oid,
                                                                     fieldList));
   // does the inserted domain have updates?
   if (registerMgr.hasRegisteredRecurse(domain)) 
   {
      updateCheck(result.get(), PSCommon::ADD_CHANGE);
   }

   // send OK last to make client aware of update errors
   FieldListtype fieldListProper;
   fieldListProper.push_back(primaryField); 
   
   PSRspXmlContentHandlers myXmlContentHandler;
   sendHttpOk(myXmlContentHandler.buildResultSetRsp(domain,
                                                    result->GetRecords(),
                                                    fieldListProper));
}

void
ProvisionClientHandler::handleSessionSetDBResult( const string& httpBody)
{
}

/**
 **  Delete Functions
 **/

void
ProvisionClientHandler::handleDeleteRecordsInDomain(const string& httpBody)
{
   TimeAccumulate t("ProvisionClientHandler::handleDeleteRecordsInDomain");
   //DebugLog(<< "handleDeleteRecordsInDomain");

   FieldListtype fieldlist;
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<PSDBRequest> request;

   try
   {
      request = proto.parseRequest(
         PSMessage::DEL_RECORDS_IN_DOMAIN_REQ, httpBody );
   }
   catch ( VException& e)
   {
      DebugLog("Caught: " << e);
      WarningLog(<< "Error parsing XML: " << httpBody);
      sendHttpFail(  e.getDescription() );
      return;
   }

   RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;

   auto_ptr<DBResultset> resultSet;
   list<FieldResult> foreignResultSets;

   try
   {
      // are there listeners for updates?
      if (registerMgr.hasRegistered(request->getDomain()))
      {
         resultSet = auto_ptr<DBResultset>(affectedRows(request->getDomain(),
                                                        request->getFilter()));
      }

      if (registerMgr.hasForeignRegistered(request->getDomain()))
      {
         // Need to call select for the update check.
         foreignResultSets = affectedForeignRows(request->getDomain(),
                                                 request->getFilter());
      }

      // Actual delete call.
      ConnectionManager::DBDelete(request->getDomain(), 
                      request->getFilter());


      // 3.  Return 200 OK w/ number of deleted records.
      const string recordCount = 
         (resultSet.get()
          ? boost::lexical_cast<string>(resultSet->GetRecordCount())
          : "0");
      PSRspXmlContentHandlers myXmlContentHandler;

      string myXmlResponse =
         myXmlContentHandler.buildSetRecordsRsp(request->getDomain(),
                                                recordCount);
      if (resultSet.get())
      {
         updateCheck(resultSet.get(), PSCommon::DELETE_CHANGE);
      }

      if (registerMgr.hasForeignRegistered(request->getDomain()))
      {
         foreignUpdateCheck(foreignResultSets, PSCommon::DELETE_CHANGE);
      }

      // send OK after updates -- client sees problems
      sendHttpOk(myXmlResponse);
   }
   catch (VDbException& e)
   {
      DebugLog(<< "Caught: " << e);
      InfoLog(<< "Error deleting records: " << httpBody);
      sendHttpFail(  e.getDescription() );
   }
}

/// !dlb! this should make a generic DBConnection call someday
// one for direct modifications, one for foreign mods
DBResultset* 
ProvisionClientHandler::affectedRows(const string& domain,
                                     const string& filter)
{
   TimeAccumulate t("ProvisionClientHandler::affectedRows:" + domain);
   if (domain == "##registrationbindings") //!dcm! hack, need to make this generic
                                         //somehow
   {
         const string primaryKey = ConnectionManager::DBPrimaryKey(domain);
         FieldListtype fieldList;
         fieldList.push_back(primaryKey);
         fieldList.push_back("'registrationbindings'");

         return ConnectionManager::DBSelect(domain,
                                            filter,
                                            fieldList);
   } 
   else
   {
      // Need to call select for the update check.
      const string primaryKey = ConnectionManager::DBPrimaryKey(domain);
      
      FieldListtype fieldList;
      fieldList.push_back(primaryKey);
      fieldList.push_back("relname"); // !dlb! postgres specific
      
      // !dlb! postgres specific
      // (modify DBConnection interface when the dust settles)
      // hack up the select arguments for now
      // get the key and leaf-domain from the to-be-modified rows
      
      // add filter code to bind the tableoid of the selected rows
      strstream modifiedFilter;
      if (!filter.empty())
      {
         modifiedFilter << "(" << filter << ") AND ";
      }
      modifiedFilter << domain << ".tableoid = pg_class.oid" << ends;
      modifiedFilter.freeze(0);
      
      return ConnectionManager::DBSelect(domain + ", pg_class",
                                         modifiedFilter.str(),
                                         fieldList);
   }
}

// ============================================================
// !dlb! needs to be pushed to DBConnection -- all sorts of SQL here
/// Insert "_t." before every field name in a filter string.
/// Disambiguates field references in a foreign key update.
class AddTablePrefix
{
      class Reserved : private set<string>
      {
         public:
            Reserved()
            {
               insert("and");
               insert("or");
               insert("not");
               insert("like");
               insert("true");
               insert("false");
               insert("AND");
               insert("OR");
               insert("NOT");
               insert("LIKE");
               insert("TRUE");
               insert("FALSE");
            }

            bool contains(const string& word) const
            {
               return set<string>::find(word) != end();
            }

         private:
            // no value semantics
            Reserved(const Reserved&);
            Reserved& operator=(const Reserved&);
      };

      enum State {Idle, InWord, InString};

   public:
      static string
      addTablePrefix(const string& prefix,
                     const string& input)
      {
         State state = Idle;
         string word = "";
         string result = "";

         for(string::const_iterator i = input.begin();
             i != input.end(); i++)
         {
            const char c = *i;
            switch (state)
            {
               case Idle:
                  if ((c >= 'A' && c <='Z') ||
                      (c >= 'a' && c <='z') ||
                      c == '_')
                  {
                     state = InWord;
                     word = c;
                  }
                  else
                  {
                     result += c;
                     if (c == '\'')
                     {
                        state = InString;
                     }
                  }
                  break;
               case InWord:
                  // fields are followed by space (?no?) or right paren
                  if (c == ' ' || c == ')')
                  {
                     if (!reserved.contains(word))
                     {
                        result += prefix + '.' + word + c;
                     }
                     else 
                     {
                        result += word + c;
                     }
                     word = "";
                     state = Idle;
                  }
                  else
                  {
                     word += c;
                  }
                  break;
               case InString:
                  result += c;
                  if (c == '\'')
                  {
                     state = Idle;
                  }
                  break;
               default:
                  assert(false);
            }
         }

         // special case for just field input
         if (!word.empty() && !reserved.contains(word))
         {
            result += prefix + word;
         }
         return result;
      }

   private:
      static const Reserved reserved;

      // no value semantics
      AddTablePrefix(const AddTablePrefix&);
      AddTablePrefix& operator=(const AddTablePrefix&);
};
const AddTablePrefix::Reserved AddTablePrefix::reserved; 

list<ProvisionClientHandler::FieldResult> 
ProvisionClientHandler::affectedForeignRows(const string& foreignDomain,
                                            const string& filter)
{
   TimeAccumulate t("ProvisionClientHandler::affectedForeignRows:" + foreignDomain);
   if (ProvisionServer::Instance().myRegisterManager.hasForeignRegistered(foreignDomain))
   {
      list<FieldResult> resultsList;
      // get the foreign root for primary and foreign key info
      const string& foreignRootDomain = ConnectionManager::DBRoot(foreignDomain);

      const set<string>& foreignFields = 
         ProvisionServer::Instance().myRegisterManager.getForeignFields(foreignDomain);
      for (set<string>::const_iterator i = foreignFields.begin();
           i != foreignFields.end(); i++) 
      {
         // get the field that maps from foreign to primary
         const string foreignField = *i;

         // get the primary root domain for select
         const string& primaryRootDomain = 
            ProvisionServer::Instance().myRegisterManager.getPrimaryDomain(foreignRootDomain, foreignField);

         // key column for primary domain
         const string primaryKey = ConnectionManager::DBPrimaryKey(primaryRootDomain);

         if (foreignDomain == "##registrationbindings") //!dcm! -- might work?
         {
            FieldListtype fieldList;
            fieldList.push_back(foreignField);
            auto_ptr<DBResultset> result(ConnectionManager::DBSelect(foreignDomain,
                                                                     filter,
                                                                     fieldList));  
            if (result->GetRecordCount() != 0)
            {
               stringstream inClause;
               inClause << "_p." << primaryKey << " IN (";
               ResultTabletype rows = result->GetRecords();
               for (ResultTabletype::const_iterator it = rows.begin();
                    it != rows.end(); it++)
               {
                  if (it != rows.begin())
                  {
                     inClause << ", ";
                  }
                  inClause << "'" << it->at(0) << "'";
               }
               inClause << ") AND _p.tableoid = _pc.oid";
               
               stringstream generateFrom;
               generateFrom << primaryRootDomain << " _p, pg_class _pc";
               
               FieldListtype forFieldList;
               stringstream foreignDomainQuoted;
               foreignDomainQuoted << "'" << foreignDomain << "'";
               
               forFieldList.push_back(foreignDomainQuoted.str());
               forFieldList.push_back("_pc.relname");
               forFieldList.push_back(primaryKey);
               
               auto_ptr<DBResultset> result1(ConnectionManager::DBSelect(generateFrom.str(),
                                                                         inClause.str(),
                                                                         forFieldList));
               // collect up each result with its corresponding field
               resultsList.push_back(FieldResult(foreignField,
                                                 result1));
            }
         }
         else
         {
            FieldListtype fieldList;
            // awful hack -- fake out the back end to produce sql
            fieldList.push_back("distinct _fc.relname");
            fieldList.push_back("_pc.relname"); 
            fieldList.push_back("_f." + foreignField); 
            
            // add filter code to bind the tableoid/relname of the affected foreign
            // rows
            
            // e.g.
            // SELECT DISTINCT _p.userid, _fc.relname, _pc.relname 
            //   FROM features _f, pg_class _fc, pg_class _pc, users _p 
            //   WHERE name = 'feature3' // original filter
            //   AND _f.userid = _p.userid  // foreign field
            //   AND _f.tableoid = _fc.oid AND _p.tableoid = _pc.oid; // leaf domains
            strstream modifiedFilter;
            if (!filter.empty())
            {
               // name the primary table to disambiguate
               modifiedFilter << "(" << AddTablePrefix::addTablePrefix("_f", filter) << ") AND ";
            }
            modifiedFilter << "_f." << foreignField << " = _p." << primaryKey << " AND "
                           << "_f.tableoid = _fc.oid AND _p.tableoid = _pc.oid" << ends;
            modifiedFilter.freeze(0);
            
            // fake out the back end some more to get the from clause
            strstream generateFrom;
            generateFrom << foreignDomain 
                         << " _f , pg_class _pc, pg_class _fc,"
                         << primaryRootDomain << " _p" << ends;
            generateFrom.freeze(0);
            
            auto_ptr<DBResultset> result(ConnectionManager::DBSelect(generateFrom.str(),
                                                                     modifiedFilter.str(),
                                                                     fieldList));
            // collect up each result with its corresponding field
            resultsList.push_back(FieldResult(foreignField,
                                              result));
         }
      }
      return resultsList;
   }
   static list<FieldResult> empty;
   return empty;
}
// ============================================================

void
ProvisionClientHandler::handleSessionDeleteAll( const string& httpBody)
{
}

void
ProvisionClientHandler::updateCheck(DBResultset* resultSet,
                                    PSCommon::ChangeType changeType)
{
   DebugLog(<< "updateCheck");

   RegisterManager& registerMgr = ProvisionServer::Instance().myRegisterManager;
   // need to check the register list to send update.
   registerMgr.checkRegisterList(resultSet, changeType);
}

void
ProvisionClientHandler::foreignUpdateCheck(const list<FieldResult>& results,
                                           PSCommon::ChangeType changeType)
{
   for (list<FieldResult>::const_iterator i = results.begin();
        i != results.end(); i++) 
   {
      RegisterManager &registerMgr = ProvisionServer::Instance().myRegisterManager;
      // check the register list to send update.
      registerMgr.checkForeignRegisterList(i->_fieldName,
                                           i->_result.get(),
                                           changeType);
   }
}

void
ProvisionClientHandler::sendHttpFail(const string& retData)
{
   Http::Status errStatus("400");
   list<Http::Header> headers;
   string errBegin = "<error>";
   string errEnd   = "</error>";

   string tempData = errBegin + retData + errEnd;

   Http::Header contentLength(Http::Header::ContentLength, boost::lexical_cast<string>(tempData.size()) );
   headers.push_back(contentLength);
   Http::Message myHttpResponse(errStatus, headers, tempData );
   string tempString = myHttpResponse.encode();

    
   DebugLog(<< "Send 400 Fail Message: " << tempString);
   if (peer().send(tempString.c_str(), tempString.size()) < 0)
   {
      WarningLog (<< "Failed writing data: " << tempString);
   }
}


void
ProvisionClientHandler::sendHttpOk(const string& retData)
{
   TimeAccumulate t("ProvisionClientHandler::sendHttpOk");
   // Build http status message.
   Http::Status myStatus ("200");
   list<Http::Header> headers;

   Http::Header contentLength(Http::Header::ContentLength, boost::lexical_cast<string>(retData.size()) );

   headers.push_back(contentLength);
   Http::Message myHttpResponse(myStatus, headers, retData );
   string tempString = myHttpResponse.encode();

   DebugLog(<< "Send 200 OK response: " << tempString);

   TimeAccumulate s("ProvisionClientHandler::sendHttpOk:send");
   if (peer().send(tempString.c_str(), tempString.size()) < 0)
   {
      WarningLog (<< "Failed writing data: " << tempString);
   }
}

