/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const RegisterManager_cxx_Version = "$Id: RegisterManager.cxx,v 1.1 2002/09/30 04:55:16 jason Exp $";

#include "RegisterManager.hxx"
#include "Sptr.hxx"
#include "NetworkAddress.h"
#include "Domain.hxx"
#include "Mutex.hxx"
#include "Lock.hxx"
#include "Logger.hxx"
#include "TimeAccumulate.hxx"

// to determine inheritance and primary keys
#include "ConnectionManager.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace Epoisses;

RegisterManager::RegisterManager()
   : errorMsg("")
{
}

RegisterManager::~RegisterManager()
{
}

bool
RegisterManager::registerDomainForUpdate( const string& domain,
                                          const string& hostport,
                                          const string& uid)
{
   assert(!domain.empty());
   assert(!hostport.empty());
   assert(!uid.empty());

   DebugLog(<<  "RegisterManager::registerDomainForUpdate (" << domain << "," << hostport << "," << uid << ")");

   if (!ConnectionManager::Connection(domain).DBVerify(domain)) 
   {
      return false;
   }
   else
   {
      Threads::WriteLock lock(_mapMutex);
      uIdMap[domain][hostport].push_front(uid);
      return true;
   }
}

bool
RegisterManager::unregisterDomainForUpdate( const string& domain,
                                            const string& hostport,
                                            const string& uid)
{
   DebugLog(<<  "RegisterManager::unregisterDomainForUpdate (" << domain << "," << hostport << "," << uid << ")");

   Threads::WriteLock lock(_mapMutex);
   if (uIdMap.find(domain) == uIdMap.end())
   {
      return false;
   }
       
   assert(uIdMap[domain].find(hostport) != uIdMap[domain].end());

   list<string>& l = uIdMap[domain][hostport];
   // remove the uid from the domain+hostport
   l.remove(uid);

   // if there are no longer any functors associated, remove the hostport
   if (l.empty())
   {
      uIdMap[domain].erase(hostport);
      _streamMap.remove(hostport);
   }

   // if there are no hostports left, drop the domain
   if (uIdMap[domain].empty()) 
   {
      uIdMap.erase(domain);
   }
    
   return true;
}

void
RegisterManager::checkRegisterList(DBResultset* resultset,
                                   PSCommon::ChangeType changeType)
{
   TimeAccumulate t("RegisterManager::checkRegisterList");
   DebugLog(<<  "** checkDomainList **");

   if (resultset == NULL ||
       resultset->GetRecords().empty())
   {
      DebugLog(<<  "table is null or empty.  nothing to do.");
      return;
   }

   // keep a set of failed host:port
   set<string> failedHostPorts;

   {
      // temporary to hold keys by domains to update
      DomainToKeyMap domainToKey;

      Threads::ReadLock lock(_mapMutex);

      ResultTabletype restable = resultset->GetRecords();
      for (ResultTabletype::const_iterator tabIt = restable.begin();
           tabIt != restable.end(); tabIt++)
      {
         assert(tabIt->size() == 2);
         // first column is primary key value
         const string& key = (*tabIt)[0];
         // second column is leaf-domain
         const string& leafDomain = (*tabIt)[1];

         // loop over the parents of the leafDomain
         string domain = leafDomain;
         string lastDomain = leafDomain;
         do
         {
            if (hasRegistered(domain))
            {
               DebugLog(<<  "found update for " << domain << " key= " << key);
               // add the record to the collection for domain
               domainToKey[domain].push_back(key);
            }

            // iterate over domain parents
            lastDomain = domain;
            domain = ConnectionManager::DBParent(domain);
         }
         // keep going until the domain root
         while (domain != lastDomain);
      }
      DebugLog(<<  "RegisterManager::checkDomainList, partitioned");
    
      // domainToKey is filled with domains and corresponding records for updates
      for (DomainToKeyMap::const_iterator dkIt = domainToKey.begin();
           dkIt != domainToKey.end(); dkIt++)
      {
         const string& domain = dkIt->first;
         const vector<string>& keys = dkIt->second;

         UIdMap::const_iterator domainUIdMapIt = uIdMap.find(domain);
       
         // send the update to all the registered hosts.
         for (HostUIdMap::const_iterator hostUidMapIt = domainUIdMapIt->second.begin();
              hostUidMapIt != domainUIdMapIt->second.end(); hostUidMapIt++)
         {
            DebugLog(<<  "*** Send domain update to host: " << hostUidMapIt->first << " ***");
            if (!sendUpdateDomain(domain,
                                  keys,
                                  changeType,
                                  hostUidMapIt->first))
            {
               failedHostPorts.insert(hostUidMapIt->first);
            }
         }
      }
   }
   // unregister all failed host:port
   for (set<string>::const_iterator i = failedHostPorts.begin();
        i != failedHostPorts.end(); i++) 
   {
      unregisterUpdates(*i);
   }
}

bool
RegisterManager::sendUpdateDomain(const string& domain,
                                  const vector<string>& keyList,
                                  PSCommon::ChangeType changeType,
                                  const string& hostport )
{
   TimeAccumulate t("RegisterManager::sendUpdateDomain");
   
   TimeAccumulate l("RegisterManager::sendUpdateDomain:postLock");

   try
   {
      TimeAccumulate w("RegisterManager::sendUpdateDomain:buildAndWriteMessage");
      
      // build http message
      PSProtocolService& proto = PSProtocolServiceHolder::Instance();
      
      Sptr<Http::Message> httpMsg;

      httpMsg = proto.buildUpdateReq(PSCommon::UPDATE_DOMAIN_STR,
                                     domain, keyList, changeType);

      string tempString = httpMsg->encode();

      DebugLog(<< "sendUpdateDomain: " << tempString);
      
      _streamMap.write(hostport, tempString);
   }
   catch (VNetworkException& e)
   {
      WarningLog(<< "Caught: " << e);
      WarningLog(<< "sendUpdateDomain(" << domain << ", " << hostport << ") failed to contact client -- dropping from updates");
      return false;
   }
   catch(VException& e)
   {
      ErrLog(<< "Caught: " << e);
      assert(false);
   }
   return true;
}

bool
RegisterManager::registerDomainForForeignUpdate(const string& foreignDomain,
                                                const string& foreignField,
                                                const string& primaryDomain,
                                                const string& hostport,
                                                const string& uid)
{
   DebugLog(<<  "RegisterManager::registerDomainForForeignUpdate(" 
            << foreignDomain << "," 
            << foreignField << "," 
            << primaryDomain << "," 
            << hostport << ")");

   if (foreignField.empty() ||
       hostport.empty() ||
       uid.empty())
   {
      return false;
   }
   
   ConnectionManager::Connection dbconnFor(foreignDomain);
   ConnectionManager::Connection dbconnPrim(primaryDomain);
      
   // if either domain does not exist, bail
   if (!dbconnFor.DBVerify(foreignDomain) ||
       !dbconnPrim.DBVerify(primaryDomain))
   {
      return false;
   }

   // make sure that the foreignField exists in the foreign domain
   ResultTabletype fields = dbconnFor.DBDescribe(foreignDomain);
   bool found = false;
   for (ResultTabletype::const_iterator r = fields.begin();
        r != fields.end(); r++)
   {
      if ((*r)[0] == foreignField)
      {
         found = true;
         break;
      }
   }
   if (!found)
   {
      return false;
   }

   const string foreignRoot = dbconnFor.DBRoot(foreignDomain);
   const string primaryRoot = dbconnPrim.DBRoot(primaryDomain);

   Threads::ReadLock lock(_mapMutex);

   // make sure the foreign/primary is consistent
   ForeignDomainToPrimaryDataMap::const_iterator it = primaryRootDomainMap.find(foreignRoot);
   found = false;
   if (it != primaryRootDomainMap.end())
   {
      for (list<PrimaryData>::const_iterator i = it->second.begin();
           i != it->second.end(); i++)
      {
         if (i->_foreignField == foreignField)
         {
            assert(i->_primaryDomain == primaryRoot);
            found = true;
            break;
         }
      }
   }
   if (!found)
   {
      // !dlb! could make sure the foreignField is a field
      // remember the foreign domain root to primary domain root relation
      primaryRootDomainMap[foreignRoot].push_back(PrimaryData(primaryRoot, foreignField));
   }

   // remember the particular mapping as well
   primaryMap[foreignDomain].push_back(PrimaryData(primaryDomain, foreignField, 
                                                   hostport, uid));
   
   // remember the host for the uid
   hostportByUId[uid] = hostport;

   return true;
}

struct PrimaryDataUidComparitor
{
      PrimaryDataUidComparitor(const string &uid)
         : _uid(uid)
      {}
      
      bool operator()(const PrimaryData& p)
      {
         return p._uid == _uid;
      }
      
      const string _uid;
};

bool
RegisterManager::unregisterDomainForForeignUpdate(const string& foreignDomain,
                                                  const string& hostport,
                                                  const string& uid)
{
   assert(!foreignDomain.empty());
   assert(!hostport.empty());
   assert(!uid.empty());

   Threads::WriteLock lock(_mapMutex);

   // don't forget the foreign domain root to primary domain root relation -- shared
   // don't forget the foreign domain root to foreign key field -- shared

   // remove the foreign domain's primary data
   list<PrimaryData> prim = primaryMap[foreignDomain];
   prim.remove_if(PrimaryDataUidComparitor(uid));

   if (prim.empty())
   {
      primaryMap.erase(foreignDomain);
   }

   // ?dlb? not tracking last use of hostport???
   // does foreign update imply update?
   
   return true;
}

struct PrimaryDataHostportComparitor
{
      PrimaryDataHostportComparitor(const string &hostport)
         : _hostport(hostport)
      {}
      
      bool operator()(const PrimaryData& p)
      {
         return p._hostport == _hostport;
      }
      
      const string _hostport;
};

void
RegisterManager::unregisterUpdates(const string& hostport)
{
   Threads::WriteLock lock(_mapMutex);

   // domain updates
   for (UIdMap::iterator i = uIdMap.begin();
        i != uIdMap.end(); i++) 
   {
      DebugLog(<< "removing domain registration for " << hostport << "for domain " << i->first);
      i->second.erase(hostport);
   }

   // foreign updates
   for (ForeignDomainToPrimaryDataMap::iterator i = primaryMap.begin();
        i != primaryMap.end(); i++) 
   {
      unsigned int tmp = i->second.size();
      
      // remove all updates for the given hostport
      i->second.remove_if(PrimaryDataHostportComparitor(hostport));

      // was there a chance to this foreign domain's updates?
      if (tmp != i->second.size()) 
      {
         DebugLog(<< "remove foreign registration for " << hostport << "for foreign domain  " << i->first);
      }
   }
   
   for (hash_map<string, string>::iterator i = hostportByUId.begin();
        i != hostportByUId.end(); )
   {
      if (i->second == hostport)
      {
         DebugLog(<< "remove foreign update uid " << i->first << " for " << hostport);
         hostportByUId.erase(i++);
      }
      else
      {
         i++;
      }
   }

   _streamMap.remove(hostport);
}

bool
RegisterManager::hasRegistered(const string& domain) const
{
   return uIdMap.find(domain) != uIdMap.end();
}

bool
RegisterManager::hasRegisteredRecurse(const string& leafDomain) const
{
   // check the given domain and all parents
   string domain = leafDomain;
   string lastDomain = leafDomain;
   do
   {
      if (hasRegistered(domain))
      {
         return true;
      }

      // iterate over domain parents
      lastDomain = domain;
      domain = ConnectionManager::DBParent(domain);
   }
   // keep going until the domain root
   while (domain != lastDomain);
   return false;
}

// weak check, but most tables will not pass, so worth doing
bool
RegisterManager::hasForeignRegistered(const string& foreignDomain) const
{
   const string foreignRoot = ConnectionManager::DBRoot(foreignDomain);
   ForeignDomainToPrimaryDataMap::const_iterator it = primaryMap.find(foreignRoot);
   if (it != primaryMap.end())
   {
      return true;
   }
   return false;
}

// called with table of primaryDomain leaf, key
// Work up the primary domain hierarchy, stopping after the registered
// primary domain is reached.

// resultset looks like this:
// foreignDomainLeaf primaryDomainLeaf primaryKey
void
RegisterManager::checkForeignRegisterList(const string& foreignField,
                                          DBResultset* resultset,
                                          PSCommon::ChangeType changeType)
{
   TimeAccumulate t("RegisterManager::checkForeignRegisterList");
   DebugLog(<<  "RegisterManager::checkForeignRegisterList");

   if (resultset == NULL ||
       resultset->GetRecords().empty())
   {
      DebugLog(<<  "table is null or empty.  nothing to do.");
      return;
   }

   // collect keys by uid
   hash_map<string, vector<string> > keys;
   
   {
      
      Threads::ReadLock lock(_mapMutex);

      ResultTabletype restable = resultset->GetRecords();
      for (ResultTabletype::const_iterator tabIt = restable.begin();
           tabIt != restable.end(); tabIt++)
      {
         assert(tabIt->size() == 3);
         // first column is foreign leaf-domain
         const string& foreignLeafDomain = (*tabIt)[0];
         // second column is primary leaf-domain
         const string& primaryLeafDomain = (*tabIt)[1];
         // third column is primary key value
         const string& key = (*tabIt)[2];

         // loop over the parents of the leaf foreign domain
         string foreignDomain = foreignLeafDomain;
         string lastForeignDomain;
         do
         {
            lastForeignDomain = foreignDomain;
            foreignDomain = ConnectionManager::DBParent(foreignDomain);
            // loop over the parents of the leaf primary domain
            string primaryDomain = primaryLeafDomain;
            string lastPrimaryDomain;
            do
            {
               // is there an updator for the (foreignDomain -(foreignField)-> primaryDomain)?
               DebugLog(<<  "checking for updator for " << foreignLeafDomain << "->" << primaryDomain);
               ForeignDomainToPrimaryDataMap::const_iterator it = primaryMap.find(foreignDomain);
               if (it != primaryMap.end())
               {
                  // loop over the list of update data
                  for(list<PrimaryData>::const_iterator pIt = it->second.begin();
                      pIt != it->second.end(); pIt++) 
                  {
                     DebugLog(<<  "  to " <<  pIt->_primaryDomain);
                     if (pIt->_primaryDomain == primaryDomain &&
                         pIt->_foreignField == foreignField)
                     {
                        DebugLog(<<  "    found key= " << key);
                        // add to keys for this uid
                        keys[pIt->_uid].push_back(key);
                     }
                  }
               }

               lastPrimaryDomain = primaryDomain;
               primaryDomain = ConnectionManager::DBParent(primaryDomain);
            }
            // keep going until the primary domain root has been processed
            while (primaryDomain != lastPrimaryDomain);
         }
         // keep going until the foreign domain root has been processed
         while (foreignDomain != lastForeignDomain);
      }
   }
   
   // send collected keys to hosts
   for (hash_map<string, vector<string> >::const_iterator hIt = keys.begin();
        hIt != keys.end(); hIt++)
   {
      assert(!hostportByUId[hIt->first].empty());
      sendUpdateForeignDomain(hIt->first, //uid
                              hIt->second, // keys 
                              changeType,
                              hostportByUId[hIt->first]);
   }
}

void
RegisterManager::sendUpdateForeignDomain(const string& uid,
                                         const vector<string>& keyList,
                                         PSCommon::ChangeType changeType,
                                         const string& hostport)
{
   TimeAccumulate t("RegisterManager::sendUpdateForeignDomain");

   try
   {
      // build http message
      PSProtocolService& proto = PSProtocolServiceHolder::Instance();

      Sptr<Http::Message> httpMsg;

      {
         TimeAccumulate t("RegisterManager::sendUpdateForeignDomain:construct");
         httpMsg = proto.buildForeignUpdateReq(PSCommon::FOREIGN_UPDATE_DOMAIN_STR, 
                                               uid, keyList, changeType);
      }
      
      string tempString = httpMsg->encode();

      DebugLog(<< "sendUpdateForeignDomain: " << tempString);

      {
         TimeAccumulate t("RegisterManager::sendUpdateForeignDomain:send");
         _streamMap.write(hostport, tempString);
      }
      
   }
   catch (VNetworkException& e)
   {
      WarningLog(<<  "Caught: " << e);
      WarningLog(<< "sendUpdateForeignDomain(" << uid << ", " << hostport << ") failed to contact client -- removing updates");
      unregisterUpdates(hostport);
   }
   catch(VException& e)
   {
      ErrLog(<<  "Caught: " << e);
      assert(false);
   }
}

string
RegisterManager::getPrimaryDomain(const string& foreignRootDomain,
                                  const string& foreignField)
{
   Threads::ReadLock lock(_mapMutex);

   ForeignDomainToPrimaryDataMap::const_iterator it = primaryRootDomainMap.find(foreignRootDomain);
   assert(it != primaryRootDomainMap.end());

   for (list<PrimaryData>::const_iterator i = it->second.begin();
        i != it->second.end(); i++)
   {
      if (i->_foreignField == foreignField)
      {
         return i->_primaryDomain;
      }
   }
   assert(false);
}

set<string>
RegisterManager::getForeignFields(const string& foreignRootDomain)

{
   set<string> foreignFields;
   Threads::ReadLock lock(_mapMutex);

   ForeignDomainToPrimaryDataMap::const_iterator it = primaryMap.find(foreignRootDomain);
   if (it != primaryRootDomainMap.end())
   {
      for (list<PrimaryData>::const_iterator i = it->second.begin();
           i != it->second.end(); i++)
      {
         foreignFields.insert(i->_foreignField);
      }
   }

   return foreignFields;
}

