#ifndef REGISTERMANAGER_HXX
#define REGISTERMANAGER_HXX

static const char* const RegisterManager_hxx_Version = "$Id: RegisterManager.hxx,v 1.1 2002/09/30 04:55:16 jason Exp $";

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



#include <list>
#include <ext/hash_map>

#include "global.h"
#include "DBResultset.hxx"
#include "PSMessage.hxx"
#include "RWMutex.hxx"
#include "Lock.hxx"
#include "Sptr.hxx"
#include "StringHash.hxx"
#include "PSProtocolService.hxx"
#include "LockedStreamMap.hxx"

namespace Epoisses {

// storage for set of keys and functor id
struct KeyListUId
{
      KeyListUId(const FieldListtype &_keyList,
                 const string &_uid)
         : keyList(_keyList),
           uid(_uid)
      {}
         
      FieldListtype keyList;
      string uid;
};

// storage for primary domain, foreign field, hostport, uid 
struct PrimaryData
{
      // to store leaf relations
      PrimaryData(const string& primaryDomain,
                  const string& foreignField,
                  const string& hostport,
                  const string& uid)
         : _primaryDomain(primaryDomain),
           _foreignField(foreignField),
           _hostport(hostport),
           _uid(uid)
      {}

      // to store root relations
      PrimaryData(const string& primaryDomain,
                  const string& foreignField)
         : _primaryDomain(primaryDomain),
           _foreignField(foreignField)
      {}

      const string _primaryDomain;
      const string _foreignField;
      const string _hostport;
      const string _uid;
};

// domain => (hostport => list<KeyListUId>) // !dlb! pair<Mutex*, list<KeyListUId>> ; use Mutex for send..
typedef hash_map <string, hash_map<string, list<Sptr<KeyListUId> > > > KeyListUIdMap;
// hostport => list<UId>
typedef hash_map<string, list<string> > HostUIdMap;
// domain => (hostport => UIds)
typedef hash_map <string, HostUIdMap > UIdMap;
// key => list<hostport>
typedef hash_map<string, list<string> > KeyToHosts;
// domain => (key => list<hostport>)
typedef hash_map <string, KeyToHosts >  DomainToKeyToHostsMap;
// domain => list<key>
typedef hash_map<string, list<string> > DomainToKeysMap;
// host => domain => keylist
typedef hash_map<string, DomainToKeysMap > HostToDomainToKeysMap;
// domain => vector<keys> (used locally in checkDomainList)
typedef hash_map<string, vector<string> > DomainToKeyMap;

// domain => parent
typedef hash_map<string, string > DomainMap;

// foreign domain => foreign key
typedef hash_map<string, string > ForeignKeyMap;
// foreign domain => list<(primary domain, hostport, UId)>
typedef hash_map<string, list<PrimaryData> > ForeignDomainToPrimaryDataMap;

/** RegisterManager manages the registrations and updates for the clients
 *  of the provisioning server.
 */
class RegisterManager
{
   public:
      /** Default Constructor. Does nothing for now.
       */
      RegisterManager();

      /** Default Destructor.  Does nothing for now.
       */ 
      ~RegisterManager();

      /** Registers a host to receive updates on a given domain.
       */
      bool registerDomainForUpdate( const string& domain,
                                    const string& hostport,
                                    const string& uid);

      /** UnRegisters a host from receiving updates on a given domain.
       */
      bool unregisterDomainForUpdate( const string& domain,
                                      const string& hostport,
                                      const string& uid);

      bool registerDomainForForeignUpdate(const string& foreignDomain,
                                          const string& foreignKeyField,
                                          const string& primaryDomain,
                                          const string& hostport,
                                          const string& uid);

      bool unregisterDomainForForeignUpdate(const string& foreignDomain,
                                            const string& hostport,
                                            const string& uid);

      /// Remove all registered updates for this host port (clean up lost connection)
      void unregisterUpdates(const string& hostport);

      /// Is anyone listening to this domain for updates?
      bool hasRegistered(const string& domain) const;
      /// Is anyine listening to this or parent domain for updates?
      bool hasRegisteredRecurse(const string& domain) const;
      /// Is anyone listening to this domain for foreign updates?
      bool hasForeignRegistered(const string& foreignDomain) const;

      string getPrimaryDomain(const string& foreignRootDomain, 
                              const string& foreignField);

      /**
         Get the column names that contain a foreign key. */
      set<string> getForeignFields(const string& foreignDomain);

      /** Checks the domain and record maps to see if a registered host needs
       *  to be updated.
       */
      void checkRegisterList(DBResultset* resultset,
                             PSCommon::ChangeType changeType);

      /**
         Determines the updates to send for foreign key updates. */
      void checkForeignRegisterList(const string& foreignField,
                                    DBResultset* resultset,
                                    PSCommon::ChangeType changeType);


      //
      string errorMsg;

   private:

      /// registered keysets by domain then hostport
      KeyListUIdMap keyListUIdMap;
      /// registered domains by domain then hostport
      UIdMap uIdMap;
      /// maping from domain to key to hostports
      DomainToKeyToHostsMap domainToKeyToHostsMap;

      /// map from foreign domain root to primary domain root plus foreignField
      ForeignDomainToPrimaryDataMap primaryRootDomainMap;
      /// map from foreign domain to (primary domain, foreign field, hostport, UId)
      ForeignDomainToPrimaryDataMap primaryMap;
      /// map from foreign domain to foreign key field name
      ForeignKeyMap foreignKeyMap;
      /// domain => hostport => uid for primary domains
      UIdMap primaryUIdMap;
      // uid => hostport
      hash_map<string, string > hostportByUId;

      /// Read/Write mutex for the all maps
      mutable Vocal::Threads::RWMutex _mapMutex;

      /// mutex for sending updates to clients (one per host spomeday) !dlb!
      mutable Vocal::Threads::Mutex _hostMutex;

      LockedStreamMap _streamMap;

      /** Checks the domain map to see whether a registered host needs
       *  to be updated.
       */
      void checkRecordList(DBResultset* resultset,
                           PSCommon::ChangeType changeType);

      /** Sends an UpdateRecord request to a registered host.
      void sendUpdateRecords(const string& domain, 
                             const list<string>& keys,
                             const string& hostport );
       */

      /** Sends an UpdateDomain request to a registered host.
       */
      bool sendUpdateDomain (const string& domain,
                             const vector<string>& keyList,
                             PSCommon::ChangeType changeType,
                             const string& hostport );

      /** Sends an UpdateForeignDomain request to a registered host.
       */
      void sendUpdateForeignDomain(const string& domain,
                                   const vector<string>& keyList,
                                   PSCommon::ChangeType changeType,
                                   const string& hostport);

      // .dlb. unused
      /** Sends an UpdateRecord request to a registered host.
      void sendRecordDelete (const string& domain,
                             const string& key,
                             const string& hostport );
       */

      // .dlb. unused
      /** Sends an UpdateDomain request to a registered host.
      void sendDomainDelete( const string& domain,
                             const vector<string>& keyList,
                             const string& hostport );

       */
};

} // namespace Epoisses
   
#endif
