/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const SessionManager_cxx_Version = "$Id: SessionManager.cxx,v 1.1 2002/09/30 04:55:16 jason Exp $";

#include <stdio.h>
#include <unistd.h>

#include "SessionManager.hxx"
#include "Logger.hxx"
#include "RandomHex.hxx"

#define NUM_RANDOMNESS 16  // 128 bits of randomness
#define CACHE_DELETE_TIME  1800     // in seconds (1800 sec = 30 min) 
#define CACHE_POLLING_TIME 1        // in seconds
#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace Epoisses;

void
SessionManager::thread()
{
   DebugLog (<< "Start SessionManager");
   return;
   
   while ( true )
   {
      try
      {
         cleanCache();
      }
      catch ( VException& e )
      {
         ErrLog (<< "Fatal exception. Ignoring");
         DebugLog (<< "caught: " << e);
      }
      catch ( ... )
      {
         ErrLog (<< "Fatal unknown exception. Ignoring");
      }

      if ( isShutdown() == true )
      {
         DebugLog (<< "Shutting down SessionManager");
         return;
      }
       
      sleep( CACHE_POLLING_TIME); 
   }
}

SessionManager::SessionManager()
{
}

SessionManager::~SessionManager()
{
   shutdown();
   join();
}

timeval
SessionManager::getTimeStamp()
{
   timeval now;

   int err = gettimeofday( &now, NULL );
   assert( !err );

   return now;
}


string
SessionManager::genSessionId()
{
   return RandomHex::get(NUM_RANDOMNESS).convertString();
}

void
SessionManager::addCacheRequest( const string sessionId,
                                 const string domain,
                                 const string filter,
                                 const string sortType,
                                 const string sortField,
                                 DBResultset* resultset)
{
   // Populate the data structure to be cached.

   // Create a new ProvData... need to delete when removing from cache.

   ProvData* pData  = new ProvData;
   pData->sessionId = sessionId;
   pData->domain    = domain;
   pData->filter    = filter;
   pData->sortType  = sortType;
   pData->sortField = sortField;
   pData->resultset = resultset;
   pData->timeStamp = getTimeStamp();

   DebugLog( <<  "TimeStamp():  sec = " << pData->timeStamp.tv_sec << ", usec = " << pData->timeStamp.tv_usec );

   Vocal::Threads::WriteLock lock(cacheLock);
   myCache[ sessionId ] = pData;
}


int
SessionManager::setCache( const string sessionId,
                          const string domain,
                          const string filter,
                          const string sortType,
                          const string sortField )
{
   int retCode = 1;

   ProvData *pData;

   Vocal::Threads::WriteLock lock(cacheLock);
   ProvRequestMap::iterator itr = myCache.find ( sessionId );

   if ( itr!= myCache.end() )
   {
      pData = itr->second;
      pData->domain     = domain;
      pData->filter     = filter;
      pData->sortType   = sortType;
      pData->sortField  = sortField;
      pData->timeStamp  = getTimeStamp();
   }
   else
   {
      DebugLog( <<  "Error setting cache...  Session ID not found");        
      retCode = 0;
   }

   return retCode; 
}

int
SessionManager::modifyCache( const string sessionId,
                             const string domain,
                             const string filter,
                             const string sortType,
                             const string sortField )
{
   int retCode = 1;
   Vocal::Threads::WriteLock lock(cacheLock);

   ProvRequestMap::iterator itr = myCache.find ( sessionId );
   ProvData* pData = itr->second;

   if ( itr != myCache.end() )
   {
      DebugLog( <<  "*** Modifying Cache Entry <" << sessionId << "> ***");
       
      if ( domain != "")
      {
         DebugLog( <<  "domain (old) = " << pData->domain );
         DebugLog( <<  "domain (new) = " << domain );  
         pData->domain = domain;
      }

      if ( filter != "")
      {
         DebugLog( <<  "filter (old) = " << pData->filter );  
         DebugLog( <<  "filter (new) = "<< filter );

         pData->filter = filter;
      }

      if ( sortType != "")
      {
         DebugLog( <<  "sortType (old) = " << pData->sortType ); 
         DebugLog( <<  "sortType (new) = " << sortType );

         pData->sortType = sortType;
      }

      if ( sortField != "")
      {
         DebugLog( <<  "sortField (old) = " << pData->sortField );        
         DebugLog( <<  "sortField (new) = " << sortField );

         pData->sortField = sortField;
      }

      pData->timeStamp = getTimeStamp();

      DebugLog (<<  "***  End Modify ***");
   }
   else
   {
      DebugLog( <<  "Error modifing cache... Session ID not found");
      retCode = 0;
   }

   return retCode;

}

void
SessionManager::getCacheRequest( const string sessionId, ProvData& pData )
   throw (VMissingDataException&)
{
   // NOTE:  This only locks and unlocks the cache.  The struct still
   //        needs to be lock when accessing.

   ProvRequestMap::iterator itr;
   {
      Vocal::Threads::ReadLock lock(cacheLock);
      itr = myCache.find( sessionId );
   }
   
   if ( itr != myCache.end() )
   {
      // returning an actual copy
      pData.domain    = (itr->second)->domain;
      pData.filter    = (itr->second)->filter;
      pData.sortType  = (itr->second)->sortType;
      pData.sortField = (itr->second)->sortField;

      pData.resultset = (itr->second)->resultset;  // not safe... copying the ptr
      pData.sessionId = (itr->second)->sessionId;
      pData.timeStamp = (itr->second)->timeStamp;
   }
   else
   {
      throw VMissingDataException (
         "CacheError:  SessionID not found", __FILE__, __LINE__, 0 );
   }
}

int 
SessionManager::deleteCacheRequest( string sessionId )
{
   Vocal::Threads::WriteLock lock(cacheLock);
   ProvRequestMap::iterator itr = myCache.find ( sessionId );

   if ( itr != myCache.end() )
   {
      DebugLog( <<  "*** Session id <" << sessionId << "> delete ***");

      // delete the DBResult
      if ( (itr->second)->resultset != 0 )
      {
         delete (itr->second)->resultset;
      };

      // delete ProvData
      delete itr->second;

      // remove from cache.
      myCache.erase( itr );

      return 1;
   }
   else
   {
      DebugLog( <<  "*** Session id <" << sessionId << "> not in cache ***");
      return 0;
   }
}


void
SessionManager::showCacheEntry(const string sessionID)
{
   ProvRequestMap::iterator itr;
   {
      Vocal::Threads::ReadLock lock(cacheLock);
      itr = myCache.find( sessionID );
   }
   
   if ( itr != myCache.end() )
   {
      DebugLog(<<  "--=--=--=--=--=--=--=--=--=--=--=--=--");
      DebugLog(<<  " Cache Entry");
      DebugLog(<<  " sessionID:     " << sessionID );
      DebugLog(<<  " domain:        " << (itr->second)->domain ); 
      DebugLog(<<  " filter:        " << (itr->second)->filter ); 
      DebugLog(<<  " sortType:      " << (itr->second)->sortType ); 
      DebugLog(<<  " sortField:     " << (itr->second)->sortField ); 
      DebugLog(<<  "--=--=--=--=--=--=--=--=--=--=--=--=--");
   }
   else
   {
      DebugLog(<<  "*** session ID <" << sessionID << "> not found ***");
   }
}

void
SessionManager::cleanCache()
{
   //DebugLog( <<  "*** clean cache ***");

   Vocal::Threads::WriteLock lock(cacheLock);
   ProvData* pData;
   ProvRequestMap::iterator itr = myCache.begin();

   timeval now;
   int err = gettimeofday( &now, NULL);
   assert (!err);

   while( itr != myCache.end() )
   {
      timeval now;
      int err = gettimeofday( &now, NULL);
      assert (!err);

      pData = itr->second;

      if ( (now.tv_sec - pData->timeStamp.tv_sec) > CACHE_DELETE_TIME)
      {
         DebugLog( <<  "Clean Moo:  ** sec " << pData->timeStamp.tv_sec << "  **, Now: ** sec = " << now.tv_sec << " **");

         ProvRequestMap::iterator itr2 = itr ++;

         // deletes DBResultset
         if ( (itr2->second)->resultset != 0)
         {
            delete (itr2->second)->resultset;
         }

         // deletes ProvData
         delete itr2->second;

         // remove from cache
         myCache.erase( itr2 );
         DebugLog( <<  "Cache size (after delete):  " << myCache.size() );            
      }
      else
      {
         itr ++;
      }
   }
}

