#ifndef DBCONNECTION_HXX
#define DBCONNECTION_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <string>
#include <vector>
#include <map>
#include "DBTypes.h"

namespace Epoisses {

/// Forward declare DBResultset
class DBResultset;

/// DBConnection class- Abstract client interface class
class DBConnection
{
   public:
      /// Constructor
      DBConnection(){};

      /// Destructor
      virtual ~DBConnection(){};

      /// Return a resultset object
      virtual DBResultset* DBSelect (const string& Domain, 
                                     const string& SearchFilter, 
                                     const FieldListtype& FieldList, 
                                     const Ordertype& Order, 
                                     const string &OrderField) = 0;

      /// Return a resultset object
      virtual DBResultset* DBSelect (const string& Domain, 
                                     const string& SearchFilter, 
                                     const FieldListtype& FieldList) = 0;

      /// Return a resultset object
      virtual DBResultset* DBSelect(const string& Domain, 
                                    const string& SearchFilter) = 0;

      /// Return a resultset object
      virtual DBResultset* DBSelect (const string& Domain,
                                     const string& SearchFilter,
                                     const string& start,
                                     const string& end,
                                     const FieldListtype& FieldList) = 0;

      /// Return a resultset object
      virtual DBResultset* DBSelect( const string& Domain,
                                     const string& SearchFilter,
                                     const string& start,
                                     const string& end,
                                     const Ordertype& Order,
                                     const string &OrderField,
                                     const FieldListtype& FieldList) = 0;

      /// Return a string containing primary key of Domain
      virtual string DBPrimaryKey ( const string& Domain) = 0;

      /// Return the parent of the domain
      virtual string DBParent(const string& domain)
      {
         // by default, there is no inheritance
         return domain;
      }

      /// Return the children of the domain
      virtual vector<string> DBChildren(const string& domain)
      {
         // by default, there is no inheritance
         vector<string> empty;
         return empty;
      }

      /// Return the root of the domain
      virtual string DBRoot(const string& domain)
      {
         // by default, there is no inheritance
         return domain;
      }

      /// Return a table describing fields/types/keys in Domain
      virtual ResultTabletype DBDescribe (const string& Domain) = 0;

      /// Return a list of domains in database
      virtual ResultRecordtype DBDescribe (void) = 0;

      /// Update Domain, based on searchfilter
      virtual void DBUpdate (const string& Domain, 
                             const string& SearchFilter, 
                             const FieldValuetype& FieldValueSet) = 0;

      /// Update Domain, based on previous resultset
      virtual void DBUpdate (const string& Domain, DBResultset* Resultset, 
                             const FieldValuetype& FieldValueSet) = 0;

      /// Insert record
      virtual int DBInsert (const string& Domain, 
                            const FieldValuetype& FieldValueSet) = 0;

      /// Delete rows based on a seachfilter
      virtual void DBDelete (const string& Domain, 
                             const string& SearchFilter) = 0;

      /// Delete rows based on a previous resultset
      virtual void DBDelete (const string& Domain, 
                             DBResultset* Resultset) = 0;

      /// Create a new domain, use STRINGTYPE for all columns
      /// with or without checking current existence
      virtual void DBCreate (const string& Domain, 
                             const FieldListtype& FieldsList, 
                             const string& PrimaryKey,
                             bool ExistenceCheck ) = 0;

      /// Create a new domain, specify types for each column
      /// with or without checking current existence
      virtual void DBCreate (const string& Domain, 
                             const FieldListtype& FieldsList, 
                             const FieldListDatatype& Datatypelist, 
                             const string& PrimaryKey,
                             bool ExistenceCheck ) = 0;

      /// Create a new domain that inherits from an existing domain,
      /// with or without checking current existence
      // Note: primary key is inherited and may not be specified here
      virtual void DBCreate(const string& domain, 
                            const string& parentDomain,
                            const FieldListtype& fieldsList, 
                            const FieldListDatatype& datatypelist, 
                            bool existenceCheck)
      {
         // not implemented by default -- could be pure virtual
         // when other implementations do inheritance
         assert(false);
      }

      /// Alter a domain, drop field
      virtual void DBAlter(const string& Domain, 
                           const string& Fieldname) = 0;
         
      /// Alter a domain, add a field
      virtual void DBAlter(const string& Domain, const string& Fieldname,
                           const Datatype Newtype) = 0;

      /// Drop a domain, with or without checking current existence
      virtual void DBDrop(const string& Domain, 
                          bool ExistenceCheck ) = 0;

      /// Returns true if Domain exists, false otherwise
      virtual bool DBVerify(const string& Domain) = 0;

      virtual DBResultset* DBRetrieveByOid(const string& Domain, 
                                           uint oid,
                                           const FieldListtype& FieldList) = 0;
      
};

} // namespace Epoisses

#endif
