// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//


#include <iostream>
#include <strstream>
#include <string>
#include <stdio.h>
#include <boost/lexical_cast.hpp>

#include "FastDBConnection.hxx"
#include "FastDBResultSet.hxx"
#include "Lock.hxx"
#include "Logger.hxx"
#include "TimeAccumulate.hxx"
#include "VDbException.hxx"
#include <fastdb/fastdb.h>
// .dlb. force loading exception.h -- must have fastdb.h before
#include <fastdb/exception.h>
#include <fastdb/reference.h>

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace std;
using namespace Epoisses;

class FastDbRegistrationBinding
{
   public:
      char const* bindingid;
      char const* groupid;
      char const* contact;
      float qvalue;
      char const* callid;
      int cseq;
      int expires;
      
      TYPE_DESCRIPTOR((KEY(bindingid, INDEXED),
                       FIELD(groupid),
                       FIELD(contact),
                       FIELD(qvalue),
                       FIELD(callid),
                       FIELD(cseq),
                       FIELD(expires)));

      string dumpRecord()
      {
         stringstream stream;
         
         stream << "bindingid: " << bindingid << endl
                << "groupid: " << groupid << endl
                << "contact: " << contact << endl
                << "qvalue: " << qvalue << endl
                << "callid: " << callid << endl
                << "cseq: " << cseq << endl
                << "expires: " << expires << endl;
         
         return stream.str();

      }
};

REGISTER(FastDbRegistrationBinding);


ResultRecordtype buildRecord(dbCursor<FastDbRegistrationBinding>& cursor, const FieldListtype& fieldList)
{
   ResultRecordtype record;
   
   for (FieldListtype::const_iterator it = fieldList.begin(); 
        it != fieldList.end(); it++)
   {
      string col = *it;
      transform(col.begin(),col.end(),col.begin(), ::tolower);
      
      if (col[0] == '\'')
      {
         record.push_back(col.substr(1, col.size() - 2));
      }
      else if (col == "groupid")
      {
         record.push_back(cursor->groupid);
      }
      else if (col == "contact")
      {
         record.push_back(cursor->contact);
      }
      else if (col == "qvalue")
      {
         record.push_back(boost::lexical_cast<string>(cursor->qvalue));
      }
      else if (col == "callid")
      {
         record.push_back(cursor->callid);
      }
      else if (col == "cseq")
      {
         record.push_back(boost::lexical_cast<string>(cursor->cseq));
      }
      else if (col == "expires")
      {
         record.push_back(boost::lexical_cast<string>(cursor->expires));
      }
      else if (col == "bindingid")
      {
         record.push_back(cursor->bindingid);
      }
      else
      {
         ErrLog( << "Unknown column: " << col);
         assert(0);
      }
   }
   return record;
}

FastDBConnection* FastDBConnection::Instance = 0;

FastDBConnection* FastDBConnection::instance()
{
   if (FastDBConnection::Instance != 0)
   {
      return FastDBConnection::Instance;
   }
   static Vocal::Threads::Mutex mutex;
   Vocal::Threads::Lock lock(mutex);
   if (FastDBConnection::Instance == 0)
   {
      FastDBConnection::Instance = new FastDBConnection;
   }
   return FastDBConnection::Instance;
}
   

FastDBConnection::FastDBConnection() 
   : db(dbDatabase::dbAllAccess, 1024*1024*10, 4*1024*1024, 1024*1024*2)
{
   if (!db.open("/tmp/registrationdb") )
   { 
      ErrLog(<< "Could not create in memory database.");
      assert(0);
   }
   InfoLog(<< "Database opened.");
   db.setConcurrency(20);
}

FastDBConnection::~FastDBConnection()
{
   db.close();
}


DBResultset* 
FastDBConnection::DBSelect (const string& Domain, 
                            const string& SearchFilter, 
                            const FieldListtype& FieldList)
{
   string domain(Domain);
   transform(domain.begin(), domain.end(), domain.begin(), ::tolower);

   if (domain != "registrationbindings")
   {
      throw VDbException("Domain not found: " + Domain, __FILE__, __LINE__);
   }

   dbCursor<FastDbRegistrationBinding> cursor;
   
   dbQuery query;
   query.pos = 0;
   query = SearchFilter.c_str();

   ResultTabletype* values = new ResultTabletype();

   try 
   {
      db.attach();
      if (cursor.select(query, dbCursorViewOnly) > 0)
      {
         ResultRecordtype record;
         do { 
            values->push_back(buildRecord(cursor, FieldList));
         } while (cursor.next());
      }
      db.detach();
   }
   catch(dbException& e)
   {
      ErrLog( << "FastDbException::DBSelect: ErrorCode: " << e.getErrCode() << ", " << e.getMsg() << " " << e.getArg());
      ErrLog( << "Domain: " << domain << " SearchFilter: " << SearchFilter);
      db.detach();
      assert(0);
   }

   return new FastDBResultset(SearchFilter, FieldList, values);
}

int 
FastDBConnection::DBInsert (const string& Domain, 
                            const FieldValuetype& FieldValueSet)
{
   TimeAccumulate t("FastDBConnection::DBInsert");

   FieldValuetype fieldValueSet(FieldValueSet);
   FastDbRegistrationBinding record;
   
   for (FieldValuetype::iterator p = fieldValueSet.begin(); p != fieldValueSet.end(); p++)
   {
      string col = p->first;
      transform(col.begin(), col.end(), col.begin(), ::tolower);

      if (col == "groupid")
      {
         record.groupid = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
      }
      else if (col == "contact")
      {
         record.contact = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
      }
      else if (col == "qvalue")
      {
         record.qvalue = boost::lexical_cast<float>(p->second);
      }
      else if (col == "callid")
      {
         record.callid = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
      }
      else if (col == "cseq")
      {
         record.cseq = boost::lexical_cast<int>(p->second);
      }
      else if (col == "expires")
      {
         record.expires = boost::lexical_cast<int>(p->second);
      }
      else 
      {
         WarningLog(<< "no match for field: " << col);
      }
   }
      
   int localId;
   {
      static int id = 0;
      static Vocal::Threads::Mutex mutex;
      Vocal::Threads::Lock lock(mutex);
      localId = id++;
   }
   DebugLog( << "id: " << localId);
   string tmp = boost::lexical_cast<string>(localId);
   record.bindingid = tmp.c_str();
   
   DebugLog( << "Trying to insert: " << record.dumpRecord());

   try 
   {
      db.attach();
      int ret = insert(record).getOid();
      db.detach();
      return ret;
   }
   catch(dbException& e)
   {
      ErrLog( << "FastDBConnection::DBInsert: ErrorCode: " << e.getErrCode() << ", " << e.getMsg() << " " << e.getArg());
      db.detach();
      assert(0);
   }
}
   
void 
FastDBConnection::DBUpdate (const string& Domain, 
                            const string& SearchFilter, 
                            const FieldValuetype& FieldValueSet)
{
   TimeAccumulate t("FastDBConnection::DBUpdate");

   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);

   FieldValuetype fieldValueSet(FieldValueSet);

   if (domain != "registrationbindings")
   {
      throw VDbException("Domain not found: " + Domain, __FILE__, __LINE__);
   }

   dbCursor<FastDbRegistrationBinding> cursor;
   
   dbQuery query;
   query.pos = 0;
   query = SearchFilter.c_str();

   try
   {
      db.attach();
      if (cursor.select(query, dbCursorForUpdate) > 0)
      {
         do
         {
            for (FieldValuetype::iterator p = fieldValueSet.begin(); p != fieldValueSet.end(); p++)
            {
               string col = p->first;
               transform(col.begin(), col.end(), col.begin(), ::tolower);
            
               if (col == "groupid")
               {
                  cursor->groupid = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
               }
               else if (col == "contact")
               {
                  cursor->contact = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
               }
               else if (col == "qvalue")
               {
                  cursor->qvalue = boost::lexical_cast<float>(p->second);
               }
               else if (col == "callid")
               {
                  cursor->callid = (p->second = p->second.substr(1, p->second.size() - 2)).c_str();
               }
               else if (col == "cseq")
               {
                  cursor->cseq = boost::lexical_cast<int>(p->second);
               }
               else if (col == "expires")
               {
                  cursor->expires = boost::lexical_cast<int>(p->second);
               }
            }
            cursor.update();
         } while (cursor.next());
      }
      db.detach();
   }
   catch(dbException& e)
   {
      ErrLog( << "FastDBConnection::DBUpdate ErrorCode: " << e.getErrCode() << ", " << e.getMsg() << " " << e.getArg());
      db.detach();
      assert(0);
   }
}

/// Delete rows based on a seachfilter
void 
FastDBConnection::DBDelete (const string& Domain, 
                            const string& SearchFilter)
{
   TimeAccumulate t("FastDBConnection::DBDelete");

   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);

   if (domain != "registrationbindings")
   {
      throw VDbException("Domain not found: " + Domain, __FILE__, __LINE__);
   }

   dbCursor<FastDbRegistrationBinding> cursor;
   
   dbQuery query;
   query.pos = 0;
   query = SearchFilter.c_str();

   DebugLog( << "trying to select with: " << SearchFilter);
   try
   {
      db.attach();
      if (cursor.select(query, dbCursorForUpdate) > 0)
      {
         DebugLog( << "trying to delete all elements in cursor");
         cursor.removeAllSelected();
      }
      db.detach();
   }
   catch(dbException& e)
   {
      ErrLog( << "FastDBConnection::DBDelete ErrorCode: " << e.getErrCode() << ", " << e.getMsg() << " " << e.getArg());
      db.detach();
      assert(0);
   }
}

ResultTabletype 
FastDBConnection::DBDescribe (const string& Domain)
{
   db.attach();

   ResultTabletype res;
   ResultRecordtype record;
   record.push_back("bindingid");
   res.push_back(record);
   record[0] = "groupid";
   res.push_back(record);
   record[0] = "contact";
   res.push_back(record);
   record[0] = "qvalue";
   res.push_back(record);
   record[0] = "cseq";
   res.push_back(record);
   record[0] = "expires";
   res.push_back(record);

   db.detach();
   
   return res;
}

DBResultset* 
FastDBConnection::DBRetrieveByOid(const string& Domain, 
                                  oid_t oid,
                                  const FieldListtype& FieldList)
{
   DebugLog( << "FastDBConnection::DBRetrieveByOid");

   TimeAccumulate t("FastDBConnection::DBRetrieveByOid");

   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);
   
   if (domain != "registrationbindings")
   {
      throw VDbException("Domain not found: " + Domain, __FILE__, __LINE__);
   }
   
   dbCursor<FastDbRegistrationBinding> cursor;
   dbReference<FastDbRegistrationBinding> ref(oid);

   DebugLog( << "Retrieving using oid: " << ref.getOid());
   
   ResultTabletype* values = new ResultTabletype();
   
   try 
   {
      db.attach();
      if (cursor.at(ref) > 0)
      {
         values->push_back(buildRecord(cursor, FieldList));
      }
      db.detach();
   }
   catch(dbException& e)
   {
      ErrLog( << "FastDbException::DBRetrieveByOid: ErrorCode: " << e.getErrCode() << ", " << e.getMsg() << " " << e.getArg());
      db.detach();
      assert(0);
   }
   
   return new FastDBResultset("byOid", FieldList, values);
}

   
