/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



#include <iostream>
#include <string>
#include "Filterscanner.hxx"

using namespace std;
using namespace Epoisses;

// Character Code Map shared between Scanner and Token classes
CharCode charCodeMap[128];

// Token constructor
Token::Token(void)
{
    myTokencode = tcDummy;
    myTokenstring.erase();
};

// Get the Token Code
TokenCode
Token::GetCode() const 
{ 
    return myTokencode;   
}

// Get the Token String
string&
Token::GetTokenString()
{ 
    return myTokenstring; 
}
 
// Word Token
void
WordToken::Get(istrstream& buffer)
{
    char ch; 
    char tokenbuff[MAXTOKENSIZE];
    int i = 0;
 
    //--Extract the word.
    while ( ch = buffer.peek(),
          ( charCodeMap[ch] == ccLetter) || (charCodeMap[ch] == ccDigit) )
    {
        tokenbuff[i++] = buffer.get();
    } 
    tokenbuff[i] = '\0';

    myTokenstring = tokenbuff; 
    myTokencode = tcWord;
}

// String Token
void
StringToken::Get(istrstream& buffer)
{
    // Read to pass the 1st (" or ') quote.
    char ch = buffer.get();
    char tokenbuff[MAXTOKENSIZE];
    unsigned int i = 0;
 
    // Extract a string, read until another quote (" or ') is found
    while (ch = buffer.get(), true)
    {
        if ((charCodeMap[ch] == ccDQuote) || (charCodeMap[ch] == ccSQuote))
	  break;
	
        if (i >= MAXTOKENSIZE)
	  {
	    tokenbuff[i] = '\0';
	    myTokenstring += tokenbuff;
	    memset(tokenbuff, '\0', MAXTOKENSIZE);
	    i = 0;
	  }
        tokenbuff[i++] = ch;
    } 
    tokenbuff[i] = '\0';
 
    myTokenstring = tokenbuff; 
    myTokencode = tcString;
}

//  Extract one, or two character special symbols
void
SpecialToken::Get(istrstream &buffer)
{
    char ch = buffer.peek();
    char tokenbuff[MAXTOKENSIZE];
    int i = 0;

    tokenbuff[i++] = ch;

    switch (ch) 
    {
    case '(':
        myTokencode = tcLParen;
        buffer.get();  
        break;

    case ')':
        myTokencode = tcRParen;
        buffer.get();
        break;

    case '&':
        myTokencode = tcAnd;
        buffer.get();
        break;

    case '|':
        myTokencode = tcOr;
        buffer.get();
        break;

    case '!':
        myTokencode = tcNot;
        buffer.get();
        break;

    case '*':
        myTokencode = tcStar;
        buffer.get();
        break;

    case '=':
        myTokencode = tcEqual;
        buffer.get();
        break;

    case '<':   
        ch = buffer.get(); // < or <= 
        ch = buffer.peek();
        if (ch == '=') 
        {
            tokenbuff[i++] = '=';
            myTokencode  = tcLe;
            buffer.get();
        }
        else
        {
            myTokencode = tcLt;
        }
        break;

    case '>':
        ch = buffer.get();  // > or >=
        ch = buffer.peek(); 
        if (ch == '=') 
        {
            tokenbuff[i++] = '=';
            myTokencode  = tcGe;
            buffer.get();
        }
        else 
        {
            myTokencode = tcGt;
        }
        break;

    default:    
        myTokencode = tcError; // error
        buffer.get();
        break;
    }
    tokenbuff[i] = '\0';
    myTokenstring = tokenbuff; 
}

// EOS - End of String
void
EOSToken::Get(istrstream &buffer)
{
    myTokenstring.erase();
    myTokencode = tcEOS;
}

// ErrorToken -  Anything not recognized
void
ErrorToken::Get(istrstream &buffer)
{
    char tokenbuff[MAXTOKENSIZE];

    tokenbuff[0] = buffer.get();
    tokenbuff[1] = '\0';

    myTokenstring = tokenbuff;
}

// Scanner Constructor
Scanner::Scanner(const string& buffer):myInbuffer(buffer.c_str())
{
    myInbuffer.unsetf(ios::skipws); // Let Scanner class handle spaces
    int i;
 
    //Initialize the character code map
    for (i = 0; i <= 127; ++i) 
        charCodeMap[i] = ccError;
    for (i = 'a'; i <= 'z'; ++i) 
        charCodeMap[i] = ccLetter;
    for (i = 'A'; i <= 'Z'; ++i) 
        charCodeMap[i] = ccLetter;
    for (i = '0'; i <= '9'; ++i) 
        charCodeMap[i] = ccDigit;

    charCodeMap['+'] = ccSpecial;
    charCodeMap['-'] = ccSpecial;
    charCodeMap['/'] = ccSpecial;
    charCodeMap['^'] = ccSpecial;
    charCodeMap[','] = ccSpecial;
    charCodeMap['['] = ccSpecial;
    charCodeMap[']'] = ccSpecial;
    charCodeMap['{'] = ccSpecial;
    charCodeMap['}'] = ccSpecial;
    charCodeMap[':'] = ccSpecial;
    charCodeMap[';'] = ccSpecial;
    charCodeMap['.'] = ccSpecial;
    charCodeMap['*'] = ccSpecial;
    charCodeMap['!'] = ccSpecial;
    charCodeMap['&'] = ccSpecial;
    charCodeMap['|'] = ccSpecial;
    charCodeMap['='] = ccSpecial;
    charCodeMap['<'] = ccSpecial;
    charCodeMap['>'] = ccSpecial;
    charCodeMap['('] = ccSpecial;
    charCodeMap[')'] = ccSpecial;
    charCodeMap['*'] = ccSpecial;
    charCodeMap['\''] = ccSQuote;
    charCodeMap['\"'] = ccDQuote;
    charCodeMap[' ' ] = ccWhitespace;
    charCodeMap['\t'] = ccWhitespace;
    charCodeMap['\n'] = ccWhitespace;
    charCodeMap['\0'] = ccWhitespace;
    charCodeMap[-1] = ccEOS;    // End of String
}

// GetToken - Return ptr to new Token
Token* Scanner::GetToken(void)
{
    SkipWhitespace();
    char ch = myInbuffer.peek();
    Token* tp;

    // Determine the token class, based on the current character.
    switch (charCodeMap[ch]) 
    {
        case ccLetter:
            tp = &myWordToken;
            break;
        case ccDQuote:
            tp = &myStringToken;  
            break;
        case ccSQuote:
            tp = &myStringToken;  
            break;
        case ccSpecial:
            tp = &mySpecialToken; 
            break;
        case ccEOS:
            tp = &myEOSToken;
            break;
        default:
            tp = &myErrorToken;   
            break;
    }
 
    // Extract a token of that class, and return a pointer to it
    tp->Get(myInbuffer);

    return tp;
}

// SkipWhitespace
void
Scanner::SkipWhitespace(void)
{
    char ch;

    while (ch = myInbuffer.peek(),
           charCodeMap[ch] == ccWhitespace)
    {
        ch = myInbuffer.get();  // Skip whitespaces
    }
}
