#ifndef FILTERPARSER_HXX
#define FILTERPARSER_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



#include <vector>
#include <stack>
#include <string>
#include <strstream>
 
namespace Epoisses {
   
/// Type Definitions
/// Character Codes
enum CharCode
{
    ccDummy,
    ccLetter,
    ccDigit,
    ccSpecial,
    ccSQuote,
    ccDQuote,
    ccWhitespace,
    ccEOS,
    ccError,
};

/// Token Codes
enum TokenCode
{
    tcDummy,
    tcWord,
    tcString,
    tcLParen, 
    tcRParen,
    tcEqual,
    tcLt,
    tcGt,
    tcLe,
    tcGe,
    tcAnd,
    tcOr,
    tcNot,
    tcStar,
    tcEscape,
    tcEOS,
    tcError,
};

// Character Code Map shared between Scanner and Token classes
extern CharCode charCodeMap[128];

// Maximum size of a single Token
const unsigned int MAXTOKENSIZE = 64;

/// Token - Base class
class Token
{
    protected:
        ///
        TokenCode myTokencode;
        ///
        string myTokenstring;

    public:
        ///
        Token(void);

        ///
        virtual ~Token(){};

        ///
        TokenCode GetCode() const; 

        ///
        string& GetTokenString();

        /// Get- Abstract function implemented in each derived Token class
        virtual void Get(istrstream &buffer) = 0;
};

/// WordToken- associated with the key in key/value pairs
class WordToken : public Token
{
    public:
        ///
        WordToken(){};
        ///
        virtual ~WordToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// StringToken- associated with a value in key/value pairs
class StringToken : public Token
{
    public:
        ///
        StringToken(){};
        ///
        virtual ~StringToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// EscapeToken- Backslash '\'
class EscapeToken : public Token
{
    public:
        ///
        EscapeToken(){};
        ///
        virtual ~EscapeToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// SpecialToken- i.e. ~ ! @ # $ % ^ & * ( ) _ + ? > < | } { ... 
class SpecialToken : public Token
{
    public:
        ///
        SpecialToken(){};
        ///
        virtual ~SpecialToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// EOSToken- End of String
class EOSToken : public Token
{
    public:
        ///
        EOSToken(){};
        ///
        virtual ~EOSToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// ErrorToken - Anything else
class ErrorToken : public Token
{
    public:
        ///
        ErrorToken(){};
        ///
        virtual ~ErrorToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// Scanner class, used to scan and call correct Token class
class Scanner
{
    private:
        ///
        void SkipWhitespace(void);
        ///
        istrstream    myInbuffer;
        ///
        WordToken     myWordToken;
        ///
        StringToken   myStringToken;
        ///
        SpecialToken  mySpecialToken;
        ///
        EOSToken      myEOSToken;
        ///
        ErrorToken    myErrorToken;

    public:
        ///
        Scanner(const string& buffer);

        ///
        virtual ~Scanner(){};
        
        ///
        Token* GetToken(void);
};

} // namespace Epoisses

#endif
