/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <iostream>
#include <strstream>
#include <string>
#include <stdio.h>
#include "VDbException.hxx"
#include "PGSQLResultset.hxx"
#include "PGSQLConnection.hxx"
#include "Logger.hxx"
#include "Lock.hxx"
#include "TimeAccumulate.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace std;
using namespace Epoisses;

const string SEQ("_seq");

Vocal::Threads::RWMutex PGSQLConnection::parentMapMutex;
DomainToParentMap PGSQLConnection::domainToParent(1000);

// PGSQLConnection Constructor
PGSQLConnection::PGSQLConnection(const string& dbname, const string& dbhost,
                                 const string& dbuser, const string& dbpassword)
{
   char* pgport = NULL;      // port of the server
   char* pgoptions = NULL;   // special options to start up the server
   char* pgtty = NULL;       // debugging tty for the server

   InfoLog (<< "Connecting to database: host=" << dbhost
            << " name=" << dbname
            << " user=" << dbuser);

   myPgSQLConnectionP = PQsetdbLogin(dbhost.c_str(), 
                                     pgport, pgoptions, pgtty,
                                     dbname.c_str(), 
                                     dbuser.c_str(), 
                                     dbpassword.c_str());
    
   // check connection
   if (PQstatus(myPgSQLConnectionP) == CONNECTION_BAD)
   {
      ErrLog(<< "PGSQLConnection Error: Connection failed dbname=" 
             << dbname
             << ", dbhost=" << dbhost
             << ", error= " <<  PQerrorMessage(myPgSQLConnectionP));
      throw VDbException("PgSQLConnection Error: Connection failed to database ", 
                         __FILE__, __LINE__, 0);
   }
}

PGSQLConnection::~PGSQLConnection()
{ 
   PQfinish(myPgSQLConnectionP);
}


/// Validate the result, cleanup, and return it or throw
DBResultset*
PGSQLConnection::DBSelectHelper(const char* Sqlcmd)
{
   DebugLog(<< "DBSelectHelper(" << Sqlcmd << ")");

   PGresult*  pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
   PGSQLResultset* resultSet;

   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBSelect Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
                   
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog(<< "PGSQLConnection Error: select failed " << err);

      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   else
   {
      resultSet = new PGSQLResultset(pgresult, Sqlcmd);
   }
    
   // Free memory used by PGresult pointer
   //PQclear(pgresult);
   return resultSet;
}

// DBSelect- Returns a DBResultset
DBResultset*
PGSQLConnection::DBSelect(const string& Domain,
                          const string& SearchFilter, const FieldListtype& FieldList,
                          const Ordertype& Order, const string &OrderField)
{
   TimeAccumulate t("PGSQLConnection::DBSelect");

   char Sqlcmd[MAXQUERYLENGTH];

   // Create SQL statement
   CreateSQLselect(
      Sqlcmd, 
      Domain, 
      SearchFilter, 
      FieldList,
      Order, 
      OrderField);

   return DBSelectHelper(Sqlcmd);
}

// DBSelect- Returns a DBResultset
DBResultset*
PGSQLConnection::DBSelect(const string& Domain,
                          const string& SearchFilter, 
                          const FieldListtype& FieldList)
{
   TimeAccumulate t("PGSQLConnection::DBSelect");

   char Sqlcmd[MAXQUERYLENGTH];

   // Create SQL statement
   CreateSQLselect(Sqlcmd, Domain, SearchFilter, FieldList);

   return DBSelectHelper(Sqlcmd);
}

DBResultset*
PGSQLConnection::DBSelect(const string& Domain,
                          const string& SearchFilter )
{
   TimeAccumulate t("PGSQLConnection::DBSelect");

   char Sqlcmd[MAXQUERYLENGTH];

   // Create SQL statement
   CreateSQLselect(Sqlcmd, 
                   Domain, 
                   SearchFilter );

   return DBSelectHelper(Sqlcmd);
}

DBResultset*
PGSQLConnection::DBSelect(const string& Domain,
                          const string& SearchFilter,
                          const string& start,
                          const string& end,
                          const Ordertype& Order, 
                          const string &OrderField,
                          const FieldListtype& FieldList)
{
   TimeAccumulate t("PGSQLConnection::DBSelect");

   char Sqlcmd[MAXQUERYLENGTH];

   // Create SQL statement
   CreateSQLselect( Sqlcmd, 
                    Domain, 
                    SearchFilter, 
                    start, 
                    end, 
                    Order, 
                    OrderField, 
                    FieldList );

   return DBSelectHelper(Sqlcmd);
}

DBResultset*
PGSQLConnection::DBSelect(const string& Domain,
                          const string& SearchFilter, 
                          const string& start,
                          const string& end,
                          const FieldListtype& FieldList)
{
   TimeAccumulate t("PGSQLConnection::DBSelect");

   char Sqlcmd[MAXQUERYLENGTH];

   // Create SQL statement
   CreateSQLselect( Sqlcmd, 
                    Domain, 
                    SearchFilter, 
                    start, 
                    end, 
                    FieldList );

   return DBSelectHelper(Sqlcmd);
}

// DBPrimaryKey - returns the primary key for a domain
// NOTE:  assumes only one primary key exists.  If more
//        than one, then it'll get the first one from
//        the list. If none exists, don't throw exception,
//        just return null
string
PGSQLConnection::DBPrimaryKey (const string& Domain)
{
   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);

   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   string resultString;
   PGresult* pgresult;

   // traverse to the root of the domain inheritance
   string toRoot = domain;
   string lastDomain;
   do 
   {
      lastDomain = toRoot;
      toRoot = DBParent(toRoot);
   }
   while (toRoot != lastDomain);
    
   // This is PostgreSQL equivalent of mySQL's 'DESCRIBE TABLE' 
   buff << "SELECT pg_class.relname, pg_attribute.attname "
        << "FROM pg_class, pg_attribute, pg_index "
        << "WHERE pg_class.oid = pg_attribute.attrelid AND "
        << " pg_class.oid = pg_index.indrelid AND "
        << " pg_index.indkey[0] = pg_attribute.attnum AND "
        << " pg_class.relname = '" << toRoot << "' AND "
        << " pg_index.indisprimary = 't'; "
        << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBPrimaryKey Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      ErrLog(<< err);
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   // The tablename, index are the only value returned from 
   // this query. This query will be valid even if there
   // are no records returned, so return an empty string if 
   // no primary key exists
   if (PQntuples(pgresult) > 0 )
   {
      resultString = PQgetvalue(pgresult, 0, 1);
   }
   else
   {
      ErrLog(<< "DBPrimaryKey failed for " << domain);
      PQclear(pgresult);
      throw VDbException("DBPrimaryKey failed for : " + domain,
                         __FILE__, __LINE__);
   }

   // Free memory used by PGresult pointer
   PQclear(pgresult);

   // return string by value, an empty string if nothing found
   return resultString; 
}

string
PGSQLConnection::DBParent(const string& Domain)
{
   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);

   string resultString;

   {
      Vocal::Threads::ReadLock lock(parentMapMutex);

      // check cache
      DomainToParentMap::iterator dpIt = domainToParent.find(domain);
      if (dpIt != domainToParent.end())
      {
         DebugLog(<< "cached DBParent(" << domain << ") = " << dpIt->second);
         return dpIt->second;
      }
   }
   
   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);

   // postgres for extracting the parent domain assumes that there is one parent
   buff << "SELECT p.relname "
      "FROM pg_class p, pg_class c, pg_inherits i "
      "WHERE c.relname = '" << domain << "' "
      "AND i.inhrelid = c.oid and p.oid = i.inhparent" << ends;

   PGresult* pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBParent Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog(<< "DBParent failed: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   
   // The tablename is the only value returned from 
   // this query. This query will be valid even if there
   // are no records returned, so return the domain if
   // no parent exists
   if (PQntuples(pgresult) == 1)
   {
      resultString = PQgetvalue(pgresult, 0, 0);
   } 
   else
   {
      // why is there multiple inheritance?
      assert(PQntuples(pgresult) == 0);
      resultString = domain;
   }
   
   // Free memory used by PGresult pointer
   PQclear(pgresult);

   // cache the result for future requests
   {
      Vocal::Threads::WriteLock lock(parentMapMutex);
      domainToParent[domain] = resultString;
   }
   
   DebugLog(<< "DBParent(" << domain << ") = " << resultString);

   return resultString; 
}

// Return the parent table or the given domain if it is a root table.
vector<string>
PGSQLConnection::DBChildren(const string& Domain)
{
   string domain(Domain);
   transform(domain.begin(),domain.end(),domain.begin(), ::tolower);

   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);

   // postgres for extracting the parent domain assumes that there is one parent
   buff << "SELECT c.relname FROM pg_class p, pg_class c, pg_inherits i "
        << "WHERE p.relname = '" << domain << "' "
        << "AND i.inhrelid = c.oid and p.oid = i.inhparent" << ends;
   
   PGresult* pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBChildren Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   vector<string> children;

   for (int i = 0; i < PQntuples(pgresult); i++)
   {
      children.push_back(PQgetvalue(pgresult, i, 0));
   } 
   
   // Free memory used by PGresult pointer
   PQclear(pgresult);
   
   return children; 
}

string
PGSQLConnection::DBRoot(const string& LeafDomain)
{
   string leafDomain(LeafDomain);
   transform(leafDomain.begin(),leafDomain.end(), leafDomain.begin(), ::tolower);

   // loop over the parents of the leafDomain
   string domain = leafDomain;
   string lastDomain;
   do
   {
      // iterate over domain parents
      lastDomain = domain;
      domain = DBParent(domain);
   }
   // keep going until the domain root
   while (domain != lastDomain);
   
   //DebugLog(<< "DBRoot(" << leafDomain << ") = " << domain);

   return domain;
}

// DBDescribe- Returns information on one domain
ResultTabletype
PGSQLConnection::DBDescribe(const string& Domain)
{
   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   ResultRecordtype recordbuff;
   string fieldbuff;
   unsigned short recordnumber;
   PGresult* pgresult;

   ResultTabletype describeTable;

   // SQL statement for getting table information
   buff << "SELECT a.attname, a.atttypid, a.attnotnull "
        << " FROM pg_class c, pg_attribute a WHERE c.relname ilike '"
        << Domain << "' AND a.attnum > 0 AND a.attrelid = c.oid" << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDescribe Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog(<< "DBDescribe failed: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   // 
   for (recordnumber = 0; recordnumber < PQntuples(pgresult); recordnumber++)
   {
      // First column is Fieldname
      fieldbuff = PQgetvalue(pgresult, recordnumber, 0);
      recordbuff.push_back(fieldbuff);

      // Second column is integer type id format.
      // Convert the PgSQL datatype, to the DBmodule
      // supported Datatype.
      char* typebuff = PQgetvalue(pgresult, recordnumber, 1);
      switch (atoi(typebuff))
      {
         case 1043:   //varchar
            recordbuff.push_back("String");
            break;
         case 23:     //integer
            recordbuff.push_back("Integer");
            break;
         case 701:    //float  
            recordbuff.push_back("Float");
            break;
         case 16:     //bool
            recordbuff.push_back("Boolean");
            break;
         case 25:     //text
            recordbuff.push_back("Text");
            break;
         case 26:     //blob
            recordbuff.push_back("Blob");
            break;
         case 1184:   //timestamp
            recordbuff.push_back("Timestamp");
            break;
         case 1082:   //date
            recordbuff.push_back("Date");
            break;
         case 1083:   //time
            recordbuff.push_back("Time");
            break;
         default:
            ErrLog (<< "DBDescribe error due to unknown datatype");
            throw VDbException("DBDescribe Error: Unkown datatype", __FILE__, __LINE__, 0);
            break;
      }

      // Use the DBPrimaryKey method to determine if this field
      // is the primary key.
      string primarykey = DBPrimaryKey(Domain);
      if (fieldbuff == primarykey)
      {
         recordbuff.push_back("PRIMARYKEY");
      }


      // Insert record buffer into result table
      describeTable.push_back(recordbuff);

      // Clear recordbuff for next record
      recordbuff.clear();
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);

   return describeTable;
}

// DBDescribe- Returns a list of all domains in database
ResultRecordtype
PGSQLConnection::DBDescribe()
{
   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   string fieldbuff;
   unsigned short recordnumber;
   PGresult* pgresult;

   // Clear contents of DescribeRecord for each query

   ResultRecordtype describeRecord;

   // This is mysql's equivalent of "SHOW TABLES"
   buff  << "SELECT c.relname as \"Name\" "
         << "FROM pg_class c, pg_user u "
         << "WHERE c.relowner = u.usesysid AND c.relkind = 'r' "
         << "  AND c.relname !~ '^pg_' "
         << "  ORDER BY c.relname " << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDescribe Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBDescribe error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   //  Loop for each row
   for (recordnumber = 0; recordnumber < PQntuples(pgresult); recordnumber++)
   {
      // Only one column is retreived, the table name
      fieldbuff = PQgetvalue(pgresult, recordnumber, 0);
      describeRecord.push_back(fieldbuff);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);

   return describeRecord;
}

// DBUpdate- Updates a record in an existing domain
void
PGSQLConnection::DBUpdate(const string& Domain, const string& SearchFilter,
                          const FieldValuetype& FieldValueSet)
{
   TimeAccumulate t("PGSQLConnection::DBUpdate");

   char Sqlcmd[MAXQUERYLENGTH];
   PGresult* pgresult;

   // Create SQL statement
   CreateSQLupdate(Sqlcmd, Domain, SearchFilter, FieldValueSet);

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBUpdate Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBUpdate error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);
}

// DBUpdate- Updates a record in an existing domain, based on a
//           previous resultset.
void
PGSQLConnection::DBUpdate(const string& Domain, DBResultset* Resultset,
                          const FieldValuetype& FieldValueSet)
{
   TimeAccumulate t("PGSQLConnection::DBUpdate");

   char Sqlcmd[MAXQUERYLENGTH];
   PGresult* pgresult;

   // Get SearchFilter from resultset, use it to generate SQL command
   const string searchfilter = Resultset->GetSearchFilter();

   // Create SQL statement
   CreateSQLupdate(Sqlcmd, Domain, searchfilter, FieldValueSet);

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBSelect Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBUpdate error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);
}

// DBInsert- Adds a record to an existing domain
int
PGSQLConnection::DBInsert(const string& Domain,
                          const FieldValuetype& FieldValueSet)
{
   TimeAccumulate t("PGSQLConnection::DBInsert");

   char Sqlcmd[MAXQUERYLENGTH];
   PGresult* pgresult;

   // Create SQL statement
   CreateSQLinsert(Sqlcmd, Domain, FieldValueSet);

   //DebugLog(<< "DBInsert: " << Sqlcmd);

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBInsert Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBInsert error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   int r = PQoidValue(pgresult);
   // Free memory used by PGresult pointer
   PQclear(pgresult);
   return r;
}

// DBDelete- Deletes a record from an existing domain, based on a
//           searchfilter
void
PGSQLConnection::DBDelete(const string& Domain, const string& SearchFilter)
{
   TimeAccumulate t("PGSQLConnection::DBDelete");

   char Sqlcmd[MAXQUERYLENGTH];
   PGresult* pgresult;

   // Create SQL statement
   CreateSQLdelete(Sqlcmd, Domain, SearchFilter);

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDelete Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBDelete error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);
}

// DBDelete- Deletes a record from an existing domain, based on a
//           previous resultset
void
PGSQLConnection::DBDelete(const string& Domain, DBResultset* Resultset)
{
   TimeAccumulate t("PGSQLConnection::DBDelete");

   char Sqlcmd[MAXQUERYLENGTH];
   PGresult* pgresult;

   // Get SearchFilter from resultset, use it to generate SQL command
   const string searchfilter = Resultset->GetSearchFilter();

   // Create SQL statement
   CreateSQLdelete(Sqlcmd, Domain, searchfilter);

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDelete Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBDelete error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);
}

// DBCreate- Creates a new domain. This version assumes all columns to
//           be of datatype STRING, which is msyql CHAR(255)
void
PGSQLConnection::DBCreate(const string& Domain, const FieldListtype& FieldList,
                          const string& PrimaryKey, bool ExistenceCheck)
{
   assert(0);
    
   if (ExistenceCheck || !DBVerify(Domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      PGresult* pgresult;
      // Create SQL statement
      CreateSQLcreate(Sqlcmd, 
                      Domain, 
                      FieldList, 
                      PrimaryKey);
      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBCreate Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBCreate error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      } 
      PQclear(pgresult);
   }

   if (!DBVerifySequence(Domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      PGresult* pgresult;
      // Create SQL statement
      CreateSQLsequence(Sqlcmd,
                        Domain, 
                        PrimaryKey);      
      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBCreate Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBCreate error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      }
      PQclear(pgresult);
   }
}

// DBCreate- Creates a new domain. This version requires the client to
//           specify the datatype of each column
void
PGSQLConnection::DBCreate(const string& Domain, const FieldListtype& FieldList,
                          const FieldListDatatype& DatatypeList, const string& PrimaryKey, 
                          bool ExistenceCheck)
{   
   // Create the SQL command to create the sequence    
   if (ExistenceCheck || !DBVerify(Domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      PGresult* pgresult;

      // Create SQL statement
      CreateSQLcreate(Sqlcmd, 
                   Domain,
                   FieldList, 
                   DatatypeList, 
                   PrimaryKey);

      //DebugLog(<< Sqlcmd);
      
      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBCreate Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBCreate error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      } 
      PQclear(pgresult);
   }

   if (!DBVerifySequence(Domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      PGresult* pgresult;
      // Create SQL statement
      CreateSQLsequence(Sqlcmd,
                        Domain, 
                        PrimaryKey);      
      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBCreate Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBCreate error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      }
      PQclear(pgresult);
   }
}

void
PGSQLConnection::DBCreate(const string& domain, 
                          const string& parentDomain,
                          const FieldListtype& fieldList, 
                          const FieldListDatatype& datatypeList, 
                          bool existenceCheck)
{

   // Check if a data type exists for each field
   if (fieldList.size() != datatypeList.size())
   {
      ErrLog (<< "DBCreate error: number of datatypes does not match number of fields");
      throw VDbException("DBCreate Error: \
            Number of Datatypes does not match number of Fields",
                         __FILE__, __LINE__, 0);
   }

   if (existenceCheck || !DBVerify(domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      PGresult* pgresult;
      CreateSQLcreate(Sqlcmd, 
                      domain, 
                      parentDomain,
                      fieldList, 
                      datatypeList);

      //DebugLog(<< Sqlcmd);

      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBCreate(with parent) Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBCreate error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      }
      PQclear(pgresult);
   }   
}

class RowComparitor
{
   public:
      bool operator()(const vector<string>& r1,
                      const vector<string>& r2) const
      {
         return r1[0] < r2[0];
      }
};

// !dcm! -- temp table use is not threadsafe.  Cannot delete columns of domains
// that have children because the drop column workaround drops the base domain.
// DBAlter- Drop a column in a table
void
PGSQLConnection::DBAlter(const string& DomainIn, const string& FieldnameIn)
{
   string Domain = DomainIn;
   string Fieldname = FieldnameIn;   
   transform(Domain.begin(),Domain.end(),Domain.begin(), ::tolower);
   transform(Fieldname.begin(),Fieldname.end(),Fieldname.begin(), ::tolower);

   //  Check if table exists first
   if (!DBVerify(Domain))
   {
      string err = "DBPrimaryKey Error: Table does not exist: " + Domain;
      ErrLog (<< "DBAlter error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   
   ResultTabletype res;
   ResultTabletype all = DBDescribe(Domain);   

   string parent = DBParent(Domain);

   if (parent != Domain)
   {
      ResultTabletype derived = all;      
      ResultTabletype parentRes = DBDescribe(parent);
      static RowComparitor rowComp;      

      sort(derived.begin(), derived.end(), rowComp);
      sort(parentRes.begin(), parentRes.end(), rowComp);

      set_difference(derived.begin(), derived.end(), 
                     parentRes.begin(), parentRes.end(),
                     back_inserter(res),
                     rowComp);
   }
   else
   {
      res = DBDescribe(Domain);
   }
      
   // remove the column to drop (complain if primary)

   bool found = false;   
   for (ResultTabletype::iterator i = res.begin();
        i != res.end(); i++)
   {
      if ((*i)[0] == Fieldname)
      {
         if (i->size() == 3)
         {
            ErrLog (<< "Tried to remove primary key " << Fieldname);
            throw VDbException("Tried to remove primary key", __FILE__, __LINE__, 0);
         }
         res.erase(i);
         found = true;
         break;         
      }      
   }

   if (!found)
   {
      ErrLog (<< "Tried to remove non-existent column " << Fieldname);
      throw VDbException("Tried to remove non-existent column", __FILE__, __LINE__, 0);
   }
   for (ResultTabletype::iterator i = all.begin();
        i != all.end(); i++)
   {
      if ((*i)[0] == Fieldname)
      {
         all.erase(i);
         break;         
      }      
   }
   try
   {      
      // step 1, create a temp copy from all but dropped column
      {     
         char Sqlcmd[MAXQUERYLENGTH];
         ostrstream buff(Sqlcmd, MAXQUERYLENGTH);

         buff << "CREATE TABLE temp" << Domain << " AS SELECT ";
         for (ResultTabletype::const_iterator i = all.begin();
              i != all.end(); i++)
         {
            buff << (*i)[0];
            if (i+1 != all.end()) 
            {
               buff << ", ";
            }
         }
         buff << " FROM " << Domain << ends;      
         PGresult* pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
         if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
         {
            string Pgerr = PQresultErrorMessage(pgresult);
            string err = "DBAlter (drop) Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
            // Free memory used by PGresult pointer
            PQclear(pgresult);
            ErrLog (<< "DBAlter error: " << err);
            throw VDbException(err, __FILE__, __LINE__, 0);
         }
         PQclear(pgresult);
      }
   
      // step 2, drop the original table
      {
         char Sqlcmd[MAXQUERYLENGTH];
         ostrstream buff(Sqlcmd, MAXQUERYLENGTH);

         buff << "DROP TABLE " << Domain << ends;      
         PGresult* pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
         if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
         {
            string Pgerr = PQresultErrorMessage(pgresult);
            string err = "DBAlter (drop) Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
            // Free memory used by PGresult pointer
            PQclear(pgresult);
            ErrLog (<< "DBAlter error: " << err);
            throw VDbException(err, __FILE__, __LINE__, 0);
         }
         PQclear(pgresult);
      }

      // step 3, create the new version of the original table
      {
         if (parent != Domain)
         {
            FieldListtype fields;
            FieldListDatatype fieldTypes;      
         
            for (ResultTabletype::const_iterator i = res.begin();
                 i != res.end(); i++)
            {
               fields.push_back((*i)[0]);
               fieldTypes.push_back(stringToType((*i)[1]));         
            }  
            DBCreate(Domain, parent, fields, fieldTypes, true);         
         }
         else
         {         
            string primKey;
            FieldListtype fields;
            FieldListDatatype fieldTypes;      
         
            for (ResultTabletype::const_iterator i = res.begin();
                 i != res.end(); i++)
            {
               fields.push_back((*i)[0]);
               fieldTypes.push_back(stringToType((*i)[1]));         
               if (i->size() == 3)
               {
                  primKey = (*i)[0];            
               }
            }  
            assert(!primKey.empty());         
            DBCreate(Domain, fields, fieldTypes, primKey, true);
         }    
      }

      // step 4, insert data into new table
      {
         char Sqlcmd[MAXQUERYLENGTH];
         ostrstream buff(Sqlcmd, MAXQUERYLENGTH);

         buff << "INSERT INTO " << Domain << " SELECT * FROM temp" << Domain << ends;      
         PGresult* pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
         if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
         {
            string Pgerr = PQresultErrorMessage(pgresult);
            string err = "DBAlter (drop) Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
            // Free memory used by PGresult pointer
            PQclear(pgresult);
            ErrLog (<< "DBAlter error: " << err);
            throw VDbException(err, __FILE__, __LINE__, 0);
         }
         PQclear(pgresult);
      }

      // step 5, drop the temp table
      {
         PGresult* pgresult = PQexec(myPgSQLConnectionP, ("DROP TABLE temp" + Domain).c_str());
         if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
         {
            string Pgerr = PQresultErrorMessage(pgresult);
            string err = "DBAlter (drop) Error: in (DROP TABLE temp" + Domain + ")";
            // Free memory used by PGresult pointer
            PQclear(pgresult);
            ErrLog (<< "DBAlter error: " << err);
            throw VDbException(err, __FILE__, __LINE__, 0);
         }
         PQclear(pgresult);      
      }
   }
   catch (...)
   {
      PGresult* pgresult = PQexec(myPgSQLConnectionP, ("DROP TABLE temp" + Domain).c_str());
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         ErrLog(<< "Failed to drop table temp" << Domain << ", please do it for me.");
      }
      PQclear(pgresult);      
      throw;      
   }   
}

// DBAlter- Add a column to a table
void
PGSQLConnection::DBAlter(const string& Domain, const string& Fieldname,
                         const Datatype Newtype)
{
   char Sqlcmd[MAXQUERYLENGTH];
   string Sqltype;
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   PGresult*  pgresult=0;

   //  Check if table even exists first
   if(!DBVerify(Domain))
   {
      string err = "DBPrimaryKey Error: Table does not exist: " + Domain;
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBVerify error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   GetSQLtypestring(Sqltype, Newtype);

   buff << "ALTER TABLE " << Domain << " ";
   buff << "ADD COLUMN " << Fieldname << " ";
   buff << Sqltype << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBAlter Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBVerify error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }
   // Free memory used by PGresult pointer
   PQclear(pgresult);
}

// DBDrop- Delete a domain.  If ExistenceCheck is true, use default
// SQL DROP TABLE which errors if table does not exist. If ExistenceCheck
// is false, use LOCK TABLE to check if table exists.
void
PGSQLConnection::DBDrop(const string& Domain, bool ExistenceCheck)
{      
   //if ExistenceCheck is true run the command even if the table doesn't exist
   if (ExistenceCheck || DBVerify(Domain))
   {      
      char Sqlcmd[MAXQUERYLENGTH];
      ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);
      PGresult* pgresult;
      
      Sqlbuff << "DROP TABLE " << Domain << ends;

      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBDrop Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBVerify error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      }
      // Free memory used by PGresult pointer
      PQclear(pgresult);
   }

   if (DBVerifySequence(Domain))
   {
      char Sqlcmd[MAXQUERYLENGTH];
      ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);
      PGresult* pgresult;
      
      Sqlbuff << "DROP SEQUENCE " << Domain + SEQ << ends;
      pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);
      
      if (!pgresult || PQresultStatus(pgresult) != PGRES_COMMAND_OK)
      {
         string Pgerr = PQresultErrorMessage(pgresult);
         string err = "DBDrop Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
         // Free memory used by PGresult pointer
         PQclear(pgresult);
         ErrLog (<< "DBVerify error: " << err);
         throw VDbException(err, __FILE__, __LINE__, 0);
      }
      // Free memory used by PGresult pointer
      PQclear(pgresult);
   }   
}

// DBVerify- Verify the existence of a domain.
bool
PGSQLConnection::DBVerify(const string& Domain)
{
   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   PGresult* pgresult;

   // generate query specific to the domain
   buff  << "SELECT COUNT(*) "
         << "FROM pg_class c, pg_user u "
         << "WHERE c.relowner = u.usesysid AND c.relkind = 'r' "
         << "  AND c.relname ilike '" << Domain << "'" << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDescribe Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBVerify error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   // check result count
   bool found = (strcmp(PQgetvalue(pgresult, 0, 0), "1") == 0);
   
   // Free memory used by PGresult pointer
   PQclear(pgresult);
   return found;
}

bool
PGSQLConnection::DBVerifySequence(const string& Domain)
{
   char Sqlcmd[MAXQUERYLENGTH];
   ostrstream buff(Sqlcmd, MAXQUERYLENGTH);
   PGresult* pgresult;

   // generate query specific to the domain
   buff  << "SELECT COUNT(*) "
         << "FROM pg_class c, pg_user u "
         << "WHERE c.relowner = u.usesysid AND c.relkind = 'S' "
         << "  AND c.relname ilike '" << Domain << "_seq'" << ends;

   pgresult = PQexec(myPgSQLConnectionP, Sqlcmd);

   if (!pgresult || PQresultStatus(pgresult) != PGRES_TUPLES_OK)
   {
      string Pgerr = PQresultErrorMessage(pgresult);
      string err = "DBDescribe Error: PostgreSQL error: " + Pgerr + "(in " + Sqlcmd + ")";
      // Free memory used by PGresult pointer
      PQclear(pgresult);
      ErrLog (<< "DBVerify error: " << err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   // check result count
   bool found = (strcmp(PQgetvalue(pgresult, 0, 0), "1") == 0);
   
   // Free memory used by PGresult pointer
   PQclear(pgresult);
   return found;
}

// Convert from public enumerated data type to private data type
// that this dbmodule supports
void
PGSQLConnection::GetSQLtypestring(string& Sqltype, Datatype type)
{
   /// PgSQL Equivalent Data types
   const char SQLtypestring[][16] =
      {
         "varchar(255)",   // STRING
         "integer",        // INTEGER
         "float",          // FLOAT
         "boolean",        // BOOLEAN
         "text",           // TEXTBLOB
         "blob",           // BINBLOB
         "timestamp",      // TIMESTAMP
         "date",           // DATE
         "time",           // TIME
      };

   Sqltype = SQLtypestring[type];
}

// Generate an SQL statement based on the DBSelect arguments
void 
PGSQLConnection::CreateSQLselect(char* Sqlcmd, const string& Domain,
                                 const string& SearchFilter, const FieldListtype& FieldList)
{
   // Use a strstream to create SQL command
   ostrstream strbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldListtype::const_iterator p;

   // All SQL SELECT statements begin with SELECT
   strbuff << "SELECT ";

   // Add FieldList elements to SQL statement
   for (p = FieldList.begin(); p != FieldList.end(); p++)
   {
      strbuff << *p ;
      if ((p + 1) != FieldList.end())
      {
         // Insert a comma between field elements
         strbuff << ",";
      }
      else
      {
         // Insert a space, instead of a comma after the last element
         strbuff << " ";
      }
   }

   // Add Domain to SQL statement
   strbuff << "FROM " << Domain  << " ";

   if (!SearchFilter.empty())
   {
      strbuff << "WHERE " << SearchFilter;
   }
   strbuff << ends;
}

// Generate an SQL statement based on the DBSelect arguments
void
PGSQLConnection::CreateSQLselect(char* Sqlcmd, 
                                 const string& Domain,
                                 const string& SearchFilter,
                                 const string& start,
                                 const string& end,
                                 const FieldListtype& FieldList)
{
   // Use a strstream to create SQL command
   ostrstream strbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldListtype::const_iterator p;

   // All SQL SELECT statements begin with SELECT
   strbuff << "SELECT ";

   // Add FieldList elements to SQL statement
   for (p = FieldList.begin(); p != FieldList.end(); p++)
   {
      strbuff << *p ;
      if ((p + 1) != FieldList.end())
      {
         // Insert a comma between field elements
         strbuff << ",";
      }
      else
      {
         // Insert a space, instead of a comma after the last element
         strbuff << " ";
      }
   }

   // Add Domain to SQL statement
   strbuff << "FROM " << Domain  << " ";

   if (!SearchFilter.empty())
   {
      strbuff << "WHERE " << SearchFilter << " ";
   }

   int diff = atoi(end.c_str()) - atoi(start.c_str());
   char buf [16];
   sprintf(buf, "%d", diff + 1);

   strbuff << "LIMIT " << start << "," << buf << ends;
}

// Generate an SQL statement based on the DBSelect arguments
void
PGSQLConnection::CreateSQLselect(char* Sqlcmd,
                                 const string& Domain,
                                 const string& SearchFilter)
{
   // Use a strstream to create SQL command
   ostrstream strbuff(Sqlcmd, MAXQUERYLENGTH);

   // All SQL SELECT statements begin with SELECT
   strbuff << "SELECT COUNT(*) FROM " << Domain << " ";

   if (!SearchFilter.empty())
   {
      strbuff << "WHERE " << SearchFilter;
   }
   strbuff << ends;
}


// Generate an SQL statement based on the DBSelect arguments
void
PGSQLConnection::CreateSQLselect( char* Sqlcmd, 
                                  const string& Domain,
                                  const string& SearchFilter,
                                  const string& start,
                                  const string& end,
                                  const Ordertype& Order, 
                                  const string& OrderField, 
                                  const FieldListtype& FieldList )
{
   // Use a strstream to create SQL command
   ostrstream sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldListtype::const_iterator p;

   // SQL Order string
   string SQLorderstring;

   // All SQL SELECT statements begin with SELECT
   sqlbuff << "SELECT ";

   // Add FieldList elements to SQL statement
   for (p = FieldList.begin(); p != FieldList.end(); p++)
   {
      sqlbuff << *p ;
      if ((p + 1) != FieldList.end())
      {
         // Insert a comma between field elements
         sqlbuff << ",";
      }
      else
      {
         // Insert a space, instead of a comma after the last element
         sqlbuff << " ";
      }
   }

   // Add Domain to SQL statement
   sqlbuff << "FROM " << Domain  << " ";

   if (!SearchFilter.empty())
   {
      sqlbuff << "WHERE " << SearchFilter << " ";
   }

   // Set SQL Order string, based on Order being ASCENDING OR DESCENDING
   if (Order == ASCENDING)
   {
      SQLorderstring = "ASC";
   }
   else
   {
      if (Order == DESCENDING)
      {
         SQLorderstring = "DESC";
      }
      else
      {
         ErrLog (<< "Bad order: " << Order);
         throw VDbException("Order must be ASCENDING or DESCEDING",
                            __FILE__, __LINE__, 0);
      }
   }
   
   // Add Order clause
   sqlbuff << "ORDER BY " << OrderField << " " << SQLorderstring << " ";

   // Add Limit

   int diff = atoi(end.c_str()) - atoi(start.c_str());
   char buf [16];
   sprintf(buf, "%d", diff + 1);

   sqlbuff << "LIMIT " << start << "," << buf << ends;
}



// Generate an SQL statement based on the DBSelect arguments
void
PGSQLConnection::CreateSQLselect(char* Sqlcmd, const string& Domain,
                                 const string& SearchFilter, const FieldListtype& FieldList,
                                 const Ordertype& Order, const string& OrderField )
{
   // Use a strstream to create SQL command
   ostrstream sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldListtype::const_iterator p;

   // SQL Order string
   string SQLorderstring;

   // All SQL SELECT statements begin with SELECT
   sqlbuff << "SELECT ";

   // Add FieldList elements to SQL statement
   for (p = FieldList.begin(); p != FieldList.end(); p++)
   {
      sqlbuff << *p ;
      if ((p + 1) != FieldList.end())
      {
         // Insert a comma between field elements
         sqlbuff << ",";
      }
      else
      {
         // Insert a space, instead of a comma after the last element
         sqlbuff << " ";
      }
   }

   // Add Domain to SQL statement
   sqlbuff << "FROM " << Domain  << " ";

   if (!SearchFilter.empty())
   {
      sqlbuff << "WHERE " << SearchFilter << " ";
   }

   // Set SQL Order string, based on Order being ASCENDING OR DESCENDING
   if (Order == ASCENDING)
   {
      SQLorderstring = "ASC";
   }
   else
   {
      if (Order == DESCENDING)
      {
         SQLorderstring = "DESC";
      }
      else
      {
         throw VDbException("Order must be ASCENDING or DESCEDING",
                            __FILE__, __LINE__, 0);
      }
   }
   
   // Add Order clause
   sqlbuff << "ORDER BY " << OrderField << " " << SQLorderstring << ends;
}

// Generate an SQL statement based on the DBUpdate arguments
void
PGSQLConnection::CreateSQLupdate(char* Sqlcmd, const string& Domain,
                                 const string& SearchFilter, const FieldValuetype& FieldValueSet)
{
   // Use a strstream to create SQL command
   ostrstream sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldValuetype::const_iterator p;

   // All supported SQL UPDATE statements begin with UPDATE Domain SET
   sqlbuff << "UPDATE " << Domain << " SET " ;

   // Add FieldList elements to SQL statement
   for (p = FieldValueSet.begin(); p != FieldValueSet.end(); p++)
   {

      sqlbuff << p->first << " = ";
      // sqlbuff << "\"" << p->second << "\"" ;
      // sqlbuff << "\'" << p->second << "\'" ;
      sqlbuff << p->second;

      // p + 1 doesn't work on a STL map iterator? so increment then decrement
      if ((++p) != FieldValueSet.end())
      {
         p--;
         // Insert a comma between fieldvalue pairs
         sqlbuff << ", ";
      }
      else
      {
         p--;
         // Insert a space, instead of a comma after the last element
         sqlbuff << " ";
      }
   }
    
   if (!SearchFilter.empty())
   {
      sqlbuff << "WHERE " << SearchFilter;
   }
   sqlbuff << ends;    
}

// Generate an SQL statement based on the DBInsert arguments
void
PGSQLConnection::CreateSQLinsert(char* Sqlcmd, const string& Domain,
                                 const FieldValuetype& FieldValueSet)
{
   // Use a strstream to create SQL command
   ostrstream sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldValuetype::const_iterator p;

   // All supported SQL INSERT statements begin with INSERT INTO Domain
   sqlbuff << "INSERT INTO " << Domain ;

   // Add Field names
   sqlbuff << " ( " ;
   for (p = FieldValueSet.begin(); p != FieldValueSet.end(); p++)
   {
      sqlbuff << p->first;

      // p + 1 doesn't work on a map iterator? so increment then decrement
      if ((++p) != FieldValueSet.end())
      {
         p--;
         // Insert a comma between fieldvalue pairs
         sqlbuff << ", ";
      }
      else
      {
         p--;
         // Insert a space, instead of a comma after the last element
         sqlbuff << " ";
      }
   }
   sqlbuff << " ) " ;

   // Add Value names
   sqlbuff << "VALUES ( " ;
   for (p = FieldValueSet.begin(); p != FieldValueSet.end(); p++)
   {
      // sqlbuff << "\"" << p->second << "\"" ;
      // sqlbuff << "\'" << p->second << "\'" ;
      sqlbuff << p->second;

      // p + 1 doesn't work on a map iterator? so increment then decrement
      if ((++p) != FieldValueSet.end())
      {
         p--;
         // Insert a comma between fieldvalue pairs
         sqlbuff << ", ";
      }
      else
      {
         p--;
         // Insert a space, instead of a comma after the last element
         sqlbuff << " ";
      }
   }
   sqlbuff << " ) " << ends;
}

// Generate an SQL statement based on the DBDelete arguments
void
PGSQLConnection::CreateSQLdelete(char* Sqlcmd, const string& Domain,
                                 const string& SearchFilter)
{
   // Use a strstream to create SQL command
   ostrstream sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // All supported SQL DELETE statements begin with DELETE FROM Domain
   sqlbuff << "DELETE FROM " << Domain << " ";

   if (!SearchFilter.empty())
   { 
      sqlbuff << "WHERE " << SearchFilter;
   } 
   sqlbuff << ends;
}

// Generate an SQL statement based on the DBCreate arguments
// that produces an appropriate sequence for NEXTVAL
void
PGSQLConnection::CreateSQLsequence(char* Sqlcmd, 
                                   const string& Domain, 
                                   const string& PrimaryKey) 
{
   // Use a strstream to create SQL command
   ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   Sqlbuff << "CREATE SEQUENCE " << Domain << SEQ << ends;
}

// Generate an SQL statement based on the DBCreate arguments
// This version of DBCreate does not take FieldDataTypeList which
// allows you to specify a different data type for each field.
// Therefore the data type for each field will be VARCHAR(255) 
// which is STRINGTYPE
void
PGSQLConnection::CreateSQLcreate(char* Sqlcmd, const string& Domain,
                                 const FieldListtype& FieldList, const string& PrimaryKey) 
{
   // Use a strstream to create SQL command
   ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);

   // Field List iterator to used to fill fields in SQL statement
   FieldListtype::const_iterator p;

   Sqlbuff << "CREATE TABLE " << Domain << " ( " << endl;

   // Add FieldList elements to SQL statement
   for (p = FieldList.begin(); p != FieldList.end(); p++)
   {
      // If the field is the PRIMARY KEY, it must have the
      // NOT NULL option set.
      if ( *p == PrimaryKey )
         Sqlbuff << *p << " VARCHAR(255) DEFAULT NEXTVAL('" << Domain << SEQ << "') NOT NULL, " << endl;
      else
         Sqlbuff << *p << " VARCHAR(255), " << endl;
   }

   // Insert Primary Key
   Sqlbuff << "PRIMARY KEY ( " << PrimaryKey << " ) ) " << ends;
}

// Generate an SQL statement based on the DBCreate arguments
void
PGSQLConnection::CreateSQLcreate(char* Sqlcmd, const string& Domain,
                                 const FieldListtype& FieldList, 
                                 const FieldListDatatype& DatatypeList, 
                                 const string& PrimaryKey)
{
   // Use a strstream to create SQL command
   ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);
   string Sqltype;
   string fieldnamebuff;
   Datatype Datatypebuff;

   // Field List iterator, and DataType iterator
   FieldListtype::const_iterator p1;
   FieldListDatatype::const_iterator p2;

   Sqlbuff << "CREATE TABLE " << Domain <<  " ( " << endl;

   // Add FieldList elements to SQL statement
   for (p1 = FieldList.begin(), p2 = DatatypeList.begin();
        p1 != FieldList.end(); p1++, p2++)
   {
      fieldnamebuff = *p1;
      Datatypebuff = *p2;

      GetSQLtypestring(Sqltype, Datatypebuff);

      Sqlbuff << fieldnamebuff  << "  " << Sqltype << "  ";

      // If the field is the PRIMARY KEY, it must have the
      // NOT NULL option set.
      if ( *p1 == PrimaryKey )
      {
         Sqlbuff << "DEFAULT NEXTVAL('" << Domain << SEQ << "') NOT NULL, " << endl;
      }
      else
      {
         Sqlbuff << ", " << endl;
      }
   }

   // Insert Primary Key
   Sqlbuff << "PRIMARY KEY ( " << PrimaryKey << " ) ) " << ends;
}

// Generate an SQL statement based on the DBCreate arguments,
// use inheritance
void
PGSQLConnection::CreateSQLcreate(char* Sqlcmd, 
                                 const string& domain,
                                 const string& parentDomain,
                                 const FieldListtype& fieldList, 
                                 const FieldListDatatype& datatypeList)
{
   // check for existence of parent
   if (!DBVerify(parentDomain))
   {
      string err = "Domain '" + domain + "' tried to inherit from non-existent parent: " + parentDomain;
      ErrLog (<< err);
      throw VDbException(err, __FILE__, __LINE__, 0);
   }

   // get the primarykey from the parent -- need to respecify
   const string primaryKey(DBPrimaryKey(parentDomain));
   
   // Use a strstream to create SQL command
   ostrstream Sqlbuff(Sqlcmd, MAXQUERYLENGTH);
   string Sqltype;
   string fieldnamebuff;
   Datatype Datatypebuff;

   // Field List iterator, and DataType iterator
   FieldListtype::const_iterator p1;
   FieldListDatatype::const_iterator p2;

   Sqlbuff << "CREATE TABLE " << domain <<  " ( " << endl;

   // Add FieldList elements to SQL statement
   for (p1 = fieldList.begin(), p2 = datatypeList.begin();
        p1 != fieldList.end(); p1++, p2++)
   {
      fieldnamebuff = *p1;
      Datatypebuff = *p2;

      GetSQLtypestring(Sqltype, Datatypebuff);
      Sqlbuff << fieldnamebuff  << " " << Sqltype << ", ";
   }

   // Insert inheritance
   Sqlbuff << "PRIMARY KEY ( " << primaryKey << " ) )  INHERITS (" << parentDomain << ")" << ends;
}

DBResultset* 
PGSQLConnection::DBRetrieveByOid(const string& Domain, 
                                 uint oid,
                                 const FieldListtype& FieldList)
{
   strstream filter;
   filter << "oid = " << oid << ends;
   filter.freeze(0);
   
   return DBSelect(Domain, filter.str(), FieldList);
}







