/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


#include <iostream>
#include "PGSQLResultset.hxx"
#include "VDbException.hxx"

using namespace std;
using namespace Epoisses;

// SQLResetset constructor, initialize attributes
PGSQLResultset::PGSQLResultset(PGresult* result, const string& searchfilter):
mySearchFilter(searchfilter)
{
    myPgSQLresultP = result;
    myRecordCount = PQntuples(myPgSQLresultP);
    myFieldList = GetFields();
};

// SQLResetset destructor, free attributes 
PGSQLResultset::~PGSQLResultset(void)
{
    // Free memory used by PQresult pointer
    PQclear(myPgSQLresultP);
};

// GetRecords[offset] - returns the record at offset
ResultRecordtype
PGSQLResultset::operator[](unsigned short offset) const
{
    string fieldbuff;
    unsigned short i;
    ResultRecordtype returnResult;

    // Check offsets are within range of results
    if (offset > myRecordCount) 
    {
        throw VDbException("Record Offset out of range", __FILE__, __LINE__, 0);
    }

    // If Records exist, iterate through fields filling returnResult
    if (myRecordCount)
    {
        for (i = 0; i < myFieldList.size(); i++)
        {
            fieldbuff = PQgetvalue(myPgSQLresultP, offset, i);
            returnResult.push_back(fieldbuff);
        }
    }

    return returnResult;
}

// GetRecord(offset) - returns the record at offset
ResultRecordtype
PGSQLResultset::GetRecord(unsigned short offset) const
{
    string fieldbuff;
    unsigned short i;
    ResultRecordtype returnResult;

    // Check offsets are within range of results
    if (offset > myRecordCount) 
    {
        throw VDbException("Record Offset out of range", __FILE__, __LINE__, 0);
    }

    // If Records exist, iterate through fields filling returnResult
    if (myRecordCount)
    {
        for (i = 0; i < myFieldList.size(); i++)
        {
            fieldbuff = PQgetvalue(myPgSQLresultP, offset, i);
            returnResult.push_back(fieldbuff);
        }
    }

    return returnResult;
};

// GetRecords(offset1, offset2) - returns all records between offset1
// and offset2
ResultTabletype
PGSQLResultset::GetRecords(unsigned short offset1, unsigned short offset2) const
{
    string fieldbuff;
    ResultRecordtype recordbuff;
    ResultTabletype returnResult;
    unsigned short recordnumber;
    unsigned short i;

    // Check offsets are within range of results
    if ((offset1 > offset2) || (offset1 > myRecordCount) 
            || (offset2 > myRecordCount))
    {
        throw VDbException("Record Offsets out of range", __FILE__, 
                __LINE__, 0);
    }

    for (recordnumber = offset1; recordnumber <= offset2; recordnumber++)
    {
        for (i = 0; i < myFieldList.size(); i++)
        {
            fieldbuff = PQgetvalue(myPgSQLresultP, recordnumber, i);
            recordbuff.push_back(fieldbuff);
        }

        // Insert record buffer into result table
        returnResult.push_back(recordbuff);

        // Clear recordbuff for next record
        recordbuff.clear();
    }
    return returnResult;
};
 
// GetRecords - returns all records from a search
ResultTabletype
PGSQLResultset::GetRecords() const
{
    ResultRecordtype recordbuff;
    ResultTabletype returnResult;
    string fieldbuff;
    unsigned short recordnumber;
    unsigned short i;

    for (recordnumber = 0; recordnumber < myRecordCount; recordnumber++)
    {
        for (i = 0; i < myFieldList.size(); i++)
        {
            fieldbuff = PQgetvalue(myPgSQLresultP, recordnumber, i);
            recordbuff.push_back(fieldbuff);
        }

        // Insert record buffer into result table
        returnResult.push_back(recordbuff);

        // Clear recordbuff for next record
        recordbuff.clear();
    }
    return returnResult;
};

// GetRecordCount- Return Record count
unsigned int
PGSQLResultset::GetRecordCount(void) const
{
    return myRecordCount;
};

// GetFields - Return fields on a resultset
FieldListtype
PGSQLResultset::GetFields(void) const
{
    FieldListtype returnResult;
    unsigned short i;
    unsigned short nfields = PQnfields(myPgSQLresultP);

    // If Records exist, fill myFieldList
    if (myRecordCount) 
    {
        for (i = 0; i < nfields; i++)
        {
            returnResult.push_back(PQfname(myPgSQLresultP, i));
        }
    }

    return returnResult;
}

// GetmySearchFilter- Return searchfilter that was used to create this
//    resultset.  This is normally used when a resultset is passed back 
//    to DBUpdate() or DBSelect()
const string&
PGSQLResultset::GetSearchFilter(void) const
{
    return mySearchFilter;
};
