// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

// Copyright 2002 Cathay Networks, Inc. 

#ifndef Logger_hxx
#define Logger_hxx

#include <loki/Singleton.h>
#include <loki/static_check.h>
#include <loki/Threads.h>

#include "Log.hxx"
#include "SysLogStream.hxx"
#include "Lock.hxx"
#include "Inserter.hxx"

/**
   Defines a set of logging macros, one for each level of logging.

   Example:
#include Logger.hxx
#define VOCAL_SUBSYSTEM Vocal::Subsystem::SIP
   ...
   DebugLog(<< "hi there " << mix << 5 << types);  // note leading << and no endl
*/

//#define VOCAL_SUBSYSTEM Vocal::Subsystem::NONE

// variadic to handle comma in template arguments
#define DebugLog(arg__, args__...)                                                 \
                                         /* eat the comma if no extra arguments */ \
GenericLog(VOCAL_SUBSYSTEM, Vocal::Log::DEBUG, arg__, ##args__)

#define CritLog(arg__, args__...) \
GenericLog(VOCAL_SUBSYSTEM, Vocal::Log::CRIT, arg__, ##args__)

#define ErrLog(arg__, args__...) \
GenericLog(VOCAL_SUBSYSTEM, Vocal::Log::ERR, arg__, ##args__)

#define WarningLog(arg__, args__...) \
GenericLog(VOCAL_SUBSYSTEM, Vocal::Log::WARNING, arg__, ##args__)

#define InfoLog(arg__, args__...) \
GenericLog(VOCAL_SUBSYSTEM, Vocal::Log::INFO, arg__, ##args__)

#ifdef NO_DEBUG
// Suppress debug loging at compile time
#define DebugLog(arg__, args__...)
#endif

// do/while allows a {} block in an expression
#define GenericLog(system__, level__, arg__, args__...)          \
do                                                               \
{                                                                \
  if (Vocal::GenericLogImpl::isLogging(level__))                 \
  {                                                              \
     Vocal::Threads::Lock lock(Vocal::Log::_mutex);              \
     if (Vocal::GenericLogImpl::isLogging(level__))              \
     {                                                           \
        Vocal::Log::tags(level__, system__,                      \
                          Vocal::GenericLogImpl::Instance())     \
          << __FILE__ << ':' << __LINE__ << DELIM                \
                  /* eat the comma if no extra arguments */      \
          arg__ , ##args__ << endl;                              \
     }                                                           \
  }                                                              \
} while (0)

namespace Vocal
{

class GenericLogImpl : public Loki::SingletonHolder <SysLogStream,
                                                     Loki::CreateUsingNew,
                                                     Loki::PhoenixSingleton,
                                                     Loki::ClassLevelLockable>,
                       public Log 
{
   public:
      static ostream& Instance()
      {
         if (Log::_type == Log::SYSLOG)
         {
            return Loki::SingletonHolder<SysLogStream, Loki::CreateUsingNew, 
                                         Loki::PhoenixSingleton, Loki::ClassLevelLockable>::Instance();
         }
         else if (Log::_type == Log::FILE)
         {
            assert(0);
         }
         else
         {
            return cout;
         }
      }
      
      static bool isLogging(Log::Level level) 
      {
         return (level <= Log::_level);
      }
};
 
} // namespace Vocal

#endif
