// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

static const char* const Time_cxx_Version =
    "$Id: Time.cxx,v 1.1 2002/09/30 04:55:17 jason Exp $";

#include "global.h"
#include "Time.hxx"
#include <cassert>

using namespace Epoisses;

Time::Time()
{
   clear();
}


Time::Time(milliseconds_t src)
{
   clear();
   operator+=(src);
}

Time::Time(long sec, long usec)
{
   clear();
   tv_sec = sec;
   tv_usec = usec;
}
        


Time::Time(const timeval & src)
{
   copy(src);
   normalize();
}


Time::~Time()
{
}


Time &
Time::operator=(const timeval & src)
{
   if ( this != & src )
   {
      copy(src);
      normalize();
   }
   return ( *this );
}


const Time &
Time::now()
{
   int rc = gettimeofday(this, 0);
   assert( rc == 0 );
   
   return ( *this );
}


milliseconds_t
Time::milliseconds() const
{
   milliseconds_t t = tv_sec;
   t *= 1000;
   t += (tv_usec / 1000);
   return t;
}

Time
Time::operator+(const timeval & right) const
{
   Time left(*this);
   return ( left += right );
}

Time
Time::operator+(milliseconds_t right) const
{
   Time left(*this);
   return ( left += right );
}

Time &
Time::operator+=(const timeval & src)
{
   tv_sec += src.tv_sec;
   tv_usec += src.tv_usec;

   normalize();

   return ( *this );
}

Time &
Time::operator+=(milliseconds_t src)
{
   tv_sec += src / 1000;
   tv_usec += (src % 1000) // Between [0..1000) milliseconds.
      * 1000;      // Convert to microseconds.

   normalize();

   return ( *this );
}

Time
Time::operator-(const timeval & right) const
{
   Time left(*this);
   return ( left -= right );
}

Time
Time::operator-(milliseconds_t right) const
{
   Time left(*this);
   return ( left -= right );
}


Time &
Time::operator-=(const timeval & src)
{
   tv_sec -= src.tv_sec;
   tv_usec -= src.tv_usec;

   normalize();

   return ( *this );
}

Time &
Time::operator-=(milliseconds_t src)
{
   tv_sec -= src / 1000;
   tv_usec += (src % 1000) // Between [0..1000) milliseconds.
      * 1000;      // Convert to microseconds.

   normalize();

   return ( *this );
}

bool
Time::operator==(const timeval & src) const
{
   // Create a normalize value for the timeval to simplify comparison.
   //
   Time right(src);

   return ( tv_sec == right.tv_sec
            && tv_usec == right.tv_usec
      );
}

bool
Time::operator<(const timeval & src) const
{
   // Create a normalize value for the timeval to simplify comparison.
   //
   Time right(src);

   if ( tv_sec < right.tv_sec )
   {
      return ( true );
   }
   else if ( (tv_sec == right.tv_sec) && (tv_usec < right.tv_usec) )
   {
      return ( true );
   }
   else
   {
      return ( false );
   }
}

bool
Time::operator<(const Time & right) const
{
   if ( tv_sec < right.tv_sec )
   {
      return ( true );
   }
   else if ( (tv_sec == right.tv_sec) && (tv_usec < right.tv_usec) )
   {
      return ( true );
   }
   else
   {
      return ( false );
   }
}

void
Time::clear()
{
   tv_sec = tv_usec = 0;
}


void
Time::copy(const timeval & src)
{
   tv_sec = src.tv_sec;
   tv_usec = src.tv_usec;
}


void
Time::normalize()
{
   // The purpose of normalize is to ensure that the tv_usec field
   // is in [0,100000).
   //

   if ( tv_usec < 0 )
   {
      long num_sec = ( -tv_usec / 1000000) + 1;

      assert(num_sec > 0);

      tv_sec -= num_sec;
      tv_usec += num_sec * 1000000;

      // The result of this operation may have tv_usec == 1000000.
      //
   }


   if ( tv_usec >= 1000000 )
   {
      tv_sec += tv_usec / 1000000;
      tv_usec %= 1000000;
   }
}

unsigned int
Time::TimeOfWeek()
{
   time_t t = time(0);
   struct tm* mtm = localtime(&t);
   
   unsigned int now = 0;
   now += mtm->tm_wday*(24*60*60);
   now += mtm->tm_hour*(60*60);
   now += mtm->tm_min*(60);
   now += mtm->tm_sec;

   return now;
}

ostream &
Cathay::operator<<(ostream & out, const Time& val)
{
   return ( out << "{ " << val.tv_sec << ", " << val.tv_usec << " }" );
}

