package org.vovida.prov.psLib;

import java.util.Hashtable;
import java.io.IOException;
import java.io.StringReader;
import java.io.StreamTokenizer;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

/**
 * <P>This interface defines the API for a client application to
 * communicate with the Provisioning Server.</P>
 * <P>Commands are separated into two groups: Transaction and Session. <UL>
 * <LI>Transaction - used to do a single database lookup or manipulation</LI>
 * <LI>Session - used for more complicated database manipulation that may
 * involve multiple query transactions such as search and sort. A session
 * is identified by a server generated token.</LI>
 * </UL> </P> <P>Other terms used in ProvisioningInterface are: <UL>
 * <LI>Domain -   The type or category of information. Examples would be users,
 * a particular type of server, etc. Analogous to the idea of a table in
 * SQL.</LI>
 * <LI>Field -    A particular piece of data, of which one exists for each
 * record. An example would be a phone number -- there would be one
 * associated with each user. Analogous to a column in
 * SQL.</LI> <LI>Value -     The contents of a particular field
 * -- a particular phone number in the case described above.</LI>
 * <LI>Filter -    A filter using the grammar defined in ???.</LI>
 * <LI>FieldList - A list of fields, usually specifying which fields one
 * wants to work with</LI> </UL> </P>
 */

public abstract class ProvisioningInterface
{
  // Remove this when testing is complete
  public static boolean templateTesting = false;

  /**
   * Sort a query by ascending values.
   */
  public static final String SORT_TYPE_ASC = "ascend";

  /**
   * Sort a query by descending values.
   */
  public static final String SORT_TYPE_DESC = "descend";

  /**
   * The reference to the instance of this class
   */
  private static ProvisioningInterface instance = null;

  /**
   * Returns an instance of this class.  All calls to this method receive the same object
   */
  public static ProvisioningInterface getInstance()
  {
    if (instance == null)
    {
      instance = new SocketProvisioningInterface();
    }

    return instance;
  }


  /**
   * Close the single connection to the data source or server.
   * @throws IOException if the close operation fails
   */
  public abstract void close() throws IOException;


  /**
   * Send request to the PServer to authenticate the given account id and
   * password.
   * @param accountId the user name
   * @param password the password for that account
   * @returns true if the authentication succeeded
   */
  public abstract boolean login(String accountId, String password) throws
    IOException, OperationException;

  /**
   * Send request to the PServer to close an authenticated session. <p>
   */
  public abstract boolean logout() throws IOException, OperationException;

  /**
   * Send a request to PServer to retrieve all the matching records
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon.
   * @param fieldList list of fields to return for each record
   */
  public abstract PSDBResult getRecordsInDomain(Domain domain,
                                                Filter filter, 
                                                FieldList fieldList) 
     throws IOException, OperationException;

  /**
   * Get the field names within a domain (corresponding to the column names
   * of, say, a SQL record).
   */
  public abstract String[] getFieldsInDomain(Domain domain) 
     throws IOException, OperationException;

  /**
   * Get an XML data description. This describes both a table in the database
   * and a GUI for entering data into that table.
   * @return an XML string describing the requested data.
   */
  public abstract String getXMLFile(String category, 
                                    String filename) 
     throws IOException, OperationException;

  /**
   * Get a text file "as is" from the www directory.
   */
  public abstract String getRawFile(String filename) 
    throws IOException, OperationException;

  /**
   * Get the list of all XML data descriptions of a particular category.
   */
  public abstract String[] getListXMLDescriptions(String category) 
     throws IOException, OperationException;

  /**
   * Given a list of fields within a domain, for each matching field,
   * replace its value with that in the fieldValuePairs.
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon. 
   * @param assignmentList a list of assigments of a value to a field
   */
      public abstract boolean setRecordsInDomain(Domain domain, 
                                                 Filter filter, 
                                                 AssignmentList assignmentList)
         throws IOException, OperationException;

  /**
   * Add a new record to a domain.
   * @param domain the domain to act upon
   * @param assignmentList a list of key (eg uid) values (eg 1000)
   */
  public abstract boolean addRecordInDomain(Domain domain, 
                                            AssignmentList assignmentList)
     throws IOException, OperationException;

  /**
   * Remove all records within a domain that match the filter
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon.
   */
   public abstract boolean deleteRecordsInDomain(Domain domain, 
                                                 Filter filter) 
      throws IOException, OperationException;

  /**
   * Instruct PServer that the client is interested in monitoring and
   * data change wihtin a domain.  ProvisionigInterface listens for update
   * notification and updates the client by invoking the call back function.
   * <B>NOT IMPLEMENTED</B>
   */
  public abstract boolean registerDomainForUpdate(String domain, 
          Object huh) throws IOException, OperationException;

  /**
   * Similar to registerDomainForUpdate, except the client is interested in
   * monitoring a specific filed within a domain. <B>NOT IMPLEMENTED</B>
   */
  public abstract boolean registerRecordForUpdate(String domain, 
          Object huh) throws IOException, OperationException;

  /**
   * Similar to getRecordsInDomain(), except it caches a snapshot of the data
   * locally for the client and automatically monitors and updates the data.
   * The first invocation of this fucntion sends a request to PServer to look
   * up the value(s) associated with the field.  Subsequent calls to this
   * function will result in accessing the resultatnt data locally instead
   * of doing a "true" get via PServer.
   */
  public abstract PSDBResult getCurrentRecordsInDomain(String domain, 
          String filter, 
          String[] fieldList) throws IOException, OperationException;

  /**
   * Open a query session with Provisioning Server which returns a unique
   * indentifier that to which subsequent request can refer.
   * @param timeout how long in milliseconds the server should leave this
   * session open.  After the time has elapsed the server automatically
   * invalidates the session.
   */
  public abstract String sessionOpen(long timeout) 
    throws IOException, OperationException;

  /**
   * Terminate a query session.
   * @param token the query session indentifier.
   */
  public abstract boolean sessionClose(String token) 
    throws IOException, OperationException;

  /**
   * Query PServer to lookup all records satisfying the supplied constraints
   * within a domain.  PServer returns the number of matching records.
   * @param token the query session indentifier
   * @param filter identifies which records within the supplied domain to act
   * upon.   It is a string of boolean logic (uid=100 & name!='John')
   * @param fieldList used to identify which fields to return for each record
   * @param sortType one of SORT_TYPE_ASC | SORT_TYPE_DESC
   * @param sortField the field to on which to sort.
   */
  public abstract int sessionSearchDB(String token, 
                                      Domain domain, 
                                      Filter filter,
                                      FieldList fieldList, 
                                      String sortType, 
                                      Field sortField) 
     throws IOException, OperationException;

  /**
   * Query PServer to retrieve a range of records based on the result of
   * the previous sessionSearchDB()
   * @param token the query session indentifier
   * @param start the start number of the range to retrieve.  Inclusive.
   * @param end the end number of the range to retrieve.  Inclusive.
   */
  public abstract PSDBResult sessionGetDBResult(String token, 
                                                int start, 
                                                int end,
                                                FieldList fieldList) 
     throws IOException, OperationException;

  /**
   * Query PServer to sort the result from a previous call to sessionSearchDB()
   * @param token the query session indentifier
   * @param sortType one of SORT_TYPE_ASC | SORT_TYPE_DESC
   * @param sortField the field to on which to sort.
   */
  public abstract boolean sessionSortDBResult(String token, 
                                              String sortType, 
                                              Field sortField) 
     throws IOException, OperationException;

  /**
   * Instruct PServer to update all the records from a previous call to
   * sessionSearchDB() with the supplied fieldValuePairs
   * @param token the query session indentifier
   * @param fieldValuePairs a list of key (eg uid) values (eg 1000)
   */
  public abstract boolean sessionSetDBResult(String token, 
                                             AssignmentList assignmentList) 
     throws IOException, OperationException;

  /**
   * Remove all the records found by a previous call to sessionSearchDB()
   * @param token the query session indentifier
   */
  public abstract boolean sessionDeleteAll(String token) 
    throws IOException, OperationException;
}



