package org.vovida.prov.psLib;

import java.util.Hashtable;
import java.io.IOException;

import org.apache.log4j.Category;

/**
 * An implementation of the ProvisioningInterface interface.  Right now
 * SocketProvisioningInterface uses HttpXMLProtocolService to generate and
 * parse the HTTP/XML messages and HttpPSConnection to send and receive
 * the messages.
 *
 * @author Matt Naish
 */
public class SocketProvisioningInterface extends ProvisioningInterface
{

  /**
   * The hostname of the socket connection.
   */
  private static String myHostname = "localhost";

  /**
   * The port of the socket connection.
   */
  private static int myPort = 80;

  /**
   * Does the actual sending and receiving of the messages
   * @supplierCardinality 1
   */
  private PSInterface psInterface;

  /**
   * Does message generation and parsing
   * @supplierCardinality 1
   */
  private PSProtocolService psProtocolService;

  /**
   * @supplierCardinality 1
   */
  private PSListener psListener;

  /**
   * @supplierCardinality 0..1
   */
  private PSDataObservable PSDataObservable;

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(SocketProvisioningInterface.class.getName());

  /**
   * Private default constructor
   */
  protected SocketProvisioningInterface()
  {
    psInterface = new PSInterface(myHostname, myPort);
    psProtocolService = HttpXmlProtocolService.getInstance();
  }

  public void close() throws IOException
  {
    myHostname = null;
    myPort = -1;

    psInterface.close();
    psInterface = null;

    psProtocolService = null;
  }

  /**
   * Sets up hostname and port. This must be called before the constructor.
   */
  public static void initialize(String hostname, int port)
  {
    logCategory.info("Connection initialized with host " + hostname 
                     + " and port " + port);

    myHostname = hostname;
    myPort = port;
  }

  public String getRawFile(String filename) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_RAW_FILE, 
            filename);
    PSMessage response = psInterface.sendRequest(request);
    String result = response.getBody();

    return result;
  }

  public String[] getListXMLDescriptions(String category) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest
        (PSProtocolService.GET_LIST_XML_DESCRIPTIONS, category);
    PSMessage response = psInterface.sendRequest(request);
    String[] result = psProtocolService.parseGetDescriptionList(response);

    return result;
  }

  public String getXMLFile(String category, 
          String filename) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_XML_DESCRIPTION, 
            category, filename);
    PSMessage response = psInterface.sendRequest(request);
    String result = psProtocolService.parseXMLResultResponse(response);

    return result;
  }

  public PSDBResult getRecordsInDomain(Domain domain,
                                       Filter filter, 
                                       FieldList fieldList) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_RECORDS_IN_DOMAIN, 
            domain, filter, fieldList);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);

    return result;
  }

  public String[] getFieldsInDomain(Domain domain) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_FIELDS_IN_DOMAIN, 
                                     domain);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);
    String[] fieldNames = result.getFieldNames();

    if (logCategory.isDebugEnabled())
    {
      logCategory.debug("getFieldsInDomain returning field names: ");

      for (int i = 0; i < fieldNames.length; i++)
      {
        logCategory.debug("  " + fieldNames[i]);
      }
    }

    return fieldNames;
  }

  public boolean setRecordsInDomain(Domain domain,
                                    Filter filter, 
                                    AssignmentList assignments)
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SET_RECORDS_IN_DOMAIN, 
                                     domain, filter, assignments);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean addRecordInDomain(Domain domain, 
                                   AssignmentList assignments) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.ADD_RECORDS_IN_DOMAIN, 
                                     domain, assignments);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean login(String accountId, String password) throws
    IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean logout() throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean registerDomainForUpdate(String domain, 
          Object huh) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean registerRecordForUpdate(String domain, 
          Object huh) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public PSDBResult getCurrentRecordsInDomain(String domain, String filter, 
          String[] fieldList) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public String sessionOpen(long timeout) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_OPEN, 
            new Long(timeout));
    PSMessage response = psInterface.sendRequest(request);

    return psProtocolService.parseStringResponse(response);

  }

  public boolean deleteRecordsInDomain(Domain domain, 
                                       Filter filter) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.DELETE_RECORDS_IN_DOMAIN, 
                                     domain, filter);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean sessionClose(String token) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_CLOSE, token);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public int sessionSearchDB(String token, 
                             Domain domain, 
                             Filter filter, 
                             FieldList fieldList,
                             String sortType, 
                             Field sortField) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SEARCH_DB, 
            token, domain, filter, fieldList, sortType, sortField);
    PSMessage response = psInterface.sendRequest(request);
    Integer result = psProtocolService.parseIntegerResponse(response);

    return result.intValue();
  }

  public PSDBResult sessionGetDBResult(String token, 
                                       int start, int end, 
                                       FieldList fieldList) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_GET_DB_RESULT, 
            token, new Integer(start), new Integer(end), fieldList);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);

    return result;
  }

  public boolean sessionSortDBResult(String token, 
                                     String sortType, 
                                     Field sortField) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SORT_DB_RESULT, 
            token, sortType, sortField);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean sessionSetDBResult(String token, 
                                    AssignmentList assignmentList) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SET_DB_RESULT, 
                                     token, assignmentList);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }


  public boolean sessionDeleteAll(String token) 
     throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_DELETE_ALL, 
            token);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  private static void printPSDBResult(PSDBResult result)
  {
    String[] fieldNames = result.getFieldNames();

    for (int i = 0; i < result.getNumRecords(); i++)
    {
      for (int j = 0; j < fieldNames.length; j++)
      {
        logCategory.debug(fieldNames[j] + "=" 
                          + result.getFieldValue(i, fieldNames[j]) + ", ");
      }

      logCategory.debug("");
    }
  }

}


