/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const FsProvisioning_cxx_Version =
    "$Id: FsProvisioning.cxx,v 1.35 2002/12/12 17:21:19 sprajpat Exp $";

#include <vector>

#include "CPLFeatureBuilder.hxx"
#include "CPLStartStateHolder.hxx"
#include "Data.hxx"
#include "FsConfigFile.hxx"
#include "FsProvisioning.hxx"
#include "FsUsers.hxx"
#include "VNetworkException.hxx"
#include "cpLog.h"
#include "SystemInfo.hxx"
#include "NetworkConfig.hxx"

//using namespace Vocal;

using Vocal::FsProvisioning;

//////////////SINGLETON IMPLEMENTATION//////////////
FsProvisioning* FsProvisioning::myInstance = NULL;
////////////////////////////////////////////////////

//////////////CONSTANTS/////////////////////////////
static int PortNumber = 0;
////////////////////////////////////////////////////

using namespace Vocal;

///
FsProvisioning::FsProvisioning()
{
    //Verify PortNumber and Feature Server name are
    //defined.  These are needed to use Provisioning
    //Server's interface.
    PortNumber = cmdLineData.sipPort;

    if(NetworkConfig::instance().isDualStack())
    {
       //Use IPv4 address as the key
       char hAddress[256];
       theSystem.gethostAddress(hAddress, 256, PF_INET);
       myFeatureServerName = hAddress;
    }
    else
    {
       NetworkAddress na;
       myFeatureServerName = na.getIpName().c_str();
    }


    if (myFeatureServerName.length() == 0)
    {
        assert( myFeatureServerName.length() > 0 );
    }
#if 0 //Change to 1 if need to debug
    else
        cpLog( LOG_DEBUG, "FS Name=%s", myFeatureServerName.c_str() );
    cpLog( LOG_DEBUG, "FS Port Number=%d", PortNumber );
#endif

    //Retrieve pointer to Provision Interface singleton
    try 
    {
        myPif = &( 
	    ProvisionInterface::instance(cmdLineData.pServer.c_str(), 
					 cmdLineData.pServerPort, 
					 cmdLineData.backupPServer.c_str(), 
					 cmdLineData.backupPServerPort,
					 cmdLineData.readSecretFile.c_str(),
					 cmdLineData.writeSecretFile.c_str()));
    } /*try*/
    catch (VException& excpInfo )
    {
        cpLog( LOG_ERR, excpInfo.getDescription().c_str() );
        cpLog( LOG_ERR, "FS not initialized properly." ); 
        return; 
    } /*catch*/

    try
    {

        // Register for updates on cpl scripts used by current
        //  Feature Server
        myPif ->registerForUpdate( myFeatureServerName, PortNumber,
                                 notifyAddDel,
                                 ProvisionInterface::VFEATURE );
    } /*try*/
    catch (VNetworkException& e)
    {
        cpLog( LOG_ERR, "Network exception: %s", e.getDescription().c_str() );
    } /*catch*/

}  // FsProvisioning


///
FsProvisioning*
FsProvisioning::Instance()
{
    if ( myInstance == 0 )
    {
        myInstance = new FsProvisioning;
    } /*if*/

    return myInstance;

}  // Instance


///
FsProvisioning::~FsProvisioning()
{}  // FsProvisioning


///

void
FsProvisioning::initialize()
{
    //From the provisioning server, fetch users configured to
    //have a feature on this Feature Server.
    retrieveFeatureUsers(FsProvisioning::FeaturesIncoming);
    retrieveFeatureUsers(FsProvisioning::FeaturesOutgoing);

    //Go fetch any global data needed by FS
    retrieveGlobalData();

} // initialize


///
void
FsProvisioning::retrieveFeatureUsers( int scriptType )
{
    vector < Data > users;

    //Find all users registered for this feature server
    try
    {
        if ( myPif == 0 ) return ;

        if ( scriptType == FsProvisioning::FeaturesIncoming )
            myPif->getCalledFeatureScripts( myFeatureServerName,
                                           PortNumber,
                                           users );
        else
            myPif->getCallingFeatureScripts( myFeatureServerName,
                                            PortNumber,
                                            users );

    } /*try*/
    catch (VNetworkException& e)
    {
        cpLog( LOG_ERR, "Network exception: %s ", e.getDescription().c_str() );
    } /*catch*/

    //Retrieve CPL script for each user in the list
    vector < Data > ::iterator user = users.begin();
    vector < Data > ::iterator endOfUsers = users.end();

    //Iterater through all users found and try to register
    //and create a feature for each user.
    LocalScopeAllocator lo;
    while ( user != endOfUsers )
    {
        string userFilename( user->getData(lo) );
        ++user;

#if 0 //Change to 1 if needed for debugging
        cpLog(LOG_DEBUG, "Processing (%s)", userFilename.c_str() );
#endif

        //Retrieve only CPL files.  Assume filenames that
        //are not in the <name>.cpl format are not valid CPL
        //files.
        if ( isValidCPLFilename( userFilename ) == false )
        {
            cpLog( LOG_DEBUG, "Rejecting file '%s'.", userFilename.c_str() );
            continue;
        } /*if*/

        //Strip out the user name from the user filename.
        FsUserId userNameOnly;
        if (getUserName( userFilename, userNameOnly ) == false )
        {
            cpLog(LOG_ERR, "Can't find a name in '%s.'", userFilename.c_str());
            continue;
        } /*if*/

#if 0 //Change to 1 if needed for debugging
        cpLog(LOG_DEBUG, "Adding (%s)", userNameOnly.c_str() );
#endif

        //Add user name and userData (pointing to no features) to FsUsers map
        //if user isn't already found in map.
        if ( FsUsers::instance()->getUserData( userNameOnly ) == 0 )
        {
            Sptr < FsUserData > dummyUserData = new FsUserData;
            dummyUserData->setUserFileName( userFilename.c_str() );
            FsUsers::instance()->addUserData( userNameOnly, dummyUserData );
        }  /*if*/
        else
            cpLog(LOG_DEBUG, "'%s' already exists.", userNameOnly.c_str() );

    } /*while*/
}  //retrieveFeatureUsers


///
void
FsProvisioning::retrieveUserCPLFeatures( const char* userFile )
{
    //Get provisioning file name for this user
    string userFilename( userFile );

    cpLog( LOG_DEBUG, "Contact Provisioning for '%s'.", userFilename.c_str() );

    //Get user's Called feature
    string cplScript;

    if ( getProvisionedUserFeature( FsProvisioning::FeaturesIncoming,
                                    userFilename,
                                    cplScript ) == true )
    {
        //Parse the script, store any user data, and create
        //the feature.
        if (!createFeatureFromCPL( cplScript, userFilename ) )
            cpLog( LOG_ERR, "Could not create Called feature for '%s'.",
                   userFilename.c_str() );
    } /*if*/

    //Get user's Calling feature
    if ( getProvisionedUserFeature( FsProvisioning::FeaturesOutgoing,
                                    userFilename,
                                    cplScript ) == true )
    {
        //Parse the script, store any user data, and create
        //the feature.
        if (!createFeatureFromCPL( cplScript, userFilename ) )
            cpLog( LOG_ERR, "Could not create Calling feature for '%s'.",
                   userFilename.c_str() );
    } /*if*/

}  //retrieveUserCPLFeatures


///
bool
FsProvisioning::getProvisionedUserFeature( const int scriptType,
        const string& userFilename,
        string& cplScript )
{
    string relativeUserFilenamePath;
    //Need to build relative path to file which means including
    //.../Called/<filename> or .../Calling/<filename>
    if ( scriptType == FsProvisioning::FeaturesIncoming )
        relativeUserFilenamePath = "Called/";
    else
        relativeUserFilenamePath = "Calling/";
    relativeUserFilenamePath += userFilename;

    //cpLog (LOG_DEBUG, "relativeUserPath(%s) ",
    //       relativeUserFilenamePath.c_str() );

    try
    {
        cplScript = myPif->readFile( myFeatureServerName,
                                    PortNumber,
                                    ProvisionInterface::VFEATURE,
                                    relativeUserFilenamePath );
        if ( cplScript.length() > 0 )
        {
            //The CPL script must end with a null character;
            //otherwise xmlParseDocument(cplScript) crashes.
            cplScript[ cplScript.length() ] = 0;
            return true;
        } /*if*/
        else
            return false;
    } /*try*/
    catch (VException& e)
    {
        return false;
    } /*catch*/

} //getProvisionedUserFeature





///
bool
FsProvisioning::createFeatureFromCPL( const string& cplScript,
                                      const string& userFilename )
{
    //cpLog(LOG_DEBUG,"createFeatureFromCPL-userFileName(%s) cplScript(%s)",
    //                  userFilename.c_str(), cplScript.c_str() );

    string userName;

    //Get the user name from the filename (everything before .cpl).
    if (getUserName( userFilename, userName ) == false )
    {
        cpLog( LOG_ERR, "Can't parse '%s':  Missing .cpl extension.",
               userFilename.c_str() );
        return false;
    } /*if*/

    //Interpret the CPL script to feature and store user
    //data with the feature.
    Sptr < CPLStartStateHolder > featureStartState;
    featureStartState = CPLFeatureBuilder::Instance()->create(cplScript);

    if (featureStartState == 0)
        return false;

    bool wasFtrCreated = false;

    //Store the incoming feature if available
    if ( featureStartState->getIncoming() != 0 )
    {
        Sptr < FsFeature > newFeature = new FsFeature;
        newFeature->setFeatureScript( userFilename );
        newFeature->addState( featureStartState->getIncoming() );
        wasFtrCreated = addFeature( userFilename,
                                    newFeature,
                                    FsProvisioning::FeaturesIncoming );
    } /*if*/

    //Store the outgoing feature if available
    if ( featureStartState->getOutgoing() != 0 )
    {
        Sptr < FsFeature > newFeature = new FsFeature;
        newFeature->setFeatureScript( userFilename );
        newFeature->addState( featureStartState->getOutgoing() );
        wasFtrCreated |= addFeature( userFilename,
                                     newFeature,
                                     FsProvisioning::FeaturesOutgoing );
    }  /*if*/

    return wasFtrCreated;

} //createFeatureFromCPL


///
bool
FsProvisioning::addFeature( const string& userFilename,
                            Sptr < FsFeature > newFeature,
                            const int& featureType)
{
    string userNameOnly;
    if ( getUserName( userFilename, userNameOnly ) == false )
    {
        cpLog( LOG_ERR, "Can't find a name in '%s'.",
               userFilename.c_str());
        return false;
    } /*if*/

    FsUserId userID( userNameOnly );
    Sptr < FsUserData > userData;

    //Search to see if script owner is already in the database.
    userData = FsUsers::instance()->getUserData( userID );
    if ( userData == 0 )
    {
        //Script owner is NOT found in the database.  Create
        //new user information.
        Sptr < FsUserData > newUserDataEntry = new FsUserData;
        if ( featureType == FsProvisioning::FeaturesIncoming )
            newUserDataEntry->addCalledFeature( newFeature );
        else
            newUserDataEntry->addCalleeFeature( newFeature );

        //Store new user and its data into List.
        FsUsers::instance()->addUserData( userNameOnly, newUserDataEntry);

        return true;
    }  /*if*/
    else
    {
        //There is existing information for this user.  We must
        //update it with the new feature.
        if ( featureType == FsProvisioning::FeaturesIncoming )
            userData->replaceCalledFeature( newFeature );
        else
            userData->replaceCalleeFeature( newFeature );
        return true;
    }  /*else*/
}  //addFeature


///
void
FsProvisioning::updateUserAddDel( const string& cplData,
                                  const string& filename )
{
    cpLog( LOG_DEBUG, "updateUserAddDel-filename(%s) data(%s)",
           filename.c_str(), cplData.c_str() );

    int scriptType;

    //Find the appropriate subdirectory for the new user
    if ( filename.find("Called") != string::npos )
        scriptType = FsProvisioning::FeaturesIncoming;
    else if ( filename.find("Calling") != string::npos )
        scriptType = FsProvisioning::FeaturesOutgoing;
    else
    {
        cpLog (LOG_ERR, "Missing Called/Calling subdirectory." );
        return ;
    } /*else*/

    //Go through data and process each user
    //Data has users separated by a comma-delimiter
    unsigned int startIdx = 0;
    unsigned int endIdx;

    while ( startIdx < cplData.length() )
    {
        string userFilename;

        //Pull out one userFilename from a comma-delimited list
        //of filenames.
        endIdx = cplData.find(",", startIdx );
        if ( endIdx == string::npos )
        {
            userFilename += cplData.substr( startIdx,
                                            cplData.length() - startIdx );
            startIdx = cplData.length() + 1;
        } /*if*/
        else
        {
            userFilename += cplData.substr( startIdx, endIdx - startIdx );
            startIdx = endIdx + 1;
        } /*else*/

        //Do not process files that may not be CPL files.
        if ( isValidCPLFilename( userFilename ) == false )
        {
            cpLog (LOG_DEBUG, "Rejecting file '%s'.", userFilename.c_str() );
            continue;
        }

        //Fetch the user's feature script from the Provisioning Server
        string cplScript;
        if ( getProvisionedUserFeature ( scriptType, userFilename,
                                         cplScript ) == false )
            continue;

        //Parse the script and build the feature
        if ( createFeatureFromCPL( cplScript, userFilename ) == false )
            cpLog( LOG_ERR, "Failed to build feature for '%s.'",
                   userFilename.c_str() );
    }  /*while*/

} //updateUserAddDel


///
bool
FsProvisioning::getUserName( const string& fileName, string& user )
{
    int state = 1;
    int extensionIdx = -1;
    int idx = 0;
    int fileNameSize = fileName.length();

    //This state machine looks for the first occurrence
    //of '.cpl.'  It disregards whether 'cpl' is in
    //lower-, upper-, or mixed case.
    while ( (idx < fileNameSize) && (state != 5) )
    {
        switch (state)
        {
            case 1:
            if (fileName[idx] == '.')
            {
                extensionIdx = idx;
                state = 2;
            } /*if*/
            break;
            case 2:
            if ( (fileName[idx] == 'c') || (fileName[idx] == 'C') )
                state = 3;
            else if (fileName[idx] == '.')
            {
                extensionIdx = idx;
                state = 2;
            } /*if*/
            else
            {
                state = 1;
                extensionIdx = -1;
            }
            break;
            case 3:
            if ( (fileName[idx] == 'p') || (fileName[idx] == 'P') )
                state = 4;
            else if (fileName[idx] == '.')
            {
                extensionIdx = idx;
                state = 2;
            } /*if*/
            else
            {
                state = 1;
                extensionIdx = -1;
            }
            break;
            case 4:
            if ( (fileName[idx] == 'l') || (fileName[idx] == 'L') )
                state = 5;
            else if (fileName[idx] == '.')
            {
                extensionIdx = idx;
                state = 2;
            } /*if*/
            else
            {
                state = 1;
                extensionIdx = -1;
            }
            break;
            default:
                ;  //Nothing to do
        } /*switch*/
        idx++;
    } /*while*/

    //If the extension was found, return the user
    //name that appears before the extension
    if (state == 5)
    {
        user = fileName.substr(0, extensionIdx );
        // remove anything after the first '@'
        user = user.substr( 0, user.find( '@' ) );
        // and then anything before the last ':'
        user = user.substr( user.rfind( ':' ) + 1 );
        return true;
    }  /*if*/
    else
        return false;
} //getUserName


///
bool
FsProvisioning::isValidCPLFilename( const string& filename ) const
{
    int currentState = 1;
    char fileChar;

    //State machine to look for .cpl as the last
    //characters in the filename.
    for (int i = filename.length() - 1; i >= 0; i-- )
    {
        fileChar = toupper( filename[i] );

        switch ( currentState )
        {
            case 1:
            if ( fileChar == 'L')
                currentState = 2;
            else if ( fileChar != ' ' )
                currentState = 6;
            break;
            case 2:
            if ( fileChar == 'P')
                currentState = 3;
            else
                currentState = 6;
            break;
            case 3:
            if ( fileChar == 'C')
                currentState = 4;
            else
                currentState = 6;
            break;
            case 4:
            if ( fileChar == '.')
                currentState = 5;
            else
                currentState = 6;
            break;
            default:
            currentState = 6;
        } /*switch*/

        if ( currentState == 5 ) return true;
        if ( currentState == 6 ) return false;
    } /*for*/

    return false;

} //isValidCPLFilename


///
void
FsProvisioning::retrieveGlobalData()
{
    VGlobalConfigData provisioningData;

    //Fetch Global Data to determine if cc-diversion will be used.
    //addCCDiversion is defined in FsConfigFile 
    try
    {
        myPif->getConfigData( provisioningData );
        addCCDiversion = provisioningData.getRedirReasonInSIP();

    } /*try*/
    catch ( VException& excpInfo )
    {
        cpLog( LOG_ERR, excpInfo.getDescription().c_str() );
        addCCDiversion = true;  //Default to true 
    } /*catch*/

    //Print debugging information 
    if ( addCCDiversion == true )
        cpLog( LOG_DEBUG, "CC-Diversion ENABLED." );
    else if ( addCCDiversion == false )
        cpLog( LOG_DEBUG, "CC-Diversion DISABLED." );
    else
        cpLog( LOG_DEBUG, "retrieveGlobalData=%d", addCCDiversion );

} //retrieveGlobalData

   
//////////////////////////////////
/// Callback /////////////////////
//////////////////////////////////
void
Vocal::notifyAddDel(const string& data, const string& fileName, bool)
{
    cpLog( LOG_DEBUG, "notifyAddDel()" );

    FsProvisioning::Instance()->updateUserAddDel( data, fileName );

    FsUsers::instance()->showUsers();
} //notifyAddDel

// End of File

