#ifndef CPLOpAddressSwitch_Hxx
#define CPLOpAddressSwitch_Hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const CPLOpAddressSwitchVersion =
    "$Id: CPLOpAddressSwitch.hxx,v 1.13 2002/11/09 02:05:37 derekm Exp $";




#include "CPLOperator.hxx"


namespace Vocal
{


/**
 *  Decide next path to take based on either origin or destination address.
 */
class CPLOpAddressSwitch : public CPLOperator
{
    public:
        /**@name Initialization & Deallocation Functions */
        //@{
        ///Constructor
        CPLOpAddressSwitch();
        ///Destructor
        ~CPLOpAddressSwitch();
        //@}

        /**@name State Machine Building Functions */
        //@{
        ///Append an operator to this class
        bool setNextOperator( Sptr < Operator > op, int prevNode );
        ///Initialize the Attributes member variables
        bool setAttributes( const char* attrName, const char* attrValue);
        //@}

        /**@name State Machine Execution Functions */
        //@{
          
        /**Execute the operator for the state machine
          *@param event Event from the Sip stack 
          *@return Pointer to next state or 0 if the operator did not need to
          *process the event. 
         **/
        const Sptr < State > process( Sptr < SipProxyEvent > event);

        /** Get the telephone number from a tel: field or from the user field if
          * the parameter, user=phone, is defined. 
          * @param baseUrl contains the URL 
          * @param url returns the telephone number if TRUE is returned.  */
        bool getTelephoneNumber( const Sptr<BaseUrl> baseUrl, 
                                 Data& phoneNum ) const;

        /**Retrieves the area code from a NPA-Nxxx telephone number.
          *Useful for North American phone numbers. 
          *@param url is a telephone number in the format 1+Nxx+Nxx+xxxx 
          *@return url contains area code only if TRUE is returned. 
          */
        bool getNPAAreaCode( Data& url ) const;

        /**Retrieves the prefix or central office code from a NPA-Nxxx
          *telephone number.
          *Useful for North American phone numbers. 
          *@param url is a telephone number in the format 1+Nxx+Nxx+xxxx 
          *@return url contains prefix only if TRUE is returned. 
          */
        bool getNPAPrefix( Data& url ) const;

        /**Retrieves the line number from a NPA-Nxxx telephone number
          *Useful for North American phone numbers. 
          *@param url is a telephone number in the format 1+Nxx+Nxx+xxxx 
          *@return url contains the Line number only if TRUE is returned */
        bool getNPALineNumber( Data& url ) const;
        //@}

        /**@name Debugging Functions */
        //@{
        ///cpLogs values of member variables
        void logData() const;
        ///Returns the name of the class
        const char* const name() const;
        ///Calls all its operators::doCpLog()
        void doCpLog() const;
        //@}

        /**@name Constants */
        //@{
        /** NotPresentField = 0,
          * ToField = 1,
          * FromField = 2,
          * RequestURLField = 3
          */
        enum _Field { NotPresentField = 0, ToField, FromField, RequestURLField };
        ///
        typedef _Field Field;

        /** EntireAddress_SubField = 0,
          * AddressType_SubField = 1,
          * User_SubField = 2,
          * Host_SubField = 3,
          * Port_SubField = 4,
          * Tel_SubField = 5,
          * Display_SubField = 6 
          * AreaCode_Subfield = 102,
          * Prefix_Subfield = 103
          * LineNumber_Subfield = 104
          */
        enum _SubField { EntireAddress_SubField = 0,
                         AddressType_SubField,
                         User_SubField,
                         Host_SubField,
                         Port_SubField,
                         Tel_SubField,
                         Display_SubField,
                         AreaCode_SubField = 102,
                         Prefix_SubField,
                         LineNumber_SubField };
        ///
        typedef _SubField SubField;
        //@}

    private:
        /**@name CPL Attributes */
        //@{
        ///Expecting a 'Field' enum
        int myField;
        ///Expecting a 'SubField' enum
        int mySubfield;
        //@}

        /**@name Special Operator */
        //@{
        ///
        Sptr < Operator > myNotPresentOp;
        //@}

        /**@name State Machine Execution Functions */
        //@{
        /** Use the last SIP message to find the appropriate Field.
          * @see field
          */
        bool getUrl( Sptr < SipProxyEvent > event, Data& url ) const;
        //@}


};
 
} // namespace Vocal

#endif

