/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package com.vovida.telephony;

import com.vovida.telephony.callcontrol.*;
import javax.telephony.*;
import javax.telephony.callcontrol.*;

/**
 *    Class EventHandler handles all events in the JTapi Library
 *    Events are Requests, Responses, Notifications from the JTapi
 *    server, or a Timer event which is instantiated periodically 
 *    to have all pending Requests checked for a timeout situation
 *
 *    This is a thread which constantly reads the events (if any) on
 *    the incoming FIFO Queue For each event a certain action must be taken,
 *    which is carried out on this thread, before attempting to read the 
 *    incoming FIFO Queue again.
 */

class VCEventHandler extends Thread implements VCMsgTypes
{
    ///Fifo used to process incoming events to Jtapi client
    VCFifo infifo;
    ///While loop boolean value
    boolean keepRunning = true;
    /// Provider object
    Provider provider;

    /**
     * Constructor
     */
    public VCEventHandler(VCFifo iq, Provider prov) 
    { 
	this.infifo = iq;
	this.provider = prov;
    }

    /**
     * Monitors the Fifo and processes incoming events in turn
     */
    public void run()
    {
	while (keepRunning == true)
	{
	    Object data_ref = infifo.pop();
	    if (data_ref != null)
	    {
		VCEvent evt = (VCEvent)data_ref;
		int evtType = evt.getEvtType();
		switch (evtType)
		{
		    case VCEvent.REQUESTEVENT : 
			{
			    System.out.println("Request EVT");
			    try
			    {
				handleRequests(evt);
			    }
			    catch (Exception e)
			    {
				e.printStackTrace();
			    }
			}
			break;
		   case VCEvent.RESPONSEEVENT : 
			{
			    System.out.println("Response EVT");
                            try
                            {
                                handleResponses(evt);
                            }
                            catch (Exception e)
                            {
                                e.printStackTrace();
                            }
			}
			break;
		    case VCEvent.NOTIFICATIONEVENT :
			{
			    System.out.println("Notification EVT");
                            try
                            {
                                handleNotifications(evt);
                            }
                            catch (Exception e)
                            {
                                e.printStackTrace();
                            }
			}
			break;
		    case VCEvent.TIMEEVENT :
			{
			    System.out.println("Timer EVT");
			    checkRequestsForTimeout();
			}
			break;
		    default :
			{}
			break;
		}
	    }
	}
    }

    /**
     * Set the while loop boolean - can stop processing of events
     */
    public void setKeep(boolean b)
    {
	keepRunning = b;
    }

/**
 *   The pending Requests are stored by the provider and removed as
 *   a response is received for a given request Each request is 
 *   has an Expiry Timer set upon sending the request If the timer is 
 *   found to be expired, there may be a communication problem with
 *   the server or the server may be down.
 */

    private void checkRequestsForTimeout()
    {
	try
	{
	    ((VCProvider)provider).checkForExpiredRequests();
	}
	catch (ExpiryTimerException e)
	{
	    VCRequest req;
	    req = e.getRequest();
	    System.out.println("Timer Expired for Request :" + e.getRequest());
            if (req.isBlocked())
            {
		boolean result = false;
                req.inputRecd(result);
            }
	    // should have some event here indicating
	    // that the provider may be out of service
	    // or some way to let app know that server
	    // is not responding
	}
    }

/**
 *  Requests from the Server must have a request id included
 *  The type of request is determined and the appropriate action
 *  taken For each request an appropriate response must also 
 *  be sent back to the server.
 */

    private void handleRequests(VCEvent evt) throws InvalidStateException,
                    InvalidArgumentException, MethodNotSupportedException,
		    PrivilegeViolationException, ResourceUnavailableException
			
    {
	String id = evt.getEvtMsgId();
	if (id == null)
	{
	    // exception
	}
        int evtMsgType = evt.getEvtMsgType();
        switch (evtMsgType)
        {
	    case CONNECT:
 	    case DISCONNECT:
	    case ANSWER :
                {
                }
                break;
	    case CALLCREATE :
		{
		    Call call = ((VCProvider)provider).createCall();
// code to get around server not receiving response
 		    System.out.println("SET CALLID TO 99999");
		    ((VCCall)call).setName("99999");
		    String callId = ((VCCall)call).getCallId();
		    ((VCProvider)provider).addCallToList(call, callId);
// end added code
		    if (call != null)
		    {
                String[] addrs = evt.getEvtAddresses();
                if (addrs != null)
                {
                    if (addrs.length == 2)
                    {
                        VCRequestUtil.callCreateResp(call, id, addrs[0], addrs[1]);
                    }
                }
		    }
		}
		break;
            default :
                {
                }
                break;
        }

    }

/**
 *  For each response received from the server, there should be a 
 *  corresponding request from the server in the list of Pending Requests
 *  If the request has been sent in a Blocking state then it is necessary 
 *  to release the Block on the requesting thread.
 */

    private void handleResponses(VCEvent evt) throws InvalidStateException,
			InvalidArgumentException
    {
	int state, ctrlState;
	int evtMsgType = evt.getEvtMsgType();
	String msgId = evt.getEvtMsgId();
	if (msgId == null)
	{
	    throw (new InvalidArgumentException());
	}
	VCRequest req = ((VCProvider)provider).getRequest(msgId);
	if (req == null)
	{
	    throw (new InvalidArgumentException());
	}
	String callStr = evt.getEvtCallId();
	Call call = null;
	if (callStr != null)
	{
	    call = ((VCProvider)provider).findCall(callStr);
	}
        switch (evtMsgType)
	{
	    case CONNECTRESP :
		{
		    // Connecting in Progress
		    // No need for action since connect is non-blocking
		    // and so once the request has been  removed from 
		    // pendingrequests, that's all the requirements..
		}
		break;
	    case DISCONNECTRESP :
		{
		    boolean result = false;
		    int value = evt.getEvtValue();
		    if (value == SUCCESS)
		    {
		        result = true;
		    }	
		    if (req.isBlocked())
	 	    {
		        req.inputRecd(result);
		    }
		    System.out.println("HandleResponse - DisConnect");
		}
		break;
	    case ANSWERRESP :
		{
                    boolean result = false;
                    int value = evt.getEvtValue();
                    if (value == SUCCESS)
                    {
                        result = true;
                    }
                    if (req.isBlocked())
                    {
                        req.inputRecd(result);
                    }
		    System.out.println("HandleResponse - Answer");
		}
		break;
            case TRANSFERRESP :
                {
                    boolean result = false;
                    int value = evt.getEvtValue();
                    if (value == SUCCESS)
                    {
                        result = true;
                    }
                    if (req.isBlocked())
                    {
                        req.inputRecd(result);
                    }
                    System.out.println("HandleResponse - Transfer");
                }
                break;
	    case REGISTERRESP :
                {
                    boolean result = false;
                    int value = evt.getEvtValue();
                    if (value == SUCCESS)
                    {
                        result = true;
                    }
                    if (req.isBlocked())
                    {
                        req.inputRecd(result);
                    }
                    System.out.println("HandleResponse - Register");
                }
		break;
            case CHKREGISTERRESP :
                {
		    VCProvider.clearProviderDomain();
                    boolean result = false;
                    int value = evt.getEvtValue();
                    if (value == SUCCESS)
                    {
		        String[] addrs = evt.getEvtAddresses();
			if (addrs != null)
			{
			    if (addrs.length == 1)
			    {
			        VCProvider.clearProviderDomain();
			        // term is same as address
			        VCProvider.setProviderDomain(addrs[0], addrs[0]);
			        Address addr = provider.getAddress(addrs[0]);
			        if (addr != null)
			        {
				    result = true;
			        }
			    }
			}
		    }
                    if (req.isBlocked())
                    {
                        req.inputRecd(result);
                    }
		    System.out.println("HandleResponse - ChkRegister");
                }
                break;

	    default :
		{
		}
		break;
	}
	
    }

    /**
     * Handle notifications from jtapi server regarding call processing
     */
    private void handleNotifications(VCEvent evt) throws InvalidStateException,
                        InvalidArgumentException
    {
	int state, ctrlState;
	Call call = ((VCProvider)provider).findCall(evt.getEvtCallId());
        int evtMsgType = evt.getEvtMsgType();
        switch (evtMsgType)
        {
            case CALLCREATED :
                {
                }
                break;
            case CALLACTIVE :
                {
                }
                break;
            case CALLINVALID :
                {
                }
                break;
	    case CONNCREATED :
		{
		    if (call != null)
		    {
			state = Connection.CONNECTED;
			ctrlState = CallControlConnection.NETWORK_REACHED;
			System.out.println("Connections Created");
			// find the appropriate address objects..
			String[] addrs = evt.getEvtAddresses();
			if (addrs == null)
			{
			    throw (new InvalidArgumentException());
			}
			if (addrs.length == 2)
			{
                            Address addr1 = ((VCProvider)provider).getTheAddress(addrs[0]);
                            Address addr2 = ((VCProvider)provider).getTheAddress(addrs[1]);

			    // check if connections exist
			    if (call.getConnections() == null)
			    {
			        System.out.println("Address 1 " + addr1.getName());
			        System.out.println("Address 2 " + addr2.getName());
			        // create 2 connections in call
			        Connection[] conns = ((VCCall)call).createConnections(addr1, addr2);
			    }
			    Connection[] conns = call.getConnections();
			    if (conns != null)
			    {
			        if (conns.length < 2)
			        {
			            throw (new InvalidArgumentException());
			        }
			        else if (conns.length == 2)
			        {
				    System.out.println("Setting State Active");
				    System.out.println("Connections 1 & 2");
				    System.out.println(conns[0].getAddress());
				    System.out.println(conns[1].getAddress());
                                    ((VCCall)call).setState(Call.ACTIVE);
				    if ((addr1 == conns[0].getAddress()) && 
				        (addr2 == conns[1].getAddress()))
				    {  
				        setConnState(call, ORIG, state, ctrlState);
				    }
				    else
				    {
				        throw (new InvalidArgumentException());
				    }
			        }
		  	    }
			    else
			    {
				throw (new InvalidArgumentException());
			    }
			}
		    }
		}
		break;
            case CONNINPROGRESS :
                {
		    if (call != null)
		    {
			String[] addrs = evt.getEvtAddresses();
		        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
		 	}
			if (addrs.length == 1)
			{
			    if (((VCCall)call).hasConnection(addrs[0]) == true)
                            {
				if (((VCCall)call).isOrigConnection(addrs[0]) == true)
				{
				    state = Connection.CONNECTED;
				    ctrlState = CallControlConnection.NETWORK_REACHED;
				    setConnState(call, ORIG, state, ctrlState);
			        }
				else
				{
				    state = Connection.INPROGRESS;
				    ctrlState = CallControlConnection.OFFERED;
				    setConnState(call, TERM, state, ctrlState);
				}
			    }
			}
			else
			{
			    throw (new InvalidArgumentException());
			}
		    }
                }
                break;
            case CONNALERTING :
                {
		    if (call != null)
                    {
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
                            throw (new InvalidArgumentException());
                        }
                        if (addrs.length == 1)
                        {
                            if (((VCCall)call).hasConnection(addrs[0]) == true) 
                            {
				if (((VCCall)call).isOrigConnection(addrs[0]))
				{
				    state = Connection.CONNECTED;
				    ctrlState = CallControlConnection.NETWORK_ALERTING;
                                    setConnState(call, ORIG, state, ctrlState);
				}
				else
				{
				    state = Connection.ALERTING;
				    ctrlState = CallControlConnection.ALERTING;
				    setConnState(call, TERM, state, ctrlState);
				}
                            }
                        }
                        else
                        {
                            throw (new InvalidArgumentException());
                        }

                    }
                }
                break;
            case CONNCONNECTED :
                {
		    if (call != null)
                    {
			state = Connection.CONNECTED;
			ctrlState = CallControlConnection.ESTABLISHED;
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
                        if (addrs.length == 1)
                        {
                            if (((VCCall)call).hasConnection(addrs[0]))
                            {
				if (((VCCall)call).isOrigConnection(addrs[0]))
				{
                                    setConnState(call, ORIG, state, ctrlState);
				}
				else
				{
				    setConnState(call, TERM, state, ctrlState);
				}
			    }
                        }
                        else
                        {
                            throw (new InvalidArgumentException());
                        }
                    }
                }
                break;
            case CONNDISCONNECTED :
                {
                    if (call != null)
                    {
			state = Connection.DISCONNECTED;
			ctrlState = CallControlConnection.DISCONNECTED;
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
			if (addrs.length == 1)
                        {
                            if (((VCCall)call).hasConnection(addrs[0]))
                            {
				if (((VCCall)call).isOrigConnection(addrs[0]))
				{
				    setConnState(call, ORIG, state, ctrlState);
				}
				else
				{
				    setConnState(call, TERM, state, ctrlState);
				}
                            }
                            else
                            {
				// all connections already removed from call
				// OK to come here for last connection
				// since when only one connection remains
				// in call it is automatically removed
                            }
                        }
			else
			{
			    throw (new InvalidArgumentException());
			}
                    }
                }
                break;
            case CONNFAILED :
                {
		    if (call != null)
                    {
			state = Connection.FAILED;
			ctrlState = CallControlConnection.FAILED;
			int value = evt.getEvtValue();
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
			if (addrs.length == 1)
                        {
                            if (((VCCall)call).hasConnection(addrs[0]))
                            {
//                              ((VCConnection)conn).setStateReason(value);
				if (((VCCall)call).isOrigConnection(addrs[0]))
				{
				    setConnState(call, ORIG, state, ctrlState);
				}
				else
				{
				    setConnState(call, TERM, state, ctrlState);
				}
                            }
                            else
                            {
				throw (new InvalidArgumentException());
                            }
                        }
                    }
                }
                break;
            case CONNUNKNOWN :
                {
                }
                break;
            case TERMCONNCREATED :
                {
                }
                break;
            case TERMCONNRINGING :
                {
		    if (call != null)
                    {
			state = TerminalConnection.RINGING;
			ctrlState = CallControlTerminalConnection.RINGING;
                        // find the appropriate address objects..
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
                            return;
                        }
                        if (addrs.length == 1)
                        {
			    Connection conn = ((VCCall)call).findConnection(addrs[0]);
			    if (conn != null)
			    {
			        Address addr1 = conn.getAddress();
			        Terminal[] terms = ((VCAddress)addr1).getTerminals();
			        if (terms != null)
			        {
			            for (int i=0; i < terms.length; i++)
			            {
				        VCTerminalConnection tc = new VCCallControlTerminalConnection(terms[i], conn);
				        setTermConnState(tc, state, ctrlState);
			            }
			        }
			    }
			    else
			    {
				   throw (new InvalidArgumentException()); 
			    }
			}
                    }
                }
                break;
	    case TERMCONNACTIVE :
                {
                    if (call != null)
                    {
			state = TerminalConnection.ACTIVE;
			ctrlState = CallControlTerminalConnection.TALKING;
			System.out.println("TermConn - Active");
                        // find the appropriate address objects..
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
                        if (addrs.length == 1)
                        {
                            Connection conn = ((VCCall)call).findConnection(addrs[0]);
			    if (conn == null)
			    {
				throw (new InvalidArgumentException());
			    }
			    TerminalConnection[] tcs = conn.getTerminalConnections();
			    // if we have a term passed in msg we find it
			    // otherwise set all to ACTIVE

			    // if no terminalconnections -- may need to 
			    // create them..
			    if (tcs == null)
			    {
				Address addr1 = conn.getAddress();
				System.out.println("Addr :" + addr1.getName());
				Terminal[] terms = ((VCAddress)addr1).getTerminals();
				if (terms != null)
                                {
				    System.out.println("Creating TermConns");
                                    for (int i=0; i < terms.length; i++)
                                    {
                                        VCTerminalConnection tc = new VCCallControlTerminalConnection(terms[i], conn);
				    }
				    tcs = conn.getTerminalConnections();
				}
			    }
			    if (tcs != null)
			    {
			        for (int i=0; i < tcs.length; i++)
			        {
				    setTermConnState(tcs[i], state, ctrlState);
			        }
			    }	
                        }
                    }
                }
                break;
            case TERMCONNHELD :
                {
                    if (call != null)
                    {
			state = TerminalConnection.ACTIVE;
			ctrlState = CallControlTerminalConnection.HELD;
                        System.out.println("TermConn - Held");
                        // find the appropriate address objects..
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
                        if (addrs.length == 1)
                        {
                            Connection conn = ((VCCall)call).findConnection(addrs[0]);
			    if (conn == null)
			    {
				throw (new InvalidArgumentException());
			    }
                            TerminalConnection[] tcs = conn.getTerminalConnections();
                            // if we have a term passed in msg we find it
                            // otherwise set all to HELD

                            if (tcs != null)
                            {
                                for (int i=0; i < tcs.length; i++)
                                {
				    setTermConnState(tcs[i], state, ctrlState);
				}
			    }
			    else
			    {
				throw (new InvalidArgumentException());
                            }

                        }
                    }
                }
                break;
            case TERMCONNPASSIVE :
                {
                }
                break;
            case TERMCONNDROPPED :
                {
                    if (call != null)
                    {
                        state = TerminalConnection.DROPPED;
                        ctrlState = CallControlTerminalConnection.DROPPED;
                        System.out.println("TermConn - Dropped");
                        // find the appropriate address objects..
                        String[] addrs = evt.getEvtAddresses();
                        if (addrs == null)
                        {
			    throw (new InvalidArgumentException());
                        }
                        if (addrs.length == 1)
                        {
                            Connection conn = ((VCCall)call).findConnection(addrs[0]);
			    if (conn == null)
			    {
				throw (new InvalidArgumentException());
			    }
                            TerminalConnection[] tcs = conn.getTerminalConnections();
                            // if we have a term passed in msg we find it
                            // otherwise set all to DROPPED

                            if (tcs != null)
                            {
                                for (int i=0; i < tcs.length; i++)
                                {
				    setTermConnState(tcs[i], state, ctrlState);
				}

                            }
			    else
			    {
				throw (new InvalidArgumentException());
                            }
                        }
                    }

                }
                break;
            case TERMCONNUNKNOWN :
                {
                }
                break;
            default :
                {
                }
                break;
        }

    }

	
    /**
     * Set the Connection state
     */
    private void setConnState(Call call, int target, int state, int ctrlState)
	throws InvalidArgumentException, InvalidStateException
    {
	VCCallControlConnection conn;
	Connection[] conns = call.getConnections();
        if (conns.length == 2)
        {
	    if ((target == TERM) || (target == BOTH))
	    {
		conn = (VCCallControlConnection)conns[1];
		if (ctrlState > 0)
		    conn.setCallControlState(ctrlState);
                if (state > 0)
                    conn.setState(state);
            }
	    if ((target == ORIG) || (target == BOTH))
            {
                conn = (VCCallControlConnection)conns[0];
                if (ctrlState > 0)
                    conn.setCallControlState(ctrlState);
                if (state > 0)
                    conn.setState(state);
	    }
	}
        else
        {
            throw (new InvalidArgumentException());
        }
    }

    /**
     * Set the TermConn state
     */
    private void setTermConnState(TerminalConnection tc, int state, int ctrlState)
	throws InvalidArgumentException, InvalidStateException
    {
	VCCallControlTerminalConnection termConn;
	termConn = (VCCallControlTerminalConnection)tc;
	if (ctrlState > 0)
	    termConn.setCallControlState(ctrlState);
	if (state > 0)
	    termConn.setState(state);
    }

} 
	    
	
