/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package com.vovida.telephony;

import com.vovida.telephony.*;
import com.vovida.singleton.*;
import javax.telephony.*;
import java.util.*;

/**
 * The Request class contains the data wich is 
 * to be sent to the server in an individual message
 * An XML message is created based on the given
 * data values and this message is put onto
 * the FIFO queue for outgoing messages
 * Additionally the request is saved to a request
 * queue maintained by the Provider, which is used 
 * to ensure that requests are responded to within
 * a given time period
 * For requests which require to be blocked a wait
 * - notify on the request object is employed which is
 * released when the response is returned 
 * or the timer expires for this request
 */

class VCRequest implements VCMsgTypes
{
    private int type;
    private String id;
    private String call;
    private String[] addrs;
    private String term;
    private String pass;
    private String user;
    private String port;
    private boolean requestSuccess = false;
    private boolean blocked = false;
    private Provider provider;
    private long expiryTime;
    private VCIdentity idCreator;
    private VCFifo fifo;
    final static private int EXPIRYSECS = 10;

    /**
     * Constructor
     */
    public VCRequest(Provider prov)
    {
	this.provider = prov;
	type = 0;
        id = "";
	idCreator = VCProvider.getIdCreator();
        fifo = VCProvider.getOutFifo();
    }

    /**
     * Create char string for sending request to server
     * If blocking is required wait until response is
     * received from server
     */
    public boolean sendRequest(int cmd, boolean Blocking)
    {
	this.type = cmd;
        this.id = idCreator.getId();
	String data = parseCmd(cmd, id);
	byte[] marshalData = data.getBytes();
	fifo.push(marshalData);
	setTimer();
	((VCProvider)provider).addRequest(this.id, this);
	if (Blocking == true)
	{
	    waitForResponse();
	    return requestSuccess;
	}
	return true;

    }

    /**
     * Send response to the server in reply to a request
     * The id must already be set in the request object
     */
    public void sendResponse(int cmd)
    {
	this.type = cmd;
	String data = parseCmd(cmd, id);
	byte[] marshalData = data.getBytes();
	fifo.push(marshalData);
	System.out.println("Response Sent : " + data);
    }

    /**
     * Wait for response from server before continuing to process
     */
    public synchronized void waitForResponse()
    {
    	try
	{
	    blocked = true;
	    this.wait();
	}
	catch (Exception e)
	{
	    e.printStackTrace();
	}
    }

    /**
     * Indication that the waiting thread can continue to process
     * because a response has been received from server
     */
    public synchronized void inputRecd(boolean success)
    {
	blocked = false;
	requestSuccess = success;
	this.notify();
    }

    /**
     * Create char string of request/response to be sent to server
     */
    private String parseCmd(int cmd, String id)
    {
        String sendStr = "";
        if (cmd > 0)
        {
            String cmdStr = String.valueOf(cmd);
            sendStr = tag("type", cmdStr);
        }
        if (id != null)
        {
            sendStr += tag("id", id);
        }
        if (this.call != null)
        {
            sendStr += tag("call", this.call);
        }
	if (this.user != null)
	{
	    sendStr += tag("user", this.user);
	}
	if (this.pass != null)
	{
	    sendStr += tag("pass", this.pass);
	}
        if (this.addrs != null)
        {
            for (int i=0; i < this.addrs.length; i++)
            {
                sendStr += tag("addr", this.addrs[i]);
            }
        }
        if (this.term != null)
        {
            sendStr += tag("term", this.term);
        }
	if (this.port != null)
	{
	    sendStr += tag("port", this.port);
	}
	sendStr = tag("jtapi", sendStr);

	System.out.println("Request created :" + sendStr);
	return sendStr;
    }

    /**
     * Create the opentag - value - closetag string
     */
    private String tag(String tag, String val)
    {
        String tagsAndValue = "";
        tagsAndValue = "<" + tag + ">" + val + "</" + tag + ">";
        return tagsAndValue;
    }

    /**
     * Set timer for expiry of request to EXPIRYSECS number of seconds
     */
    public void setTimer()
    {
	Date now = new Date();
	expiryTime = now.getTime() + (EXPIRYSECS*1000);
    }

    /**
     * Get expiry time for request
     */
    public long getExpiryTime()
    {
	return expiryTime;
    }

    /**
     * Check if request is blocked
     */
    public boolean isBlocked()
    {
	return blocked;
    }

    /**
     * Print out the object
     */
    public String toString()
    {
	String str = "\n" + "Command: " + type + "\n";
	str += "ID: " + id;
	return str;
    }

    /**
     * Set the Request Id
     */
    public void setReqId(String idStr)
    {
	this.id = idStr;
    }

    /**
     * Set the Request Call
     */
    public void setReqCall(Call call)
    {
        this.call = ((VCCall)call).getCallId();
    }

    /**
     * Set the Request Addresses
     */
    public void setReqAddrs(String[] addresses)
    {
        this.addrs = addresses;
    }

    /**
     * Set the Request Terminal
     */
    public void setReqTerm(String term)
    {
        this.term = term;
    }

    /**
     * Set the Request Password
     */
    public void setReqPassWord(String pwd)
    {
	this.pass = pwd;
    }

    /**
     * Set the Request User
     */
    public void setReqUser(String usr)
    {
	this.user = usr;
    }

    /**
     * Set the Request Port
     */
    public void setReqPort(String portStr)
    {
	this.port = portStr;
    }

    /**
     * Get the Request Result
     */
    public boolean getRequestResult()
    {
	return requestSuccess;
    }

}
