
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const JSOpCallerReady_cxx_Version =
    "$Id$";
#include "JSOpCallerReady.hxx"
#include "SipEvent.hxx"
#include "JSStateMachine.hxx"
#include "JSCall.hxx"
#include "StatusMsg.hxx"
#include "InviteMsg.hxx"
#include "AckMsg.hxx"
#include "ByeMsg.hxx"
#include "SipSdp.hxx"
#include "SdpConnection.hxx"
#include "JSProvider.hxx"
//#include "SipProxyAuthorization.hxx"


///
JSOpCallerReady::JSOpCallerReady()
{}


///

JSOpCallerReady::~JSOpCallerReady()
{}


///

const char* const
JSOpCallerReady::name() const
{
    return "JSOpCallerReady";
}


///
const Sptr < State >
JSOpCallerReady::process( const Sptr < SipProxyEvent > event )
{
    // 1. Receive 200
    // 2. Get SDP and add it to call object
    // 3. Check if 200 is for INVITE or INVITE(hold), 200 for INVITE(hold) is
    //        handled in JSOpTransfer
    // 4. Send new ACK with local SDP to caller
    // 5. Start full duplex on local SDP
    // 6. Send INVITE(hold) to caller

    cpLog( LOG_DEBUG, "JSOpCallerReady operation" );

    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );
    if ( sipEvent == 0 )
    {
        return 0;
    }
    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
    assert( sipMsg != 0 );

    Sptr < StatusMsg > msg;
    msg.dynamicCast( sipMsg );
    if ( msg == 0 )
    {
        return 0;
    }

    if ( msg->getStatusLine().getStatusCode() != 200 )
    {
        return 0;
    }

    // Get remote SDP from message
    Sptr<SipSdp> sdp;
    sdp.dynamicCast ( sipMsg->getContentData(0) );
    if (sdp == 0)
    {
        // No remote SDP in 200 while this does not affect jS's functionaility
        // It is a violation of SIP protocol, terminate the call by
        // first ACK then BYE.
        cpLog( LOG_ERR, "No remote SDP or invalid SDP in 200, terminate call" );
        AckMsg ack( *msg );
        sipEvent->getSipStack()->sendAsync( ack );
        ByeMsg bye( *msg );
        sipEvent->getSipStack()->sendAsync( bye );
        // Change state to unknown to terminate this call
        Sptr < JSStateMachine > stateMachine;
        stateMachine.dynamicCast( event->getCallInfo()->getFeature() );
        assert( stateMachine != 0 );

        return stateMachine->findState( "JSStateUnknown" );
    }

    // WE can't differentiate between the 200 for INVITE and INVITE(hold)
    // untill someone tell me differently.
    // To solve this, I'll add another state, transfer call, which will
    // handle the 200 for invite(hold).
    // SO, we'll change state after we send out the invite hold here

    // This SDP is needed so that JS can open full duplex with caller
    // Remember caller doesn't know we are not an end point

    Sptr < JSCall > call;
    call.dynamicCast( event->getCallInfo() );
    assert( call != 0 );

    JSProvider* provider = call->getProvider();

    call.dynamicCast(event->getCallInfo());
    call->setFromSdp( new SipSdp( *sdp ) );
    call->setControl200( *msg );

    AckMsg ack( *msg );

    SipSdp localSdp( *sdp );

    // TODO We should't have to hard code the port here, use provider's config
    localSdp.setRtpPort( 23456 );
    int rtpPacketSize = 20;

    SdpSession sdpDesc = localSdp.getSdpDescriptor();
    vector < SdpMedia* > mediaList;
    mediaList = sdpDesc.getMediaList();
    vector < SdpMedia* > ::iterator mediaIterator = mediaList.begin();

    vector < int > * formatList = (*mediaIterator)->getFormatList();
    if ( formatList != 0 )
    {
        formatList->clear();
    }
    //using the default codec
    (*mediaIterator)->addFormat( 0 );

    MediaAttributes* mediaAttrib;
    mediaAttrib = (*mediaIterator)->getMediaAttributes();
    if ( mediaAttrib != 0 )
    {
        vector < ValueAttribute* > * valueAttribList = mediaAttrib->getValueAttributes();
        vector < ValueAttribute* > ::iterator attribIterator = valueAttribList->begin();
        while ( attribIterator != valueAttribList->end() )
        {
            char* attribName = (*attribIterator)->getAttribute();
            if ( strcmp( attribName, "ptime" ) == 0 )
            {
                rtpPacketSize = Data((*attribIterator)->getValue()).convertInt();
                break;
            }
            attribIterator++;
        }
        mediaAttrib->flushValueAttributes();
        mediaAttrib->flushrtpmap();

        ValueAttribute* attrib = new ValueAttribute();
        attrib->setAttribute( "ptime" );
        LocalScopeAllocator lo;
        attrib->setValue( Data( rtpPacketSize ).getData(lo) );

        //add the rtpmap attribute for the default codec
        SdpRtpMapAttribute* rtpMapAttrib = new SdpRtpMapAttribute();
        rtpMapAttrib->setPayloadType( 0 );
        rtpMapAttrib->setEncodingName( "PCMU" );
        rtpMapAttrib->setClockRate( 8000 );

        mediaAttrib->addValueAttribute( attrib );
        mediaAttrib->addmap( rtpMapAttrib );
    }
    else
    {
        mediaAttrib = new MediaAttributes();
        assert(mediaAttrib);
        (*mediaIterator)->setMediaAttributes(mediaAttrib);

        // create the new value attribute object
        ValueAttribute* attrib = new ValueAttribute();
        // set the attribute and its value
        attrib->setAttribute("ptime");
        attrib->setValue( "20" );

        // add the rtpmap attribute for the default codec
        SdpRtpMapAttribute* rtpMapAttrib = new SdpRtpMapAttribute();
        rtpMapAttrib->setPayloadType(0);
        rtpMapAttrib->setEncodingName("PCMU");
        rtpMapAttrib->setClockRate(8000);
        // add the value attribute just created to the media attribute objec
        mediaAttrib->addValueAttribute(attrib);
        mediaAttrib->addmap(rtpMapAttrib);
    }

    localSdp.setSdpDescriptor(sdpDesc);

    call->setLocalSdp( new SipSdp(localSdp) );

    ack.setContentData((SipContentData*) &localSdp);
    //ack.setAuthBasic( provider->getConfig()->getProxyAuthKey() );
    ack.setProxyAuth( provider->getConfig()->getProxyAuthKey() );

    cpLog(LOG_DEBUG, "JSOpCallerReady: ACK message for caller:");
    cpLog(LOG_DEBUG, "***************************************");
    cpLog(LOG_DEBUG, "%s", ack.encode().logData());
    cpLog(LOG_DEBUG, "***************************************");
    sipEvent->getSipStack()->sendAsync( ack );

    // use original INVITE, but use route saved, and set hold
    Sptr < InviteMsg > invite = sipEvent->getInvite();
    assert( invite != 0 );

    // Rather than playing with route manipulation, I'll use RS for now
    // Later, we should use route manipulation
    //invite->setRouteList( sipMsg->getrecordrouteList() );

    // Increment CSeq count
    SipCSeq cseq = invite->getCSeq();
    cseq.incrCSeq();
    invite->setCSeq(cseq);
    // Set SDP connection to HOLD
    SdpConnection* connectionHold = sdpDesc.getConnection();
    connectionHold->setHold();
    localSdp.setSdpDescriptor(sdpDesc);
    // invite->flushContentType();
    invite->setContentData((SipContentData*) &localSdp);

    // Add authorization
    //invite->setAuthBasic( provider->getConfig()->getProxyAuthKey() );
    invite->setProxyAuth( provider->getConfig()->getProxyAuthKey() );

    cpLog(LOG_DEBUG, "JSOpCallerReady: send INVITE(HOLD):");
    cpLog(LOG_DEBUG, "***************************************");
    cpLog(LOG_DEBUG, "%s", invite->encode().logData());
    cpLog(LOG_DEBUG, "***************************************");
    sipEvent->getSipStack()->sendAsync( *invite );

    // Change state to transfer caller
    Sptr < JSStateMachine > stateMachine;
    stateMachine.dynamicCast( event->getCallInfo()->getFeature() );
    assert( stateMachine != 0 );

    return stateMachine->findState( "JSStateTransferCaller" );
}
