
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const JSOpRedirect_cxx_Version =
    "$Id$";
#include "JSOpRedirect.hxx"
#include "SipEvent.hxx"
#include "JSStateMachine.hxx"
#include "JSCall.hxx"
#include "StatusMsg.hxx"
#include "InviteMsg.hxx"
#include "AckMsg.hxx"
#include "SipRequestLine.hxx"
#include "SipContact.hxx"
#include "JSCommon.hxx"


///
JSOpRedirect::JSOpRedirect()
{}


///

JSOpRedirect::~JSOpRedirect()
{}


///

const char* const
JSOpRedirect::name() const
{
    return "JSOpRedirect";
}


///
const Sptr < State >
JSOpRedirect::process( const Sptr < SipProxyEvent > event )
{
    // 1. Receive 302 from RS
    // 2. Send ACK to RS
    // 3. Send INVITE to next hop

    cpLog( LOG_DEBUG, "JSOpRedirect operation" );

    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );
    if ( sipEvent == 0 )
    {
        return 0;
    }
    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
    assert( sipMsg != 0 );

    Sptr < StatusMsg > statusMsg;
    statusMsg.dynamicCast( sipMsg );
    if ( statusMsg == 0 )
    {
        return 0;
    }

    if ( statusMsg->getStatusLine().getStatusCode() >= 400 )
    {
        return 0;
    }

    if ( statusMsg->getStatusLine().getStatusCode() != 302 )
    {
        return 0;
    }

    cpLog( LOG_DEBUG, "Receive 302 message" );

    Sptr < JSCall > call;
    call.dynamicCast( event->getCallInfo() );
    assert( call != 0 );

    // Check branch number make sure it is from RS
    // 302 from contact will be hanled in other operator
    SipVia via = statusMsg->getVia( 0 );
    Data branch = via.getBranch();
    int branchNum( branch.convertInt() );

    // All 302 from RS has branch numbers in the range of 200
    // Please see JSCommon.hxx
    if ( (branchNum < 200) || (branchNum > 300) )
        return 0;

    if ( branchNum == REDIRECT_CTRL )
        branch = CONTACT_CTRL;
    else if ( branchNum == REDIRECT_JSN )
        branch = CONTACT_JSN;
    else if ( branchNum == REDIRECT_JSD )
        branch = CONTACT_JSD;
    else
    {
        // impossible to anyother branch numbers, terminate this call
        cpLog( LOG_ERR, "unknown branch number: %s", branch.logData() );
        // Change state to unknow which will get rid of the call object
        Sptr < JSStateMachine > stateMachine;
        stateMachine.dynamicCast( event->getCallInfo()->getFeature() );
        assert( stateMachine != 0 );

        return stateMachine->findState( "JSStateUnknown" );
    }

    // Send ACK to RS for the 302 received
    Sptr < InviteMsg > invite = call->getRSInvite();
    assert( invite != 0 );

    Sptr <BaseUrl> baseUrl;
    baseUrl = invite->getRequestLine().getUrl();
    assert( baseUrl.getPtr() != 0 );
    assert( baseUrl->getType() == SIP_URL ); // JTAPI doesn't support TEL_URL

    AckMsg ack( *statusMsg );
    SipRequestLine ackRequestLine( SIP_ACK, baseUrl );

    ack.setRequestLine( ackRequestLine );
    cpLog(LOG_DEBUG, "ACK send to 302:");
    cpLog(LOG_DEBUG, "***************************************");
    cpLog(LOG_DEBUG, "%s", ack.encode().logData());
    cpLog(LOG_DEBUG, "***************************************");
    sipEvent->getSipStack()->sendAsync( ack );

    // Forward INVITE to next hop
    Sptr<SipContact> contact = 0;
    SipContactList contactList = statusMsg->getContactList();

    // check for empty contact list
    if ( contactList.begin() == contactList.end() )
    {
        cpLog( LOG_ERR, "Empty Contact list in 302 STATUS" );
        // Change state to unknow which will get rid of the call object
        Sptr < JSStateMachine > stateMachine;
        stateMachine.dynamicCast( event->getCallInfo()->getFeature() );
        assert( stateMachine != 0 );

        return stateMachine->findState( "JSStateUnknown" );
    }

    SipContactList::iterator iter;
    iter = contactList.begin();
    contact = *iter;

    // Set the branch
    SipVia newVia = invite->getVia( 0 );
    newVia.setBranch( branch );
    invite->removeVia( 0 );
    invite->setVia( newVia, 0 );

    // send Invite to Contact
    cpLog( LOG_DEBUG, "send Invite to Contact ..." );

    Sptr <BaseUrl> contactBaseUrl;
    contactBaseUrl = contact->getUrl();
    assert( contactBaseUrl.getPtr() != 0 );
    assert( contactBaseUrl->getType() == SIP_URL ); // JTAPI doesn't support TEL_URL
    SipRequestLine inviteRequestLine( SIP_INVITE, contactBaseUrl );
    invite->setRequestLine( inviteRequestLine );
    cpLog(LOG_DEBUG, "JSOpRedirect: INVITE send to next contact:");
    cpLog(LOG_DEBUG, "***************************************");
    cpLog(LOG_DEBUG, "%s", invite->encode().logData());
    cpLog(LOG_DEBUG, "***************************************");

    sipEvent->getSipStack()->sendAsync( *invite );

    return 0;
}
