/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const Marshal_cxx_Version =
    "$Id: Marshal.cxx,v 1.76.2.3 2003/01/23 03:00:02 sprajpat Exp $";

#include "VocalCommon.hxx"

#include <iostream.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>


#include "RoundRobinServers.hxx"
#include "Marshal.hxx"
#include "VCdrException.hxx"
#include "ProvisionInterface.hxx"
#include "InviteMsg.hxx"
#include "RegisterMsg.hxx"
#include "SystemInfo.hxx"
#include "SipAuthorization.hxx"
#include "SipProxyAuthorization.hxx"
#include "SipMaxForwards.hxx"
#include "NetworkConfig.hxx"

#if 0 //Enable this to get feature enabled Marshal
#include "BasicProxy.hxx"
#include "MrshlBuilder.hxx"
#else
#include "StatelessBasicProxy.hxx"
#endif

#include "CdrInterface.hxx"
#include "Daemon.hxx"
#include "VMrshlException.hxx"

#include "ByeMsg.hxx"
#include "MrshlUserCache.hxx"
#include "MrshlOpAck.hxx"
#include "MrshlOpBye.hxx"
#include "MrshlOpInvite.hxx"
#include "MrshlOpOptions.hxx"
#include "MrshlOpStatus.hxx"
#include "MrshlOpCancel.hxx"
#include "MrshlOpProxy.hxx"
#include "MrshlOpRegister.hxx"
#include "MrshlOpSubNotify.hxx"
#include "VPs_def.h"

//#include "BaseAuthentication.hxx"

using namespace Vocal;

#define INLINE_ inline

// static data declarations
Marshal* Marshal::myInstance = 0;
string Marshal::myGroupType = "";

/** Provisioning constants **/

const string PS_STRING_GATEWAY( "Gateway" );
const string PS_STRING_USERAGENT( "UserAgent" );
const string PS_STRING_INTERNETWORK( "Internetwork" );
const string PS_STRING_CONFERENCE( "ConferenceBridge" );
const string PS_STRING_TRUE( "true" );
const char PS_STRING_AUTH_ACCESS_LIST[] =  "Access list";
const char PS_STRING_AUTH_DIGEST[] = "Digest";
const char PS_STRING_AUTH_BASIC[] = "Basic";
const char PS_STRING_AUTH_NONE[] = "None";


Marshal&
Marshal::instance( const int sipPort ) throw (VMrshlException&)
{
    if ( myInstance == 0 )
    {
        if ( sipPort < 0 )
        {
            throw VMrshlException("SIP port number has not been defined",
                                  __FILE__, __LINE__);
        }

        myInstance = new Marshal( sipPort );

        myInstance->loadPsData();

        myInstance->doRegistrationForUpdate();

        // create the RoundRobinServers
        RoundRobinServers::instance( SERVER_RS );

        // create the user cache
        MrshlUserCache::instance(myGroupType.c_str());
    }
    
    return *myInstance;
}


Marshal::Marshal( const int sipPort ) :
        mySipPort(sipPort),
	myVia()
{
    // set myVia
    myVia.setHost( theSystem.gethostAddress() );
    myVia.setPort( mySipPort );
    myVia.setprotoVersion( DEFAULT_VERSION );

  try
  {
    list<string> marshalServers;
    ProvisionInterface::instance().getMarshalServers(marshalServers);
    list<string> redirectServers;
    ProvisionInterface::instance().getRedirectServers(redirectServers);
    list<string> fServers;
    ProvisionInterface::instance().getFeatureServers(fServers);

    for(list<string>::iterator itr = marshalServers.begin();
                  itr != marshalServers.end(); itr++)
    {
        myListOfServers.insert(*itr);
    }

    for(list<string>::iterator itr = redirectServers.begin();
                  itr != redirectServers.end(); itr++)
    {
        myListOfServers.insert(*itr);
    }

    for(list<string>::iterator itr = fServers.begin();
                  itr != fServers.end(); itr++)
    {
        myListOfServers.insert(*itr);
    }
    myCdrPort = 0;
    myCdr2Port = 0;
  }
  catch(VException& e)
  {
       cpLog(LOG_DEBUG, 
	     "Failed to read data from provisioning, reason %s", 
	     e.getDescription().c_str());
  }
}


void
Marshal::destroy()
{
    if ( myInstance != 0 )
    {
        delete myInstance;
        myInstance = 0;
    }
}
    

void
Marshal::run()
{
    // get billing server connection
    try
    {
        CdrInterface::instance( myCdrServer.c_str(), myCdrPort,
                                myCdr2Server.c_str(), myCdr2Port);
        cpLog( LOG_DEBUG, "Marshal is connected to CDR Server" );
    }
    catch ( VCdrException &e )
    {
        // if there is no connection, continue anyways
        // and try to reconnect later
        cpLog( LOG_ALERT, e.getDescription().c_str() );
    }

    if ( SipParserMode::sipParserMode() )
    {
        cpLog( LOG_DEBUG, "SIP_PARSER_STRICT_MODE is turned on" );
    }
    else
    {
        cpLog( LOG_DEBUG, "SIP_PARSER_STRICT_MODE is turned off" );
    }

    cpLog( LOG_DEBUG, "Marshal starting on port %d", mySipPort );
    
    bool filter = true;  // turn on duplicate message filter in sip stack
    bool nat = true;    // turn on network address trans functionality

#if 0  //Enable this to get feature enabled Marshal
    Sptr < MrshlBuilder > builder = new MrshlBuilder;
    BasicProxy theMarshal( builder, mySipPort,
                           "Marshal", filter, nat );
#else

    StatelessBasicProxy theMarshal(mySipPort,
                           "Marshal", filter, nat , APP_CONTEXT_PROXY);
    Sptr < MrshlOpAck > opAck           = new MrshlOpAck;
    theMarshal.addOperator( opAck );
    Sptr < MrshlOpBye > opBye           = new MrshlOpBye;
    theMarshal.addOperator( opBye );
    Sptr < MrshlOpCancel > opCancel     = new MrshlOpCancel;
    theMarshal.addOperator( opCancel );
    Sptr < MrshlOpInvite > opInvite     = new MrshlOpInvite;
    theMarshal.addOperator( opInvite );
    Sptr < MrshlOpSubNotify > opSubNotify     = new MrshlOpSubNotify;
    theMarshal.addOperator( opSubNotify );
    Sptr < MrshlOpOptions > opOptions     = new MrshlOpOptions;
    theMarshal.addOperator( opOptions );
    Sptr < MrshlOpRegister > opRegister = new MrshlOpRegister;
    theMarshal.addOperator( opRegister );
    Sptr < MrshlOpStatus > opStatus     = new MrshlOpStatus;
    theMarshal.addOperator( opStatus );
    Sptr < MrshlOpProxy > opProxy       = new MrshlOpProxy;
    theMarshal.addOperator( opProxy );    
#endif

    theMarshal.run();
    theMarshal.runHeartbeatThread();
    theMarshal.join();
    theMarshal.joinHeartbeatThread();
}


void
Marshal::showMrshlConfig() const
{
    cout << "\n-------------------------------------------------" << endl;
    cout << "Mrshl Configuration Data"  << endl;
    cout << "SIP Port#:               " << mySipPort << endl;

    (RoundRobinServers::instance( SERVER_RS )).displayServerList();

    cout << "Proxy Authorization Key: "
         << myProxyAuthorizationKey.c_str() << endl;

    cout << "Marshal Type:            " << myServerTypeStr << endl;

    {
	Lock lockable (myGatewayMutex);
	cout << "Gateway Hostname:        " << myGatewayHost << endl;
	cout << "Gateway Port:            " << myGatewayPort << endl;
    }

    if ( myServerType == MRSHL_TYPE_INTERNETWORK )
    {
        cout << "Policy Checking:         On" << endl;
    }
    else
    {
        cout << "Policy Checking:         Off" << endl;
    }

    if ( !myRequireCdr )
    {
        cout << "Require CDR Server:      false" << endl;
    }
    else
    {
        cout << "Require CDR Server:      true" << endl;
    }

    if ( myServerType == MRSHL_TYPE_CONFERENCE )
    {
        cout << "Conference number:       " 
             << myBridgeNumber.c_str() << endl;
    }

    cout << "-------------------------------------------------" << endl;

} // end showMrshlConfig()

void
Marshal::updateRegister( const string& data,
			const string& fileName,
			const bool deletedFlag )
{
    cpLog(LOG_DEBUG, "Update register");
    if(fileName == "REGISTERITEM")
    {
        if(myInstance)
            myInstance->doRegistrationForUpdate();
    }

}


void
Marshal::updateMarshal( const string& data,
			const string& fileName,
			const bool deletedFlag )
{
    cpLog(LOG_DEBUG, "got update notice");
    cpLog(LOG_DEBUG, "data: '%s' filename: '%s', deleted: %d", 
	  data.c_str(), 
	  fileName.c_str(),
	  deletedFlag);

    Marshal& instance = Marshal::instance();


    Lock lockable(instance.myGatewayMutex);

    VMarshalServerData marshalData;
    ProvisionInterface& pif = ProvisionInterface::instance();
    char buf[1024];

    if(NetworkConfig::instance().isDualStack())
    {
       //Use IPv4 address as the key
       char hAddress[256];
       theSystem.gethostAddress(hAddress, 256, PF_INET);
       sprintf( buf, "%s:%d", hAddress, instance.mySipPort );
    }
    else
    {
       NetworkAddress na;
       sprintf( buf, "%s:%d", na.getIpName().c_str(), instance.mySipPort );
    }


    try
    {
        pif.getMarshalData( buf, marshalData );
    }
    catch (VBadDataException &e)
    {
        cpLog(LOG_ERR, "*** Failed to get Marshal Data from Provisioning ***");
        cpLog(LOG_ERR, "***                  cannot update               ***");
    }

    if ( instance.myServerType != MRSHL_TYPE_USERAGENT )
    {
	instance.myGatewayHost = marshalData.getGatewayHost();
	instance.myGatewayPort = marshalData.getGatewayPort();
    }
}


void
Marshal::loadPsData() throw ( VMrshlException&, VBadDataException& )
{
    ProvisionInterface& pif = ProvisionInterface::instance();

    VMarshalServerData marshalData;
    string ipName;

    // Get self's ip address for config file name
    if(NetworkConfig::instance().isDualStack() ||
       NetworkConfig::instance().getAddrFamily() == AF_INET)
    {
        //Use IPv4 name instead
       char hAddress[256];
       theSystem.gethostAddress(hAddress, 256, PF_INET);
       ipName = hAddress;
    }
    else
    {
        NetworkAddress na;
        ipName = na.getIpName().c_str();
    }
        
    if ( ipName.length() )
    {
        char buf[1024];

        sprintf( buf, "%s:%d", ipName.c_str(), mySipPort );
        ipName = buf;
        cpLog(LOG_DEBUG, "Marshal hostname:  *** %s ***", ipName.c_str() );
    }
    else
    {
        throw VMrshlException(
	    "*** Failed to get ip address for config file name ***",
	    __FILE__, __LINE__);
    }

    // get marshal data from provisioning
    try
    {
        pif.getMarshalData( ipName.c_str(), marshalData );
    }
    catch (VBadDataException &e)
    {
        cpLog(LOG_ERR, "*** Failed to get Marshal Data from Provisioning ***");
        cpLog(LOG_ERR, "***             Marshal cannot continue          ***");
        throw;
    }

    myPolicy.initialize( marshalData.getNoResponse() );

    // Get IP and Port for both primary and secondary CDR servers
    try
    {
        list < string > cdrHosts;
        VCdrServerData cdrDat;

        pif.getCdrServers(cdrHosts);
        list < string > ::iterator cdrItr = cdrHosts.begin();

        if ( cdrItr != cdrHosts.end() )
        {
            pif.getCdrData( *cdrItr, cdrDat );
            myCdrServer = cdrDat.getHost();
            myCdrPort = cdrDat.getPort();
            cdrItr++;
            
            if ( cdrItr != cdrHosts.end() )
            {
                pif.getCdrData( *cdrItr, cdrDat );
                myCdr2Server = cdrDat.getHost();
                myCdr2Port = cdrDat.getPort();
            }
        }
    }
    catch ( VException &e )
    {
        cpLog( LOG_ERR, 
	       "Error occured while getting CDR config data from provisioning"
	    );
    }

    // marshal types as defined in Provisioning, must match what is in
    // DataDefaults.java
    myServerTypeStr = marshalData.getType();
    myGroupType = marshalData.getGroup();


    if ( myServerTypeStr == PS_STRING_GATEWAY )
    {
        myServerType = MRSHL_TYPE_GATEWAY;
    }
    else if ( myServerTypeStr == PS_STRING_USERAGENT )
    {
        myServerType = MRSHL_TYPE_USERAGENT;
    }
    else if ( myServerTypeStr == PS_STRING_INTERNETWORK )
    {
        myServerType = MRSHL_TYPE_INTERNETWORK;
    }
    else if ( myServerTypeStr == PS_STRING_CONFERENCE )
    {
        myServerType = MRSHL_TYPE_CONFERENCE;
        myBridgeNumber = marshalData.getBridgeNumber();
        marshalData.getNumberList( myAccessNumbers );
    }
    else
    {
        throw VMissingDataException ("Marshal type not specified.  Exiting...",
                                     __FILE__, __LINE__, 0 );
    }

    {
	Lock lockable(myGatewayMutex);
	if ( myServerType != MRSHL_TYPE_USERAGENT )
	{
	    myGatewayHost = marshalData.getGatewayHost();
	    myGatewayPort = marshalData.getGatewayPort();
	}
	else
	{
	    myGatewayHost = "n/a";
	    myGatewayPort = 0;
	}
    }

    string cdrNotReqStr = marshalData.getAllowUnbillable();

    if ( cdrNotReqStr == PS_STRING_TRUE )
    {
        myRequireCdr = false;
    }
    else
    {
        myRequireCdr = true;
    }

    // get proxy authorization key from global provisioning data
    VGlobalConfigData globalConfigData;
    try
    {
        pif.getConfigData( globalConfigData );
    }
    catch ( VException &e )
    {
        throw VMrshlException("Failed to get global data from provisioning",
                              __FILE__, __LINE__);
    }

    myProxyAuthorizationKey = globalConfigData.getproxyAuthKey();

    //The transport wshould also come from provisioning, for now hard-coding
    //    myTransport = "TCP";
    myTransport = "UDP";
}


INLINE_ bool
Marshal::inViaList( const Sptr < SipMsg > sipMsg , int limit)
{
     return( sipMsg->isLoopInViaList(theSystem.gethostAddress(), mySipPort, limit));
}


INLINE_ bool
Marshal::checkGatewayIp( const Data &packetIp )
{
    Lock lockable(myGatewayMutex);

    if ( packetIp == myGatewayHost )
    {
        cpLog( LOG_DEBUG, "incoming packet IP check successful" );
        return true;
    }
    else
    {
	cpLog( LOG_DEBUG, "incoming packet IP check failed" );
	cpLog( LOG_DEBUG, "source ip = %s, gateway ip = %s",
	       packetIp.logData(), myGatewayHost.logData() );
	return false;
    }
}


Data
Marshal::genMrshlNonce( const Data& user ) const
{
    timeval nonceTime;
    Data userNonce;
    char tempbuf[1024];

    gettimeofday( &nonceTime, NULL);
    long unsigned nonce = nonceTime.tv_sec ^ nonceTime.tv_usec;

    sprintf(tempbuf, "%lu", nonce );
    userNonce = tempbuf;

    return userNonce;
}


INLINE_ bool
Marshal::AccessListCheck( const Sptr < SipCommand > sipCommand )
{
    bool retCode = false;
    string receivedIp;
    string domainName;
    string userIp;
    string user;
    MrshlUserData tempUserData;

    cpLog( LOG_DEBUG, "***** Incoming Access List Check... *****" );

    // Get the user name to be checked in provisioning
    LocalScopeAllocator lo;
    user = sipCommand->getFrom().getUser().getData(lo);
    cpLog( LOG_DEBUG, "User: %s", user.c_str() );

    try
    {
        ( MrshlUserCache::instance() ).getMrshlUserData( user, tempUserData );
    }
    catch ( VException &e)
    {
         
        return false;
    }


    receivedIp = sipCommand->getReceivedIPName().getData(lo);
    userIp = tempUserData.ip.c_str();
    
    cpLog( LOG_DEBUG, "Received IP:                    %s",
           receivedIp.c_str() );

    // Match the IP from provisioning vs what we got from the packet
    if ( userIp == receivedIp )
    {
        cpLog( LOG_DEBUG, "* user access check successful *");
        retCode = true;
    }
    else
    {
        cpLog( LOG_ALERT, "* access check failed (user, packet):  %s, %s *",
               userIp.c_str(), receivedIp.c_str() );
        retCode = false;
    }

    cpLog(LOG_DEBUG, "***** End Access Check *****");
    return retCode;
}


INLINE_ bool
Marshal::BasicCheck( const Sptr < SipCommand > sipCommand )
{
    MrshlUserData tempUserData;
    string user;

    // Do this only if it's a register msg.
    // NOTE:  Will need to change later to check all msg.

    Sptr < RegisterMsg > registerMsg;
    registerMsg.dynamicCast( sipCommand );
    if (registerMsg == 0)
    {
        // Not register msg.  Don't check... just let it pass for now.
        cpLog(LOG_DEBUG, "Not Register Msg.  Basic Auth not performed");
        return true;
    }
    // Get user
    LocalScopeAllocator lo;
    user = sipCommand->getFrom().getUser().getData(lo);

    // Get user data
    MrshlUserCache & userCache = MrshlUserCache::instance();

    try
    {
        userCache.getMrshlUserData( user, tempUserData );

        // Check the basic.
        cpLog( LOG_DEBUG, "Calling checkAuthBasic with:" );
        cpLog( LOG_DEBUG, "    user    = %s", user.c_str() );
        cpLog( LOG_DEBUG, "    passwd  = %s", tempUserData.password.c_str() );
        if (sipCommand->checkAuthBasic(user, tempUserData.password) )
        {
            // Basic check passed.
            cpLog( LOG_DEBUG, "*** checkAuthBasic passed! ***" );
            return true;
        }
        else
        {
            // Basic check failed.
            cpLog( LOG_DEBUG, "*** checkAuthBasic failed!  ***" );
            return false;
        }
    }
    catch ( VException &e)
    {
        cpLog( LOG_DEBUG, "*** Basic check Failed!  No User Data ***" );
        return false;
    }
}


INLINE_ bool
Marshal::DigestCheck( const Sptr < SipCommand > sipCommand )
{
    // Do this only if it's a register or invite msg.
    // NOTE:  Will need to change later to check all msg.
    Sptr < RegisterMsg > registerMsg;
    Sptr < InviteMsg > inviteMsg;

    registerMsg.dynamicCast( sipCommand );
    inviteMsg.dynamicCast (sipCommand );
    if (registerMsg == 0 && inviteMsg == 0)
    {
        // Not register or invite msg.  Don't check... 
        cpLog(LOG_DEBUG, "Not Register or Invite.  Digest Auth not performed");
        return true;
    }

    // Set the method
    string method;

    SipAuthorization myAuth;
    SipProxyAuthorization myProxyAuth;
    Data realm;
    Data user;
    Data reqURI;
    bool authReg;
    LocalScopeAllocator lo;

    // for user, and request URI, and realm first get it from Authorization 
    //   field for the register or Proxy Authorization for the invite.
    // if there is nothing there, then take the from field's usr, and the
    //   request lines URI.
	
	if (registerMsg == 0){
	    method      = INVITE_METHOD.getData(lo);
	    myProxyAuth = sipCommand->getProxyAuthorization();
	    realm       = myProxyAuth.getTokenValue( REALM );
	    user        = myProxyAuth.getTokenValue( USERNAME );
	    reqURI      = myProxyAuth.getTokenValue( URI );
	    authReg     = false;
	}else{
	    method    = REGISTER_METHOD.getData(lo);
	    myAuth    = sipCommand->getAuthorization();
	    realm     = myAuth.getTokenValue( REALM );
	    user      = myAuth.getTokenValue( USERNAME );
	    reqURI    = myAuth.getTokenValue( URI );
	    authReg     = true;
	    
	}

    cpLog( LOG_DEBUG, "From the sender" );
    cpLog( LOG_DEBUG, "  user     = %s", user.logData() );
    cpLog( LOG_DEBUG, "  uri      = %s", reqURI.logData());
    cpLog( LOG_DEBUG, "  realm    = %s", realm.logData());

    // Get user data
    MrshlUserCache& userCache = MrshlUserCache::instance();

    if (user.length() == 0)
    {
        user = sipCommand->getFrom().getUser().getData(lo);
    }

    //get the request URI.

    if (reqURI.length() == 0)
    {
        reqURI = sipCommand->getRequestLine().encode();
    }

    // Get user data from the cache (or ps) and perform digest check.

    MrshlUserData userData;

    try
    {
        userCache.getMrshlUserData( user.getData(lo), userData );
       
        // Check the digest
        if ( sipCommand->checkAuthDigest(
                    userData.nonce,
                    user,
                    userData.password,
                    realm,
                    reqURI,
                    authReg  ) )
        {
            // Digest check passed.
            cpLog(LOG_DEBUG, "*** checkAuthDigest passed for user: <%s> ***",
                  user.logData() );

            return true;
        }
        else
        {
            // Digest check failed.
            cpLog( LOG_DEBUG, "*** checkAuthDigest failed for user: <%s>  ***",
                   user.logData() );
	    throw VMrshlException("","*** checkAuthDigest failed  ***"
                                  __FILE__, __LINE__);
            return false;
        }
    }
    catch ( VException &e)
    {
        cpLog( LOG_ERR, "*** Digest check Failed!  No User Data ***" );
	if(realm.length()){
	throw VMrshlException("","*** checkAuthDigest failed  ***"
                                  __FILE__, __LINE__);
	}
        return false;
    }
}


INLINE_ bool
Marshal::UserAgentCheck( const Sptr < SipCommand > sipCommand )
{
    Data authType;
    MrshlUserData userData;
    bool retCheck = false;

    Sptr < ByeMsg > byeMsg;
    byeMsg.dynamicCast( sipCommand );

    if ( byeMsg != 0 )
    {
        //according to the RFC, anyone can send a BYE for a call
        cpLog(LOG_DEBUG, "* Bye Message...  let it pass *");
        return true;
    }

   if ( (sipCommand->getCSeq().getMethod() == SUBSCRIBE_METHOD)|| (sipCommand->getCSeq().getMethod() == NOTIFY_METHOD))

   {
      LocalScopeAllocator lo;
        cpLog(LOG_DEBUG, "Checking User: %s",
                  sipCommand->getFrom().getUser().logData() );

        if( sipCommand->getFrom().getUser().length()== 0)
        {
           Data tmpfromuser = sipCommand->getFrom().getDisplayName();
           if(tmpfromuser.length())
           {
              SipFrom subfrom = sipCommand->getFrom();
              subfrom.setUser(tmpfromuser);
              sipCommand->setFrom(subfrom);
           }
        }
   }


  // Find user from MrshlUserCache and get authentication type.

    MrshlUserCache & userCache = MrshlUserCache::instance();

    //try
    //Let it throw the exception be caught at higherup so that 
    //we know that user is not provisioned
    {
        LocalScopeAllocator lo;
        cpLog(LOG_DEBUG, "Checking User: %s", 
                  sipCommand->getFrom().getUser().logData() );

        userCache.getMrshlUserData( 
                   sipCommand->getFrom().getUser().getData(lo),
                   userData );

        authType = userData.authType;
    }

    //catch (VException &e)
    //{
    //    // User doesn't exist.
    //    cpLog( LOG_ALERT, "*** No auth check.  User not found: %s ***, reason %s",
    //           sipCommand->getFrom().getUser().logData(), e.getDescription().c_str() );
    //    return false;
    //}

    if (authType == PS_STRING_AUTH_ACCESS_LIST)
    {
        retCheck = AccessListCheck( sipCommand );
    }
    else if (authType == PS_STRING_AUTH_DIGEST)
    {
        retCheck = DigestCheck( sipCommand );
    }
    else if (authType == PS_STRING_AUTH_BASIC)
    {
        retCheck = BasicCheck( sipCommand );
    }
    else if (authType == "None")
    {
        // No auth check.
        retCheck = true;
    }
    else
    {
        // No auth type.  Fail check.
	cpLog( LOG_ERR, "Invalid authentication scheme: %s",
               authType.logData() );
        retCheck = false;
    }

    return retCheck;
}


INLINE_ bool
Marshal::incomingCheck( const Sptr < SipCommand > sipCommand )
{
    bool retCheck = false;

    // perform authentication check based on marshal server type.

    if ( myServerType == MRSHL_TYPE_USERAGENT )
    {
        retCheck = UserAgentCheck( sipCommand );
    }

    // enable for conference support
#if 0
    else if ( myServerType == MRSHL_TYPE_CONFERENCE )
    {
        retCheck = AccessListCheck( sipCommand );
    }
#endif

    else if (myServerType == MRSHL_TYPE_GATEWAY)
    {
        // Check gateway ip only.
	retCheck = checkGatewayIp( sipCommand->getReceivedIPName() );
    }
    else
    {
#if 0
        // Check gateway ip only.
        retCheck = checkGatewayIp( sipCommand->getReceivedIPName() );
#endif
	retCheck = true;
    }

    return retCheck;
}


unsigned int
Marshal::marshalIncomingCommand( Sptr < SipCommand > sipCommand )
{
    // check for key to determine if incoming or outgoing
    if ( hasMarshalKey( sipCommand ) )
    {
        // Outgoing command, remove ProxyAuthorization key
        cpLog( LOG_DEBUG, "Outgoing, remove proxy authorization key" );
        SipProxyAuthorization proxyAuthorizationVoid;
        sipCommand->setProxyAuthorization( proxyAuthorizationVoid );

        // check Via for loops; if already present more than once, discard;
        // note, being in via once must be allowed in order to allow
        // for calls being received and terminated by the same marshal
        if ( inViaList( sipCommand , 2) )
        {
            // discard this message
            cpLog( LOG_ERR, "Already in Via twice or more, discard" );
            return 482;  // 482 Loop Detected
        }

        // done marshalling
        return MRSHL_STATUS_OUTGOING;

    } // end if outgoing

    else
    {
        // check Max-Forwards
        SipMaxForwards sipMaxForwards = sipCommand->getMaxForwards();
        cpLog( LOG_DEBUG, "Original Max-Forwards: %s",
               sipMaxForwards.encode().logData() );
        int maxForwards = sipMaxForwards.get();
        // if maxForwards strlen = 0, there is nothing to check
        if (maxForwards >= 0)
        {
            // check value encoded in Max-Forwards field
            // decrement Max-Forwards by one
            maxForwards--;
            sipMaxForwards.set( maxForwards );
            cpLog( LOG_DEBUG, "New Max-Forwards: %s",
                   sipMaxForwards.encode().logData() );
            sipCommand->setMaxForwards( sipMaxForwards );
        }
        else
        {
            cpLog( LOG_ERR, "Too many hops, return 483" );
            return 483;  // 483 Too many hops
        }

        // perform appropriate level of authentication;
	// return 401 if it fails
        try
        {
            if ( incomingCheck( sipCommand ) == false )
            {
                // authentication failed.  return error code based on message
		// xxx -- proxies should not ever return a 401 -- they should
		// always return 407
                if ( ( sipCommand->getType() == SIP_INVITE ) && 
		     ( myServerType == MRSHL_TYPE_USERAGENT ) )
                {
                    // this is an invite, send 407 with authentication
		    // headers...
                    return 407;
                }
                else
                {
                    return 401;
                }
            }
        }
	catch(VMrshlException& e){
	    cpLog( LOG_ALERT, "*** Exception [%s]",e.getDescription().c_str());
	    return 403;
	}
        catch(VException& e)
        {
            cpLog( LOG_ALERT, "*** User[%s] not found:  reason is\n [%s]",
                   sipCommand->getFrom().getUser().logData(), e.getDescription().c_str() );
            return 403;
        }

        // check Via for loops; if already present, discard
        if ( inViaList( sipCommand , 1) )
        {
            // discard this message
            cpLog( LOG_ERR, "Already in Via, discard to avoid loops" );
            return 482;  // 482 Loop Detected
        }

	// since it is OK, add proxy authorization key
        cpLog( LOG_DEBUG, "Incoming, add proxy authorization key" );
        sipCommand->setProxyAuth( myProxyAuthorizationKey );

        // done marshalling
        return MRSHL_STATUS_INCOMING;
    }
}


Sptr<StatusMsg>
Marshal::generateChallenge(const Sptr < SipCommand > command, 
			   int statusErr,
			   ChallengeType type /* default value */)
{
    // verify the toUrl
    Sptr < SipUrl > fromSipUrl;
    fromSipUrl.dynamicCast( command->getFrom().getUrl() );
    
    Data user = fromSipUrl->getUserValue();
    
    MrshlUserCache &myUserCache = MrshlUserCache::instance();
    MrshlUserData userData;

    // assert this is 401 or 407
    assert( (statusErr == 401) || (statusErr == 407) );

    Sptr<StatusMsg> statusMsg = new StatusMsg(*command, statusErr);

    //check what kind of authentication we want.
    try
    {
	myUserCache.getMrshlUserData( user.convertString(), userData );
	Data nonce;

	if ( userData.authType == AUTH_BASIC || 
	     userData.authType == AUTH_DIGEST )
	{
	    bool useDigest;

	    if(userData.authType == AUTH_BASIC)
	    {
		cpLog( LOG_DEBUG, " send %d requesting BASIC authentication", 
		       statusErr );
		useDigest = false;
	    }
	    else
	    {
		cpLog( LOG_DEBUG, " send %d requesting DIGEST authentication",
		       statusErr);
		nonce = genMrshlNonce( user.convertString() );
		useDigest = true;
	    }
	    
	    myUserCache.addUserAuthData (user.convertString(),
					 theSystem.gethostAddress(),
					 nonce.convertString());

	    if(type == CHALLENGE_WWW_AUTH)
	    {
		if(useDigest)
		{
		    statusMsg->challengeDigestWwwAuthenticate( 
			theSystem.gethostAddress(),
			nonce);
		}
		else
		{
		    statusMsg->challengeBasicWwwAuthenticate( 
			theSystem.gethostAddress());
		}
	    }
	    else
	    {
		if(useDigest)
		{
		    statusMsg->challengeDigestProxyAuthenticate(
			theSystem.gethostAddress(),
			nonce);
		}
		else
		{
		    statusMsg->challengeBasicProxyAuthenticate(
			theSystem.gethostAddress());
		}
	    }
	}
	else
	{
	    // return 401 without any challenge
	    cpLog( LOG_DEBUG, "send %d, unauthorized without challenge", 
		   statusErr );
	}
    }
    catch ( VException &e )
    {
	cpLog( LOG_ERR, "User does not exist. Send  without challenge" );
    }

    return statusMsg;
}


bool
Marshal::hasMarshalKey( const Sptr < SipCommand > sipCommand ) const
{
    // check is command contains ProxyAuthorization header with key
    if ( sipCommand->checkProxyAuth( myProxyAuthorizationKey ) )
    {
        return true;
    }
    return false;
}
    

void
Marshal::addViaWithBranch( SipCommand& sipCommand,
			   const Data &viaBranch )
{
    //As per RFC 2543 bis4 , section H bullet 4, page 178
    //the transport parameter of the requewst
    //line should be the transport paramter in Via
    Data trpt = sipCommand.getRequestLine().getTransportParam();
    Data tParam("UDP");
    SipVia tVia;
    tVia.setHost( theSystem.gethostAddress() );
    tVia.setPort( mySipPort );
    tVia.setprotoVersion( DEFAULT_VERSION );

    if((trpt == "tcp")  || (trpt == "TCP"))
    {
        tParam = "TCP";
    }
    tVia.setTransport(tParam);
    if ( viaBranch.length() > 0 )
    {
	// set branch with '.<viaBranch>' appended
	tVia.setBranch( sipCommand.computeBranch() + "." + viaBranch );
    }
    else
    {
	// set branch without any applciation specific addition
	tVia.setBranch( sipCommand.computeBranch() );
    }
    sipCommand.setVia( tVia, 0 );

}


void
Marshal::proxyCommandWithoutRoute( const Sptr < SipEvent > sipEvent,
                                   const bool trueCancel )
{
    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
    Sptr < SipCommand > commandMsg;
    commandMsg.dynamicCast( sipMsg );
    Data callId = commandMsg->getCallId().getCallIdData();
    Sptr < SipCommand > commandToCancel;
    commandToCancel = sipEvent->getPendingCommand( commandMsg );
    // check if pending command found for cancel
    if ( commandToCancel == 0 )
    {
        // reply with 481, Transaction does not exist
        cpLog( LOG_ERR, "no matching request for cancel, return 481" );
        StatusMsg statusMsg( *commandMsg, 481 );
        sipEvent->getSipStack()->sendReply( statusMsg );

        return;
    }

    if ( serverType() == MRSHL_TYPE_INTERNETWORK )
    {
        // delete pending policy
        if ( policy().inList( callId ) )
        {
            cpLog( LOG_DEBUG, "Disabling QoS and aborting policy" );

            // disableQoS (if exists)
            policy().disableQoS( callId );

            // abort the policy (if exists)
            policy().policyAbort( callId );
        }
    }
    else if ( serverType() == MRSHL_TYPE_CONFERENCE )
    {
        // check conference map and remove caller from call
        MrshlConfCallMap::iterator iter = myMrshlConfCallMap.find( callId );

        // check that a conference with this callId exists
        if ( iter != myMrshlConfCallMap.end() )
        {
            // remove user from conference list
            MrshlConfMapEntry &confMapEntry = myMrshlConfCallMap[ callId ];
            if( confMapEntry.removeCallFromConf( 
		commandMsg->getFrom().getUser()) )
            {
                unsigned int numParticipants =
		    ( confMapEntry.getNumParticipants() );
                if ( numParticipants == 0 )
                {
                    cpLog( LOG_DEBUG, "Removing conference from conf map" );
                    myMrshlConfCallMap.erase( callId );
                }
            }
            else
            {
                cpLog( LOG_ERR, 
		       "User not found in conference, returning 404" );
		// 404 not found 
                StatusMsg errorStatusMsg ( *commandMsg, 404 );
                sipEvent->getSipStack()->sendReply( errorStatusMsg );
                return;
            }
        }
        else
        {
            cpLog( LOG_ERR, 
		   "CallId not found in conference map, returning 404" );
            StatusMsg errorStatusMsg ( *commandMsg, 404 ); // 404 not found 
            sipEvent->getSipStack()->sendReply( errorStatusMsg );
            return;
        }
    }

    // change RequestUri to that of pending command
    if( trueCancel )
    {
        // reply with 200, and forward Cancel to receiver of commandToCancel
        StatusMsg statusMsg( *commandMsg, 200 );
        sipEvent->getSipStack()->sendReply( statusMsg );
        SipRequestLine requestLine( SIP_CANCEL,
                                    commandToCancel->getRequestLine().getUrl()
	    );

        commandMsg->setRequestLine( requestLine );
    }
    else
    {
        SipRequestLine requestLine( SIP_BYE,
                                    commandToCancel->getRequestLine().getUrl()
	    );

   commandMsg->setRequestLine( requestLine );
}

    // flush Via list and add same via used in pending command
    commandMsg->flushViaList();
    SipVia via( commandToCancel->getVia( 0 ) );
//    commandMsg->setVia( via, 0 );
    commandMsg->copyViaList(*commandToCancel);

    cpLog(LOG_DEBUG, "forwarded command: %s", commandMsg->encode().logData());
    //send to next hop
    commandMsg->setNextHopIsProxy(isNextHopAProxy(*commandMsg));

    NetworkAddress nSentNetworkAddress(commandToCancel->getSendAddress());
    cpLog(LOG_DEBUG_STACK, "SipMsg from CommandToCancel Returned SendAddress ip= %s, port= %d ", 
           nSentNetworkAddress.getHostName().c_str(), nSentNetworkAddress.getPort() ); 
    sipEvent->getSipStack()->sendAsync( commandMsg, nSentNetworkAddress.getIpName(), 
                                        Data(nSentNetworkAddress.getPort()));

    return;
}


bool
Marshal::isNextHopAProxy(const SipCommand& sipMsg)
{
    if(myTransport == "UDP")
    {
        //Use it only when transport is TCP
        return false;
    }
    Sptr<BaseUrl> dest = sipMsg.getRequestLine().getUrl();
    Sptr<SipUrl> sipDest;
    sipDest.dynamicCast(dest);
    assert(sipDest != 0);
    Data nHop(sipDest->getHost());
    nHop += ":";
    nHop += sipDest->getPort();
    LocalScopeAllocator lo;
    ProxyServerList::iterator itr = myListOfServers.find(string(nHop.getData(lo)));
    if(itr != myListOfServers.end())
    {
        //Next hop is one of the provisioned servers
        return true;
    }
    return false;
}

void 
Marshal::doRegistrationForUpdate()
{
    cpLog(LOG_INFO, "Marshal::doRegistrationForUpdate");
    // install a watch for changes in gateway
    try
    {
        ProvisionInterface& pif = ProvisionInterface::instance();
        if ( myInstance->myServerType != MRSHL_TYPE_USERAGENT )
        {
            cpLog(LOG_DEBUG, "registering for gateway change update");

            pif.registerDirForUpdate( 
		Marshal::updateMarshal,
		"Marshals"
		);
        }

        pif.registerForUpdate( Marshal::updateRegister, "REGISTERITEM", ProvisionInterface::VCONFIG);

        pif.registerDirForUpdate( MrshlUserCache::updateUsers,
                                  VACCOUNTS );
        pif.registerDirForUpdate( MrshlUserCache::updateUsers,
                                  VALIASES );
 
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server , reason %s", e.getDescription().c_str());
    }
}

// End of File
