#ifndef _MRSHL_USER_CACHE_HXX_
#define _MRSHL_USER_CACHE_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const MrshlUserCache_hxx_Version =
    "$Id: MrshlUserCache.hxx,v 1.27.4.1 2003/03/01 00:42:17 sprajpat Exp $";

//#include <string>
#include "Data.hxx"
#include "VMissingDataException.hxx"
#include "VRWLock.hxx"

#if defined(__linux__) && !defined (__INTEL_COMPILER)
#define MRSHL_USE_HASHMAP 1
#else
#define MRSHL_USE_HASHMAP 0
#endif 



namespace Vocal
{

/**
 *  Structure for storing the user data.
 *
 */
typedef struct
{
    Data authType;     // auth method for this user
    Data ip;           // ip that the user allowed to coming from
    Data password;     // password for digest auth
    Data realm;        // string for the user to recognize which identity to use
    Data nonce;        // randomly generated number (see Marshal::genMrshlNonce)
    unsigned int index;  // This is used for determining clean up. 
    Data transport;    //Transport TCP/UDP of the end-point
    bool inMarshalGroup;

}  MrshlUserData;
 
}

#if ( MRSHL_USE_HASHMAP == 1 )

#include "StringHash.hxx"
#if defined(__GNUC__) && (__GNUC__ < 3)
#include <hash_map>
#else
#include <ext/hash_map>
#endif

typedef hash_map < Data, Vocal::MrshlUserData, StringHash > MrshlUserMap;

#else

#include <map>
typedef map < Data, Vocal::MrshlUserData > MrshlUserMap;

#endif


namespace Vocal
{

/**  MrshlUserCache is a singleton.  It is used to cache user info so that
 *   it can be referenced without calls to provisioning.  This is needed to
 *   optimize the performance.
 */

class MrshlUserCache
{
    public:
        /**
         */
        ~MrshlUserCache();

        /**
         */
        static MrshlUserCache& instance( const char* msGroup = 0);

        /**  Get user data.  Looks in cache first, provisioning second.  
         *   If the user is not in the cache, then it will add the info from
         *   provisioning.
         *   If the user is not in provisioning, then nothing is added.
         */
        void getMrshlUserData( const Data& user, 
                               MrshlUserData& userData) 
        throw ( VMissingDataException& );

        /**  Adds the info used for authentication (digest or basic) for the 
         *   user. 
         */
        void addUserAuthData(const Data& user,
                             const Data& realm,
                             const Data& nonce )
        throw ( VMissingDataException& );        

        ///Updates the user transport
        void updateUserTransport( const Data& user,
                                 const Data& transport)
                    throw ( VMissingDataException& );


        /**  Callback function used by provisioning to update user info in
         *   the cache when there are changes.
         */
        static void
        updateUsers( const string& data,
                     const string& fileName,
                     const bool deletedFlag );


        /**  Returns and increments the index.  This is used to keep track
         *   the order that the users got into the cache.  If the index
         *   exceeds the cache size, then getIndex() will remove an existing 
         *   user from the cache to create space for the new user.
         */
        int getIndex();

    private:
        /**
         */
        MrshlUserCache( const Data msGroup );


        /**  Finds user data from cache.  throws an exception if user is not 
         *   found.
         */
        bool findUserData( const Data& user, 
                           MrshlUserData& userData ); 

        /**  Look up user in provisioning.  If exist, add user to cache.
         */
        void createUserData( const Data& user,
                             MrshlUserData& userData ) 
        throw ( VMissingDataException& );

        /**  Update user info from provisioning into the cache entry.
         */
        void updateUserData( const Data& user,
                             const Data& userAuthType,
                             const Data& userIp,
                             const Data& userPassword ) 
        throw ( VMissingDataException& );
        
        /**  Lock used for locking the map for read or write.
         */
        static VRWLock cacheLock;

        /**  Removes users from the cache to make room for a new user.  
         *   This function is only called when map size exceeds MS_CACHE_SIZE.
         */
        void cleanCache();

        /**  Removes one particular user.  This function is called when an
         *   update from provisioning notifies the marshal that a user
         *   has been removed.
         */
        void removeUser( const Data& user );

        /**
         */
        static MrshlUserCache* myInstance;

        /**
         */
        Data myMarshalGroup;

        /**  
         */ 
        MrshlUserMap myUserCache;

        /**  This counter is used to determine the order that a user is
         *   inserted into the cache.  It is also used for determining
         *   which user is deleted during cleanup.
         */
        static unsigned int indexCounter;

        /**  This marker is used to keep track of which users to delete
         *   when the cache becomes full.
         */
        static unsigned int indexMarker;

        /**  Used for determining when to clean the cache
         */
        static bool firstPass;

};
 
}

#endif // _MRSHL_USER_CACHE_HXX_
