
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const OpOptions_cxx_Version =
    "$Id: OpOptions.cxx,v 1.7 2002/11/12 20:25:22 veer Exp $";


#include "SipContact.hxx"
#include "OptionsMsg.hxx"
#include "SipEvent.hxx"
#include "StatusMsg.hxx"
#include "OpOptions.hxx"
#include "RSStateWait.hxx"
#include "SipContactIf.hxx"
#include "RSCommon.h"
#include "RSProxy.hxx"
#include "SipVia.hxx"

using namespace Vocal;

/**
 * Constructor
 */
OpOptions::OpOptions()
{
}

/**
 * Destructor
 */
OpOptions::~OpOptions()
{
}

/**
 * Return an identifying name for this object
 * @return  String of characters to identify this object.
 */
const char* const
OpOptions::name() const
{
   return "OpOptions";
}

/**
 * Decode the event, trigger the process of handling the options, and return a
 * status response containing a contact list.
 * @param  event  Sip event containing a sip message which may be an Options.
 * @return  The next state
 */
const Sptr < State >
OpOptions::process( const Sptr < SipProxyEvent > event)
{
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast(event);

    if (sipEvent == 0)
    {
        cpLog(LOG_ERR, "Event is not a sipEvent");
        return PROXY_CONTINUE;
    }

    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();

    if (sipMsg == 0)
    {
        cpLog( LOG_DEBUG, "In OpOptions::process, sipEvent->getSipMsg() returned 0" );
        return PROXY_CONTINUE;
    }

    Sptr < OptionsMsg > sipOptions;
    sipOptions.dynamicCast(sipMsg);

    if ( sipOptions == 0)
    {
        return PROXY_CONTINUE;
    }

    doCpLog();

    // If the rs is not active, return here
    RSSTATE rsState = RSProxy::instance().getRsState();

    Sptr <BaseUrl> baseReqUrl( sipOptions->getRequestLine().getUrl() );
    Data uri;
    Data uriUserName;
    Data uriUserParm;
    if (baseReqUrl.getPtr() != 0) 
    { 
       if (baseReqUrl->getType() == SIP_URL) 
       { 
          Sptr <SipUrl> sipUrl; 
          uriUserName = (sipUrl.dynamicCast(baseReqUrl))->getUserValue(); 
          uriUserParm = (sipUrl.dynamicCast(baseReqUrl))->getUserParam(); 
          uri = (sipUrl.dynamicCast(baseReqUrl))->getNameAddr();
       }
    }
    
    /// Obtain the last two Via entries
    int viaSize = sipOptions->getNumVia();
    Data viaHostPort = sipOptions->getVia(0).getHost();
    viaHostPort += ":";
    viaHostPort += Data(sipOptions->getVia(0).getPort());
    cpLog(LOG_DEBUG, " 1st via's host and port = %s", viaHostPort.logData());

    vector < Data > lastTwoVias;
    lastTwoVias.push_back(viaHostPort);
    if (viaSize > 1)
    {
        viaHostPort = sipOptions->getVia(1).getHost();
        viaHostPort += ":";
        viaHostPort += Data(sipOptions->getVia(1).getPort());
        cpLog(LOG_DEBUG, " 2nd via's host and port = %s", viaHostPort.logData());
        lastTwoVias.push_back(viaHostPort);
    }

    /// Determine if To URL is a phone number or IP address
    bool isPhoneNumber = (uriUserParm == "phone") ? true : false;



    int responseCode = 302;
    SipContactIf sipContactIf;
    vector < Data >  workingContacts;
    vector < Data >  finalContacts;

    if (rsState == RSINACTIVE)
    {
	cpLog(LOG_NOTICE, "returning 503: the RS is in the inactive state");
	responseCode = 503;
    }
    else
    {
       bool uriRegistered = true;
       responseCode = sipContactIf.generateTermOnlyContactList(uri,
                                                               uriRegistered,
                                                               workingContacts);

       if (302 == responseCode)
       {
          vector < Data > ::iterator contactIter = workingContacts.begin();
          Data validContact = "";

          /// Loop through the recently composed contact list in search of
          /// a match on the top VIA.
          if (sipContactIf.locateVia( workingContacts, lastTwoVias, contactIter))
          {
             responseCode = sipContactIf.obtainNextContact(workingContacts, 
                                                           contactIter, 
                                                           isPhoneNumber, 
                                                           uriRegistered, 
                                                           validContact);
             if (validContact != "")
             {
                finalContacts.push_back(validContact);
             }
          } 
          else
          {
             // Unable to determine location in contact list. Fail call.
             responseCode = 404;
          }
       }
    }
    
    cpLog(LOG_DEBUG, " Status Response code: [%d]", responseCode);

    /// Generate the Status Response message
    StatusMsg statusMsg( *sipOptions, responseCode );

    string url;
    LocalScopeAllocator lo;
    for ( vector < Data > ::iterator i = finalContacts.begin(); i != finalContacts.end(); ++i )
    {
       cpLog(LOG_DEBUG, " ------> Validated CONTACT: [%s]", i->logData());
       
       url = i->getData(lo);
       Sptr<SipUrl> sipUrl = new SipUrl(url);
       SipContact sipContact;
       sipContact.setUrl(sipUrl);
       statusMsg.setContact( sipContact );
    }
    
    sipEvent->getSipStack()-> sendReply( statusMsg );
    return PROXY_CONTINUE;
}

// End of File
