#ifndef RSServerContainer_HXX
#define RSServerContainer_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include "RSRedirectServer.hxx"
#include "BaseServerContainer.hxx"
#include "VRWLock.hxx"

static const char* const RSServerContainerVersion = "$Id: RSServerContainer.hxx,v 1.19 2001/02/06 23:39:44 marcl Exp $";

/**
 * The Provisioning Server is notifying us of a change in RS's.
 * @param  data  Unused parameter
 * @param  filename  Unused parameter
 * @param bool  Unused parameter
 */
void updateRSRedirectServers(const string& data, const string& fileName, bool );
/**
 * The Provisioning Server is notifying us of a change in MS's.
 * @param  data  Unused parameter
 * @param  filename  Unused parameter
 * @param bool  Unused parameter
 */
void updateMarshalServers(const string& data, const string& fileName, bool );
/**
 * The Provisioning Server is notifying us of a change in FS's.
 * @param  data  Unused parameter
 * @param  filename  Unused parameter
 * @param bool  Unused parameter
 */
void updateFeatureServers(const string& data, const string& fileName, bool );
/**
 * The Provisioning Server is notifying us of a change in JS's.
 * @param  data  Unused parameter
 * @param  filename  Unused parameter
 * @param bool  Unused parameter
 */
void updateJtapiServers(const string& data, const string& fileName, bool );
/**
 * The Provisioning Server is notifying us of a change in Heartbeat Parms.
 * @param  data  Unused parameter
 * @param  filename  Unused parameter
 * @param bool  Unused parameter
 */
void updateRSHeartbeatParms( const string& data, const string& fileName, bool );

/**
 *  This class defines a singleton container of Servers. 
 *  Servers are stored in individual lists according to type.
 *  Some types of servers are part of groups and some are not. 
 */

class RSServerContainer : public BaseServerContainer
{
    public:
        /**
         * Return the one RS Server container or create it if it doesn't already exist. 
         * @return  Reference to the singleton
         */
        static RSServerContainer& instance();

        /// destructor
        virtual ~RSServerContainer();

        /**
         * Add a server object that has groups associated with the server.   
         * @param  key  Server's address   
         * @param  groups  all the groups this servers belongs to.  
         * @param  serverType  type of server.  
         */
        void addServerWithGroup( const Data& key,
                                 const Sptr < vector < Data > > groups,
                                 const ServerType serverType);

        /**
         * Add a redirect server object.   
         * @param  key  Server's address   
         */
        void addRSRedirectServer( const Data& key);

        /**
         * Get List of Redirect Servers   
         * @param  listofRS  list of all redirect servers in RSServerontainer.  
         */
        void getListofRS( Sptr < vector < Data > > listofRS);

        /**
         * Get List of Active Servers of a certain type  
         * @param  type  the type of server to search for   
         * @param  listofServers  list of all servers in container that matches the type.  
         */
        void getListofActive( const ServerType type, Sptr < vector < Data > > &listofActiveServers);

        /**
         * Determine if specified server is a valid server configured in the system 
         * @param  key  Server's address   
         * @param  type  type of server.  
         * @param  listofGroups  list of groups this server belongs to.  
         */
        bool isValid( const Data& key, ServerType& type, vector < Data > & listofGroups);
        /**
         *  get the sync port for a particular RS
         *  @param  key  Identifier of server
         *  @return  Port number for specified server
         */
        unsigned short getRSSyncPort(const Data& key);

        /**
         * Deal with a heartbeat event from the given server
         * @param  key  Server who sent the heartbeat
         * @return  True if server is recognized, false if not
         */
        bool processHeartbeat( const Data& key);

        /**
         * Check all servers to see if they are still actively heartbeating
         */
        void doHouseKeeping();

        /**
         * A change was made to the RS's in the provisioning server. 
         * Delete the stored RS server list and recreate it with the new data.
         */
        void updateRSsFromPS();

        /**
         * A change was made to the MS's in the provisioning server. 
         * Delete the stored MS server list and recreate it with the new data.
         */
        void updateMSsFromPS();

        /**
         * A change was made to the FS's in the provisioning server. 
         * Delete the stored FS server list and recreate it with the new data.
         */
        void updateFSsFromPS();

        /**
         * A change was made to the JS's in the provisioning server. 
         * Delete the stored JS server list and recreate it with the new data.
         */
        void updateJSsFromPS();

        /**
         * A change was made to the Global Config Parms Data on the provisioning server. 
         * Delete the stored parameter(s) and replace it with new values.
         */
        void updateHBParmsFromPS();

        /**
         * Display a server given the pointer to a server object  
         * @param  serverPtr  pointer to server in server container   
         */
        void displayServer( const Server& server);

        /// Display list of servers by type 
        void displayLists();

    protected:
        /**
         * Constructor
         */
        RSServerContainer();

        /// This is the single instance of RSServerContainer
        static Sptr < RSServerContainer > theRSSrvContainer;

        /// Configure Redirect Servers with Provisioning data.
        void configRedirectServers();

        /// Configure Marshal Servers with Provisioning data.
        void configMarshalServers();

        /// Configure Feature Servers with Provisioning data.
        void configFeatureServers();

        /// Configure Jtapi Servers with Provisioning data.
        void configJtapiServers();

        /**
         * Get Servers of the specified type from Provisioning   
         * @param  type  Type of servers requested
         * @param  retList  List of servers of the requested type
         */
        void getServersFromPS( const ServerType type, vector < Data > & retList);

        /**
         * Remove Duplicates entries in the List of Servers passed in.   
         * @param  list  list of servers to be added to server container.  
         */
        void removeDuplicatesFromList( Sptr < vector < Data > > list);

    private:

        /// Data member list of Redirect Servers
        Sptr < vector < Data > > itsRSList;
        /// Mutex for RS List
        VRWLock rsLock;

        /// Data member list of Marshal Servers
        Sptr < vector < Data > > itsMSList;
        /// Mutex for MS List
        VRWLock msLock;

        /// Data member list of Feature Servers
        Sptr < vector < Data > > itsFSList;
        /// Mutex for FS List
        VRWLock fsLock;

        /// Data member list of Jtapi Servers
        Sptr < vector < Data > > itsJSList;
        /// Mutex for JS List
        VRWLock jsLock;
};

#endif
