
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const RedundancyWorkerThread_cxx_Version =
    "$Id$";


#include "RSCommon.h"
#include "global.h"
#include "RedundancyWorkerThread.hxx"
#include "RSProxy.hxx"
#include "SubscriberContainer.hxx"
#include "SystemInfo.hxx"
#include "Tcp_ClientSocket.hxx"
using namespace Vocal;

#define MIN_NUMSUBSCRIBERS 20
#define MAX_ATTEMP_RETRY   5


/**
 * Constructor  
 * @param  inputFifo 
 * @param  udpStack 
 * @param  sipPort  The port on which this server receives SIP messages.
 * @param  redundancyPort  The port on which this server receives redundancy messages.
 */
RedundancyWorkerThread::RedundancyWorkerThread( const Sptr < Fifo < Sptr < SipProxyEvent > > > inputFifo, 
						Sptr < UdpStack > udpStack, 
						unsigned short sipPort, 
						unsigned short redundancyPort )
        : itsFifo( inputFifo ), 
	  itsUdpStack( udpStack ), 
	  itsSipPort( sipPort ), 
	  itsRedundancyPort( redundancyPort )
{
    assert(itsFifo != 0 );
    itsListOfRSTried = new vector < Data > ;
    itsRetryCounter = 0;
    itsRedundancyState = Unsynchronized ;
}

/**
 * Destructor
 */
RedundancyWorkerThread::~RedundancyWorkerThread()
{
   if(tcpSyncServerSocket != NULL)
   {
      delete tcpSyncServerSocket;
      tcpSyncServerSocket = NULL;
   }
}

/**
 * Process Redundancy related events
 * The two main types are messages and timers.
 */
void
RedundancyWorkerThread::thread()
{
    /// schedule first timer
    Sptr < RSTimerEvent > rsTimerEvent = new RSTimerEvent (itsFifo, StartupDelayTimer);
    if ( rsTimerEvent != 0 )
    {
        /// Put a timed event on the fifo to trigger synchronization after RS has had
        /// a chance to collect heartbeats for the servers configured in this system.
        rsTimerEvent->startTimer( rsTimerEvent, 2000 );   // 2 sec
        rsTimerEvent = 0;  // cleanup rsTimerEvent
        cpLog( LOG_DEBUG, "Collecting heartbeats..... wait 2 seconds" );
    }
    else
    {
        cpLog( LOG_ERR, "RedundancyWorkerThread::RedundancyWorkerThread(), new == 0" );
    }

    while (true)
    {
        assert( itsFifo != 0 );

        Sptr < SipProxyEvent > nextEvent = itsFifo->getNext();
        assert( nextEvent != 0 );

        Sptr < RedundancyEvent > redundancyEvent;
        redundancyEvent.dynamicCast(nextEvent);

	if (redundancyEvent != 0)
	{
	    processRedundancyEvent(redundancyEvent);
	}
	else
	{
	  Sptr < RSTimerEvent > rsTimerEvent;
	  rsTimerEvent.dynamicCast(nextEvent);
	  processTimerEvent(rsTimerEvent);
	}

    }
}


void
RedundancyWorkerThread::processRedundancyEvent(Sptr <RedundancyEvent> redundancyEvent)
{
    assert (redundancyEvent != 0);

    switch ( redundancyEvent->getInfo() )
      {
      case SyncReq:
	{
	  /// Handle a Sync Request
	  
	  if (Synchronized == itsRedundancyState)
	    {
	      Data fromServer = *(redundancyEvent->getSender());
	      LocalScopeAllocator lo; 
	      string server = fromServer.getData(lo);
	      string::size_type pos = server.find(":");
	      if (pos != string::npos)
		{
		  Data requestingHost = server.substr(0, pos);
		  Data requestingPort = server.substr(pos + 1, server.size());

		  cpLog( LOG_DEBUG, "Requesting RS Host: [%s]", requestingHost.logData());
		  cpLog( LOG_DEBUG, "Requesting RS Port: [%s]", requestingPort.logData());
		  
		  /// Send SyncRequest Ack to the requesting RS
		  if (sendRedundancyMsg(fromServer, SyncReqAck, redundancyEvent->getEventId() ))
		  {
		      /// Prepare to start a Sync Complete Ack timer
		      Sptr < RSTimerEvent > nextEvent = new RSTimerEvent (itsFifo, SyncDelayTimer);
		      
		      int count = 0;
                      unsigned short targetRedundancyPort(RSServerContainer::instance().getRSSyncPort(fromServer));

                      //connect to the TCP Server of the Sync requesting server
                      TcpClientSocket tcpClientSocket(server.substr(0, pos), targetRedundancyPort);
                      tcpClientSocket.connect();

		      /// Trigger synchronization of registered subscribers
		      if ( SubscriberContainer::instance().synchronize( requestingHost, requestingPort, count) )
		      {
			  cpLog( LOG_DEBUG, "Redundancy succeeded" );

			  /// Start a Sync Delay timer to allow the receiving RS to
			  /// receive and process the Register messages.
			  nextEvent->setCount( 0 );
			  nextEvent->setFlag( true );
			  nextEvent->setData( fromServer );
			  nextEvent->startTimer( nextEvent, (80 * count));
		      }
		      else
		      {
			  cpLog( LOG_DEBUG, "Redundancy failed" );
			  
			  /// Start a Sync Complete Ack timer
			  nextEvent->setCount( 0 );
			  nextEvent->setFlag( false );
			  nextEvent->setData( fromServer );
			  nextEvent->startTimer( nextEvent, 1000 );   // 1 sec
			  FifoEventId timerId = nextEvent->getTimerId();
			  sendRedundancyMsg(fromServer, SyncCompFail, timerId);
		      }
		  } // end sendRedundancyMsg
		}
	    }
	} // case SyncReq
	
	break;
      case SyncReqAck:
	{
	  /// Handle a SyncRequest Ack
	  cpLog( LOG_DEBUG, "SyncReq Ack Received" );
	  
	  if (WaitingForSyncReqAck == itsRedundancyState)
	    {
	      /// Take Sync Request timer off queue
	      if (( itsSyncRequestTimerId != 0 ) &&
		  ( itsSyncRequestTimerId == redundancyEvent->getEventId() ))
		{
		  
		  itsFifo->cancel(itsSyncRequestTimerId);
		  itsSyncRequestTimerId = 0;
		  
		  /// update redundancy state
		  itsRedundancyState = SyncReqAckReceived;
		  
		  /// start sychronizing timer
		  Sptr < RSTimerEvent > nextEvent = new RSTimerEvent (itsFifo, SynchronizingTimer);
		  if ( nextEvent != 0 )
		    {
		      /// Put a timed event on the fifo to make sure that synchronization completes
		      unsigned int numSubscribers(SubscriberContainer::instance().getSize());
		      cpLog( LOG_DEBUG, "Subscriber Container size = %d", numSubscribers);

		      /// Ensure a minimum of 10 seconds for Synchronizing Timer
		      if (numSubscribers < MIN_NUMSUBSCRIBERS) numSubscribers = MIN_NUMSUBSCRIBERS;
		      
		      nextEvent->startTimer( nextEvent, (numSubscribers * 500) );   // .5sec per subscriber
		      cpLog( LOG_ALERT, "Please wait while synchronization takes place...");
		      
		      itsSynchronizingTimerId = nextEvent->getTimerId();
		    }
		  else
		    {
		      cpLog( LOG_DEBUG, "RedundancyWorkerThread::Thread(), new RSTImer == 0" );
		    }
                  Connection conn;
                  char buf[10];
                  int nread = 0, len =0;

                  //loop for accpting connection for sync data, will timeout.
                  for (; ; )
                  {
                     if(tcpSyncServerSocket->accept(conn))
                     {
                        cpLog(LOG_DEBUG, "Accepted connection:%s" , conn.getDescription().c_str() );
                        break;
                     }
                 cpLog(LOG_DEBUG, " Waiting for connection");

                  }

                  if (conn.readn(buf, sizeof(buf), nread) <= 0)
                  {
                     cpLog(LOG_ERR,
                           "Failed to read content len. -- Closing connection");
                     break;
                  }

                  cpLog(LOG_ERR, "Data is: (%s), size: %d", buf, nread);

                  nread =0;

                  // get the content length
                  sscanf(buf, "%d", &len);

                  if (len == 0)
                  {
                     // no data on remote server, so just return OK
                     break;
                  }

                  // allocate buffer for contents
                  char *bigbuf;
                  LocalScopeAllocator lAllocator(&bigbuf, len+1);

                  conn.readn((void*)bigbuf, len, nread);
                  cpLog(LOG_ERR, "Number of byte read is: (%d) out of %d", nread, len);

                  if (nread < len)
                  {
                     cpLog(LOG_ERR, "Insufficent data on socket - closing");
                     break;
                  }

                  bigbuf[len] = 0;

                  //synchronize the data
                  SubscriberContainer::instance().sync(bigbuf);

		}
	      else
              {
                 cpLog( LOG_DEBUG, "msg eventId: %lu != itsSyncRequestTimerId: %lu, ignore SyncReqAck",
                        (unsigned long)(redundancyEvent->getEventId()),
                        (unsigned long)(itsSyncRequestTimerId) );
              }
	    }
	  else
	    {
	      cpLog( LOG_ERR, "SyncReq Ack Received in state: %d Ignore", itsRedundancyState );
	    }

	} // SyncReqAck
	break;

      case SyncReqFail:
	{
	  /// Handle a SyncRequest Fail
	  cpLog( LOG_DEBUG, "SyncReq Fail Received" );
	  
	  if (WaitingForSyncReqAck == itsRedundancyState)
	    {
	      /// Take Sync Request timer off queue
	      if (( itsSyncRequestTimerId != 0 ) &&
		  ( itsSyncRequestTimerId == redundancyEvent->getEventId() ))
		{
		  itsFifo->cancel(itsSyncRequestTimerId);
		  itsSyncRequestTimerId = 0;
		  
		  /// change redundancy state back to unsynchronized
		  itsRedundancyState = Unsynchronized ;
		  
		  cpLog( LOG_DEBUG, "Retry SyncReq on new RS" );
		  
		  /// Set its retry counter to the max # of attempts allowed.
		  /// This will force beginSync() to find a new RS.
		  itsRetryCounter = MAX_ATTEMP_RETRY;
		  beginSync();
		}
	      else
		{
		  cpLog( LOG_DEBUG, "msg eventId: %lu != itsSyncRequestTimerId: %lu, ignore SyncReqFail", 
			 (unsigned long)(redundancyEvent->getEventId()), 
			 (unsigned long)(itsSyncRequestTimerId) );
		}
	    }
	  else
	    {
	      cpLog( LOG_ERR, "SyncReq Fail Recvd in state: %d Ignore", itsRedundancyState );
	    }

	} // SyncReqFail
	break;

      case SyncCompFail:
	{
	  /// Handle a SyncComplete Fail
	  cpLog( LOG_DEBUG, "SyncComplete Fail Received" );
	  
	  if (SyncReqAckReceived == itsRedundancyState)
	    {
	      /// Cancel Synchronizing Timer
	      if ( itsSynchronizingTimerId != 0 )
		{
		  itsFifo->cancel(itsSynchronizingTimerId);
		  itsSynchronizingTimerId = 0;
		  
		  /// change redundancy state back to unsynchronized
		  itsRedundancyState = Unsynchronized ;
		  
		  cpLog( LOG_DEBUG, "Retry SyncReq on new RS" );
		  
		  /// Set its retry counter to the max # of attempts allowed.
		  /// This will force beginSync() to find a new RS.
		  itsRetryCounter = MAX_ATTEMP_RETRY;
		  beginSync();
		}
	      else
		{
		  cpLog( LOG_DEBUG, "no timer active, ignore SyncComplete Fail");
		}
	    }
	  else
	    {
	      cpLog( LOG_ERR, "SyncComp Fail Recvd in state: %d Ignore", itsRedundancyState );
	    }
	  
	} // SyncCompFail
	break;

      case SyncCompSucc:
	{
	  /// Handle a SyncComplete Succ
	  cpLog( LOG_DEBUG, "SyncComplete Succ Received" );
	  
	  if (SyncReqAckReceived == itsRedundancyState)
	    {
	      /// Cancel Synchronizing Timer
	      if ( itsSynchronizingTimerId != 0 )
		{
		  itsFifo->cancel(itsSynchronizingTimerId);
		  itsSynchronizingTimerId = 0;
		  
		  /// Make this RS active.
		  activateRS();
		}
	      else
		{
		  cpLog( LOG_DEBUG, "no timer active, ignore cancel timer");
		}
	    }

	  Data fromServer = *(redundancyEvent->getSender());
	  cpLog( LOG_DEBUG, "Send SyncCompAck to Host: [%s]", fromServer.logData());
	  
	  /// Send SyncComplete Ack
	  sendRedundancyMsg(fromServer, SyncCompAck, redundancyEvent->getEventId());

	} // SyncCompSucc
	break;

      case SyncCompAck:
	{
	  /// Handle a Sync Complete Ack
	  cpLog( LOG_DEBUG, "SyncComplete Ack Received" );
	  
	  /// Cancel the SyncCompleteAckTimer
	  if (itsFifo != 0)
	    {
	      itsFifo->cancel( redundancyEvent->getEventId() );
	    }

	} // SyncCompAck
	break;
	
      default:
	break;
	
      } // end of switch

    if(tcpSyncServerSocket != NULL)
    {
       delete tcpSyncServerSocket;
       tcpSyncServerSocket = NULL;
    }
}


void
RedundancyWorkerThread::processTimerEvent(Sptr <RSTimerEvent> rsTimerEvent)
{
    assert (rsTimerEvent != 0 );

    switch ( rsTimerEvent->getTimerEventType() )
      {
      case StartupDelayTimer:
	{
	  /// Begin Synchronization for the first time
	  cpLog( LOG_DEBUG, "StartupDelayTimer timed out. Begin Sync" );
	  beginSync();
	  
	}
	break;
	
      case SyncRequestTimer:
	{
	  /// Did not receive a Sync Request Ack in time. Assume it was not
	  /// received by the active RS and try to send a Sync Request again.
	  cpLog( LOG_DEBUG, "SyncRequestTimer timed out to %s ", itsListOfRSTried->back().logData() );
	  beginSync();
	}
	break;
	
      case SynchronizingTimer:
	{
	  /// Did not receive a Sync Complete Succeed or Fail.
	  /// Need to begin synchronization over again to be certain
	  /// we have all of the RS subscriber register data.
	  cpLog( LOG_DEBUG, "SyncronizingTimer timed out to %s ", itsListOfRSTried->back().logData() );
	  beginSync();
	}
	break;
	
      case SyncDelayTimer:
	{
	  /// Waited long enough for all Register messages to reach
	  /// the other RS and get processed. May now send the Sync
	  /// complete message.
	  cpLog( LOG_DEBUG, "SyncDelayTimer timed out");
	  
	  /// Prepare to start a Sync Complete Ack timer
	  Sptr < RSTimerEvent > nextEvent = new RSTimerEvent (itsFifo, SyncCompleteAckTimer);
	  
	  /// Start a Sync Complete Ack timer
	  Data fromServer(rsTimerEvent->getData());
	  nextEvent->setData( fromServer );
	  nextEvent->setCount( 0 );
	  nextEvent->setFlag( true );
	  nextEvent->startTimer( nextEvent, 1000);  /// 1 second
	  FifoEventId timerId = nextEvent->getTimerId();
	  
	  sendRedundancyMsg(fromServer, SyncCompSucc, timerId);
	  
	}
	break;
      case SyncCompleteAckTimer:
	{
	  /// Retry sending the sync complete only a few times before giving up
	  unsigned short count = rsTimerEvent->getCount();
	  if (count < 3)
	    {
	      /// Sync Comp Ack Timer timed out. Try again.
	      bool passed = rsTimerEvent->getFlag();
	      Data hostPort(rsTimerEvent->getData());
	      
	      Sptr < RSTimerEvent > nextEvent = new RSTimerEvent (itsFifo, SyncCompleteAckTimer);
	      FifoEventId timerId = 0;
	      if ( nextEvent != 0 )
		{
		  nextEvent->setFlag( passed );
		  nextEvent->setData( hostPort );
		  nextEvent->setCount( count + 1 );
		  nextEvent->startTimer( nextEvent, 1000 );   // 1 sec
		  timerId = nextEvent->getTimerId();
		}
	      
	      if (passed)
		{
		  sendRedundancyMsg(hostPort, SyncCompSucc, timerId);
		}
	      else
		{
		  sendRedundancyMsg(hostPort, SyncCompFail, timerId);
		}
	    }
	}
	break;
	
      default:
	break;
	
      } // end of switch

}



/**
 * Compose and send a redundancy message via UDP to the specified server.
 * @param  server  Address of destination server  
 * @param  msgType  Type of message to be sent.
 * @param  eventId  Unique identifier of event.
 * @return  True if message was sent, false otherwise
 */
bool
RedundancyWorkerThread::sendRedundancyMsg( const Data & server, const InfoType msgType, const FifoEventId eventId)
{

    cpLog(LOG_DEBUG, "**target host:port [%s]", server.logData());

    unsigned short targetRedundancyPort(RSServerContainer::instance().getRSSyncPort(server));
    cpLog(LOG_DEBUG, "**Redundancy port [%d]", targetRedundancyPort);

    if (targetRedundancyPort != 0)
    {
        LocalScopeAllocator lo;
        string serverHost = server.getData(lo);
        string::size_type pos = serverHost.find(":");
        if (pos != string::npos)
        {
            Data targetHost (serverHost.substr(0, pos) );

            NetworkAddress requestorAddr = NetworkAddress(targetHost.getData(lo), targetRedundancyPort);

            redundancyPortData rsData;
            rsData.dataType = htons(static_cast <u_int16_t> (RedundancyData));
            rsData.port = htons(itsSipPort);
            rsData.info = htons(static_cast <u_int16_t> (msgType));
	    // Note: eventId is a typedef of Vocal::TimeAndDate::TimerEntryId
            rsData.misc = htonl((u_int32_t)eventId);
            itsUdpStack->transmitTo( (char*)&rsData, sizeof(redundancyPortData), &requestorAddr);

            cpLog(LOG_DEBUG, "send To: %s:%d", targetHost.logData(), targetRedundancyPort );
            return true;
        }
        else
        {
            cpLog(LOG_ERR, "Redundancy Message failed to send for server [%s]", server.logData());
            return false;
        }
    }
    else
    {
        cpLog(LOG_ERR, "Redundancy Message failed to send for server [%s]", server.logData());
        return false;
    }

} /// end of sendRedundancyMsg

/**
 * Begin the process of synchronizing subscriber registration data
 * by finding an active RS and sending a Sync Request to it.
 */
void
RedundancyWorkerThread::beginSync()
{
    Data server;

    if ( getActiveRS(server) )
    {
        /// update redundancy state
        itsRedundancyState = WaitingForSyncReqAck;

        /// increment counter
        itsRetryCounter++;
        cpLog(LOG_DEBUG, "[%d] Attempt(s)", itsRetryCounter);

        /// Start a Synchronization timer
        assert(itsFifo != 0 );

        Sptr < RSTimerEvent > nextEvent = new RSTimerEvent (itsFifo, SyncRequestTimer);
        if ( nextEvent != 0 )
        {
            cpLog(LOG_DEBUG, "send Sync Request");

            nextEvent->startTimer( nextEvent, 1000 );   // 1 sec
            itsSyncRequestTimerId = nextEvent->getTimerId();

            //Create a TCP server for receiving the sync data

            unsigned short targetRedundancyPort(RSServerContainer::instance().getRSSyncPort(server));

            if(tcpSyncServerSocket!=NULL)
               delete tcpSyncServerSocket;

            tcpSyncServerSocket = new TcpServerSocket(targetRedundancyPort);

            /// send Sync Request
            sendRedundancyMsg(server, SyncReq, itsSyncRequestTimerId);

        }

    }
    else
    {
        if ( itsListOfRSTried->empty() )
        {
            /// No other active RS's exist. This RS is as synchronized as it can get at this moment
            cpLog(LOG_DEBUG, "no other active RS's exist");
            activateRS();
        }
        else
        {
            cpLog(LOG_ALERT, "********** ALERT: Cannot Synchronize with Any Active RS!!!");
            cpLog(LOG_ALERT, "********** ALERT: Cannot Synchronize with Any Active RS!!!");
            cpLog(LOG_ALERT, "********** Check configuration to make sure RSSyncPort is not 0 ");
            assert(0);
        }
    }
} /// end beginSync


/**
 * Obtain an active RS. Allow a number of retries on the same RS before 
 * looking for an alternate active RS.
 * @param  activeRS  Address of a currently active Redirect Server.
 * @return  True if an active RS was found, false otherwise.
 */
bool
RedundancyWorkerThread::getActiveRS( Data & activeRS )
{
    /// Assuming that each RS tried is pushed to the end of the
    /// vector, the back() command will get the most recent RS tried.

    RSServerContainer& srvContainer = RSServerContainer::instance();

    /// check if we already have an RS to retry.
    if ((itsRetryCounter > 0) && (itsRetryCounter < MAX_ATTEMP_RETRY))
    {
        Data tryingRS( itsListOfRSTried->back() );
        cpLog(LOG_DEBUG, "tryingRS= %s", tryingRS.logData() );
        if ( (tryingRS != Data("") ) &&
                (Active == srvContainer.getStatus(tryingRS)) )
        {
            /// return hostname of active RS found
            activeRS = tryingRS;
            cpLog( LOG_DEBUG, "requesting server: [%s]", activeRS.logData() );
            return true;
        }
    }

    /// get any activeRS that has not been tried.
    Sptr < vector < Data > > allRSs = new vector < Data > ;
    srvContainer.getListofRS( allRSs );

    bool alreadyTriedThisRS = false;
    
    for ( vector < Data > ::iterator nextRS = allRSs->begin(); 
	  nextRS != allRSs->end() ; 
	  ++nextRS )
    {
        alreadyTriedThisRS = false;
        for (vector < Data > ::iterator i = itsListOfRSTried->begin(); 
	     i != itsListOfRSTried->end(); 
	     ++i)
        {
            if (*nextRS == *i)
            {
                alreadyTriedThisRS = true;
                break;
            }
        }

        if (alreadyTriedThisRS)
        {
            cpLog( LOG_DEBUG, "alreadyTriedThisRS: %s", (*nextRS).logData() );
            continue;
        }

        if (Active == srvContainer.getStatus(*nextRS) )
        {
            /// return hostname of active RS found
            activeRS = *nextRS;
            cpLog( LOG_DEBUG, "requesting server: [%s]", activeRS.logData() );

            /// add this RS to itsListOfRSTried
            itsListOfRSTried->push_back(*nextRS);
            itsRetryCounter = 0;

            return true;
        }
    }

    return false;
} /// end of getActiveRS

/**
 * Activate RS by changing its state and have it begin heartbeating.
 */
void
RedundancyWorkerThread::activateRS()
{
    cpLog( LOG_INFO, "Activating this RS now" );
    itsRetryCounter = 0;
    itsListOfRSTried->clear();

    /// change redundancy state to synchronized
    itsRedundancyState = Synchronized;

    /// Make this RS active.
    RSProxy::instance().setRsState(RSACTIVE);

    /// start heartbeating.
    RSProxy::instance().runHeartbeatThread();

} /// end of activateRS

/**
 * Update the Redundancy port in response to a change in provisioning.
 * @param  oldSyncPort  Value of previously stored sync port
 * @param  newSyncPort  Value of sync port to be stored.
 */
void
RedundancyWorkerThread::setRedundancyPort(unsigned short oldSyncPort, unsigned short newSyncPort)
{
    /// set the local copy of the redundancy port
    itsRedundancyPort = newSyncPort;

    /// Generate a maintenance Redundancy message to change the sync port
    NetworkAddress requestorAddr = NetworkAddress(theSystem.gethostAddress(), oldSyncPort);
    cpLog(LOG_DEBUG, "send To: [%s:%d]", theSystem.gethostAddress(), oldSyncPort );

    redundancyPortData rsData;
    rsData.dataType = htons(static_cast <u_int16_t> (MaintenanceData));
    rsData.port = htons(itsSipPort);
    rsData.info = htons(static_cast <u_int16_t> (NA));
    rsData.misc = htonl(static_cast <u_int32_t> (newSyncPort));
    itsUdpStack->transmitTo( (char*)&rsData, sizeof(redundancyPortData), &requestorAddr);

} /// end of setRedundancyPort


// End of File
