#ifndef SubscriberContainer_HXX
#define SubscriberContainer_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include "RSProxy.hxx"
#include "Subscriber.hxx"
#include "VMissingDataException.hxx"
#include "VRWLock.hxx"
#if defined(__GNUC__)
#if (__GNUC__ < 3)
#include <hash_map>
#else
#include <ext/hash_map>
#endif
#include "DataHash.hxx"
#else
#include <map>
#endif

static const char* const SubscriberContainerVersion = "$Id: SubscriberContainer.hxx,v 1.47.2.1 2003/01/23 02:51:46 sprajpat Exp $";

/**
 * The Provisioning Server is notifying us of a change in the 
 * Accounts Directory. A user could have been Added, Changed or Deleted.
 * @param  data  Subscriber's identifier
 * @param  fileName  Name of file as it appears in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void updateAccounts(const string& data, const string& fileName, bool deleteFlag );

/**
 * The Provisioning Server is notifying us of a change in
 * a specific subscriber. A user could have been Changed or Deleted.
 * @param  data  Contents of xml file. 
 * @param  fileName  Subscriber's identifier
 * @param  deleteFlag  Indicator of whether or not subscriber was deleted.
 */
void updateSubscriber(const string& data, const string& fileName, bool deleteFlag );

/**
 *  Obtain the subscriber information from provisioning
 *  This should only be called for master users, not aliases.
 * @param  userName  Master user identifier 
 * @param  deleteFlag  Indicator of whether subscriber was deleted or not.
 */
void updateUser(const string& userName, bool deleteFlag);

/**
 * The Provisioning Server is notifying us of a change in the 
 * Aliases Directory. There will be no changes to an alias.
 * They can only be added and deleted. If it is deleted, remove
 * it. If it was added, rely upon the master user to add it.
 * @param  data  Alias's identifier 
 * @param  fileName  Name of file modified in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void updateAliases(const string& data, const string& fileName, bool deleteFlag);

/**
 * The maximum expire time was updated in provisioning. Store the new maximum.
 * @param  data  Contents of xml file.
 * @param  fileName  Name of file modified in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void updateExpireTime(const string& data, const string& fileName, bool deleteFlag);

/**
 * Called when PS got restarted, time to update the registration requests
 * @param  data  REGISTER_ITEM_REQ.
 * @param  all
 * @param  ignore
 */
void updateRegister(const string& data, const string& fileName, bool deleteFlag);

/**
 *  This class defines a container of Subscribers. Subscribers are
 *  stored in a map with an identifier (usually a phone number) as the key.
 */

class SubscriberContainer
{
    public:

        /**
         * Return the singleton subscriber container or create it
         * if it doesn't already exist. 
         * @return Reference to the subscriber container singleton object.
         */
        static SubscriberContainer& instance();

        /// Destructor
        virtual ~SubscriberContainer();

        /**
         * Add a master user to the Subscriber Container
         * @param  key  Unique identifier for this subscriber
         * @param  calling  Contact list used for calls made from this subscriber.
         * @param  called  Contact list used for calls terminating to this subscriber.
         * @param  term  Contact list used for calls terminating to this subscriber.
         * @param  expireReq  Time in seconds originator of Register message would
         *                    like before having to register again.
         * @param  staticFlag  Flag indicating if subscriber is statically registered.
         * @return Maximum time in seconds before another register message
         *         is required.
         */
        unsigned long add( const Data& key,
                           const Sptr < vector < Data > > calling,
                           const Sptr < vector < Data > > called,
                           const Sptr < vector < Data > > term,
                           const unsigned long expireReq,
                           const bool staticFlag);

        /**
         * Upon startup, add each user with the same empty subscriber object.
         * Aliases and masters come through this function independent of
         * each other, so they will have different blank users.
         * Aliases at this time are not linked to their master users. That is
         * done later when their provisioning data is obtained.
         * @param  users  List of all valid user names to be added.
         * @param  aliasFlag  Flag indicating whether users in list are aliases.
         */
        void add(vector < Data > & users, const bool aliasFlag);

        /**
         * Using the key, find the Subscriber object in the map and delete it
         * @param  key  Unique identifier for subscriber.
         */
        void remove( const Data& key );
        /**
         * Delete the subscriber key from the container and make the
         * smart pointer point to a brand new subscriber in case this master key
         * has aliases pointing to the same subscriber object.
         * @param  key  Unique identifier for subscriber.
         */
        void removeMaster( const Data& key );

        /**
         * Using the key, find the subscriber and combine its Called and
         * Term contacts to be passed back.
         * @param  key  Unique identifier for subscriber.
         * @param  contacts  List of contacts obtained from specified subscriber.
         * @param  isExpired  Flag indicating if registration has expired.
         * @param  termOnly  Flag indicating to not return called features
         * @return  true if subscriber was found, false if not
         */
        bool searchTo( const Data& key,
                       vector < Data > & contacts,
                       bool& isExpired, 
                       bool termOnly=false) ;

        /**
         * Using the key, find the subscriber and obtain it's Calling contacts
         * @param  key  Unique identifier for subscriber.
         * @param  contacts  List of contacts obtained from specified subscriber.
         * @return  true if subscriber was found, false if not
         */
        bool searchFrom( const Data& key, vector < Data > & contacts ) ;

        /**
         * Diagnostic - Display all information stored against a subscriber
         * @param  key  Unique identifier for subscriber.
         */
        void display( const Data& key) ;

        /**
         * Display all subscribers.
         */
        void displayAll( ) ;

        /**
         * Store registration specific data for subscriber.
         * @param  to  Unique identifier for registering subscriber.
         * @param  from  Unique identifier for subscriber which triggered the registration.
         * @param  newContacts  Contact list used for terminating calls to this subscriber.
         * @param  reqExpire  Requested time in seconds before next registration must be sent.
         * @param  expire  Time in seconds granted before next registration must be sent.
         * @param  existingContacts  Previously stored terminating contact list.
         * @return  Status response value to be sent back to originator of Register message.
         */
        int updateOnRegister(const Data& to,
                             const Data& from,
                             const Sptr < vector < Data > > & newContacts,
                             const unsigned long &reqExpire,
                             unsigned long &expire,
                             vector < Data > &existingContacts);

        /**
         * Store the data against the specified subscriber. Obtain the provisioning
         * data if it has not already been done. Aliases are not updated here.
         * @param  identifier  Unique identifier for subscriber.
         * @param  newCalled  Contact list for terminating a call to this subscriber.
         * @param  newCalling  Contact list for calls originating from this subscriber.
         * @param  newTerm  Contact list for terminating a call to this subscriber.
         * @param  staticRegFlag  Flag to indicate if subscriber is statically registered.
         */
        void updateMasterUser(const Data& identifier,
                              const Sptr < vector < Data > > & newCalled,
                              const Sptr < vector < Data > > & newCalling,
                              const Sptr < vector < Data > > & newTerm,
                              const bool staticRegFlag);

        /**
         * Retrieve the port on which this RS sends/receives SIP messages
         * @return  Port number
         */
        unsigned short getPort() const
        {
           return localSipPort;
           //return mySipPort;
        }

        /**
         * Store port on which RS sends/receives SIP messages
         * @param  sipPort  A port used for sending/receiving SIP messages
         */
        void setPort (const unsigned short sipPort)
        {
           localSipPort = sipPort;
           //mySipPort = sipPort;
        }

        /**
         * Store reference to this RS's local proxy
         * @param  myProxy  Reference to this RS's proxy object
         */
        void setProxy (RSProxy& myProxy)
        {
            localRSProxy = &myProxy;
        }

        /**
         * Loop through each subscriber and generate a Register 
         * message for each one that is currently registered to be
         * sent to all other RS's configured in the system
         * @param  remoteRSHost  Host of RS in need of synchronization.
         * @param  remoteRSPort  Port of RS on which to send Register messages.
         * @param  count  Number of registered users currently in the system.
         * @return  true if synchronization occurred correctly.
         */
        bool synchronize(const Data& remoteRSHost, const Data& remoteRSPort, int& count) ;

        /**
         * Generate and send a Register message for specified subscriber.
         * @param  subscriber  Subscriber object containing info to be sent.
         * @param  userName  Unique identifier for subscriber.
         * @param  remoteRSHost  Host of RS in need of synchronization.
         * @param  remoteRSPort  Port of RS on which to send Register messages.
         * @param  localRSHost  Host of this RS which is sending the Register message.
         * @param  localRSPort  Port of this RS on which responses should be sent.
         * @return  true if message sent, false if not.
         */
        bool generateRegisterMsg( Subscriber& subscriber,
                                  const Data& userName,
                                  const Data& remoteRSHost,
                                  const Data& remoteRSPort,
                                  const Data& localRSHost,
                                  const Data& localRSPort) ;

        /**
         * Obtain the number of subscribers currently in the container
         * @return  Number of subscribers stored in the container.
         */
        unsigned int getSize() const
        {
            return subList.size();
        }

        /**
         * Obtain the list of all Subscribers from PS and
         * store them in the Subscriber Container.   
         * Begin with the master users and then get the aliases.
         */
      void updateAllSubscribers();


      // From the syncAgent, called when a sync request is received
      bool sendRequest(Connection &conn);

      /// Request sync info from another rs
      void sync(const string& s);

      string serialize();

      void unserialize(const string& s);

        /**
         * Step through the Alias list checking to see if each one
         * already exists in the Subscriber Container. If it does not
         * it will be added and it will register an interest in it
         * with provisioning.
         * @param  masterKey  Unique identifer for this master subscriber.
         * @param  aliasList  List of alias identifiers for this master subscriber.
         */
        void updateAliasList(const Data& masterKey, list < string > & aliasList);

        /**
         * Obtain data from provisioning for specified subscriber if it has not been done already.
         * If key is an alias it will be linked to its master user.
         * @param  key  Unique identifier for subscriber.
         */
        void checkProvisioned(const Data& key);

        /**
         * Store and retrieve functions for global maximum registration time.
         * @return  Maximum expire time in seconds allowed for all subscriber registrations.
         */
        unsigned long getMaxExpireTime() const
        {
            return MaxExpireTime;
        }

        /**
         * Store the maximum amount of time a subscriber can be registered.
         * @param  expTime  Maximum time in seconds that a subscriber can be registered.
         */
        void setMaxExpireTime (const unsigned long expTime)
        {
            MaxExpireTime = expTime;
        }

      static const string syncServiceName;
    protected:

        /// Constructor
        SubscriberContainer();

        /**
         * Store the given contacts for the specified subscriber.
         * @param  user  Unique identifier for subscriber.
         * @param  contacts  Contact list used for terminating a call 
         *                   on this subscriber.
         */
        void updateTerm( Subscriber& user, const Sptr < vector < Data > > contacts);

        /**
         * Store the given data for the specified subscriber, usually due to a provisioning change.
         * @param  user  Unique identifier for subscriber.
         * @param  calling  Contact list used for outgoing calls from subscriber.
         * @param  called  Contact list used for terminating calls to subscriber.
         * @param  expire  Time in seconds before registration expires.
         */
        void updateConfig( Subscriber& user, const Sptr < vector < Data > > calling,
                           const Sptr < vector < Data > > called, const unsigned long expire );

        /// This is the single instance of SubscriberContainer
        static Sptr < SubscriberContainer > theSubContainer;

        /**
         * Obtain subscriber object for specified subscriber.
         * @param  key  Unique identifier for subscriber.
         * @param  subFound  Pointer to requested subscriber object.
         * @return True if found, false if not.
         */
        bool search( const Data& key, Sptr < Subscriber > & subFound );

    private:

#if defined (__linux__) && !defined (__INTEL_COMPILER)
        /// Use hash_map on Linux since it faster than regular map
        typedef hash_map < Data, Sptr < Subscriber > , DataHash > Table;
#else
        /**
         * Some STL implementations do not support the hash map eg. Solaris
         * so use map instead. Note: map is slower
         */
        typedef map < Data, Sptr < Subscriber > > Table;
#endif

        typedef Table::iterator TableIter;

        /// Data member instance of the multimap
        mutable Table subList;

        /// Sip port for local RS.
        unsigned short localSipPort;
      //unsigned short mySipPort;
        /// Maximum expire time allowed for any subscriber. Specified in provisioning.
        unsigned long MaxExpireTime;

        RSProxy* localRSProxy;

        VRWLock SubContainerMutex;

};

#endif
