/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2001 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const DefaultResourcesAvailablePolicy_cxx_Version = 
    "$Id: DefaultResourcesAvailablePolicy.cxx,v 1.4 2002/05/15 11:35:43 icahoon Exp $";


#include "DefaultResourcesAvailablePolicy.hxx"
#include "GatekeeperInfo.hxx"
#include "EndpointInfoMap.hxx"
#include "RASGKProtocol.hxx"
#include "VLog.hxx"


using Vocal::H323GK::DefaultResourcesAvailablePolicy;
using Vocal::H323GK::ResourcesAvailablePolicy;
using Vocal::H323GK::RASPolicy;
using Vocal::H323GK::RASGKProtocol;
using Vocal::SipH323Info::GatekeeperInfo;
using Vocal::SipH323Info::EndpointInfoMap;
using Vocal::SipH323Info::EndpointInfo;
using Vocal::Services::Service;
using Vocal::Transport::IPAddress;
using Vocal::Logging::VLog;


DefaultResourcesAvailablePolicy::DefaultResourcesAvailablePolicy(
    Service  	    	&   service,
    RASGKProtocol   	&   rasGkProtocol,
    GatekeeperInfo  	&   gatekeeperInfo,
    EndpointInfoMap 	&   endpointInfoMap
)
    :	ResourcesAvailablePolicy(service, rasGkProtocol, 
                                 gatekeeperInfo, endpointInfoMap)
{
}


DefaultResourcesAvailablePolicy::~DefaultResourcesAvailablePolicy()
{
}


void 	    	    
DefaultResourcesAvailablePolicy::processResourcesAvailableIndication(
    Sptr<H225_RasMessage>   rasPDU, 
    Sptr<IPAddress>         remoteAddr
)
{
    const string    fn("DefaultResourcesAvailablePolicy::processResourcesAvailableIndication");
    VLog    	    log(fn);
    
    H225_ResourcesAvailableIndicate	&   rai = *rasPDU;
    Sptr<EndpointInfo>      	            endpointInfo = 0;

    // Let's first see if we can make some sense of it.
    //
    if ( !validateResourcesAvailableIndicate(rai, endpointInfo) )
    {
        // This isn't an RAI that we can make sense of so we will just
        // ignore it. 
        
        VDEBUG(log) << fn << ": Cannot validate RAI message. Ignoring." 
                    << VDEBUG_END(log);
        return;
    }

    // Let the derived class decide how to process it.
    //
    Proceed proceed = onResourcesAvailable(rasPDU, remoteAddr, 
                                           endpointInfo, rai);

    switch ( proceed )
    {
	case wait:
	{
	    // The gatekeeper is probably waiting for an external 
	    // event to drive the discovery forward. 
	    //
	    VDEBUG(log) << fn << ": waiting." << VDEBUG_END(log);
	    return;
	}

	default:
	{
            remoteAddr = responseAddress(endpointInfo, remoteAddr);
	    buildResourcesAvailableConfirm(rai, remoteAddr);

	    break;
	}
    }
}


ostream &           
DefaultResourcesAvailablePolicy::writeTo(ostream & out) const
{
    return ( out << "DefaultResourcesAvailablePolicy" );
}

				
RASPolicy::Proceed	    
DefaultResourcesAvailablePolicy::onResourcesAvailable(
    Sptr<H225_RasMessage>  	            rasPDU, 
    Sptr<IPAddress> 	    	            remoteAddr,
    Sptr<EndpointInfo>  	            endpointInfo,
    H225_ResourcesAvailableIndicate     &   rai
)
{
    const string    fn("DefaultResourcesAvailablePolicy::onResourcesAvailable");
    VLog            log(fn);

    endpointInfo->recordResourcesAvailable();

    bool change 
            = ( endpointInfo->outOfResources() != rai.m_almostOutOfResources);
    
    endpointInfo->outOfResources(rai.m_almostOutOfResources);

    if ( change )
    {    
        VINFO(log)  << fn << "\nChanged resource available for endpoint =\n{" 
                    << *endpointInfo << '}' << VINFO_END(log);
    }
    
    return ( confirm );
}


void	    
DefaultResourcesAvailablePolicy::finalizeResourcesAvailableConfirm(
    H225_ResourcesAvailableIndicate     &   rai,
    H225_ResourcesAvailableConfirm      &   rac
)
{
    // By default do nothing.
}


Sptr<IPAddress>     
DefaultResourcesAvailablePolicy::responseAddress(
    Sptr<EndpointInfo>  	    	    endpointInfo,
    Sptr<IPAddress>     	    	    remoteAddr
)
{
    // First try to get it from the endpointInfo
    //
    if  (   myGatekeeperInfo.returnToSender() == false
        &&  endpointInfo != 0 
        )
    {
    	remoteAddr = new IPAddress(endpointInfo->rasAddress());
    }

    // The last resort is to use the address obtained from the socket.
    //
    return ( remoteAddr );
}


bool  	    
DefaultResourcesAvailablePolicy::validateResourcesAvailableIndicate(
    H225_ResourcesAvailableIndicate     &   rai,
    Sptr<EndpointInfo>  	        &   endpointInfo
)
{
    string  endpointIdentifier;

    endpointIdentifier = rai.m_endpointIdentifier.GetValue();
    endpointInfo = myEndpointInfoMap.findByEndpointId(endpointIdentifier);

    // If they don't have an endpointIdentifier, they aren't validated.
    //
    if ( endpointInfo == 0 )
    {
    	return ( false );
    }

    return ( true );
}


void	    
DefaultResourcesAvailablePolicy::buildResourcesAvailableConfirm(
    H225_ResourcesAvailableIndicate     &   rai, 
    Sptr<IPAddress>     	            remoteAddr
)
{
    const string    fn("DefaultResourcesAvailablePolicy::buildResourcesAvailableConfirm");
    VLog    	    log(fn);

    //  ResourcesAvailableConfirm ::= SEQUENCE --(RAC)
    // {
    //      requestSeqNum           := [from rai]
    //      protocolIdentifier      := [from openh323]
    //  }

    H225_RasMessage      pdu;
    pdu.SetTag(H225_RasMessage::e_resourcesAvailableConfirm);
    
    H225_ResourcesAvailableConfirm  &   rac = pdu;
    
    rac.m_requestSeqNum = rai.m_requestSeqNum;
    rac.m_protocolIdentifier.SetValue(myGatekeeperInfo.h225ProtocolID().c_str());

    finalizeResourcesAvailableConfirm(rai, rac);

    myRasGkProtocol.resourcesAvailableConfirm(pdu, remoteAddr);
}
