#if !defined(VOCAL_ENDPOINT_INFO_HXX)
#define VOCAL_ENDPOINT_INFO_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000, 2001 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const EndpointInfo_hxx_Version = 
    "$Id: EndpointInfo.hxx,v 1.17 2002/05/15 09:43:31 icahoon Exp $";


#include "ptlib.h"
#include "h323pdu.h"

#include "Writer.hxx"
#include "IPAddress.hxx"
#include "FifoBase.h"
#include "Sptr.hxx"
#include "Event.hxx"
#include "Alias.hxx"
#include "TimeVal.hxx"
#include "RWMutex.hxx"

#include <string>
#include <list>


/** Vovida Open Communication Application Library<br><br>
 */
namespace Vocal
{


/** SIP to H.323 Object Information<br><br>
 */
namespace SipH323Info
{


using Vocal::Transport::IPAddress;
using Vocal::TimeAndDate::TimeVal;
using Vocal::Threads::RWMutex;
using Vocal::Services::Event;
using Vocal::FifoBase;
using std::string;
using std::list;


/** Information for each endpoint.<br><br>
 *
 *  @see    Vocal::Transport::IPAddress
 */
class EndpointInfo : public Vocal::IO::Writer
{

    public:


    	/** Create an empty endpoint record 
    	 */
    	EndpointInfo();


    	/** Create endpoint record given an RRQ
    	 */
    	EndpointInfo(H225_RegistrationRequest & rrq);


        /** Create and endpoint record for a statically registered
         *  gateway.
         */
        EndpointInfo(
            const string        &   endpointId,
            const IPAddress     &   callSignalAddress,
            const list<string>  &   prefixes,
            bool                    stripPrefix);

        
    	/** Copy constructor
	 */
	EndpointInfo(const EndpointInfo &);


    	/** Virtual destructor
	 */
	virtual     	~EndpointInfo();


    	/** Assignment operator
	 */
    	const EndpointInfo &  operator=(const EndpointInfo &);


        /** Equality operator
         */
        const bool  operator==(const EndpointInfo &) const;
        

        /** Less than relational operator
         */
        const bool  operator<(const EndpointInfo &) const;


    	/** Update the endpoint record given a RRQ. Returns true
	 *  if the resulting endpoint record is valid.
	 */
    	virtual bool	update(H225_RegistrationRequest & rrq);


    	/** Returns true is the endpoint record is valid.
	 *  An endpoint record is valid if it has a phone number in
	 *  the alias address, the endpoint identifier is not a null
	 *  string, it has a RAS address, and it has a call signal address.
	 */
    	virtual bool	valid() const;


    	/** Set the endpoint identifier.
	 */
    	virtual void    endpointIdentifier(const string &);


    	/** Access the endpoint identifier.
	 */
    	virtual string  endpointIdentifier() const;


        /** The type of endpoint.
         */
        enum Type
        {
            UNDEFINED_TYPE = 0,
            TERMINAL_TYPE,
            GATEWAY_TYPE,
            MCU_TYPE,
            TYPE_SIZE
        };
        
        
        /** The type as a string.
         */
        static  const char * typeName(Type);
        
        
        /** Set the endpoint type.
         */
        virtual void    endpointType(Type);
        

        /** Access the endpoint type.
         */
        virtual Type    endpointType() const;
        
        
    	/** Set the alias address.
	 */
    	virtual void    alias(const H225_ArrayOf_AliasAddress &);


    	/** Access the alias address.
	 */
	virtual H225_ArrayOf_AliasAddress   alias() const;


    	/** Set the gateway prefix.
	 */
    	virtual void    prefix(const H225_GatewayInfo &);


    	/** Set the gateway prefix.
	 */
    	virtual void    prefix(const list<string> &);


    	/** Access the gateway prefix.
	 */
	virtual H225_ArrayOf_AliasAddress   prefix() const;


        /** If set, the gateway prefix will be stripped from the
         *  destination phone number when calling out to a gateway.
         */
        virtual void    stripPrefix(bool);
        
        
        /** Access the stripPrefix flag.
         */
        virtual bool    stripPrefix() const;
        

        /** Set to true if the registration is done statically.
         */
        virtual void    staticRegistration(bool);
        

        /** Determine whether the prefix was set statically.
         */
        virtual bool    staticRegistration() const;


    	/** Set the ras address.
	 */
    	virtual void    rasAddress(const IPAddress &);


    	/** Access the ras address.
	 */
	virtual IPAddress   rasAddress() const;
	

    	/** Set the call signal address.
	 */
	virtual void    callSignalAddress(const IPAddress &);


    	/** Access the call signal address.
	 */
	virtual IPAddress   callSignalAddress() const;


    	/** Set the time to live.
	 */
    	virtual void    timeToLive(u_int32_t);


    	/** Access the time to live.
	 */
	virtual u_int32_t   timeToLive() const;
		

    	/** Set the event id for the keep alive timer.
	 */
	virtual void    
        timeToLiveTimerEventId(FifoBase< Sptr<Event> >::EventId);


    	/** Access the event id for the keep alive timer.
	 */
	virtual FifoBase< Sptr<Event> >::EventId 
        timeToLiveTimerEventId() const;


        /** When set, the endpoint is almost out of resources.
         */         
        virtual void    outOfResources(bool);


        /** When set, the endpoint is almost out of resources.
         */         
        virtual bool    outOfResources() const;
        


        /** Record the registration.
         */
        virtual void        recordRegistration();
        

        /** Return the number of registrations for this endpoint
         *  (not including keep alive registrations).
         */
        virtual u_int32_t   registrationCount() const;
        
        
        /** Access the time of the first registration. This is also
         *  the time of the first activity for this endpoint.
         */
        virtual TimeVal     firstRegistrationTime() const;
        
        
    	/** Access the time of the last registration.
	 */
	virtual TimeVal     lastRegistrationTime() const;
		   


        /** Record the keep alive registration
         */
        virtual void        recordKeepAlive();
        

        /** Access the time of the last keep alive.
         */
        virtual TimeVal     lastKeepAliveTime() const;



        /** Record the admission of this endpoint into a call.
         */
        virtual void        recordAdmission();
                 

        /** Return the number of calls this endpoint has been 
         *  admitted into.
         */
        virtual u_int32_t   admissionCount() const;


        /** Return the number of active admissions for this endpoint.
         */
        virtual u_int32_t   activeAdmissionCount() const;
          

        /** Access the time of the first admission.
         */
        virtual TimeVal     firstAdmissionTime() const;
        
        
    	/** Access the time of the last admission.
	 */
	virtual TimeVal     lastAdmissionTime() const;
		   

        /** Record the removal of this endpoint from a call.
         */
        virtual void        recordDisengage();


        /** Return the number of calls this endpoint has been removed from.
         */
        virtual u_int32_t   disengageCount() const;


        /** Access the time of the first disengage.
         */
        virtual TimeVal     firstDisengageTime() const;
        
        
    	/** Access the time of the last disengage.
	 */
	virtual TimeVal     lastDisengageTime() const;



        /** Record the resource availability indication for this endpoint.
         */
        virtual void        recordResourcesAvailable();


        /** Return the number of resource available indications.
         */
        virtual u_int32_t   resourcesAvailableCount() const;


        /** Access the time of the first resource available indication.
         */
        virtual TimeVal     firstResourcesAvailableTime() const;
        
        
    	/** Access the time of the last resource available indication.
	 */
	virtual TimeVal     lastResourcesAvailableTime() const;



    	/** Access the time of the last activity.
	 */
	virtual TimeVal     lastActivityTime() const;

    
        /** Resets a statically registered endpoint
         */
        virtual void        clear();
        
         
    	/** Write the endpoint information to an ostream.
	 */
	virtual ostream &       writeTo(ostream &) const;


    protected:	

        void	    	updatePrefix(const H225_GatewayInfo &);

	mutable RWMutex	    	    	    myMutex;

    private:

    	string	    	    	    	    myEndpointIdentifier;
        Type                                myEndpointType;
    	Alias 	    	    	    	    myAlias;
    	H225_ArrayOf_AliasAddress           myPrefix;
        bool                                myStripPrefix;
        bool                                myStaticRegistration;
	IPAddress   	    	    	    myRasAddress;
	IPAddress   	    	    	    myCallSignalAddress;
	u_int32_t   	    	    	    myTimeToLive;
	FifoBase< Sptr<Event> >::EventId    myTimeToLiveTimerEventId;
        bool                                myOutOfResources;

        u_int32_t                           myRegistrationCount;
        TimeVal                             myFirstRegistrationTime;
        TimeVal                             myLastRegistrationTime;

        TimeVal                             myLastKeepAliveTime;
                                    
        u_int32_t                           myAdmissionCount;
        TimeVal                             myFirstAdmissionTime;
        TimeVal                             myLastAdmissionTime;
        
        u_int32_t                           myDisengageCount;
        TimeVal                             myFirstDisengageTime;
        TimeVal                             myLastDisengageTime;

        u_int32_t                           myResourcesAvailableCount;
        TimeVal                             myFirstResourcesAvailableTime;
        TimeVal                             myLastResourcesAvailableTime;

	TimeVal   	    	    	    myLastActivityTime;
};


} // namespace SipH323Info
} // namespace Vocal


#endif // !defined(VOCAL_ENDPOINT_INFO_HXX)
