/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const H323SipOpRegisterOut_cxx_Version = 
    "$Id: H323SipOpRegisterOut.cxx,v 1.17 2002/12/10 04:15:00 icahoon Exp $";


#include "H323SipOpRegisterOut.hxx"
#include "OutgoingRegEvent.hxx"
#include "RegisterMsg.hxx"
#include "SipProxyInfo.hxx"
#include "EndpointInfo.hxx"
#include "AddressToFrom.hxx"
#include "VLog.hxx"
#include "CallTrace.hxx"
#include <strstream>


using namespace Vocal;
using namespace Vocal::SipH323Proxy;
using namespace Vocal::SipH323Info;
using namespace Vocal::Transport;
using namespace Vocal::Logging;
    

H323SipOpRegisterOut::H323SipOpRegisterOut(
    SipProxyInfo    	    	    &   sipProxyInfo,
    H323SipRegistrationMap  	    &   registrationMap
) 
    :	H323SipOpOut(sipProxyInfo),
    	registrationMap_(registrationMap)
{
}


H323SipOpRegisterOut::~H323SipOpRegisterOut() 
{
}

   
const char* const
H323SipOpRegisterOut::name() const 
{
    return ( "H323SipOpRegisterOut" );
}


const Sptr<State>
H323SipOpRegisterOut::process(const Sptr<SipProxyEvent> event)
{
    const string    fn("H323SipOpRegisterOut::process");
    VLog    	    log(fn);

    Sptr<OutgoingRegEvent>          outRegEvent;
    Sptr<H323RegistrationEvent>     h323RegistrationEvent;

    if  (   outRegEvent.dynamicCast(event) == 0
        ||  ( h323RegistrationEvent = outRegEvent->h323RegEvent() ) == 0
        )
    {
        return ( PROXY_CONTINUE );
    }
        
    // If we are talking to an SUA, don't register.
    //
    if ( sipProxyInfo_.remoteClient() )
    {
    	h323RegistrationEvent->returnCode(SUCCESS);
    	h323RegistrationEvent->replyFifo().add(h323RegistrationEvent);

    	return ( PROXY_CONTINUE );
    }

    if ( sipStack_ == 0 )
    {
    	VWARN(log)  << fn << ": Received sip msg from Router "
                    << "before initialization. \n"
		    << VWARN_END(log);

	return ( PROXY_CONTINUE );
    }

    // We are talking to a sip proxy.
    //    
    Sptr<EndpointInfo>	endpointInfo = h323RegistrationEvent->endpointInfo();
    assert( endpointInfo != 0 );

    Translator::Address     	aliasA = endpointInfo->alias();
    
    
    RegisterMsg  registerMsg;

    // The request line, to and from use the ip address of the remote
    // proxy.
    //
    aliasA.setTransport(sipProxyInfo_.sipRemoteAddr());
    Sptr<BaseUrl> remoteUrl = new SipUrl(aliasA);

    Translator::AddressToFrom	aliasToFrom(aliasA);

    // Set RequestUri to remote sip proxy server.
    //
    SipRequestLine  requestLine = registerMsg.getRequestLine();
    
    requestLine.setUrl(remoteUrl);
    registerMsg.setRequestLine(requestLine);

    // Set the via field.
    //
    updateVia(registerMsg);

    // Set the from url.
    //
    registerMsg.setFrom(aliasToFrom);

    // Set the to field. Same as from.
    //
    registerMsg.setTo(aliasToFrom);


    // Set the contact using the ip address of the translator's proxy.
    //
    aliasA.setTransport(sipProxyInfo_.sipProxy());
    Sptr<BaseUrl> localUrl = new SipUrl(aliasA);

    SipContact	contact;
    contact.setUrl(localUrl);
    registerMsg.setContact(contact, 0);


    // Set the expires. Use same url as contact field.
    //
    SipExpires	sipExpires;
    strstream	delta;
    if ( h323RegistrationEvent->registration() )
    {
    	delta << ( endpointInfo->timeToLive() ? endpointInfo->timeToLive()
					  : (60*60*24*365*10) ) << ends;
    }
    else
    {
    	delta << 0 << ends;
    }
    sipExpires.setDelta(delta.str());
    delta.freeze(false);
    registerMsg.setExpires(sipExpires);

    // Add the h323RegistrationEvent to the registration map.
    //
    SipCallId	    callId = registerMsg.getCallId();
    registrationMap_[callId] = h323RegistrationEvent;

    CALLTRACE(log, CT_OUT, sipProxyInfo_.sipRemoteAddr(), CT_SIP, "REGISTER")
	<< "{\n" 
    	<< registerMsg.encode().logData()
	<< "}" << CALLTRACE_END(log);

    sipStack_->sendAsync(registerMsg);

    return ( PROXY_CONTINUE );
}
