/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const H323SipOpStatusIn_cxx_Version = 
  "$Id: H323SipOpStatusIn.cxx,v 1.19 2002/12/10 04:15:00 icahoon Exp $";


#include "H323SipOpStatusIn.hxx"
#include "SipEvent.hxx"
#include "StatusMsg.hxx"
#include "SipProxyInfo.hxx"
#include "VLog.hxx"
#include "CallTrace.hxx"
#include "IncomingSipEvent.hxx"


using namespace Vocal;
using namespace Vocal::SipH323Proxy;
using namespace Vocal::SipH323Info;
using namespace Vocal::SipEvents;
using namespace Vocal::Logging;
    

H323SipOpStatusIn::H323SipOpStatusIn(
    H323SipRegistrationMap   &	registrationMap,
    SipProxyInfo             &  sipProxyInfo
)
    :	H323SipOpIn(sipProxyInfo),
        registrationMap_(registrationMap)
{
}


H323SipOpStatusIn::~H323SipOpStatusIn() 
{
}

   
const char* const
H323SipOpStatusIn::name() const 
{
    return ( "H323SipOpStatusIn" );
}


const Sptr<State>
H323SipOpStatusIn::process(const Sptr<SipProxyEvent> event)
{
    const string    fn("H323SipOpStatusIn::process");
    VLog    	    log(fn);
    
    Sptr<SipEvent>  sipevent;
    if ( sipevent.dynamicCast(event) == 0 )
    {
        return ( PROXY_CONTINUE );
    }

    Sptr<SipMsg> sipMsg = sipevent->getSipMsg();
    Sptr<StatusMsg> statusMsg;
    statusMsg.dynamicCast(sipMsg);

    if ( statusMsg == 0 )
    {
        return ( PROXY_CONTINUE );
    }

    int statusCode = statusMsg->getStatusLine().getStatusCode();

    VDEBUG(log) << fn << ": Received " << statusCode << VDEBUG_END(log);

    // Ignore 100.
    //
    if ( statusCode == 100 )
    {
	VDEBUG(log) << fn << ": Received 100, no processing required." 
	    	    << VDEBUG_END(log);

    	return ( PROXY_CONTINUE );
    }

    // Handle registration as special case.
    //
    if ( statusMsg->getCSeq().getMethod() == REGISTER_METHOD )
    {
	// Identify registration event.
	//
	SipCallId callId = statusMsg->getCallId();

	H323SipRegistrationMap::iterator it = registrationMap_.find(callId);

	Sptr<H323RegistrationEvent> h323RegEventPtr 
	    = ( it == registrationMap_.end() ? 0 : (*it).second );
	
        if ( h323RegEventPtr == 0 )
	{
	    VWARN(log) << fn << ": Received unknown REGISTER RESPONSE" 
	    		<< VWARN_END(log);

	    return ( PROXY_CONTINUE );
	}

        // Registration event exists, so set return code appropriately.
	//
    	if ( statusCode == 200 )
	{
            h323RegEventPtr->returnCode(SUCCESS);
    	}
	else
	{
            h323RegEventPtr->returnCode(!SUCCESS);
	}

    	registrationMap_.erase(callId);

	CALLTRACE(log, CT_IN, sipProxyInfo_.sipRemoteAddr(), CT_SIP, "STATUS")
	    << "{\n" 
	    << statusMsg->encode().logData()
	    << "}" << CALLTRACE_END(log);

    	h323RegEventPtr->replyFifo().add(h323RegEventPtr);
	
	return ( PROXY_CONTINUE );
    }

    // Forward the rest to the call translator.
    //
    if ( fifo_ != 0 || serviceFifo_ != 0 )
    {
	CALLTRACE(log, CT_IN, sipProxyInfo_.sipRemoteAddr(), CT_SIP, "STATUS")
	    << "{\n" 
    	    << statusMsg->encode().logData()
	    << "}" << CALLTRACE_END(log);

    	fifo_->add(new IncomingSipEvent(sipevent, *serviceFifo_));
    }
    else
    {	
    	VWARN(log)  << fn << ": Received sip msg from SipStack "
                    << "before initialization. \n"
		    << VWARN_END(log);
    }

    return ( PROXY_CONTINUE );
}
