/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Q931ProtocolBase_cxx_Version = 
    "$Id$";


#include "VLog.hxx"

#include "Q931ProtocolBase.hxx"
#include "RouterInfo.hxx"
#include "PollFifoService.hxx"
#include "Q931Base.hxx"
#include "H323CodecException.hxx"
#include "ConnectionCallSig.hxx"
#include "EvtTCPConnected.hxx"
#include "ControlEvent.hxx"
#include "CallTrace.hxx"
#include "H245Defs.hxx"


using namespace Vocal;
using Vocal::Logging::VLog;


Q931ProtocolBase::Q931ProtocolBase(
    PollFifoService 	    &	service,
    Sptr<TPKTClientSocket>  	q931Socket,
    router::SesQ931         &   q931EventSrc,
    RouterInfo	    	    &	routerInfo, 
    const char      	    *	name = 0
    )
    : 	TPKTClientProtocol(name ? name : "Q931ProtocolBase"),
    	service_(service),
    	q931Socket_(q931Socket),
        q931EventSrc_(q931EventSrc),
        routerInfo_(routerInfo),
        connection_(0)
{
    awaitIncoming(true);
    service_.registerProtocol(*this);
    getTPKTClientSocket().setNonblocking();
}


Q931ProtocolBase::~Q931ProtocolBase()
{
    VLog log;
    VDEBUG(log) << "Q931ProtocolBase::~Q931ProtocolBase" << VDEBUG_END(log);

    service_.unregisterProtocol(*this);
    onDestroyCall();
}


const FileDescriptor &	    
Q931ProtocolBase::getFileDescriptor() const
{
    assert( q931Socket_ != 0 );
    return ( *q931Socket_ );
}


TPKTClientSocket &     	    
Q931ProtocolBase::getTPKTClientSocket()
{
    assert( q931Socket_ != 0 );
    return ( *q931Socket_ );
}


void Q931ProtocolBase::
onOutgoingAvailable()
{
    assert(this);
    assert(connection_);

    awaitOutgoing(false);
    awaitIncoming(true);
    connection_->getSM()->getFifo()->updateProtocol(*this);

    // Generate exception if connection was not opened
    getTPKTClientSocket().completeConnection();

    // Generate 'TCP Connected' event
    router::EvtTCPConnected evt(connection_->getSM());
    connection_->connected();
    evt.start();
}


void Q931ProtocolBase::
onTPKTArrived(vector<u_int8_t> & message)
{
    VLog log;
    VDEBUG(log) << "Q931ProtocolBase::onTPKTArrived, size="
                << message.size()
                << VDEBUG_END(log);

    assert(this);

    try
    {
        PASN_OctetString pos;
        pos.SetValue(&message[0], message.size());

        router::Q931Base msg;

        if (!msg.getQ931().Decode(pos.GetValue()))
        {
            throw router::H323CodecException(
                "Problem decoding Q.931 pdu",
                __FILE__, __LINE__,
                router::H323CodecException::Q931_DECODE_ERROR);
        }
        else
        {
            if (!msg.getQ931().HasIE(Q931::UserUserIE)) {
                VERR(log) << "No Q931 User-User field for Q.931 PDU:\n  "
                            << setprecision(2) << msg.getQ931()
                            << VERR_END(log);
                throw router::H323CodecException(
                    "No Q931 User-User field",
                    __FILE__, __LINE__,
                    router::H323CodecException::H225_DECODE_ERROR);
            }

            PPER_Stream strm = msg.getQ931().GetIE(Q931::UserUserIE);
            if (!msg.Decode(strm)) {
                VERR(log)
                    << "Read error: "
                    << "PER decode failure in Q.931 User-User field\n"
                    << "Partial PDU:\n  "
                    << setprecision(2) << msg
                    << VERR_END(log);
                throw router::H323CodecException(
                    "Problem decoding H.225 pdu",
                    __FILE__, __LINE__,
                    router::H323CodecException::H225_DECODE_ERROR);
            }
        }

        if (!connection_)
        {
            /*
            //TODO***:
            static int x = 0;
            x++;
            if(x>1)
            {
                service_.getServiceManagerFifo().add(new ControlEvent(ControlEvent::SHUTDOWN));
                return;
            }
            */
            makeNewCall();
        }
        assert(connection_);

#ifndef DISABLE_WRITER
    CALLTRACE(log, CT_IN, getTPKTClientSocket().getRemoteAddress(),
	      CT_Q931, string(msg.m().GetTagName())) 
        << msg << setprecision(0) << "\n(Q.931)= " << msg.getQ931()
	<< CALLTRACE_END(log);
#endif

        q931EventSrc_.generate(msg, connection_);
    }
    catch(router::H245EndSession&)
    {
        VINFO(log) << "Terminating the call on EndSession command"
                   << VINFO_END(log);
    }
    catch(...)
    {
        VWARN(log) << "Terminating the call" << VWARN_END(log);

        terminate();
    }
}


void Q931ProtocolBase::
onDisconnect(ConnectionBrokenException*)
    throw ( Vocal::Transport::ProtocolException )
{
    VLog log;

    VWARN(log) << "Got TCP disconnect for call signalling channel, "
              << "terminating the call" << VWARN_END(log);
    // TODO*: try to close socket and ignore

    terminate();
}


void Q931ProtocolBase::
onError(SystemException*)
    throw ( Vocal::Transport::ProtocolException )
{
    VLog log;

    VWARN(log) << "Got TCP error for call signalling channel, "
              << "terminating the call" << VWARN_END(log);

    terminate();
}


void Q931ProtocolBase::
handleUnknownException()
{
    VLog log;

    VWARN(log) << "Got unknown exception for call signalling channel, "
              << "terminating the call" << VWARN_END(log);

    terminate();
}


void Q931ProtocolBase::
terminate()
{
    if(connection_)
    {
        // We have a call object - destroy it
	connection_->disable();
        connection_->getSM()->terminate();
    }
    else
    {
        // No call object, only destroy self
        delete this;
    }
}


void Q931ProtocolBase::
setConnection(router::ConnectionCallSig* connection) throw()
{
    assert(this);
    assert(connection);

    connection_ = connection;
}
