/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2001 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const RasClient_cxx_Version = 
    "$Id: RasClient.cxx,v 1.2 2002/05/15 07:33:56 icahoon Exp $";


#include "RasCommon.hxx"
#include "RasClient.hxx"
#include "RasConfig.hxx"
#include "Poll.hxx"
#include "SocketOptions.hxx"
#include "SystemException.hxx"
#include "SystemStatus.hxx"
#include "TimeVal.hxx"
#include "Alias.hxx"
#include "vstring.hxx"
#include <iostream>
#include <cstring>
#include <cassert>


using namespace Vocal;
using Vocal::SipH323Info::Alias;
using Vocal::TimeAndDate::TimeVal;


RasClient::RasClient(const RasConfig & config)
    :   myConfig(config),
        myEncoded(false),
        myInSize(1024)
{
    IPAddress   addr;

    if ( !myConfig.address().isAny() )
    {
        addr = myConfig.address();
    }
    else
    {
        addr = myConfig.rasAddr();
    }
    
    try 
    {
        mySocket = new UDPSocket(addr);

        SocketOptions   sockopt(*mySocket);
        sockopt.bsdCompatible();
    }
    catch ( SystemException & e )
    {
        cerr << "Error creating UDP socket on " << addr << NL
             << e.getDescription() << NL
             << "Error: " << strerror(e.getError()) << endl;
        throw;
    }

    if ( myConfig.multicast() )
    {
        myGkAddr.setIP("224.0.1.41:1718");
    }
    else
    {
        myGkAddr = myConfig.gkAddr();
    }
}


RasClient::~RasClient()
{
    delete mySocket;
}


ReturnCode          
RasClient::discoverGatekeeper()
{
    outPdu.SetTag(H225_RasMessage::e_gatekeeperRequest);

    H225_GatekeeperRequest	&   grq = outPdu;
    
    grq.m_requestSeqNum = myConfig.seqNum();
    grq.m_protocolIdentifier.SetValue(PROTOCOL_ID);
    grq.m_rasAddress = getIPAddress(myConfig.rasAddr());
    grq.m_endpointType = buildEndpointType();
    if ( myConfig.gkId().size() )
    {
        grq.IncludeOptionalField(H225_GatekeeperRequest::e_gatekeeperIdentifier);
        grq.m_gatekeeperIdentifier = myConfig.gkId().c_str();
    }
    grq.m_endpointAlias = buildAlias();

    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( discoverGatekeeperResponse() );
}


ReturnCode          
RasClient::registerEndpoint()
{
    outPdu.SetTag(H225_RasMessage::e_registrationRequest);

    H225_RegistrationRequest	&   rrq = outPdu;
    
    rrq.m_requestSeqNum = myConfig.seqNum();
    rrq.m_protocolIdentifier.SetValue(PROTOCOL_ID);
    rrq.m_discoveryComplete = myConfig.discoveryComplete();
    appendAddress(rrq.m_callSignalAddress, myConfig.callSigAddr());
    appendAddress(rrq.m_rasAddress, myConfig.rasAddr());
    rrq.m_terminalType = buildEndpointType();

    if ( myConfig.phone().size() || myConfig.user().size() )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_terminalAlias);
        rrq.m_terminalAlias = buildAlias();
    }

    if ( myConfig.gkId().size() )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_gatekeeperIdentifier);
        rrq.m_gatekeeperIdentifier = myConfig.gkId().c_str();
    }
    
    rrq.m_endpointVendor = buildVendorIdentifier();

    if ( myConfig.timeToLive() )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_timeToLive);
        rrq.m_timeToLive = myConfig.timeToLive();
    }

    rrq.IncludeOptionalField(H225_RegistrationRequest::e_keepAlive);
    rrq.m_keepAlive = myConfig.keepAlive();

    if ( myConfig.endpointId().size() )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_endpointIdentifier);
        rrq.m_endpointIdentifier = myConfig.endpointId().c_str();
    }

    rrq.IncludeOptionalField(H225_RegistrationRequest::e_willSupplyUUIEs);
    rrq.m_willSupplyUUIEs = false;

    rrq.IncludeOptionalField(H225_RegistrationRequest::e_maintainConnection);
    rrq.m_maintainConnection = true;
    
    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( registerEndpointResponse() );
}


ReturnCode          
RasClient::unregisterEndpoint()
{
    outPdu.SetTag(H225_RasMessage::e_unregistrationRequest);

    H225_UnregistrationRequest	&   urq = outPdu;
    
    urq.m_requestSeqNum = myConfig.seqNum();
    appendAddress(urq.m_callSignalAddress, myConfig.callSigAddr());
    if ( myConfig.endpointId().size() )
    {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointIdentifier);
        urq.m_endpointIdentifier = myConfig.endpointId().c_str();
    }

    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( unregisterEndpointResponse() );
}


ReturnCode          
RasClient::admitCall()
{
    outPdu.SetTag(H225_RasMessage::e_admissionRequest);

    H225_AdmissionRequest	&   arq = outPdu;
    
    arq.m_requestSeqNum = myConfig.seqNum();

    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( admitCallResponse() );
}


ReturnCode          
RasClient::disengageCall()
{
    outPdu.SetTag(H225_RasMessage::e_disengageRequest);

    H225_DisengageRequest	&   drq = outPdu;
    
    drq.m_requestSeqNum = myConfig.seqNum();

    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( disengageCallResponse() );
}


ReturnCode          
RasClient::locateEndpoint()
{
    outPdu.SetTag(H225_RasMessage::e_locationRequest);

    H225_LocationRequest	&   lrq = outPdu;
    
    lrq.m_requestSeqNum = myConfig.seqNum();

    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( registerEndpointResponse() );
}


ReturnCode          
RasClient::reportResources()
{
    outPdu.SetTag(H225_RasMessage::e_resourcesAvailableIndicate);

    H225_ResourcesAvailableIndicate	&   rai = outPdu;
    
    rai.m_requestSeqNum = myConfig.seqNum();
    rai.m_protocolIdentifier.SetValue(PROTOCOL_ID);
    rai.m_endpointIdentifier = myConfig.endpointId().c_str();
    rai.m_almostOutOfResources = myConfig.outOfResources();
    
    ReturnCode rc = SUCCESS;
    
    if  (   ( rc = send() ) != SUCCESS 
        ||  ( rc = receive() ) != SUCCESS
        )
    {
        return ( rc );
    }
    
    return ( reportResourcesResponse() );
}


ReturnCode          
RasClient::discoverGatekeeperResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_gatekeeperConfirm )
    {
        return ( GRJ );
    }
    
    const H225_GatekeeperConfirm & gcf = pdu;
    
    if ( gcf.HasOptionalField(H225_GatekeeperConfirm::e_gatekeeperIdentifier) )
    {
        cout << zGatekeeper_Id << ": " << gcf.m_gatekeeperIdentifier.GetValue()
             << endl;
    }
    cout << zRas_Address << ": " << getTransportAddress(gcf.m_rasAddress) << endl;

    return ( GCF );
}


ReturnCode          
RasClient::registerEndpointResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_registrationConfirm )
    {
        return ( RRJ );
    }
    
    const H225_RegistrationConfirm & rcf = pdu;

    IPAddress call_sig_addr = getIPAddress(rcf.m_callSignalAddress);
    if ( !call_sig_addr.isAny() )
    {
        cout << zCall_Signaling_Address << ": " << call_sig_addr << endl;
    }

    if ( rcf.HasOptionalField(H225_RegistrationConfirm::e_terminalAlias) )
    {
        Alias   alias(rcf.m_terminalAlias);
        if ( alias.hasPhone() )
        {
            cout << zPhone_Number << ": " << alias.phone() << endl;
        }
        if ( alias.hasUser() )
        {
            cout << zUser << ": " << alias.user() << endl;
        }
    }

    if ( rcf.HasOptionalField(H225_RegistrationConfirm::e_gatekeeperIdentifier) )
    {
        cout << zGatekeeper_Id << ": " << rcf.m_gatekeeperIdentifier.GetValue()
             << endl;
    }
    
    cout << zEndpoint_Id << ": " << rcf.m_endpointIdentifier.GetValue() << endl;
    
    if ( rcf.HasOptionalField(H225_RegistrationConfirm::e_timeToLive) )
    {
        cout << zTime_To_Live << ": " << rcf.m_timeToLive.GetValue() << endl;
    }
    
    return ( RCF );
}


ReturnCode          
RasClient::unregisterEndpointResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_unregistrationConfirm )
    {
        return ( URJ );
    }
    return ( UCF );
}


ReturnCode          
RasClient::admitCallResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_admissionConfirm )
    {
        return ( ARJ );
    }
    return ( ACF );
}


ReturnCode          
RasClient::disengageCallResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_disengageConfirm )
    {
        return ( DRJ );
    }
    return ( DCF );
}


ReturnCode          
RasClient::locateEndpointResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_locationConfirm )
    {
        return ( LRJ );
    }
    return ( LCF );
}


ReturnCode          
RasClient::reportResourcesResponse()
{
    const H225_RasMessage & pdu = inPdu;

    if ( pdu.GetTag() != H225_RasMessage::e_resourcesAvailableConfirm )
    {
        return ( HUH );
    }
    return ( RAC );
}


UDPSocket &                 
RasClient::socket()
{
    assert(mySocket);
    return ( *mySocket );
}


H225_TransportAddress
RasClient::getIPAddress(const IPAddress & ipAddress) const
{
    H225_TransportAddress   transAddr;
    transAddr.SetTag(H225_TransportAddress::e_ipAddress);

    H225_TransportAddress_ipAddress & ipAddr = transAddr;
    memcpy( &ipAddr.m_ip[0], 
    	    &(ipAddress.getIPAddress().sin_addr.s_addr), 
	    4);

    ipAddr.m_port = ipAddress.getPort();
    
    return ( transAddr );
}


IPAddress                   
RasClient::getTransportAddress(const H225_TransportAddress & transAddr) const
{
    IPAddress ip;
    
    if ( transAddr.GetTag() == H225_TransportAddress::e_ipAddress )
    {
        // @#$%*%$!
        //
    	const H225_TransportAddress_ipAddress & cipAddr = transAddr;
        H225_TransportAddress_ipAddress ipAddr 
            = const_cast<H225_TransportAddress_ipAddress &>(cipAddr);

        ip.setIP(ipAddr.m_ip.GetPointer());
	ip.setPort(ipAddr.m_port);
    }
    return ( ip );
}


IPAddress
RasClient::getIPAddress(const H225_ArrayOf_TransportAddress & transAddrs) const
{
    IPAddress   ip;
    
    for ( size_t i = 0; i < transAddrs.GetSize(); ++i )
    {
        if ( transAddrs[i].GetTag() == H225_TransportAddress::e_ipAddress )
        {
            ip = getTransportAddress(transAddrs[i]);
            break;
        }
    }
    return ( ip );
}



H225_VendorIdentifier
RasClient::buildVendorIdentifier() const
{
    u_int8_t    cc = myConfig.countryCode() 
                        ?   myConfig.countryCode() 
                        :   181;                    // US by default
                        
    u_int8_t    ext = myConfig.extension();
    
    u_int16_t   mc = myConfig.manufacturerCode() 
                        ? myConfig.manufacturerCode() 
                        : 22399;                    // Vovida by default

    H225_VendorIdentifier   vendId;
    
    vendId.m_vendor.m_t35CountryCode = cc;
    vendId.m_vendor.m_t35Extension = ext;
    vendId.m_vendor.m_manufacturerCode = mc;
    
    if ( myConfig.productId().size() )
    {
        vendId.IncludeOptionalField(H225_VendorIdentifier::e_productId);
        vendId.m_productId.SetValue(myConfig.productId().c_str());
    }
    if ( myConfig.versionId().size() )
    {
        vendId.IncludeOptionalField(H225_VendorIdentifier::e_versionId);
        vendId.m_versionId.SetValue(myConfig.versionId().c_str());
    }
    
    return ( vendId );
}


H225_EndpointType
RasClient::buildEndpointType() const
{
    H225_EndpointType   eptType;
    
    eptType.IncludeOptionalField(H225_EndpointType::e_vendor);
    eptType.m_vendor = buildVendorIdentifier();
    
    switch ( myConfig.endpointType() )
    {
        case RasConfig::Terminal:
        {
            eptType.IncludeOptionalField(H225_EndpointType::e_terminal);
            break;
        }
        case RasConfig::Gateway:
        {
            eptType.IncludeOptionalField(H225_EndpointType::e_gateway);
            break;
        }
        case RasConfig::MCU:
        {
            eptType.IncludeOptionalField(H225_EndpointType::e_mcu);
            break;
        }
        default:
        {
            assert(0);
        }
    }
    
    eptType.m_mc = myConfig.mc();
    eptType.m_undefinedNode = false;
    
    return ( eptType );
}



H225_ArrayOf_AliasAddress
RasClient::buildAlias() const
{
    H225_ArrayOf_AliasAddress   result;
    
    if ( myConfig.phone().size() || myConfig.user().size() )
    {
        Alias alias;
        
        if ( myConfig.user().size() )
        {
            alias.user(myConfig.user());
        }
        if ( myConfig.phone().size() )
        {
            alias.phone(myConfig.phone());
        }
        result = alias.alias();
    }
    return ( result );
}


void    
RasClient::appendAddress(H225_ArrayOf_TransportAddress & src, IPAddress ip) const
{
    PINDEX size = src.GetSize();
    src.SetSize(size + 1);
    
    H225_TransportAddress & transAddr = src[size];

    transAddr.SetTag(H225_TransportAddress::e_ipAddress);

    H225_TransportAddress_ipAddress & ipAddr = transAddr;
    
    memcpy( &ipAddr.m_ip[0], 
    	    &(ip.getIPAddress().sin_addr.s_addr), 
	    4);

    ipAddr.m_port = ip.getPort();
}


void
RasClient::encode()
{
    if ( !myEncoded )
    {
        outPdu.Encode(myOut);
        myOut.CompleteEncoding();

        myEncoded = true;
    }

    cout.precision(0);
    log_debug(cout) << outPdu << endl;

}


ReturnCode
RasClient::decode()
{
    if ( !inPdu.Decode(myIn) )
    {
        return ( DECODE_FAILED );
    }

    cout.precision(0);
    log_debug(cout) <<'\n' << inPdu << endl;
    
    return ( SUCCESS );
}


ReturnCode  	    	    
RasClient::send()
{
    log_debug(cout) << "Sending message to: " << myGkAddr << endl;

    encode();
    
    bool sent = false;
    
    while ( !sent )
    {
        try
        {    
            size_t bytesSent = socket().sendTo(myOut.GetPointer(), 
                                               myOut.GetSize(),
                                               myGkAddr);
                                            
            assert( bytesSent == myOut.GetSize() );

            if ( myConfig.debug() )
            {
	        cout << "Packet dump, size = " << bytesSent;

                u_int8_t    *   bytes = myOut.GetPointer();

	        for ( size_t i = 0; i < bytesSent; i++ )
	        {
   	            if( i % 16 == 0 )
	            {
		        cout << "\n\t";
	            }
	            cout << setfill('0') << setw(2) << hex 
                         << (unsigned)bytes[i] << ' ';
	        }
	        cout << setfill(' ') << dec << endl;
            }
        }
        catch ( SystemStatus & status )
        {
            if ( status.getStatus() == EINTR )
            {
                continue;
            }
            
            cerr << "Error: Sending on socket: " << socket() 
                 << " to address: " << myGkAddr << NL
                 << status.getDescription() << NL << NL;
                 
            return ( SEND_FAILED );
        }
        sent = true;
    }

    return ( SUCCESS );
}


ReturnCode
RasClient::receive()
{
    ReturnCode  rc = SUCCESS;
    
    Poll    poll;
    poll.registerFD(socket());
    
    u_int16_t   retry_count = myConfig.retryCount();
    u_int16_t   retry_delay = myConfig.retryDelay();

    size_t  i    = 0;
    bool    done = false;
        
    for ( ; !done && i <= retry_count; ++i )
    {
        TimeVal whenToRetry 
            = TimeVal().now() + TimeVal(retry_delay, 0);
        
        bool    poll_done = false;
        
        while ( !poll_done )
        {
            TimeVal now;
            now.now();
            
            TimeVal delay = whenToRetry - now;

            log_debug(cout) << "Before poll: " << TimeVal().now() 
                 << ": delay: " << delay << endl;
                        
            ReturnCode  numberActive = poll.poll(delay.milliseconds());

            log_debug(cout) << "After poll: " << TimeVal().now() 
                 << ": fds active: " << numberActive << endl;

            // Timed out
            //
            if ( numberActive < 1 && TimeVal().now() > whenToRetry )
            {
                log_debug(cout) << "Timed out." << endl;
                
                if ( i != retry_count )
                {
                    rc = send();
                
                    if ( rc != SUCCESS )
                    {
                        done = true;
                    }
                }
                else
                {
                    done = true;
                }
                
                poll_done = true;
            }

            // Read the new message
            //
            else if ( numberActive > 0 )
            { 
                IPAddress   remoteAddr;
                
                size_t  bytesReceived = socket().receiveFrom(
    	    	    	                            myIn.GetPointer(myInSize),
			                            myInSize,
			                            remoteAddr);

                log_debug(cout) << "Received message from: " << remoteAddr 
                                << endl;

	        log_debug(cout) << "Packet dump, size = " << bytesReceived;

                u_int8_t    *   bytes = myIn.GetPointer();

                if ( myConfig.debug() )
                {
	            for ( size_t i = 0; i < bytesReceived; i++ )
	            {
   	                if( i % 16 == 0 )
	                {
		            cout << "\n\t";
	                }
	                cout << setfill('0') << setw(2) << hex 
                             << (unsigned)bytes[i] << ' ';
	            }
	            cout << setfill(' ') << dec << endl;
                }

                // If we receive an empty message, poll again.
                //
                if  (   bytesReceived == 0 
                    ||  ( !myConfig.multicast() && remoteAddr != myGkAddr )
                    )
                {
                    log_debug(cout) << "Received empty message. Ignoring" 
                                    << endl;
    	            continue;
                }

                bool    grow_buffer = false;
                
                // See if we need to grow the buffer.
                //
                if ( bytesReceived == myInSize )
                {
                    grow_buffer = true;
                    
    	            // Double the size.
	            //
	            myInSize *= 2;
                }

                rc = decode();

                if ( rc == DECODE_FAILED  )
                {
                    log_debug(cout) << "Decode failed." << endl;
                    
                    if ( grow_buffer )
                    {
                        // Maybe the buffer was too small, retransmit the
                        // message in hopes of decoding the next response.
                        //
                        rc = send();

                        if ( rc != SUCCESS )
                        {
                            done = true;
                        }
                        poll_done = true;
                    }
                    else
                    {
                        done = true;
                    }
                }
                else if ( !sequenceNumberMatch() )
                {
                    log_debug(cout) << "Sequence number mismatch." << endl;
                    continue;
                }
                else if ( inPdu.GetTag() == H225_RasMessage::e_requestInProgress )
                {
                    log_debug(cout) << "Sequnec number mismatch." << endl;
                    H225_RequestInProgress & src = inPdu;
                    whenToRetry = TimeVal().now() + TimeVal(src.m_delay, 0);

                    log_debug(cout) << "Received RIP. Rety in "
                                    << src.m_delay << " seconds." << endl;

                    continue;
                }
                else
                {
                    // We have received a message and have decoded it.
                    // We are good to go.
                    //
                    poll_done = true;
                    done = true;
                }
            }
        }
    }
    
    poll.unregisterFD(socket());
    
    if ( rc == SUCCESS && i > retry_count )
    {
        rc = RETRY_EXPIRY;
    }
    
    return ( rc );    
}


bool                        
RasClient::sequenceNumberMatch() const
{
    H225_RequestSeqNum      in, out, tmp;
    
    for ( size_t i = 0; i < 2; ++i )
    {
        const H225_RasMessage & pdu = ( i ? outPdu : inPdu );

        switch ( pdu.GetTag() )
        {
            case H225_RasMessage::e_gatekeeperRequest:
            {
                const H225_GatekeeperRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_gatekeeperConfirm:
            {
                const H225_GatekeeperConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_gatekeeperReject:
            {
                const H225_GatekeeperReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_registrationRequest:
            {
                const H225_RegistrationRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_registrationConfirm:
            {
                const H225_RegistrationConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_registrationReject:
            {
                const H225_RegistrationReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_unregistrationRequest:
            {
                const H225_UnregistrationRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_unregistrationConfirm:
            {
                const H225_UnregistrationConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_unregistrationReject:
            {
                const H225_UnregistrationReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_admissionRequest:
            {
                const H225_AdmissionRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_admissionConfirm:
            {
                const H225_AdmissionConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_admissionReject:
            {
                const H225_AdmissionReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_bandwidthRequest:
            {
                const H225_BandwidthRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_bandwidthConfirm:
            {
                const H225_BandwidthConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_bandwidthReject:
            {
                const H225_BandwidthReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_disengageRequest:
            {
                const H225_DisengageRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_disengageConfirm:
            {
                const H225_DisengageConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_disengageReject:
            {
                const H225_DisengageReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_locationRequest:
            {
                const H225_LocationRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_locationConfirm:
            {
                const H225_LocationConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_locationReject:
            {
                const H225_LocationReject & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_infoRequest:
            {
                const H225_InfoRequest & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_infoRequestResponse:
            {
                const H225_InfoRequestResponse & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_nonStandardMessage:
            {
                const H225_NonStandardMessage & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_unknownMessageResponse:
            {
                const H225_UnknownMessageResponse & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_requestInProgress:
            {
                const H225_RequestInProgress & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_resourcesAvailableIndicate:
            {
                const H225_ResourcesAvailableIndicate & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_resourcesAvailableConfirm:
            {
                const H225_ResourcesAvailableConfirm & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_infoRequestAck:
            {
                const H225_InfoRequestAck & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            case H225_RasMessage::e_infoRequestNak:
            {
                const H225_InfoRequestNak & src = pdu;
                tmp = src.m_requestSeqNum;
                break;
            }
            default:
            {
                return ( false );
            }
        }
        
        if ( i )
        {
            out = tmp;
        }
        else
        {
            in = tmp;
        }
    }
    return ( in == out );
}
