/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const TrConnectionSIP_cxx_Version = 
    "$Id: TrConnectionSIP.cxx,v 1.23 2002/02/25 05:57:17 sprajpat Exp $";


#include "VLog.hxx"
#include "PollFifoService.hxx"

#include "SipMsg.hxx"
#include "SipSdp.hxx"
#include "InviteMsg.hxx"
#include "IncomingSipEvent.hxx"
#include "OutgoingSipEvent.hxx"

#include "TrConnectionSIP.hxx"
#include "TrSMCall.hxx"
#include "TrRouterSIP.hxx"
#include "EvtGotSDP.hxx"
#include "EvtProcessSDP.hxx"

using namespace translator;
using Vocal::Logging::VLog;
using Vocal::SipEvents::OutgoingSipEvent;

TrConnectionSIP::
TrConnectionSIP(TrSMCall* sm)
    : ConnSIP(sm),
      haveContents_(false),
      active_(true),
      registered_(false),
      nuked_(0)
{
    const string    fn("TrConnectionSIP::TrConnectionSIP");
    VLog            log(fn);
    
    VDEBUG(log) << fn << ": this = " << this << VDEBUG_END(log);
}

TrConnectionSIP::
~TrConnectionSIP()
{
    const string    fn("TrConnectionSIP::~TrConnectionSIP");
    VLog            log(fn);
    
    VDEBUG(log) << fn << ": this = " << this << VDEBUG_END(log);
    
    active_= false;
    
    // This is a huge, but necessary kludge. This is to avoid the
    // delete this problem in routeImpl.
    //
    if ( nuked_ != 0 )
    {
        *nuked_ = true;
    }
}

void TrConnectionSIP::
disable() throw()
{
    active_ = false;
}


void TrConnectionSIP::
setContents(SipSdp contents)
{
    const string    fn("TrConnectionSIP::setContents");
    VLog            log(fn);

    if (haveContents_)
    {
        VWARN(log) << fn << ": Rewritting old contents for outgoing SIP"
                   << VWARN_END(log);
    }

    contents_ = contents;
    haveContents_ = true;
}


bool TrConnectionSIP::
hasContents() const
{
    return haveContents_;
}


void TrConnectionSIP::
send(Sptr<SIPBase> message)
{
    const string    fn("TrConnectionSIP::send");
    VLog            log(fn);
    
    VDEBUG(log) << fn << ": TrConnectionSIP::send" << VDEBUG_END(log);

    if (!active_)
    {
        // Connection was shut down from outside
        return;
    }

    // Before sending SIP message check if some SDP data is to be sent
    // in reply to incoming SDP data
    // TODO: make sure all methods calling send() keep proper state and
    // no additional SIP message will be created
    EvtProcessSDP sdpEventP(getSM()->getSDPConnection());
    sdpEventP.start();

    // Remove contents if message contained something
    message->flushContentData();

#ifndef DISABLE_WRITER
    Data msgData = message->encode();
    VDEBUG(log) << fn << ": Sending SIP message =\n"
                << msgData.logData()
                << VDEBUG_END(log);
#endif

    // Add contents
    if (haveContents_)
    {
        message->setContentData(&contents_);
        //contents_.clear();
        haveContents_ = false;
    }

#ifndef DISABLE_WRITER
    msgData = message->encode();
    VDEBUG(log) << fn << ": SIP message after contents insertion =\n"
                << msgData.logData()
                << VDEBUG_END(log);
#endif

    // Send it to proxy
    Sptr<SipEvent> sipEvent= new SipEvent(0);
    sipEvent->setSipMsg(message);
    Sptr<OutgoingSipEvent> ose = new OutgoingSipEvent(sipEvent);
    assert(getSM()->getProxyFifo());
    getSM()->getProxyFifo()->add(ose);
    VDEBUG(log) << fn << ": Sip message sent, proxy=" << getSM()->getProxyFifo()
		<< VDEBUG_END(log);

    Sptr<InviteMsg> s;
    s.dynamicCast(message);
    if (!registered_ && s != 0)
    {
        VDEBUG(log) << fn << ": Registering connection..." << VDEBUG_END(log);
	
        // Fill in SIPID
        id_ = message->getCallId();
        setProxy(getSM()->getProxyFifo());

        assert(getSM()->getSipRouter());
        getSM()->getSipRouter()->add(*this);
        registered_ = true;
        VDEBUG(log) << fn << ": Sip connection registered" << VDEBUG_END(log);
    }

    VDEBUG(log) << fn << ": exit" << VDEBUG_END(log);
}


void TrConnectionSIP::
routeImpl(EvtRouterType& evtRouter, SrcType& evtSource)
{
    const string    fn("TrConnectionSIP::routeImpl");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    const SipMsg* msg = &(evtSource.getSelMessage());

    VDEBUG(log) << fn << ": msg=" << msg << VDEBUG_END(log);

    if (msg)
    {
        if (msg->getNumContentData() > 1)
        {
            VWARN(log) << fn 
                << ": Multipart SIP content is not supported: using last part"
                << VWARN_END(log);
        }

        // Construct SDP message to getSM().getSDPConnection() and start it

        //SipSdp* remoteSdp = dynamic_cast<SipSdp*>(msg->getContentData());
	
	Sptr<SipSdp> remoteSdp;
	remoteSdp.dynamicCast(msg->getContentData());

        if (remoteSdp != 0)
        {
            EvtGotSDP sdpEvent(getSM()->getSDPConnection());
            sdpEvent.sdpSession_ = remoteSdp->getSdpDescriptor();
            sdpEvent.emptySdp_ = false;
            sdpEvent.start();
        }

        // Process empty SDP only for Invite (ignore for others)
        if ((!remoteSdp || msg->getNumContentData() < 1)
            && evtSource.getType() == SIP_INVITE)
        {
            EvtGotSDP sdpEvent(getSM()->getSDPConnection());
            sdpEvent.emptySdp_ = true;
            sdpEvent.start();
        }

        VDEBUG(log) << fn << ": SDP end" << VDEBUG_END(log);
    }

    VDEBUG(log) << fn << ": processing" << VDEBUG_END(log);

    // This is an ugly, but necessary kludge. This is to avoid the
    // "delete this" problem.
    //
    // evtRouter.process(...) below can potentially delete this object
    // so we need some way to find out if this object has been deleted. 
    // Using the nuked variable, which is on the stack, should work since 
    // it will survive even if the object has been deleted. 
    //
    bool    nuked = false;
    nuked_ = &nuked;

    evtRouter.process(*getSM());

    if ( nuked )
    {
        VVERBOSE(log) << fn << ": NUKED!" << VVERBOSE_END(log);
    }
    
    VDEBUG(log) << fn << ": processed" << VDEBUG_END(log);

    if ( !nuked && active_ )
    {
        EvtProcessSDP sdpEventP(getSM()->getSDPConnection());
        sdpEventP.start();
    }
}


#ifndef DISABLE_WRITER

ostream& TrConnectionSIP::
writeTo(ostream& out) const
{
    out << "TrConnectionSIP: " << this
        << ", sm=" << getSM();
    return out;
}

#endif
