#ifndef LSApiServer_hxx
#define LSApiServer_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSApiServer_hxx_Version = 
"$Id: LSApiServer.hxx,v 1.8 2001/06/29 21:02:44 bko Exp $";

#include "global.h"
#include <vector>
#include <string>

#include "vtypes.h"
#include "Sptr.hxx"
#include "VThread.hxx"
#include "Connection.hxx"
#include "TRIPDef.hxx"
#include "TRIPWorkSet.hxx"
#include "TRIPMsg.hxx"

class LSPeerSession;

/** 
    Singelton class, implements the  server side API to serve the 
    route request from LSApiClient called by the proxy.When instantiated
    create a seprate thread to start listening the client's request.
*/
class LSApiServer {
    public:
        ///List of connections with Proxy Servers
        typedef vector<Connection>  ConnectionList;

        ///Map of ProtoKey(Address Family, Protocol) and associated connections
        typedef map<TRIPProtoKey, Sptr <ConnectionList> > ConnectionListMap;

        ///Instance of LSApiServer initiated with listening port
        static LSApiServer& instance(int port);

        ///Instance of LSApiServer
        static LSApiServer& instance();

        ///ClassName of LSApiServer
        string className() { return "LSApiServer"; }

        ///Update all relevant proxies with the updated workSet
        void updateProxies(Sptr <TRIPWorkSet> workSet);

        ///Indication of whether or not a proxy is connected
        bool isProxyConnected() { return itsProxyConnectedFlag; }


        /**
           Frees the memory associated with singelton instance.
           gets register to atexit() function at the time of creation.
        */
        static void destroy(void);

    protected:
        ///Destructor
        virtual ~LSApiServer();

        ///Start LSApiServer running in its own thread
        static void* run(void* args);

        ///Set indication of Proxy being connected to LSApiServer
        void setProxyConnected(bool isProxyConnected)
        {
            itsProxyConnectedFlag = isProxyConnected;
        };

        ///Loop continuously while monitoring the Proxy Connections for messages
        void processMsg();

        ///Process a request from the associated Proxy Server
        void processRequest(Connection& conn) throw (VNetworkException&);

        ///Process an Open Msg from Proxy Server
        void processOpenMsg(Sptr<TRIPMsg> msg, Connection& conn);

        ///Process a Notification Msg from Proxy Server
        void processNotificationMsg(Sptr<TRIPMsg> msg, Connection& conn);

        ///Provide initial route data to appropriate Proxy Server
        void initialiseProxyRouteSet(Connection& conn, ProtoKeyList& pList);

        ///Get list of Update Msgs containing route data for Proxy
        Sptr <UpdateMsgList> getProxyMsgList(const TRIPRouteSet& routeSet, 
                                                                bool initial);

        ///Send Update Msgs from MsgList on the given Connection
        void sendProxyUpdates(Sptr <UpdateMsgList> msgList, 
                                                    Connection& conn);

        ///Send Update Msg on given Connection
        void sendUpdate(const TRIPUpdateMsg& uMsg, Connection& conn) 
                                                throw (VNetworkException&);

        ///Send Update Msgs from MsgList on all appropriate connections
        void sendAllProxiesUpdates(Sptr <UpdateMsgList> msgList,
                                                const TRIPProtoKey& protoKey);

        ///Remove the given connection 
        void removeConnection(Connection& conn);

        ///Constructor 
        LSApiServer(int port );


        ///instance of LSApiServer Class
        static LSApiServer* _instance;

        ///TCP/IP Server side socket
        Sptr<TcpServerSocket>  _serverSocket;

        ///LSApiServer thread
        Sptr<VThread>          _thread;

        ///List of associated Proxy Connections
        ConnectionList _connectionList;

        ///Map of Connection Lists and associated Protokeys
        ConnectionListMap itsConnListMap;

        ///Port where LSApiServer listens for incoming connections
        int _port;

        ///indicates if the LSApiServer is being destroyed
        bool _shutdown;

        ///boolean value indicating whether a proxy is currently connected
        bool itsProxyConnectedFlag;
};

#endif
