#ifndef LSConfig_hxx
#define LSConfig_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSConfig_hxx_Version = 
"$Id: LSConfig.hxx,v 1.2 2001/04/06 21:52:33 kevinmc Exp $";

#include "global.h"
#include <vector>
#include <string>

#include "TRIPDef.hxx"
#include "TRIPRouteElement.hxx"
#include "TRIPTransCapability.hxx"
#include "VLocationServerData.hxx"
#include "VLocationRouteData.hxx"
#include "ProvisionInterface.hxx"


/**
 * The Provisioning Server is notifying us of a change in LS data.
 */
void notifyLSServerUpdate(const string& data, const string& fileName, bool );

/**
 * The Provisioning Server is notifying us of a change in LS Routes data.
 */
void notifyLSRoutesUpdate(const string& data, const string& fileName, bool );


/** 
    LSConfig implements the Lookup of data from the Provisioning Server.
    The values for Timers, Peers, supported Protocols, supported Address
    Families, and Send/Receive capability are set through this interface.
    Also LSConfig takes care of the injection of local routes which are
    also read from the Provisioning Server.
*/
class LSConfig {
    public:
        ///
        typedef vector<string>  LSList;

        ///
        LSConfig();

        ///
        string className() { return "LSConfig"; }

        ///
        friend ostream& operator<<(ostream& os, const LSConfig& lsconfig);

        ///return the holdTimer for this LS
        int getHoldTimer() const { return itsHoldTimer; };

        ///return the RestartTimer for this LS
        int getRestartTimer() const { return itsRestartTimer; };

        ///return the MaxRestartTimer for this LS
        int getMaxRestartTimer() const { return itsMaxRestartTimer; };

        ///return the ConnectRetryTimer of this LS
        int getConnectRetryTimer() const { return itsConnectRetryTimer; };

        ///return the Max ConnectRetryTimer of this LS
        int getMaxConnectRetryTimer() const { return itsMaxConnectRetryTimer; };

        ///return the AdvertiseTimer for this LS
        int getAdvertiseTimer() const { return itsAdvertiseTimer; };

        ///return the tranport mode of operation for this LS
        const TRIPTransMode& getTransMode() const { return itsTransMode; };

        ///return the itad value of this LS
        const u_int16_t getItadValue() const { return itsItad; };

        ///return the list of LS Peers for this LS
        const LSList& getLSList() const { return itsLsList; };

        ///return the list of supported Protocols for this LS
        const ProtoCapabilityList& getProtoCapabilityList() const
        {
            return itsProtoCapabilityList;
        };

        ///return the list of supported address families for this LS
        const AddrFamilyList& getAddrFamilyList() const
        {
            return itsAddrFamilyList;
        };

        /**Takes an IP address and checks if the address is in the 
        * list of LSPeers that, this session is supposed to be peering with.
        * returns true if find the matching Ip.
        */
        bool isTrustedLs(const string& lsIp) const; 

        ///
        void init(const char* psHost = 0, int psPort = 0, 
                                                const char* altHost = 0, int altPort = 0 );
        
        ///Create Update msg(s) of local routes and pass to DecisionProxy
        void updateLocalRoutes();

        ///Provisioning data for PS has changed - update LS Data
        void updateLSFromPS();

        ///Provisioning data for PS has changed - update LS Routes Data
        void updateLSRoutesFromPS();

    protected:
        ///
        void init();

        ///default config uses lsconfig.cfg for peer sessions and default values
        void defaultConfig();

        ///
        TRIPAppProto getAppProto(const string& proto);

        ///
        TRIPAddrFamily getAddrFamily(const string& addr);

        ///Find the changes which have occurred to LSList during update
        void getChangesInList( const LSList& srcList, 
                                            const LSList& destList, LSList& retList);
        
        ///update the localadjTribIn with a given set of routes
        void updateRoutes(TRIPGeneralRouteSet& updateRouteSet);

        ///
        void injectUpdateMsg(Sptr<TRIPWithdrawnRoute> wRoute,
                                            Sptr<TRIPReachableRoute> rRoute, Data& nHop);

        ///Gets all the appropriate data from the VLocationServerData
        void getLocationData(const VLocationServerData& lsData, bool update);

        ///Retrieves the local Routes data for this LS
        void getLocationRouteData(const VLocationRouteData& lsRouteData);

        ///Provisioned HoldTimer in Seconds
        int itsHoldTimer;

        ///Provisioned RestartTimer in Seconds
        int itsRestartTimer;

        ///Max RestartTimer - restartTimer continues to increase up to this value
        int itsMaxRestartTimer;

        ///Provisioned ConnectRetryTimer in Seconds
        int itsConnectRetryTimer;

        ///Max ConnectRetryTimer - connectRetryTimer continues to increase up to this value
        int itsMaxConnectRetryTimer;

        ///Provisioned AdvertisementTimer in Seconds
        int itsAdvertiseTimer;

        ///Transport Mode of the LSSessionManager - Send, Receive, Send-Receive
        TRIPTransMode itsTransMode;

        ///Itad value for this LS
        u_int16_t itsItad;

        ///List of Protocols supported by this LS
        ProtoCapabilityList itsProtoCapabilityList;

        ///List of Address families supported by this address family
        AddrFamilyList itsAddrFamilyList;

        ///
        TRIPGeneralRouteSet itsLocalRouteSet;

        ///List of all Peers for LS
        LSList itsLsList;

        ///ip name for ls server
        string itsIpName;

        ///ip name with routes extension for LS Server (access to local routes)
        string itsIpNameWithRoute;
};


#endif
